//@***h* BDBP/MakeTier1Databases.dpr
// 
// NAME
//   MakeTier1Databases.dpr
//
// PURPOSE
//  Runs a multiple-linear regression with the available monthly mean, zonal mean 
//  data to create a gap-free data set. This data set is a combination of different 
//  basis function contributions.
//
// DESCRIPTION
//  The program runs a multiple-linear regression with the available monthly mean, 
//  zonal mean data to create a gap-free data set. This data set is a combination of 
//  different basis function contributions. Depending on the combination of the 
//  different basis function contributions, the different Tiers of the data set 
//  are created.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


program MakeTier1Databases;

{$APPTYPE CONSOLE}
{$I UNITDIREC.INC}

uses
  Classes,
  DateUtils,
  Dialogs,
  SysUtils,
  gb_Array,
  gb_AutocorrelationModel,
  gb_BasisFunctions,
  gb_ExcelUtils,
  gb_Files,
  gb_GeneralConstantsAndTypes,
  gb_LeastSquaresRegression,
  gb_RegressionBasisFunction,
  gb_strings;

const
  DensDesc:array[1..2] of string=('NumDens','MixRatio');
  SetDesc:array[1..2] of string=('LatAlt','LatPress');
  DataStartYear=1978;
  DataEndYear=2006;
  FitStartYear=1979;
  FitEndYear=2007;
  NumMonths=(DataEndYear-DataStartYear+1)*12;
  NumFitMonths=(FitEndYear-FitStartYear+1)*12;
  LatWidth=5;
  NumLevels=70;
  NumLat=180 div LatWidth;

type
  TLAMRec=record
    MonthlyMean,
    MonthlyMeanError,
    MonthlyMeanCosineError:single;
  end;

var
  Parameters:TStringList;
  Reg:TLeastSquaresRegression;
  DensCount,
  LevelCount,
  SetCount:integer;
  OutputDir:string;

  {This array holds all of the data that are a function of latitude, altitude}
  {and month number.}
  LAMData:array[1..NumLat,1..NumLevels,1..NumMonths] of TLAMRec;

  {This array holds the Tier 1.1 to Tier 1.4 data as a function of latitude, level,}
  {and month.}
  TierLAMData:array[1..4,1..NumLat,1..NumLevels,1..NumFitMonths] of single;

procedure ShowErrorAndHalt(ErrorMessage:string);
begin
  ShowMessage(ErrorMessage);
  Halt;
end;

function TransformOzone(Alt:word;Ozone:double):double;
begin
  if ((Alt<=10) or (Alt>=50) or (Ozone>=1.0)) then Result:=Ozone else Result:=ln(Ozone)+1.0;
end;

function InverseTransformOzone(Alt:word;Value:double):double;
begin
  if ((Alt<=10) or (Alt>=50) or (Value>=1.0)) then Result:=Value else Result:=exp(Value-1.0);
end;

procedure LoadParameters;
var
  OutputDir:string;
  I,J:integer;
begin
  if ParamCount<>1 then ShowErrorAndHalt('A parameter file was not specified.');
  if not FileExists(ParamStr(1)) then ShowErrorAndHalt('The parameter file could not be found.');
  Parameters:=TStringList.Create;
  Parameters.LoadFromFile(ParamStr(1));
  OutputDir:=Parameters.Values['OutputDirectory'];
  if pos('+VersionStr+',OutputDir)>0 then OutputDir:=ReplaceChars(OutputDir,'+VersionStr+',Parameters.Values['VersionStr']);
  if not DirectoryExists(OutputDir) then ForceDirectories(OutputDir);
  for I:=1 to 2 do
    for J:=1 to 2 do
      if not DirectoryExists(OutputDir+'\'+DensDesc[I]+'\'+SetDesc[J]) then ForceDirectories(OutputDir+'\'+DensDesc[I]+'\'+SetDesc[J]);
end;

procedure PrepareEESC(Hold:TEESCBasisFunctionWithAoA_SpecifyAlt);
var
  EESCStartYear,
  EESCEndYear:word;
begin
  if DataStartYear<FitStartYear then EESCStartYear:=DataStartYear else EESCStartYear:=FitStartYear;
  if DataEndYear>FitEndYear then EESCEndYear:=DataEndYear else EESCEndYear:=FitEndYear;
  Hold.SetApproximateStartAndEndYear(EESCStartYear,EESCEndYear);
  Hold.WithRespectTo1980:=true;
end;

procedure CreateBasisFunctions;
{------------------------------------------------------------------------------}
{ This is where the basis functions are created. The number of fourier expans- }
{ ions is not set here as these will be set individually for each level.       }
{------------------------------------------------------------------------------}
begin
  TOffsetBasisFunction.Create;
  PrepareEESC(TEESCBasisFunctionWithAoA_SpecifyAlt.Create(Parameters.Values['EESCPath']));
  TTrendBasisFunction.Create;
  TQBOBasisFunction.Create(Parameters.Values['QBOFile']);
  TQBOOrthogBasisFunction.Create(Parameters.Values['QBOFile']);
  TSOIBasisFunction.Create(Parameters.Values['SOIFile']);
  TSolarBasisFunction.Create(Parameters.Values['SolarPath']);
  TPinatuboBasisFunctionWithPresOffset.Create(Parameters.Values['OptimalPinatuboOffsetFilename']);
  TElChichonBasisFunction.Create;
end; {End of CreateBasisFunctions.}

procedure WriteParameters;
var
  Source,
  Dest:textfile;
  Hold:TBasisFunction;
  I,J:integer;
  OutputDir,
  Linein,
  NameStr,
  HeaderStr:string;
  Str:array[1..70] of string;
begin
  OutputDir:=Parameters.Values['OutputDirectory'];
  if pos('+VersionStr+',OutputDir)>0 then OutputDir:=ReplaceChars(OutputDir,'+VersionStr+',Parameters.Values['VersionStr']);
  if not DirectoryExists(OutputDir) then ForceDirectories(OutputDir);
  assignfile(Dest,OutputDir+'\Parameters.txt');
  rewrite(Dest);
  for I:=1 to Parameters.Count do writeln(Dest,Parameters[I-1]);
  writeln(Dest);
  HeaderStr:='';
  for I:=1 to 70 do Str[I]:='';
  for I:=1 to BasisFunctionList.Count do
  begin
    Hold:=BasisFunctionList[I-1] as TBasisFunction;
    NameStr:=copy(Hold.ClassName,2,length(Hold.ClassName));
    assignfile(Source,Parameters.Values['BasisFunctionParameterPath']+'\'+NameStr+'.txt');
    reset(Source);
    DeleteHeader(Source);
    for J:=1 to 70 do
    begin
      readln(Source,Linein);
      if I>1 then Linein:=copy(Linein,3,length(Linein));
      if J=1 then HeaderStr:=HeaderStr+BackBuffer(copy(NameStr,1,length(Linein)),' ',length(Linein))+'  ';
      Str[J]:=Str[J]+Linein+'  ';
    end;
    closefile(Source);
  end;
  writeln(Dest,HeaderStr);
  for I:=1 to 70 do writeln(Dest,Str[I]);
  closefile(Dest);
end;

function CreateRegressionModel:TLeastSquaresRegression;
begin
  Result:=TLeastSquaresRegression.Create(ResMonth,TTwoTermAutoCorrelation,true);
end;

function GetFilenameSubStr(LatIndex:integer):string;
var
  LatStr1,
  LatStr2:string;
begin
  LatStr1:=inttostrf(abs(90-(LatIndex-1)*LatWidth),2,true);
  if (90-(LatIndex-1)*LatWidth)<0.0 then LatStr1:=LatStr1+'S' else LatStr1:=LatStr1+'N';
  LatStr2:=inttostrf(abs(90-LatIndex*LatWidth),2,true);
  if (90-LatIndex*LatWidth)<0.0 then LatStr2:=LatStr2+'S' else LatStr2:=LatStr2+'N';
  Result:=LatStr1+'To'+LatStr2;
end;

procedure LoadMonthlyMeans(DataDirectory:string);
var                        
  Year,Month,
  I,J:integer;
  Source:file;
  Filename:string;
  MonthlyMeans:array[1..NumLevels,DataStartYear..DataEndYear,1..12] of TMeasRec;
begin
  for I:=1 to NumLat do
  begin
    Filename:=DataDirectory+'\BDBPMonthMeans_'+GetFilenameSubStr(I)+'.bin';
    if not FileExists(Filename) then ShowErrorAndHalt('The file named '+Filename+' doesn''t exist.');
    assignfile(Source,Filename);
    reset(Source,1);
    if filesize(Source)<>sizeof(MonthlyMeans) then ShowErrorAndHalt('The input file was not the right size.');
    blockread(Source,MonthlyMeans,sizeof(MonthlyMeans));
    closefile(Source);
    for J:=1 to NumLevels do
      for Year:=DataStartYear to DataEndYear do
        for Month:=1 to 12 do
        with LAMData[I,J,(Year-DataStartYear)*12+Month] do
        begin
          MonthlyMean:=MonthlyMeans[J,Year,Month].Value;
          if MonthlyMean<-900 then MonthlyMeanError:=-999.99 else MonthlyMeanError:=1/sqrt(MonthlyMeans[J,Year,Month].Weight);
          if MonthlyMean<-900 then MonthlyMeanCosineError:=-999.99 else MonthlyMeanCosineError:=1/sqrt(MonthlyMeans[J,Year,Month].CosWeight);
        end;
  end;
end; {End of LoadMonthlyMeans.}

procedure PassDataToRegressionModel;
const
  NumShortYears=5;
  YearsToPad=0;
var
  InfoObj:TLatitudeInfo;
  Num,
  Year,
  Month:word;
  LowerLevel,UpperLevel,
  LevelDelta,
  LevCount,
  LatCount:integer;
  Hold:TLAMRec;
  Total,
  TotalError,
  ErrorValue,
  OzValue:double;
  ShortTimeSeries:boolean;
begin
  Reg.ClearAllMeasurements;
  LevelDelta:=strtoint(Parameters.Values['Upper and lower levels to cover']);
  if LevelCount=1 then LowerLevel:=1 else LowerLevel:=LevelCount-LevelDelta;
  if LevelCount=70 then UpperLevel:=70 else UpperLevel:=LevelCount+LevelDelta;
  for LevCount:=LowerLevel to UpperLevel do
    for LatCount:=1 to NumLat do
    begin
      if ((LevCount<=20) and (abs(90-(LatCount-0.5)*LatWidth)<=35)) then ShortTimeSeries:=true
      else ShortTimeSeries:=false;
      for Year:=DataStartYear to DataEndYear do
      if Year<>2002 then
        for Month:=1 to 12 do
        begin
          Hold:=LAMData[LatCount,LevCount,(Year-DataStartYear)*12+Month];
          if Hold.MonthlyMean>0.0 then
          begin
            {The InfoObj objects will be freed by the regression model object.}
            InfoObj:=TLatitudeInfo.Create;
            InfoObj.Latitude:=90-(LatCount-0.5)*LatWidth;
            OzValue:=TransformOzone(LevCount,Hold.MonthlyMean);
            ErrorValue:=Hold.MonthlyMeanError;
            if ErrorValue<0.05*OzValue then ErrorValue:=0.05*OzValue;
            Reg.PassMeasurement(EncodeDate(Year,Month,1)+(DaysInAMonth(Year,Month)/2),OzValue,ErrorValue,InfoObj);
            if Year<(DataStartYear+NumShortYears) then ShortTimeSeries:=false;
          end;
        end;
      if (ShortTimeSeries and (Parameters.Values['Seed short time series']='YES')) then
      begin
        for Month:=1 to 12 do
        begin
          Total:=0.0;
          TotalError:=0.0;
          Num:=0;
          for Year:=DataStartYear to DataEndYear do
          begin
            Hold:=LAMData[LatCount,LevCount,(Year-DataStartYear)*12+Month];
            if Hold.MonthlyMean>0.0 then
            begin
              Total:=Total+Hold.MonthlyMean;
              TotalError:=TotalError+Hold.MonthlyMeanError;
              inc(Num);
            end;
          end;
          if Num>0 then
          for Year:=DataStartYear to DataStartYear+YearsToPad do
          begin
            {The InfoObj objects will be freed by the regression model object.}
            InfoObj:=TLatitudeInfo.Create;
            InfoObj.Latitude:=90-(LatCount-0.5)*LatWidth;
            OzValue:=TransformOzone(LevCount,Total/Num);
            {We add to the regression time series the first available non-null value}
            {and put this in the first year to 'anchor' the time series but also double}
            {the uncertainty on the measuremement so that it doesn't constrain the}
            {regression too much.}
            ErrorValue:=TotalError/Num;
            if ErrorValue<0.05*OzValue then ErrorValue:=0.05*OzValue;
            Reg.PassMeasurement(EncodeDate(Year,Month,1)+(DaysInAMonth(Year,Month)/2),
              OzValue,ErrorValue,InfoObj);
          end;
        end;
      end;
    end; {End of LatCount Loop.}
end; {End of PassDataToRegressionModel.}

procedure BasisFunctionSettings(PrevLevel:integer);
var
  Source:textfile;
  NumFourier,
  NumLegendre,
  I:integer;
  PatternFilename,
  NameStr,
  Linein:string;
  Hold:TBasisFunction;
begin
  for I:=1 to BasisFunctionList.Count do
  try
    Hold:=BasisFunctionList[I-1] as TBasisFunction;
    NameStr:=copy(Hold.ClassName,2,length(Hold.ClassName));
    assignfile(Source,Parameters.Values['BasisFunctionParameterPath']+'\'+NameStr+'.txt');
    reset(Source);
    DeleteHeader(Source);
    repeat readln(Source,Linein) until (strtoint(GetWord(Linein,' ',1))=LevelCount);
    closefile(Source);
    if GetWord(Linein,' ',2)='No' then Hold.Active:=false else
    begin
      Hold.Active:=true;
      Hold.FitOnlyPreloaded:=false;
      NumFourier:=strtoint(GetWord(Linein,' ',4));
      NumLegendre:=strtoint(GetWord(Linein,' ',5));
      Hold.NumFourier:=NumFourier;
      if NumLegendre>0 then Hold.ExtendWith(TLatitudeSphericalHarmonicExtender.Create(NumLegendre));
      if ((GetWord(Linein,' ',3)='Yes') or (GetWord(Linein,' ',3)='NegOnly')) then
      begin
        if PrevLevel=0 then PatternFilename:=Parameters.Values['PatternFilePath']+'\'+NameStr+'.dat' else
          PatternFilename:=OutputDir+'\'+Hold.Name+'Coefficients\'+Hold.Name+'CoefficientsAtLevel_'+
          inttostrf(PrevLevel,2,true)+'.dat';
        Hold.PreloadWith(TPreloadLatAndSeason.Create(PatternFilename,GetWord(Linein,' ',3)));
      end;
      if Hold is TEESCBasisFunctionWithAoA_SpecifyAlt then (Hold as TEESCBasisFunctionWithAoA_SpecifyAlt).SetAlt(LevelCount);
      if Hold is TPinatuboBasisFunctionWithPresOffset then (Hold as TPinatuboBasisFunctionWithPresOffset).SetLevel(LevelCount);
    end;
  except
    on E:Exception do
    begin
      writeln('The problem was with '+NameStr+' basis function.');
      raise;
    end;
  end;
end;

function GetDataFromRegressionModel(FixNegative:boolean):boolean;
var
  TierInc:array of array of TBasisFunctionClass;
  InfoObj:TLatitudeInfo;
  Year,
  Month:word;
  TierCount,
  LatCount,
  I:integer;
  DateTime:TDateTime;
  BasisValue,
  RegressionValue,
  Total:double;
  Hold:TBasisFunction;
begin
  Result:=false;
  setlength(TierInc,4);

  {Tier 1.1 include offset, EESC and trend.}
  setlength(TierInc[0],3);
  TierInc[0,0]:=TOffsetBasisFunction;
  TierInc[0,1]:=TEESCBasisFunctionWithAoA_SpecifyAlt;
  TierInc[0,2]:=TTrendBasisFunction;

  {Tier 1.2 include offset, QBO, ENSO and solar cycle.}
  setlength(TierInc[1],5);
  TierInc[1,0]:=TOffsetBasisFunction;
  TierInc[1,1]:=TQBOBasisFunction;
  TierInc[1,2]:=TQBOOrthogBasisFunction;
  TierInc[1,3]:=TSOIBasisFunction;
  TierInc[1,4]:=TSolarBasisFunction;

  {Tier 1.3 add volcanoes to Tier 1.2.}
  setlength(TierInc[2],7);
  for I:=0 to 4 do TierInc[2,I]:=TierInc[1,I];
  TierInc[2,5]:=TPinatuboBasisFunctionWithPresOffset;
  TierInc[2,6]:=TElChichonBasisFunction;

  {Tier 1.4 add EESC and trend Tier 1.3.}
  setlength(TierInc[3],9);
  for I:=0 to 6 do TierInc[3,I]:=TierInc[2,I];
  TierInc[3,7]:=TEESCBasisFunctionWithAoA_SpecifyAlt;
  TierInc[3,8]:=TTrendBasisFunction;

  for LatCount:=1 to NumLat do
  begin
    InfoObj:=TLatitudeInfo.Create;
    try
      InfoObj.Latitude:=90-(LatCount-0.5)*LatWidth;
      for Year:=FitStartYear to FitEndYear do
        for Month:=1 to 12 do
        begin
          DateTime:=EncodeDate(Year,Month,1)+(DaysInAMonth(Year,Month)/2);
          for TierCount:=0 to 3 do
          begin
            Total:=0.0;
            for I:=Low(TierInc[TierCount]) to High(TierInc[TierCount]) do
            try
              Hold:=Reg.GetBasisFunctionObject(TierInc[TierCount,I]);
              if Hold.Active then
              begin
                BasisValue:=Hold.GetContribution(ResFinal,DateTime,InfoObj);

                {For my analysis of assessing whether ozone is increasing faster or slower}
                {than expected I add the line below.}
                if (Hold is TEESCBasisFunctionWithAoA_SpecifyAlt) then BasisValue:=BasisValue*1.5;

                Total:=Total+BasisValue;
              end;
            except
              on EBasisFunction do; {Do nothing.}
            end;
            RegressionValue:=InverseTransformOzone(LevelCount,Total);
            if RegressionValue<0.0 then
              if FixNegative then RegressionValue:=1E-5 else Result:=true;
            TierLAMData[TierCount+1,LatCount,LevelCount,(Year-FitStartYear)*12+Month]:=RegressionValue;
          end;
        end;
    finally
      FreeAndNil(InfoObj);
    end;
  end;
end;

procedure Write2DFile(DensCount,SetCount:integer);
var
  Dest:textfile;
  TierCount:integer;
  Direc,
  UnitStr:string;
  Year,
  Month,
  LatIndex:word;
begin
  if DensCount=1 then UnitStr:='(10^18 molecules/m^3)' else UnitStr:='(ppm)';
  Direc:=OutputDir+'\AtLevel';
  if not DirectoryExists(Direc) then ForceDirectories(Direc);
  assignfile(Dest,Direc+'\BDBP_MMsFrom'+SetDesc[SetCount]+'_AtLevel_'+inttostrf(LevelCount,2,true)+'.dat');
  rewrite(Dest);
  writeln(Dest,'''This file was created by D:\projects\Ozone\current_proj\ConsensusDatabase\Source\MakeTier1\MakeTier1Databases.bdsproj');
  writeln(Dest,'''Column A) The year.');
  writeln(Dest,'''Column B) The month.');
  writeln(Dest,'''Column C) The decimal year.');
  writeln(Dest,'''Column D) The latitude (degrees).');
  writeln(Dest,'''Column E) The measured monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column F) The Tier 1.1 monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column G) The Tier 1.2 monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column H) The Tier 1.3 monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column I) The Tier 1.4 monthly mean ozone value '+UnitStr+'.');
  for Year:=FitStartYear to FitEndYear do
    for Month:=1 to 12 do
      for LatIndex:=1 to NumLat do
      begin
        write(Dest,Year:4,', ',Month:2,', ',Year+(Month-0.5)/12:8:3,', ',90-(LatIndex-0.5)*LatWidth:5:1,', ');
        if ((Year>=DataStartYear) and (Year<=DataEndYear) and (LAMData[LatIndex,LevelCount,(Year-DataStartYear)*12+Month].MonthlyMean>0.0)) then
          write(Dest,LAMData[LatIndex,LevelCount,(Year-DataStartYear)*12+Month].MonthlyMean:8:6,', ') else write(Dest,'        , ');
        for TierCount:=1 to 4 do write(Dest,TierLAMData[TierCount,LatIndex,LevelCount,(Year-FitStartYear)*12+Month]:8:6,', ');
        writeln(Dest);
      end;
  closefile(Dest);
end;

procedure WriteContributionFile;
var
  Dest:textfile;
  I:integer;
  Year,
  Month,
  LatCount:word;
  Direc:string;
  Value:double;
  InfoObj:TLatitudeInfo;
begin
  Direc:=OutputDir+'\Contributions';
  if not DirectoryExists(Direc) then ForceDirectories(Direc);
  assignfile(Dest,Direc+'\BFContributionAtLevel_'+inttostrf(LevelCount,2,true)+'.dat');
  rewrite(Dest);
  writeln(Dest,'''This file was created by D:\projects\Ozone\current_proj\ConsensusDatabase\Source\MakeTier1\MakeTier1Databases.bdsproj');
  writeln(Dest,'''Column A) Year.');
  writeln(Dest,'''Column B) Month.');
  writeln(Dest,'''Column C) Decimal year.');
  writeln(Dest,'''Column D) Latitude.');
  for I:=0 to BasisFunctionList.Count-1 do
    writeln(Dest,'''Column '+ConvertExcelNumToCol(I+5)+') '+(BasisFunctionList[I] as TBasisFunction).Name);
  for LatCount:=1 to NumLat do
    for Year:=DataStartYear to DataEndYear do
      for Month:=1 to 12 do
      begin
        write(Dest,Year:4,', ',Month:2,', ',Year+(Month-0.5)/12:8:4,', ',90-(LatCount-0.5)*LatWidth:5:1,', ');
        InfoObj:=TLatitudeInfo.Create;
        InfoObj.Latitude:=90-(LatCount-0.5)*LatWidth;
        try
          for I:=0 to BasisFunctionList.Count-1 do
          with BasisFunctionList[I] as TBasisFunction do
          if Active then
          begin
            Value:=GetContribution(ResFinal,EncodeDate(Year,Month,1)+(DaysInAMonth(Year,Month)/2),InfoObj);
            write(Dest,InverseTransformOzone(LevelCount,Value):6:3,', ');
          end
          else write(Dest,'      , ');
        finally
          FreeAndNil(InfoObj);
        end;
        writeln(Dest);
      end;
  closefile(Dest);
end;

procedure WriteResolvedBasisFunctionCoeff(BF:TBasisFunction);
//{$DEFINE DEBUG1}
var
  Dest:textfile;
  CoeffValueBeforePattern,
  CoeffValueAfterPattern:T2DDoubleArray;
  ErrorVector:TDoubleArray;
  I,J:integer;
  InfoObj:TLatitudeInfo;
  SumSquare:double;
  NewDirec:string;
  {$IFDEF DEBUG1}DebugDest:textfile;{$ENDIF}
begin
  NewDirec:=OutputDir+'\'+BF.Name+'Coefficients';
  if not DirectoryExists(NewDirec) then ForceDirectories(NewDirec);
  assignfile(Dest,NewDirec+'\'+BF.Name+'CoefficientsAtLevel_'+inttostrf(LevelCount,2,true)+'.dat');
  rewrite(Dest);
  writeln(Dest,'''This file was created by B:\Contracts\Con00004\Source\MakeTier1\MakeTier1Databases.bdsproj');
  writeln(Dest,'''Column A) The day of the year.');
  writeln(Dest,'''Column B) The latitude.');
  writeln(Dest,'''Column C) The '+BF.Name+' coefficient.');
  writeln(Dest,'''Column D) The error on the '+BF.Name+' coefficient.');
  writeln(Dest,'''Column E) The T value i.e. abs(C)/D; value greater than 2 means significant at 95%.');

  {Note: calling GetResolvedCoefficientArray returns the fit coefficient but excluding any}
  {contribution from the preloaded pattern. That must be done additionally by the calling }
  {program.}
  CoeffValueBeforePattern:=Reg.GetResolvedCoefficientArray(BF,ResFinal);
  CoeffValueAfterPattern:=T2DDoubleArray.Create(CoeffValueBeforePattern);
  try
    {Here is where we add the effects of the preloaded pattern.}
    for I:=1 to CoeffValueBeforePattern.GetNumY do
    begin
      InfoObj:=TLatitudeInfo.Create;
      InfoObj.Latitude:=CoeffValueBeforePattern[2,I];
      if BF.HasPreload then CoeffValueAfterPattern[3,I]:=CoeffValueBeforePattern[3,I]+BF.GetRegCoeffForPreload(ResFinal).Coeff*
        BF.GetPreloadValue(EncodeDate(2000,1,1)+CoeffValueBeforePattern[1,I],InfoObj)
      else CoeffValueAfterPattern[3,I]:=CoeffValueBeforePattern[3,I];
      FreeAndNil(InfoObj);
    end;

    {$IFDEF DEBUG1}
    NewDirec:=OutputDir+'\Debug';
    if not DirectoryExists(NewDirec) then ForceDirectories(NewDirec);
    assignfile(DebugDest,NewDirec+'\'+BF.Name+'CoefficientsAtLevel_'+inttostrf(LevelCount,2,true)+'.dat');
    rewrite(DebugDest);
    writeln(DebugDest,'''Column A) The day of the year.');
    writeln(DebugDest,'''Column B) The latitude.');
    writeln(DebugDest,'''Column C) The '+BF.Name+' coefficient before the preloaded pattern contribution has been added.');
    writeln(DebugDest,'''Column D) The '+BF.Name+' coefficient after the preloaded pattern contribution has been added.');
    writeln(DebugDest,'''Column E) The contribution of the preloaded pattern.');
    for I:=1 to CoeffValueBeforePattern.GetNumY do
      writeln(DebugDest,CoeffValueBeforePattern[1,I]:8:4,' ',CoeffValueBeforePattern[2,I]:8:4,' ',
      CoeffValueBeforePattern[3,I]:8:4,' ',CoeffValueAfterPattern[3,I]:8:4,' ',
      CoeffValueAfterPattern[3,I]-CoeffValueBeforePattern[3,I]:8:4);
    closefile(DebugDest);
    {$ENDIF DEBUG1}

    {This piece of code below does include the contribution of the error on the}
    {fit coefficient for any preloaded pattern. The I loop is through days and }
    {latitudes i.e. 36 steps through the year and 36 steps in latitude.}
    for I:=1 to CoeffValueAfterPattern.GetNumY do
    begin
      InfoObj:=TLatitudeInfo.Create;
      InfoObj.Latitude:=CoeffValueAfterPattern[2,I];
      ErrorVector:=BF.GetCoefficientErrorVector(EncodeDate(2000,1,1)+CoeffValueAfterPattern[1,I],ResFinal,InfoObj);
      SumSquare:=0.0;
      for J:=Low(ErrorVector) to High(ErrorVector) do SumSquare:=SumSquare+ErrorVector[J];
      begin
        write(Dest,CoeffValueAfterPattern[1,I]:8:4,' ',CoeffValueAfterPattern[2,I]:8:4,' ',
        CoeffValueAfterPattern[3,I],' ',sqrt(SumSquare),' ');
        if SumSquare=0.0 then writeln(Dest,0.0:8:4) else writeln(Dest,abs(CoeffValueAfterPattern[3,I])/sqrt(SumSquare):8:4);
      end;
      FreeAndNil(InfoObj);
    end;
    closefile(Dest);
  finally
    FreeAndNil(CoeffValueBeforePattern);
    FreeAndNil(CoeffValueAfterPattern);
  end;
end;

procedure WriteResolvedBasisFunctions;
var
  BasisFunction:TBasisFunction;
  BFCount:word;
begin
  for BFCount:=1 to Reg.NumBasisFunctions do
  begin
    BasisFunction:=Reg.GetBasisFunctionObject(BFCount);
    if BasisFunction.Active then WriteResolvedBasisFunctionCoeff(BasisFunction);
  end;
end;

procedure Write1DFile(DensCount,SetCount,LatIndex:integer);
var
  Dest:textfile;
  TierCount:integer;
  Direc,
  UnitStr:string;
  Year,
  Month:word;
begin
  if DensCount=1 then UnitStr:='(10^18 molecules/m^3)' else UnitStr:='(ppm)';
  Direc:=OutputDir+'\AtLatAndLevel';
  if not DirectoryExists(Direc) then ForceDirectories(Direc);
  assignfile(Dest,Direc+'\BDBPMMsFrom'+SetDesc[SetCount]+'_AtLatAndLevel_'+
    inttostrf(LatIndex,2,true)+'_'+inttostrf(LevelCount,2,true)+'.dat');
  rewrite(Dest);
  writeln(Dest,'''This file was created by D:\projects\Ozone\current_proj\ConsensusDatabase\Source\MakeTier1\MakeTier1Databases.bdsproj');
  writeln(Dest,'''Column A) The year.');
  writeln(Dest,'''Column B) The month.');
  writeln(Dest,'''Column C) The decimal year.');
  writeln(Dest,'''Column D) The measured monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column E) The Tier 1.1 monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column F) The Tier 1.2 monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column G) The Tier 1.3 monthly mean ozone value '+UnitStr+'.');
  writeln(Dest,'''Column H) The Tier 1.4 monthly mean ozone value '+UnitStr+'.');
  for Year:=FitStartYear to FitEndYear do
    for Month:=1 to 12 do
    begin
      write(Dest,Year:4,', ',Month:2,', ',Year+(Month-0.5)/12:8:3,', ');
      if ((Year>=DataStartYear) and (Year<=DataEndYear) and (LAMData[LatIndex,LevelCount,(Year-DataStartYear)*12+Month].MonthlyMean>0.0)) then
        write(Dest,LAMData[LatIndex,LevelCount,(Year-DataStartYear)*12+Month].MonthlyMean:8:6,', ') else write(Dest,'        , ');
      for TierCount:=1 to 4 do write(Dest,TierLAMData[TierCount,LatIndex,LevelCount,(Year-FitStartYear)*12+Month]:8:6,', ');
      writeln(Dest);
    end;
  closefile(Dest);
end;

procedure WriteFileForBob(DensCount,SetCount,LatIndex:integer);
var
  Dest:textfile;
  TierCount:integer;
  Direc,
  UnitStr:string;
  Year,
  Month,
  Level:word;
begin
  if DensCount=1 then UnitStr:='(10^18 molecules/m^3)' else UnitStr:='(ppm)';
  Direc:=OutputDir+'\ForBob';
  if not DirectoryExists(Direc) then ForceDirectories(Direc);
  assignfile(Dest,Direc+'\BDBPMMsFrom'+SetDesc[SetCount]+'_'+GetFilenameSubStr(LatIndex)+'.txt');
  rewrite(Dest);
  writeln(Dest,'''This file was created by D:\projects\Ozone\current_proj\ConsensusDatabase\Source\MakeTier1\MakeTier1Databases.bdsproj');
  writeln(Dest,'''Column A) The year.');
  writeln(Dest,'''Column B) The month.');
  writeln(Dest,'''Column C) The decimal year.');
  if SetCount=1 then
  begin
    writeln(Dest,'''Column D) Altitude (km).');
    writeln(Dest,'''Column E) The measured monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column F) The Tier 1.1 monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column G) The Tier 1.2 monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column H) The Tier 1.3 monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column I) The Tier 1.4 monthly mean ozone value '+UnitStr+'.');
  end;
  if SetCount=2 then
  begin
    writeln(Dest,'''Column D) Approximate altitude (km).');
    writeln(Dest,'''Column E) Pressure (hPa).');
    writeln(Dest,'''Column F) The measured monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column G) The Tier 1.1 monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column H) The Tier 1.2 monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column I) The Tier 1.3 monthly mean ozone value '+UnitStr+'.');
    writeln(Dest,'''Column J) The Tier 1.4 monthly mean ozone value '+UnitStr+'.');
  end;
  for Year:=FitStartYear to FitEndYear do
    for Month:=1 to 12 do
      for Level:=1 to NumLevels do
      begin
        write(Dest,Year:4,', ',Month:2,', ',Year+(Month-0.5)/12:8:3,', ',Level:2,', ');
        if SetCount=2 then write(Dest,1013.25*exp(-Level/7):8:4,', ');
        if ((Year>=DataStartYear) and (Year<=DataEndYear) and (LAMData[LatIndex,Level,(Year-DataStartYear)*12+Month].MonthlyMean>0.0)) then
          write(Dest,LAMData[LatIndex,Level,(Year-DataStartYear)*12+Month].MonthlyMean:8:6,', ') else write(Dest,'        , ');
        for TierCount:=1 to 4 do write(Dest,TierLAMData[TierCount,LatIndex,Level,(Year-FitStartYear)*12+Month]:8:6,', ');
        writeln(Dest);
      end;
  closefile(Dest);
end;

function ReduceBasisFunctionParameters(BasisFunc:TBasisFunctionClass;AllowDeactivate:boolean=true):boolean;
var
  Hold:TBasisFunction;
  Extender:TLatitudeSphericalHarmonicExtender;
begin
  Result:=false;
  Hold:=Reg.GetBasisFunctionObject(BasisFunc);
  if not Hold.Active then exit;
  Extender:=Hold.Extender as TLatitudeSphericalHarmonicExtender;
  if Extender.Degree>1 then {Reduce the Legendres.}
  begin
    writeln('Reducing Legendre expansions for '+Hold.Name+' basis function from '+
      inttostr(Extender.Degree)+' to '+inttostr(Extender.Degree-1));
    Hold.ExtendWith(TLatitudeSphericalHarmonicExtender.Create(Extender.Degree-1));
    Result:=true;
  end
  else {Start reducing the Fouriers.}
  begin
    if Hold.NumFourier>0 then
    begin
      writeln('Reducing Fourier expansions for '+Hold.Name+' basis function from '+
        inttostr(Hold.NumFourier)+' to '+inttostr(Hold.NumFourier-1));
      Hold.NumFourier:=Hold.NumFourier-1;
      Result:=true;
    end
    else
    begin
      if not Hold.FitOnlyPreloaded then {Fit only the preloaded pattern.}
      begin
        writeln('Fitting only the preloaded pattern for the '+Hold.Name+' basis function');
        Hold.FitOnlyPreloaded:=true;
        Result:=true;
      end
      else {De-activate the basis function.}
      if AllowDeactivate then
      begin
        writeln('De-activating the '+Hold.Name+' basis function');
        WriteResolvedBasisFunctionCoeff(Hold);
        Hold.Active:=false;
        Result:=true;
      end;
    end;
  end;
end;

function ReduceExpansions:boolean;
const
  NumBFToReduce=9;
  BasisFunctionToReduce:array[1..NumBFToReduce] of TBasisFunctionClass=(TTrendBasisFunction,
    TSOIBasisFunction,TElChichonBasisFunction,TPinatuboBasisFunctionWithPresOffset,
    TEESCBasisFunctionWithAoA_SpecifyAlt,TQBOOrthogBasisFunction,TQBOBasisFunction,TSolarBasisFunction,TOffsetBasisFunction);
var
  I:integer;
begin
  I:=1;
  repeat
    Result:=ReduceBasisFunctionParameters(BasisFunctionToReduce[I],BasisFunctionToReduce[I]<>TOffsetBasisFunction);
    inc(I);
  until (Result or (I>NumBFToReduce));
end;

function OffsetBasisFunctionCoeffSomewhereZero:boolean;
var
  I:integer;
  InfoObj:TLatitudeInfo;
  Hold:TBasisFunction;
  ValueAfterPatternAdd:double;
  CoeffValueBeforePattern:T2DDoubleArray;
begin
  Result:=false;
  Hold:=Reg.GetBasisFunctionObject(TOffsetBasisFunction);
  if not Hold.Active then exit;
  CoeffValueBeforePattern:=Reg.GetResolvedCoefficientArray(Hold,ResFinal);
  try
    {Here is where we add the effects of the preloaded pattern.}
    for I:=1 to CoeffValueBeforePattern.GetNumY do
    begin
      InfoObj:=TLatitudeInfo.Create;
      InfoObj.Latitude:=CoeffValueBeforePattern[2,I];
      ValueAfterPatternAdd:=CoeffValueBeforePattern[3,I]+
        Hold.GetRegCoeffForPreload(ResFinal).Coeff*Hold.GetPreloadValue(EncodeDate(2000,1,1)+
        CoeffValueBeforePattern[1,I],InfoObj);
      if ValueAfterPatternAdd<0.0 then Result:=true;
      FreeAndNil(InfoObj);
    end;
  finally
    FreeAndNil(CoeffValueBeforePattern);
  end;
end;

function EESCBasisFunctionCoeffSomewherePositive:boolean;
var
  I:integer;
  InfoObj:TLatitudeInfo;
  Hold:TBasisFunction;
  ValueAfterPatternAdd:double;
  CoeffValueBeforePattern:T2DDoubleArray;
begin
  Result:=false;
  Hold:=Reg.GetBasisFunctionObject(TEESCBasisFunctionWithAoA_SpecifyAlt);
  if not Hold.Active then exit;
  CoeffValueBeforePattern:=Reg.GetResolvedCoefficientArray(Hold,ResFinal);
  try
    {Here is where we add the effects of the preloaded pattern.}
    for I:=1 to CoeffValueBeforePattern.GetNumY do
    begin
      InfoObj:=TLatitudeInfo.Create;
      InfoObj.Latitude:=CoeffValueBeforePattern[2,I];
      if Hold.HasPreload then ValueAfterPatternAdd:=CoeffValueBeforePattern[3,I]+
        Hold.GetRegCoeffForPreload(ResFinal).Coeff*Hold.GetPreloadValue(EncodeDate(2000,1,1)+
        CoeffValueBeforePattern[1,I],InfoObj)
      else ValueAfterPatternAdd:=CoeffValueBeforePattern[3,I];
      if ValueAfterPatternAdd>0.0 then Result:=true;
      FreeAndNil(InfoObj);
    end;
  finally
    FreeAndNil(CoeffValueBeforePattern);
  end;
end;

procedure RunRegression;
var
  WorkStr:string;
  PercentToRemove:double;
  I:integer;
begin
  Reg.Run;

  {Remove outliers if this option has been selected.}
  WorkStr:=Parameters.Values['Remove data after run'];
  if GetWord(WorkStr,' ',1)='YES' then
  begin
    PercentToRemove:=strtofloat(EraseEndChar(GetWord(WorkStr,' ',2),'%'));
    for I:=1 to strtoint(GetWord(WorkStr,' ',3)) do
    begin
      Reg.RemoveXPercentOfDataByResiduals(PercentToRemove);
      Reg.Run;
    end;
  end;
end;

procedure RunAtLevel(var PrevLevel:integer);
var
  NegativeOzone,
  FixNegative:boolean;
  Hold:TBasisFunction;
begin
  FixNegative:=false;
  writeln;
  writeln('Doing analysis for DensCount=',DensCount,' SetCount=',SetCount,' Level=',LevelCount);
  PassDataToRegressionModel;
  BasisFunctionSettings(PrevLevel);
  writeln('The number of coefficients is ',Reg.GetNumCoefficients);
  RunRegression;

  {First make sure that the Offset basis function doesn't produce negative coefficients.}
  if Parameters.Values['Require resolved offset basis function coefficient to be positive']='YES' then
  while OffsetBasisFunctionCoeffSomewhereZero do
  begin
    ReduceBasisFunctionParameters(TOffsetBasisFunction,false);
    Reg.Run;
  end;

  if Parameters.Values['Require resolved EESC basis function coefficient to be negative']='YES' then
  while EESCBasisFunctionCoeffSomewherePositive do
  begin
    ReduceBasisFunctionParameters(TEESCBasisFunctionWithAoA_SpecifyAlt,false);
    Reg.Run;
  end;

  {Now make sure we're not getting negative ozone.}
  NegativeOzone:=GetDataFromRegressionModel(FixNegative);
  if (NegativeOzone and (Parameters.Values['Reduce expansions when regression ozone below 0.0']='YES')) then
  repeat
    FixNegative:=not ReduceExpansions;
    writeln('The number of coefficients is ',Reg.GetNumCoefficients);
    Reg.Run;
    NegativeOzone:=GetDataFromRegressionModel(FixNegative);
  until not NegativeOzone;
  if Parameters.Values['Write Contribution Files']='YES' then WriteContributionFile;
  if Parameters.Values['Write Resolved Basis Function Coefficients']='YES' then WriteResolvedBasisFunctions;
  PrevLevel:=LevelCount;
end;

procedure GenerateDataSet(DensCount,SetCount:integer);
const
  LowerLevel=1;
  UpperLevel=70;
var
  SeedLevel,
  PrevLevel,
  LatCount:integer;
  InputDir:string;
begin
  InputDir:=Parameters.Values['InputDirectory'];
  if pos('+VersionStr+',InputDir)>0 then InputDir:=ReplaceChars(InputDir,'+VersionStr+',Parameters.Values['VersionStr']);
  OutputDir:=Parameters.Values['OutputDirectory']+'\'+DensDesc[DensCount]+'\'+SetDesc[SetCount];
  if pos('+VersionStr+',OutputDir)>0 then OutputDir:=ReplaceChars(OutputDir,'+VersionStr+',Parameters.Values['VersionStr']);
  LoadMonthlyMeans(InputDir+'\'+DensDesc[DensCount]+'\'+SetDesc[SetCount]+'\Bin');
  SeedLevel:=strtoint(Parameters.Values['Start at Level']);

  {First we start at Seed Level and work our way up.}
  PrevLevel:=0;
  for LevelCount:=SeedLevel to UpperLevel do RunAtLevel(PrevLevel);

  {Then we go to one level below Seed level and work our way down.}
  PrevLevel:=SeedLevel;
  for LevelCount:=SeedLevel-1 downto LowerLevel do RunAtLevel(PrevLevel);

  {Write the results to files.}
  if Parameters.Values['Write 1D Files']='YES' then
    for LatCount:=1 to NumLat do
      for LevelCount:=LowerLevel to UpperLevel do Write1DFile(DensCount,SetCount,LatCount);
  if Parameters.Values['Write2DFiles']='YES' then for LevelCount:=LowerLevel to UpperLevel do Write2DFile(DensCount,SetCount);
  if Parameters.Values['WriteFilesForBob']='YES' then for LatCount:=1 to NumLat do WriteFileForBob(DensCount,SetCount,LatCount);
end;

begin
  LoadParameters;
  CreateBasisFunctions;
  WriteParameters;
  Reg:=CreateRegressionModel;
  try
    try
      for DensCount:=1 to 2 do
        for SetCount:=1 to 2 do GenerateDataSet(DensCount,SetCount);
    finally
      Reg.Free;
      Parameters.Free;
    end;
  except
    on E:Exception do
    begin
      writeln;
      writeln('The error message was: '+E.Message);
      ShowMessage('An error occurred somewhere in the program. The error messge is: '+E.Message);
    end;
  end;
end.


