//@***h* BDBP/gb_4Dsynoptic.pas
// 
// NAME
//   gb_4Dsynoptic.pas
//
// PURPOSE
//  Defines a class which manages synoptic data fields, typically geophysical data 
//  as a function of latitude and longitude, on one or more levels and at one or 
//  more times. The levels and times need not be evenly spaced.
//
// DESCRIPTION
//  This unit defines a class which manages synoptic data fields, typically 
//  geophysical data as a function of latitude and longitude, on one or more levels 
//  and at one or more times. The levels and times need not be evenly spaced. The 
//  object does not need a destructor since the ancestors destructor calls the Clear 
//  method which is all that is required to clean up the resourecs used.  
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_4Dsynoptic;

{$I UNITDIREC.INC}

interface

uses
  {The are delphi units}
  Classes,
  SysUtils,

  {These are specific units}
  gb_array,
  gb_DelphiNetCDF,
  gb_SynopticGlobal,
  gb_GlobalSynopticStack,
  gb_merctoms,
  gb_ncepncar;

type
  T4DSynoptic=class(TList)
  strict private
    {FDataType specifies whether newly created GlobalSynopticStack objects are }
    {created compressed or uncompressed. Since, once created, the compression  }
    {status of GlobalSynopticStack cannot be changed, changing the FDataType   }
    {field in this object doesn't change the compression status of existing    }
    {GlobalSynopticStack objects. It only affects newly created stack objects. }
    {The field is accessed through the DataType property.                      }
    FDataType:TStackType;

    {The NewSynopticGlobalDefaults record specifies the default values to be   }
    {used for any new SynopticGlobals that are created.                        }
    NewSynopticGlobalDefaults:TNewSynopticGlobalDefaults;

    {FMaxMemory specifies the maximum amount of memory to be used by the       }
    {object. The field is accessed through the MaxMemory property.             }
    FMaxMemory:cardinal;

    {TempFilename is the name of a temporary storage file used by the object.  }
    TempFilename:string;

    {FInterpolationFailPlan specifies how the object should behave when the    }
    {interpolation of a value fails. The type is defined in gb_synopticglobal  }
    {and the two values currently allowed are raiseexception and returnnull.   }
    {The field is accessed through the InterpolationFailPlan property.         }
    FInterpolationFailPlan:TInterpolationFailPlan;

    {FExtendToPole specifies whether data at the most poleward latitude should }
    {be used to estimate values poleward of that latitude in the case of       }
    {spatial interpolation. The field is accessed through the ExtendToPole     }
    {property and can be changed through the life of the object. The setting   }
    {is propagated through to all constituent SynopticGlobal objects.          }
    FExtendToPole:boolean;

    {FConfineToLevels specifies whether data at the uppermost and lowermost    }
    {levels should be used to estimate values above or below those levels in   }
    {the case of spatial interpolation. If FConfineToLevels is true, then the  }
    {interpolation fails if the level at which the interpolation is requested  }
    {is not spanned by the available levels. The field is accessed through the }
    {ConfineToLevels property and can be changed through the life of the       }
    {object. The setting is propagated to all constituent GlobalSynopticGlobal }
    {objects.                                                                  }
    FConfineToLevels:boolean;

    {This value is true if the memory is restricted. It is accessed via the    }
    {memory restricted property.                                               }
    FMemRestriction:boolean;

    {This value is true if all of the SynopticGlobals in the object have the   }
    {same spatial structure.                                                   }
    FCommonStructure:boolean;

    {The Int array holds the indices of the two GlobalSynopticStack objects    }
    {most recently accessed in a data interpolation call.                      }
    Int:array[1..2] of longint;

    {FSynopticGlobalBuffer holds a reference to the last accessed TSynopticGlobal}
    {field. It is used to speed access.}
    FSynopticGlobalBuffer:TSynopticGlobal;

    {This field holds the DelphiNetCDF object if the data are being read from or}
    {written to a NetCDF file.}
    NetCDF:TDelphiNetCDF;

    {These two fields hold the functions to be used to convert date and time values}
    {into TDateTime values when reading in the data from NetCDF files.}
    FFunctionToConvertStrToDateTime:TConvertStrToDateTime;
    FFunctionToConvertValueToDateTime:TConvertValueToDateTime;

    {This field holds the function that may be used to transform the levels array}
    {e.g. if a user might want to log transform pressures.}
    FFunctionToTransformLevels:TTransformLevel;

    {In the case of reading from a NetCDF file, this field specifies the string}
    {that will identify the time variable. It is set to 'time' as a default in }
    {the constructor.}
    FNetCDFTimeIDStr:string;

    {In the case of reading from a NetCDF file, this field specifies the string}
    {that will identify the data level within the NetCDF file.}
    FNetCDFLevelIDStr:string;

    {In the case of reading from a NetCDF file, this field specifies the string}
    {that will identify which variable should be extracted into the object.}
    FNetCDFVarStr:string;

    {In the case of reading from a NetCDF file, this field specifies the subset}
    {of vertical levels that should be targetted. By default, this field is    }
    {nil. If it does contain entries, only those levels in teh NetCDF file     }
    {matching entries on this list will be added to the object.}
    FNetCDFTargetLevels:TDoubleArray;

    {These are the private methods listed in alphabetical order.}
    procedure AddNewStack(Stack:TGlobalSynopticStack);
    function  AllGridsSameStructure:boolean;
    procedure ConfirmLatLongAttributesSet;
    procedure ConfirmOffsetAndScalingSet;
    function  ConfirmSynopticGlobalBuffer(ADateTime:TDateTime;ALevel:double):boolean;
    function  FindTemporaryStorageFilename:string;
    function  GetDateTimeForInterval(INum:integer):TDateTime;
    procedure GetIntervalAndLevel(ADateTime:TDateTime;ALevel:double;AllowAddition:boolean;out INum,LNum:cardinal);
    function  GetIntervalForDateTime(ADateTime:TDateTime):cardinal;
    function  GetIntervalOnOrBeforeDateTime(ADateTime:TDateTime):cardinal;
    function  GetLatitude(I:word):double;
    function  GetLevelNumForLevel(INum:cardinal;ALevel:double):cardinal;
    function  GetLongitude(I:word):double;
    function  GetStackNumForDateTime(ADateTime:TDateTime):longint;
    function  GetFirstStackNumOnDate(ADateTime:TDateTime):longint;
    function  GetStackObject(INum:integer):TGlobalSynopticStack;
    procedure ReplaceSynopticGlobalByIndices(INum,LNum:cardinal;Value:TSynopticGlobal);
    procedure RestrictMemoryTo(MaxMem:cardinal);
    procedure RestrictMemory(Value:boolean);
    procedure WriteExtendToPole(Value:boolean);
    procedure WriteConfineToLevels(Value:boolean);
    procedure SetNetCDFLevelIDStr(Value:string);
    procedure SetNetCDFTimeIDStr(Value:string);
    procedure SetNetCDFVarStr(Value:string);
  protected
    procedure Notify(Ptr:Pointer;Action:TListNotification);override;
  public
    constructor Create;

    {These are the sorts of methods you would call when creating a new object.}
    procedure SetDefaultLatLongAttr(ANumLong,ANumLat:longint;AStartLong,AStartLat,ADeltaLong,ADeltaLat:double);
    procedure SetDefaultScalingAndOffset(AScaling,AOffset:double);
    procedure SetDefaultNonUniformLatitudes(Latitudes:TDoubleArray);
    function  AddInterval(ADateTime:TDateTime;StartLevel,DeltaLevel:double;NumLevels:word):integer;overload;
    function  AddInterval(ADateTime:TDateTime;Level:double):integer;overload;
    function  AddInterval(ADateTime:TDateTime;LevelList:TDoubleArray):integer;overload;
    procedure RemoveFirstInterval;
    procedure RemoveLastInterval;
    procedure AddSynopticGlobal(ADateTime:TDateTime;Value:TSynopticGlobal);
    procedure InsertTomsData(ADateTime:TDateTime;ALevel:double;TomsData:TMercToms);
    procedure InsertNCEPNCARData(ADateTime:TDateTime;ALevel:double;NCEPNCARData:TNCEPGrid);
    procedure PutValueByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double;Value:single);
    procedure PutValueByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal;Value:single);
    procedure SaveToFile(Filename:string);
    procedure SaveToNetCDFFile(OptionsFile:string);

    {These are the sorts of methods you would call when reading from an existing object.}
    procedure ReadFromFile(Filename:string);
    procedure CombineWithFile(Filename:string);
    procedure ReadFromNetCDFFile(Filename:string);
    function  GetDateTimeList:TDoubleArray;
    function  QueryDataAvailableAtDateTime(ADateTime:TDateTime):boolean;
    function  QueryDataAvailableOnDay(ADateTime:TDateTime):boolean;
    function  DateTimeSpanned(ADateTime:TDateTime):boolean;
    function  GetLevelsForDateTime(ADateTime:TDateTime):TDoubleArray;
    function  GetArrayOfAllLevels:TDoubleArray;
    function  GetSynopticGlobal(ADateTime:TDateTime;ALevel:double;AllowAddition:boolean;Owner:boolean=false):TSynopticGlobal;
    function  GetSynopticGlobalOnDay(ADateTime:TDateTime;ALevel:double;AllowAddition:boolean;Owner:boolean=false):TSynopticGlobal;
    function  GetSynopticGlobalOnClosestDay(ADateTime:TDateTime;ALevel:double;AllowAddition:boolean;Owner:boolean=false):TSynopticGlobal;
    function  GetMeanSynopticGlobalOnDay(ADateTime:TDateTime;ALevel:double):TSynopticGlobal;
    function  GetMeanSynopticGlobalForMonth(Year,Month:word;ALevel:double):TSynopticGlobal;
    function  GetFirstDateTime:TDateTime;
    function  GetLastDateTime:TDateTime;
    function  GetNumIntervals:cardinal;
    function  GetNumLongitudes:cardinal;
    function  GetNumLatitudes:cardinal;
    function  GetStartLat:double;
    function  GetStartLong:double;
    function  GetEndLat:double;
    function  GetEndLong:double;
    function  GetDeltaLat:double;
    function  GetDeltaLong:double;
    function  GetValueByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double;Interpolate:boolean=false):single;
    function  GetValueByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal):single;
    function  InterpolateSynopticGlobal(ADateTime:TDateTime;ALevel:double;Owner:boolean=false):TSynopticGlobal;
    function  IsNullByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double):boolean;
    function  IsNullByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal):boolean;
    function  IsNotNullByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double):boolean;
    function  IsNotNullByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal):boolean;

    {These are general methods.}
    procedure Clear;override;
    function  GetCellArea(Lat:cardinal):single;
    function  GetMemoryUsed:cardinal;
    procedure ReplaceSynopticGlobal(ADateTime:TDateTime;ALevel:double;Value:TSynopticGlobal);overload;
    procedure SortByInterval;

    {Properties}
    property DataType:TStackType read FDataType write FDataType;
    property SynopticGlobalBuffer:TSynopticGlobal read FSynopticGlobalBuffer write FSynopticGlobalBuffer;
    property MaxMemory:cardinal read FMaxMemory write RestrictMemoryTo;
    property MemoryRestricted:boolean read FMemRestriction write RestrictMemory;
    property InterpolationFailPlan:TInterpolationFailPlan read FInterpolationFailPlan write FInterpolationFailPlan;
    property ExtendToPole:boolean read FExtendToPole write WriteExtendToPole;
    property ConfineToLevels:boolean read FConfineToLevels write WriteConfineToLevels;
    property Latitude[I:word]:double read GetLatitude;
    property Longitude[I:word]:double read GetLongitude;
    property NetCDFLevelIDStr:string write SetNetCDFLevelIDStr;
    property NetCDFTimeIDStr:string write SetNetCDFTimeIDStr;
    property NetCDFVarStr:string write SetNetCDFVarStr;
    property NetCDFTargetLevels:TDoubleArray write FNetCDFTargetLevels;
    property NetCDFFunctionToConvertStrToDateTime:TConvertStrToDateTime write FFunctionToConvertStrToDateTime;
    property NetCDFFunctionToConvertValueToDateTime:TConvertValueToDateTime write FFunctionToConvertValueToDateTime;
    property NetCDFFunctionToTransformLevels:TTransformLevel write FFunctionToTransformLevels;
  end;

  E4DSynoptic=class(Exception);

implementation

uses
  DateUtils,
  gb_datetime,
  gb_math,
  gb_strings;

constructor T4DSynoptic.Create;
{------------------------------------------------------------------------------}
{ This is the sole constructor for the object. It sets default values and      }
{ calls the constructor of its ancestor.                                       }
{------------------------------------------------------------------------------}
begin
  FDataType:=uncompressed;
  with NewSynopticGlobalDefaults do
  begin
    with Header do
    begin
      NumLong:=0;
      NumLat:=0;
      DeltaLong:=-999.99;
      DeltaLat:=-999.99;
      StartLong:=-999.99;
      StartLat:=-999.99;
    end;
    Scaling:=-999.99;
    Offset:=-999.99;
  end;
  FMaxMemory:=0;
  TempFilename:='';
  FInterpolationFailPlan:=raiseexception;
  FExtendToPole:=false;
  FConfineToLevels:=true;
  FMemRestriction:=false;
  Int[1]:=1;
  Int[2]:=2;
  FSynopticGlobalBuffer:=nil;
  NetCDF:=nil;
  FFunctionToConvertStrToDateTime:=nil;
  FFunctionToConvertValueToDateTime:=nil;
  FFunctionToTransformLevels:=nil;
  FNetCDFLevelIDStr:='lev';
  FNetCDFTimeIDStr:='time';
  FNetCDFTargetLevels:=nil;
  inherited Create;
end; {End of T4DSynoptic.Create.}

procedure T4DSynoptic.SetDefaultLatLongAttr(ANumLong,ANumLat:longint;AStartLong,AStartLat,ADeltaLong,ADeltaLat:double);
{------------------------------------------------------------------------------}
{ This procedure is called to set the geographical coverage and grid spacing   }
{ for new SynopticGlobal objects that are created.                             }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  with NewSynopticGlobalDefaults.Header do
  begin
    NumLong:=ANumLong;
    NumLat:=ANumLat;
    DeltaLong:=ADeltaLong;
    DeltaLat:=ADeltaLat;
    StartLong:=AStartLong;
    StartLat:=AStartLat;
  end;
  for I:=1 to Count do GetStackObject(I).SetNewSynopticGlobalDefaults(NewSynopticGlobalDefaults);
end; {End of T4DSynoptic.SetDefaultLatLongAttr.}

procedure T4DSynoptic.SetDefaultScalingAndOffset(AScaling,AOffset:double);
{------------------------------------------------------------------------------}
{ This procedure is called to set the default scaling and offset values for    }
{ new SynopticGlobal objects that are created. These only need to be set if    }
{ the DataType is compressed.                                                  }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  NewSynopticGlobalDefaults.Scaling:=AScaling;
  NewSynopticGlobalDefaults.Offset:=AOffset;
  for I:=1 to Count do GetStackObject(I).SetNewSynopticGlobalDefaults(NewSynopticGlobalDefaults);
end; {End of T4DSynoptic.SetDefaultScalingAndOffset.}

procedure T4DSynoptic.SetDefaultNonUniformLatitudes(Latitudes:TDoubleArray);
{------------------------------------------------------------------------------}
{ This procedure is called to set non-uniformly spaced latitudes for new       }
{ SynopticGlobal objects that are created.                                     }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  with NewSynopticGlobalDefaults do
  begin
    HeapSort(Latitudes);
    setlength(FLatitudes,length(Latitudes));
    FLatitudes:=Copy(Latitudes);
    Header.StartLat:=Latitudes[0];
    Header.NumLat:=length(Latitudes);
    Header.DeltaLat:=0.0;
  end;
  for I:=1 to Count do GetStackObject(I).SetNewSynopticGlobalDefaults(NewSynopticGlobalDefaults);
end; {End of T4DSynoptic.SetDefaultNonUniformLatitudes.}

function T4DSynoptic.AddInterval(ADateTime:TDateTime;StartLevel,DeltaLevel:double;NumLevels:word):integer;
{------------------------------------------------------------------------------}
{ This function adds 'NumLevels' levels to the object at the date and time     }
{ specified by 'ADateTime'. The start level and inter-level spacing control    }
{ which levels are added. The function returns the interval number (1 to Count)}
{ of the newly added interval.                                                 }
{------------------------------------------------------------------------------}
var
  I:integer;
  LevelList:TDoubleArray;
begin
  setlength(LevelList,NumLevels);
  for I:=0 to NumLevels-1 do LevelList[I]:=StartLevel+I*DeltaLevel;
  Result:=AddInterval(ADateTime,LevelList);
end; {End of T4DSynoptic.AddInterval.}

function T4DSynoptic.AddInterval(ADateTime:TDateTime;Level:double):integer;
{------------------------------------------------------------------------------}
{ This function adds a single level to the object at the date and time         }
{ specified by 'ADateTime'. The function returns the interval number (1 to     }
{ Count) of the newly added interval.                                          }
{------------------------------------------------------------------------------}
var
  LevelList:TDoubleArray;
begin
  setlength(LevelList,1);
  LevelList[0]:=Level;
  Result:=AddInterval(ADateTime,LevelList);
end; {End of T4DSynoptic.AddInterval.}

function T4DSynoptic.AddInterval(ADateTime:TDateTime;LevelList:TDoubleArray):integer;
{------------------------------------------------------------------------------}
{ This function adds the levels listed in the 'LevelList' array to the object  }
{ at the date and time specified by 'ADateTime'. The function returns the      }
{ interval number (1 to Count) of the newly added interval.                    }
{------------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variable through levels.}
  Hold:TGlobalSynopticStack; {This is the new synoptic stack object that gets created.}
begin
  {Confirm that the necessary default values have been set.}
  ConfirmLatLongAttributesSet;
  if FDataType=compressed then ConfirmOffsetAndScalingSet;

  {See if there is already a TGlobalSynopticStack at this date and time.}
  {If there is then grab it, otherwise create a new stack.}
  if QueryDataAvailableAtDateTime(ADateTime) then
    Hold:=GetStackObject(GetStackNumForDateTime(ADateTime))
  else
  begin
    Hold:=TGlobalSynopticStack.Create(FDataType,ADateTime);
    AddNewStack(Hold);
  end;
  for I:=Low(LevelList) to High(LevelList) do Hold.AddLevel(LevelList[I]);
  Result:=IndexOf(Hold)+1;
end; {End of T4DSynoptic.AddInterval.}

procedure T4DSynoptic.RemoveFirstInterval;
{------------------------------------------------------------------------------}
{ This routine removes the first interval from the object. If there are no     }
{ data in the object, the routine does nothing.                                }
{------------------------------------------------------------------------------}
begin
  if Count<>0 then
  begin
    TGlobalSynopticStack(Items[0]).Free;
    Delete(0);
  end;
end; {End of T4DSynoptic.RemoveFirstInterval.}

procedure T4DSynoptic.RemoveLastInterval;
{------------------------------------------------------------------------------}
{ This routine removes the last interval from the object. If there are no     }
{ data in the object, the routine does nothing.                                }
{------------------------------------------------------------------------------}
begin
  if Count<>0 then
  begin
    TGlobalSynopticStack(Items[Count-1]).Free;
    Delete(Count-1);
  end;
end; {End of T4DSynoptic.RemoveLasttInterval.}

procedure T4DSynoptic.AddSynopticGlobal(ADateTime:TDateTime;Value:TSynopticGlobal);
{------------------------------------------------------------------------------}
{ This procedure adds a SynopticGlobal object to the 4DSynoptic object at the  }
{ date and time specified by 'ADateTime'. None of the default values in terms  }
{ of grid spacing etc. are overwritten by the grid spacing in the              }
{ SynopticGlobal object.                                                       }
{------------------------------------------------------------------------------}
var
  IntervalNum,       {The interval number.}
  LevelNum:cardinal; {The level number.}
begin
  try ConfirmLatLongAttributesSet;
  except
    on E4DSynoptic do with Value do SetDefaultLatLongAttr(NumLong,NumLat,StartLong,StartLat,DeltaLong,DeltaLat);
  end;
  if FDataType=compressed then
  try ConfirmOffsetAndScalingSet;
  except
    on E4DSynoptic do with Value do SetDefaultScalingAndOffset(Scaling,Offset);
  end;
  GetIntervalAndLevel(ADateTime,Value.Level,true,IntervalNum,LevelNum);
  ReplaceSynopticGlobalByIndices(IntervalNum,LevelNum,Value);
end; {End of T4DSynoptic.AddSynopticGlobal.}

procedure T4DSynoptic.InsertTomsData(ADateTime:TDateTime;ALevel:double;TomsData:TMercToms);
{------------------------------------------------------------------------------}
{ This procedure extracts data from the TOMS grid specified by 'TomsData' and  }
{ inserts it into the T4DSynoptic object at the date and time specified by     }
{ 'ADateTime' and at the level specified by 'ALevel'. If a time interval and/  }
{ or a level need to be added to the object to accomplish this, this is done   }
{ automatically.                                                               }
{------------------------------------------------------------------------------}
var
  I,J:integer; {Loop control variables over the grid.}
  Value:double; {The value extracted from the TOMS grid.}
  HoldSynoptic:TSynopticGlobal; {Holds the SynopticGlobal object.}
begin
  HoldSynoptic:=GetSynopticGlobal(ADateTime,ALevel,true);
  for I:=1 to HoldSynoptic.NumLong do
    for J:=1 to HoldSynoptic.NumLat do
    begin
      Value:=TomsData.ReadOut(HoldSynoptic.GetLatitude(J),HoldSynoptic.GetLongitude(I),false);
      if Value>0.0 then HoldSynoptic.PutValue(I,J,Value);
    end;
end; {End of T4DSynoptic.InsertTomsData.}

procedure T4DSynoptic.InsertNCEPNCARData(ADateTime:TDateTime;ALevel:double;NCEPNCARData:TNCEPGrid);
{------------------------------------------------------------------------------}
{ This procedure extracts data from the NCEP/NCAR grid specified by            }
{ 'NCEPNCARData' and inserts it into the T4DSynoptic object at the date and    }
{ time specified by 'ADateTime' and at the level specified by 'ALevel'. If a   }
{ time interval and/or a level need to be added to the object to accomplish    }
{ this, this is done automatically. Note that data in TNCEPGrid go from north  }
{ to south.                                                                    }
{------------------------------------------------------------------------------}
var
  I,J:integer; {Loop control variables over the grid.}
  Header:TSynopticGlobalHeader; {The header for the new object.}
  HoldSynoptic:TSynopticGlobal; {Holds the SynopticGlobal object.}
begin
  with Header do
  begin
    NumLong:=NCEPNCARData.NumLong;
    NumLat:=NCEPNCARData.NumLat;
    DeltaLong:=2.5;
    DeltaLat:=2.5;
    StartLong:=0.0;
    StartLat:=-90.0;
  end;
  HoldSynoptic:=TSynopticGlobal.Create(Header);
  for I:=1 to Header.NumLong do
    for J:=1 to Header.NumLat do
      if not NCEPNCARData.IsNull(I,J) then HoldSynoptic.PutValue(I,Header.NumLat+1-J,NCEPNCARData[I,J]);
  HoldSynoptic.Compressed:=(FDataType=compressed);
  HoldSynoptic.Level:=ALevel;
  AddSynopticGlobal(ADateTime,HoldSynoptic);
end; {End of T4DSynoptic.InsertNCEPNCARData.}

procedure T4DSynoptic.PutValueByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double;Value:single);
{------------------------------------------------------------------------------}
{ This procedure inserts the value specified by 'Value' into the T4DSynoptic   }
{ object at the date and time specified by 'ADateTime', at the level specified }
{ by 'ALevel', and at the longitude and latitude specified by 'ALongitude' and }
{ 'ALatitude' respectively. If a time interval and/or a level need to be added }
{ to the object to accomplish this, this is done automatically.                }
{------------------------------------------------------------------------------}
begin
  GetSynopticGlobal(ADateTime,ALevel,true).PutValueByCoords(ALongitude,ALatitude,Value);
end; {End of T4DSynoptic.PutValueByCoords.}

procedure T4DSynoptic.PutValueByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal;Value:single);
{------------------------------------------------------------------------------}
{ This procedure inserts the value specified by 'Value' into the T4DSynoptic   }
{ object at the date and time specified by 'ADateTime', at the level specified }
{ by 'ALevel', and at the longitude and latitude indices by 'LongIndex' and    }
{ 'LatIndex' respectively. If a time interval and/or a level need to be added }
{ to the object to accomplish this, this is done automatically.                }
{------------------------------------------------------------------------------}
begin
  GetSynopticGlobal(ADateTime,ALevel,true).PutValue(LongIndex,LatIndex,Value);
end; {End of T4DSynoptic.PutValueByIndex.}

procedure T4DSynoptic.SaveToFile(Filename:string);
{------------------------------------------------------------------------------}
{ This routine saves the object to the file specified by 'Filename'.           }
{------------------------------------------------------------------------------}
var
  I:longint; {Loop control variable through GlobalSynopticStacks.}
  Dest:file; {The destination file handle.}
begin
  assignfile(Dest,Filename);
  rewrite(Dest,1);
  for I:=1 to Count do GetStackObject(I).SaveToFile(Dest);
  closefile(Dest);
end; {End of T4DSynoptic.SaveToFile.}

procedure T4DSynoptic.SaveToNetCDFFile(OptionsFile:string);
{------------------------------------------------------------------------------}
{ This procedure writes to a NetCDF file using the options listed in the file  }
{ named 'OptionsFile'.                                                         }
{------------------------------------------------------------------------------}
var
  NumLat,
  NumLong,
  I,J,K,L:integer;
  LevelNum:cardinal;
  TestSyn:TSynopticGlobal;
  Buffer:TSingleArray;
  Levels,
  Hold:TDoubleArray;
  HoldStack:TGlobalSynopticStack;
  StartPosition,
  CountsPerDim:TIntegerArray;
  LevelDesc,
  DataDesc,
  WorkStr:string;
  NullValue:single;
  IncludeLevels:boolean;
  Data:TDelphiNetCDF; {The delphi NetCDF object instance.}
  Options:TGBStringList; {The list of string options read in from OptionsFile.}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'saving to NetCDF doesn''t work unless all grids have the same structure.');
  if not FileExists(OptionsFile) then raise E4DSynoptic.Create(
    'the options file name '+OptionsFile+' couldn''t be found.');
  try
    TestSyn:=GetStackObject(1).GetSynopticGlobal(1);
    NumLat:=TestSyn.NumLat;
    NumLong:=TestSyn.NumLong;

    Options:=TGBStringList.Create;
    Options.LoadFromFile(OptionsFile);
    LevelDesc:=Options.Values['Level descriptor'];
    DataDesc:=Options.Values['Data descriptor'];

    DeleteFile(Options.Values['Output file name']);
    Data:=TDelphiNetCDF.Create(Options.Values['Output file name']);
    if Data=nil then raise E4DSynoptic.Create('the TDelphiNetCDF object couldn''t be created.');

    {Set the global text attributes.}
    if Options.YesNo('Include creation date and time') then
      Data.PutGlobalTextAttribute('Created',DateTimeToStr(Now,'HH:TT DD MMM YYYY'));
    I:=Options.IndexOf('[Global text attributes]');
    repeat
      inc(I);
      if Options[I]<>'' then Data.PutGlobalTextAttribute(Options.Names[I],Options.ValueFromIndex[I]);
    until Options[I]='';

    {Define the dimensions.}
    Hold:=GetDateTimeList;
    Data.DefineDimension('time',length(Hold));
    Levels:=GetArrayOfAllLevels;
    if ((length(Levels)=1) and (not Options.YesNo('Include level if only one level'))) then IncludeLevels:=false else
    begin
      IncludeLevels:=true;
      Data.DefineDimension(LevelDesc,length(Levels));
    end;
    Data.DefineDimension('latitude',NumLat);
    Data.DefineDimension('longitude',NumLong);

    {Define the variables.}
    Data.DefineFloatVariable('time');
    if IncludeLevels then Data.DefineFloatVariable(LevelDesc);
    Data.DefineFloatVariable('latitude');
    Data.DefineFloatVariable('longitude');
    if IncludeLevels then WorkStr:='time,'+LevelDesc+',latitude,longitude' else WorkStr:='time,latitude,longitude';
    Data.DefineFloatVariable(DataDesc,WorkStr);

    {Define the variable text attributes.}
    Data.PutVariableTextAttribute('time','long_name','time');
    Data.PutVariableTextAttribute('time','units','days since 1950-01-01 00:00:00');
    Data.PutVariableTextAttribute('time','standard_name','time');
    Data.PutVariableTextAttribute('time','calendar','standard');
    Data.PutVariableTextAttribute('time','axis','T');
    if IncludeLevels then
    begin
      I:=Options.IndexOf('[Level description]')+1;
      repeat
        inc(I);
        if Options[I]<>'' then Data.PutVariableTextAttribute(LevelDesc,Options.Names[I],Options.ValueFromIndex[I]);
      until Options[I]='';
    end;
	  Data.PutVariableTextAttribute('latitude','long_name','Latitude');
    Data.PutVariableTextAttribute('latitude','units','degrees_north');
    Data.PutVariableTextAttribute('latitude','standard_name','latitude');
    Data.PutVariableTextAttribute('latitude','axis','Y');
	  Data.PutVariableTextAttribute('longitude','long_name','Longitude');
    Data.PutVariableTextAttribute('longitude','units','degrees_east');
    Data.PutVariableTextAttribute('longitude','standard_name','longitude');
    Data.PutVariableTextAttribute('longitude','axis','X');
    I:=Options.IndexOf('[Data description]')+1;
    repeat
      inc(I);
      if Options[I]<>'' then Data.PutVariableTextAttribute(DataDesc,Options.Names[I],Options.ValueFromIndex[I]);
    until ((Options[I]='') or (I=Options.Count-1));

    Data.SwitchToDataMode;

    {Pass values to the dimensions.}
    Hold:=GetDateTimeList;
    setlength(Buffer,length(Hold));
    for I:=Low(Hold) to High(Hold) do Buffer[I]:=Hold[I]-EncodeDate(1950,1,1);
    Data.PutVariableFloat('time',Buffer);
    if IncludeLevels then
    begin
      setlength(Buffer,length(Levels));
      for I:=Low(Levels) to High(Levels) do Buffer[I]:=Levels[I];
      Data.PutVariableFloat(LevelDesc,Buffer);
    end;
    Hold:=TestSyn.GetLatitudes;
    setlength(Buffer,length(Hold));
    for I:=Low(Hold) to High(Hold) do Buffer[I]:=Hold[I];
    Data.PutVariableFloat('latitude',Buffer);
    setlength(Buffer,NumLong);
    for I:=0 to NumLong-1 do Buffer[I]:=TestSyn.StartLong+I*TestSyn.DeltaLong;
    Data.PutVariableFloat('longitude',Buffer);

    {Start writing the data. We write one TSynopticGlobal at a time.}
    if IncludeLevels then
    begin
      setlength(StartPosition,4);
      StartPosition[2]:=0; {Latitude}
      StartPosition[3]:=0; {Longitude.}
      setlength(CountsPerDim,4);
      CountsPerDim[0]:=1; {Time}
      CountsPerDim[1]:=1; {Level}
      CountsPerDim[2]:=NumLat;
      CountsPerDim[3]:=NumLong;
    end
    else
    begin
      setlength(StartPosition,3);
      StartPosition[1]:=0; {Latitude}
      StartPosition[2]:=0; {Longitude.}
      setlength(CountsPerDim,3);
      CountsPerDim[0]:=1; {Time}
      CountsPerDim[1]:=NumLat;
      CountsPerDim[2]:=NumLong;
    end;
    setlength(Buffer,NumLat*NumLong);
    NullValue:=strtofloat(Options.Values['Null data value']);
    for I:=1 to Count do
    begin
      StartPosition[0]:=I-1;
      HoldStack:=GetStackObject(I);
      for J:=1 to length(Levels) do
      begin
        if IncludeLevels then StartPosition[1]:=J-1;
        try
          LevelNum:=HoldStack.GetLevelNum(Levels[J-1]);
          TestSyn:=HoldStack.GetSynopticGlobal(LevelNum);
          for K:=1 to NumLat do
            for L:=1 to NumLong do
            if TestSyn.IsNull(L,K) then Buffer[(K-1)*NumLong+L-1]:=NullValue
            else Buffer[(K-1)*NumLong+L-1]:=TestSyn[L,K];
        except
          on exception do for K:=1 to NumLat*NumLong do Buffer[K-1]:=NullValue;
        end;
        Data.PutVariableFloatArray(DataDesc,StartPosition,CountsPerDim,Buffer);
      end; {End of loop through levels.}
    end; {End of loop through time intervals.}
  finally
    FreeAndNil(Data);
    FreeAndNil(Options);
  end;
end; {End of T4DSynoptic.SaveToNetCDFFile.}

procedure T4DSynoptic.ReadFromFile(Filename:string);
{------------------------------------------------------------------------------}
{ This procedure reads the data from the file specified by 'Filename'.         }
{------------------------------------------------------------------------------}
var
  Source:file;
  Hold:TGlobalSynopticStack;
begin
  Clear;
  FileMode:=0;
  assignfile(Source,Filename);
  reset(Source,1);
  repeat
    Hold:=TGlobalSynopticStack.Create(Source);
    AddNewStack(Hold);
  until eof(Source);
  closefile(Source);
  FileMode:=2;
end; {End of T4DSynoptic.ReadFromFile.}

procedure T4DSynoptic.CombineWithFile(Filename:string);
{------------------------------------------------------------------------------}
{ This procedure reads data from the file specified by 'Filename' and adds it  }
{ to the data currently in the object. This is done on a grid by grid basis    }
{ i.e. if there is already data available at any time and level, it is         }
{ replaced in its entirity by the data read in from the file.                  }
{------------------------------------------------------------------------------}
var
  I:integer;
  Source:file;
  DateTime:TDateTime;
  Hold:TGlobalSynopticStack;
begin
  FileMode:=0;
  assignfile(Source,Filename);
  reset(Source,1);
  repeat
    Hold:=TGlobalSynopticStack.Create(Source);
    try
      DateTime:=Hold.GetDateStamp;
      for I:=1 to Hold.Count do AddSynopticGlobal(DateTime,Hold.GetSynopticGlobal(I,true));
    finally
      FreeAndNil(Hold);
    end;
  until eof(Source);
  closefile(Source);
  FileMode:=2;
end; {End of T4DSynoptic.CombineWithFile.}

procedure T4DSynoptic.ReadFromNetCDFFile(Filename:string);
{------------------------------------------------------------------------------}
{ This procedure reads the data from the NetCDF file specified by 'Filename'   }
{------------------------------------------------------------------------------}
var
  NetCDF:TDelphiNetCDF;
  DimIndicesForVar:gb_array.TByteArray;
  Dimensions,
  Variables:TStringList;
  LevelDimIndex,
  TimeDimIndex,
  DimIndex,
  VarIndex,
  LatIndex,
  I,J,K,L:integer;
  Latitudes,
  Longitudes,
  LevelList,
  DateTime:TDoubleArray;
  Hold:TSynopticGlobal;
  Header:TSynopticGlobalHeader;
  DataArray:gb_array.TDoubleArray;
  StartPosition,
  CountsPerDim:gb_array.TIntegerArray;
  PrimaryCoord:string;
  DecreasingLatitudes,
  SkipLevel:boolean;
begin
  NetCDF:=TDelphiNetCDF.Create(Filename);
  try
    if ((@FFunctionToConvertStrToDateTime=nil) and (@FFunctionToConvertValueToDateTime=nil)) then
      raise E4DSynoptic.Create('no function has been specified to indicate how the date and time values'+
      ' from the NetCDF file should be converted to TDateTime values. Please specify a function using'+
      ' either FunctionToConvertNetCDFStrToDateTime or FunctionToConvertNetCDFValueToDateTime.');
    if @FFunctionToConvertStrToDateTime<>nil then NetCDF.FunctionToConvertStrToDateTime:=FFunctionToConvertStrToDateTime;
    if @FFunctionToConvertValueToDateTime<>nil then NetCDF.FunctionToConvertValueToDateTime:=FFunctionToConvertValueToDateTime;
    if FNetCDFVarStr='' then raise E4DSynoptic.Create('the variable string has not been set. Please set the NetCDFVarStr property.');

    {Set the time ID and variables strings.}
    NetCDF.NetCDFTimeIDStr:=FNetCDFTimeIDStr;
    NetCDF.NetCDFLevelIDStr:=FNetCDFLevelIDStr;
    NetCDF.NetCDFVarStr:=FNetCDFVarStr;

    Dimensions:=NetCDF.GetListOfDimensions;
    Variables:=NetCDF.GetListOfVariables;

    {Find out which dimension holds the time.}
    TimeDimIndex:=-1;
    repeat inc(TimeDimIndex) until ((pos(FNetCDFTimeIDStr,LowerCase(Dimensions[TimeDimIndex]))>0) or (TimeDimIndex=Dimensions.Count-1));
    if pos(FNetCDFTimeIDStr,LowerCase(Dimensions[TimeDimIndex]))=0 then raise E4DSynoptic.Create('the dimension holding the time could not be found.');

    {Find out which variable holds the time.}
    VarIndex:=-1;
    repeat inc(VarIndex) until ((pos(FNetCDFTimeIDStr,LowerCase(Variables[VarIndex]))>0) or (VarIndex=Variables.Count-1));
    if pos(FNetCDFTimeIDStr,LowerCase(Variables[VarIndex]))=0 then raise E4DSynoptic.Create('the variable holding the time could not be found.');

    {Get the dates and times.}
    DateTime:=NetCDF.GetDatesAndTimes(TimeDimIndex,VarIndex);

    {Get the latitudes.}
    DimIndex:=-1;
    repeat inc(DimIndex) until ((pos('lat',LowerCase(Dimensions[DimIndex]))>0) or (DimIndex=Dimensions.Count-1));
    if pos('lat',LowerCase(Dimensions[DimIndex]))=0 then raise E4DSynoptic.Create('the dimension holding the latitude could not be found.');
    VarIndex:=-1;
    repeat inc(VarIndex) until ((pos('lat',LowerCase(Variables[VarIndex]))>0) or (VarIndex=Variables.Count-1));
    if pos('lat',LowerCase(Variables[VarIndex]))=0 then raise E4DSynoptic.Create('the variable holding the latitude could not be found.');
    Latitudes:=NetCDF.GetDimensionArray(DimIndex,VarIndex);
    if length(Latitudes)<2 then raise E4DSynoptic.Create('there were too few latitudes.');
    if (Latitudes[1]-Latitudes[0])<0 then
    begin
      DecreasingLatitudes:=true;
      HeapSort(Latitudes);
    end
    else DecreasingLatitudes:=false;

    {Get the longitudes. Note that unlike the latitudes, there may not be any longitude}
    {dimension in the file e.g. in the case of a zonal mean files. The code has been   }
    {written to cope with this by introducing a single longitude at 0.0.}
    DimIndex:=-1;
    repeat inc(DimIndex) until ((pos('lon',LowerCase(Dimensions[DimIndex]))>0) or (DimIndex=Dimensions.Count-1));
    if pos('lon',LowerCase(Dimensions[DimIndex]))=0 then
    begin
      setlength(Longitudes,1);
      Longitudes[0]:=0.0;
    end
    else
    begin
      VarIndex:=-1;
      repeat inc(VarIndex) until ((pos('lon',LowerCase(Variables[VarIndex]))>0) or (VarIndex=Variables.Count-1));
      if pos('lon',LowerCase(Variables[VarIndex]))=0 then raise E4DSynoptic.Create('the variable holding the longitude could not be found.');
      Longitudes:=NetCDF.GetDimensionArray(DimIndex,VarIndex);
      if length(Longitudes)<1 then raise E4DSynoptic.Create('there were too few longitudes.');
    end;

    {Find out which dimension, if any, holds the levels.}
    LevelDimIndex:=-1;
    repeat inc(LevelDimIndex) until ((pos(FNetCDFLevelIDStr,LowerCase(Dimensions[LevelDimIndex]))>0) or (LevelDimIndex=Dimensions.Count-1));
    if pos(FNetCDFLevelIDStr,LowerCase(Dimensions[LevelDimIndex]))=0 then LevelDimIndex:=-1;

    {Find out which variable, if any, holds the levels.}
    VarIndex:=-1;
    repeat inc(VarIndex) until ((pos(FNetCDFLevelIDStr,LowerCase(Variables[VarIndex]))>0) or (VarIndex=Variables.Count-1));
    if pos(FNetCDFLevelIDStr,LowerCase(Variables[VarIndex]))=0 then VarIndex:=-1;

    {Generate the Levels array.}
    if ((LevelDimIndex=-1) and (VarIndex=-1)) then
    begin
      setlength(LevelList,1);
      LevelList[0]:=0;
    end
    else LevelList:=NetCDF.GetDimensionArray(LevelDimIndex,VarIndex);
    if @FFunctionToTransformLevels<>nil then for I:=0 to length(LevelList)-1 do
      LevelList[I]:=FFunctionToTransformLevels(LevelList[I]);

    {Create the header for the generic TSynopticGlobal object.}
    if not EqualSpacing(Longitudes) then raise E4DSynoptic.Create('the longitudes were not equally spaced.');
    with Header do
    begin
      NumLong:=length(Longitudes);
      NumLat:=length(Latitudes);
      TagValue:=0;
      if length(Longitudes)=1 then DeltaLong:=0.0 else DeltaLong:=Longitudes[1]-Longitudes[0];
      DeltaLat:=Latitudes[1]-Latitudes[0];
      StartLong:=Longitudes[0];
      StartLat:=Latitudes[0];
    end;

    {Set up the loops.}
    DimIndicesForVar:=NetCDF.GetListOfDimIndicesForVar(FNetCDFVarStr);
    if length(DimIndicesForVar)>4 then raise E4DSynoptic.Create('there were too many dimensions for the '+FNetCDFVarStr+' variable.');
    setlength(StartPosition,length(DimIndicesForVar));
    setlength(CountsPerDim,length(DimIndicesForVar));
    for I:=Low(DateTime) to High(DateTime) do
      for J:=Low(LevelList) to High(LevelList) do
      begin
        {If the target level list has been set, check first to see that we are}
        {at one of those levels.}
        if FNetCDFTargetLevels=nil then SkipLevel:=false else
        begin
          SkipLevel:=true;
          for L:=Low(FNetCDFTargetLevels) to High(FNetCDFTargetLevels) do
            if EquivDouble(FNetCDFTargetLevels[L],LevelList[J]) then SkipLevel:=false;
        end;
        if not SkipLevel then
        begin
          Hold:=TSynopticGlobal.Create(Header); {We don't need to free this since we pass it to the object.}
          Hold.Level:=LevelList[J];
          Hold.DateTime:=DateTime[I];
          if not EqualSpacing(Latitudes) then Hold.SpecifyNonuniformLatitudes(Latitudes);
          PrimaryCoord:='';
          for K:=Low(DimIndicesForVar) to High(DimIndicesForVar) do
          begin
            if ((pos('lat',Dimensions[DimIndicesForVar[K]])>0) and (PrimaryCoord='')) then PrimaryCoord:='latitude';
            if ((pos('lon',Dimensions[DimIndicesForVar[K]])>0) and (PrimaryCoord='')) then PrimaryCoord:='longitude';
            StartPosition[K]:=0;
            CountsPerDim[K]:=NetCDF.GetDimensionLength(DimIndicesForVar[K]);
            if DimIndicesForVar[K]=TimeDimIndex then
            begin
              StartPosition[K]:=I;
              CountsPerDim[K]:=1;
            end;
            if DimIndicesForVar[K]=LevelDimIndex then
            begin
              StartPosition[K]:=J;
              CountsPerDim[K]:=1;
            end;
          end;
          DataArray:=NetCDF.GetDataArray(FNetCDFVarStr,StartPosition,CountsPerDim);
          for K:=1 to Header.NumLong do
            for L:=1 to Header.NumLat do
            begin
              if DecreasingLatitudes then LatIndex:=Header.NumLat+1-L else LatIndex:=L;
              if PrimaryCoord='latitude' then Hold[K,LatIndex]:=DataArray[(L-1)*Header.NumLong+K-1]
              else Hold[K,LatIndex]:=DataArray[(K-1)*Header.NumLat+L-1];
            end;

          {Insert the SynopticGlobal into the grid.}
          AddSynopticGlobal(DateTime[I],Hold);
        end;
      end;
  finally
    NetCDF.Free;
  end;
end; {End of T4DSynoptic.ReadFromNetCDFFile.}

function T4DSynoptic.GetDateTimeList:TDoubleArray;
{------------------------------------------------------------------------------}
{ This function returns an array of TDateTime values for each interval in the  }
{ file.                                                                        }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  setlength(Result,Count);
  for I:=1 to Count do Result[I-1]:=GetDateTimeForInterval(I);
end; {End of T4DSynoptic.GetDateTimeList.}

function T4DSynoptic.QueryDataAvailableAtDateTime(ADateTime:TDateTime):boolean;
{------------------------------------------------------------------------------}
{ This function returns true if there are data available at the date and time  }
{ specified by 'ADateTime'.                                                    }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  Result:=false;
  for I:=1 to Count do if SameDateTime(ADateTime,GetDateTimeForInterval(I)) then Result:=true;
end; {End of T4DSynoptic.QueryDataAvailableAtDateTime.}

function T4DSynoptic.QueryDataAvailableOnDay(ADateTime:TDateTime):boolean;
{------------------------------------------------------------------------------}
{ This function returns true if there are data available on the day specified  }
{ by 'ADateTime'.                                                              }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  Result:=false;
  for I:=1 to Count do if IsSameDay(ADateTime,GetDateTimeForInterval(I)) then Result:=true;
end; {End of T4DSynoptic.QueryDataAvailableOnDay.}

function T4DSynoptic.DateTimeSpanned(ADateTime:TDateTime):boolean;
{------------------------------------------------------------------------------}
{ This routine returns true if the data in the object span the data and time   }
{ specified by 'ADateTime'.                                                    }
{------------------------------------------------------------------------------}
begin
  if ((Count>0) and DateTimeBetween(ADateTime,GetDateTimeForInterval(1),GetDateTimeForInterval(Count)))
  then Result:=true else Result:=false;
end; {End of T4DSynoptic.DateTimeSpanned.}

function T4DSynoptic.GetLevelsForDateTime(ADateTime:TDateTime):TDoubleArray;
{------------------------------------------------------------------------------}
{ This function returns an array containing the levels at which data are       }
{ available at the time interval specified by 'ADateTime'. If there are no     }
{ data available at this date and time, the function returns an empty array.   }
{------------------------------------------------------------------------------}
var
  INum,
  NumLevels,
  I:integer;
begin
  if QueryDataAvailableAtDateTime(ADateTime) then
  begin
    INum:=GetIntervalForDateTime(ADateTime);
    NumLevels:=GetStackObject(INum).Count;
  end
  else
  begin
    NumLevels:=0;
    INum:=-1;
  end;
  setlength(Result,NumLevels);
  for I:=0 to NumLevels-1 do Result[I]:=GetStackObject(INum).GetLevel(I+1);
end; {End of T4DSynoptic.GetLevelsForInterval.}

function T4DSynoptic.GetArrayOfAllLevels:TDoubleArray;
{------------------------------------------------------------------------------}
{ This function returns an array of all levels at all time intervals in the    }
{ object. The list is sorted in increasing level order and there are no        }
{ duplicates.                                                                  }
{------------------------------------------------------------------------------}
var
  AlreadyThere:boolean;
  I,J,K:integer;
  Level:double;
  Hold:TGlobalSynopticStack;
begin
  setlength(Result,0);
  for I:=1 to Count do
  begin
    Hold:=GetStackObject(I);
    for J:=1 to Hold.Count do
    begin
      Level:=Hold.GetLevel(J);
      AlreadyThere:=false;
      for K:=1 to length(Result) do
        if EquivDouble(Result[K-1],Level) then AlreadyThere:=true;
      if not AlreadyThere then
      begin
        setlength(Result,length(Result)+1);
        Result[High(Result)]:=Level;
      end;
    end;
  end;
  HeapSort(Result);
end; {End of T4DSynoptic.GetArrayOfAllLevels.}

function T4DSynoptic.GetSynopticGlobal(ADateTime:TDateTime;ALevel:double;
  AllowAddition:boolean;Owner:boolean=false):TSynopticGlobal;
{------------------------------------------------------------------------------}
{ This function returns the SynopticGlobal object at the date and time         }
{ specified by 'ADateTime' and at the level specified by 'ALevel'.             }
{------------------------------------------------------------------------------}
var
  IntervalNum,
  LevelNum:cardinal;
begin
  if not ConfirmSynopticGlobalBuffer(ADateTime,ALevel) then
  begin
    GetIntervalAndLevel(ADateTime,ALevel,AllowAddition,IntervalNum,LevelNum);
    FSynopticGlobalBuffer:=GetStackObject(IntervalNum).GetSynopticGlobal(LevelNum);
  end;
  if Owner then Result:=TSynopticGlobal.Create(FSynopticGlobalBuffer) else Result:=FSynopticGlobalBuffer;
end; {End of T4DSynoptic.GetSynopticGlobal.}

function T4DSynoptic.GetSynopticGlobalOnDay(ADateTime:TDateTime;ALevel:double;AllowAddition:boolean;Owner:boolean=false):TSynopticGlobal;
{------------------------------------------------------------------------------}
{ This function returns any SynopticGlobal object available on the date        }
{ specified by 'ADateTime' and at the level specified by 'ALevel'.             }
{------------------------------------------------------------------------------}
var
  IntervalNum,
  LevelNum:cardinal;
begin
  if not ConfirmSynopticGlobalBuffer(ADateTime,ALevel) then
  begin
    if QueryDataAvailableOnDay(ADateTime) then IntervalNum:=GetFirstStackNumOnDate(ADateTime) else
    if AllowAddition then IntervalNum:=AddInterval(ADateTime,ALevel) else
      raise E4DSynoptic.Create('there were no data available on '+DateTimeToStr(ADateTime,'DD/MM/YYYY'));
    if GetStackObject(IntervalNum).DataAvailableForLevel(ALevel) then
      LevelNum:=GetLevelNumForLevel(IntervalNum,ALevel)
    else if AllowAddition then LevelNum:=GetStackObject(IntervalNum).AddLevel(ALevel);
    FSynopticGlobalBuffer:=GetStackObject(IntervalNum).GetSynopticGlobal(LevelNum);
  end;
  if Owner then Result:=TSynopticGlobal.Create(FSynopticGlobalBuffer) else Result:=FSynopticGlobalBuffer;
end; {End of T4DSynoptic.GetSynopticGlobal.}

function T4DSynoptic.GetSynopticGlobalOnClosestDay(ADateTime:TDateTime;ALevel:double;AllowAddition:boolean;Owner:boolean=false):TSynopticGlobal;
{------------------------------------------------------------------------------}
{ This function returns the SynopticGlobal object available on the date        }
{ closest to 'ADateTime' and at the level specified by 'ALevel'.               }
{------------------------------------------------------------------------------}
var
  OptimalDateTime:TDateTime;
  I:integer;
  MinimumTimeDifference:double;
begin
  MinimumTimeDifference:=1E8;
  for I:=1 to Count do
  if abs(ADateTime-GetDateTimeForInterval(I))<MinimumTimeDifference then
  begin
    OptimalDateTime:=GetDateTimeForInterval(I);
    MinimumTimeDifference:=abs(ADateTime-GetDateTimeForInterval(I));
  end;
  Result:=GetSynopticGlobal(OptimalDateTime,ALevel,AllowAddition,Owner);
end; {End of T4DSynoptic.GetSynopticGlobalOnClosestDay.}

function T4DSynoptic.GetMeanSynopticGlobalOnDay(ADateTime:TDateTime;ALevel:double):TSynopticGlobal;
var
  Hold:TSynopticGlobalArray;
  I:integer;
  LevelNum:cardinal;
begin
  for I:=1 to Count do if IsSameDay(ADateTime,GetDateTimeForInterval(I)) then
  begin
    try
      LevelNum:=GetLevelNumForLevel(I,ALevel);
    except
      on Exception do raise E4DSynoptic.Create('no data were available on this level.');
    end;
    setlength(Hold,length(Hold)+1);
    Hold[High(Hold)]:=GetStackObject(I).GetSynopticGlobal(LevelNum);
  end;
  Result:=SynopticGlobalMean(Hold);
end;

function T4DSynoptic.GetMeanSynopticGlobalForMonth(Year,Month:word;ALevel:double):TSynopticGlobal;
var
  Hold:TSynopticGlobalArray;
  I:integer;
  LevelNum:cardinal;
begin
  for I:=1 to Count do
  if ((MonthOf(GetDateTimeForInterval(I))=Month) and (YearOf(GetDateTimeForInterval(I))=Year)) then
  begin
    try
      LevelNum:=GetLevelNumForLevel(I,ALevel);
    except
      on Exception do raise E4DSynoptic.Create('no data were available on this level.');
    end;
    setlength(Hold,length(Hold)+1);
    Hold[High(Hold)]:=GetStackObject(I).GetSynopticGlobal(LevelNum);
  end;
  Result:=SynopticGlobalMean(Hold);
end;



function T4DSynoptic.GetFirstDateTime:TDateTime;
{------------------------------------------------------------------------------}
{ This function returns the first i.e earliest date and time for which data    }
{ are available.                                                               }
{------------------------------------------------------------------------------}
begin
  if Count=0 then raise E4DSynoptic.Create('there are no data in the object.');
  Result:=GetDateTimeForInterval(1);
end; {End of T4DSynoptic.GetFirstDateTime.}

function T4DSynoptic.GetLastDateTime:TDateTime;
{------------------------------------------------------------------------------}
{ This function returns the last i.e latest date and time for which data are   }
{ available.                                                                   }
{------------------------------------------------------------------------------}
begin
  if Count=0 then raise E4DSynoptic.Create('there are no data in the object.');
  Result:=GetDateTimeForInterval(Count);
end; {End of T4DSynoptic.GetLastDateTime.}

function T4DSynoptic.GetNumIntervals:cardinal;
{------------------------------------------------------------------------------}
{ This function returns the number of time intervals in the object.            }
{------------------------------------------------------------------------------}
begin
  Result:=Count;
end; {End of T4DSynoptic.GetNumIntervals.}

function T4DSynoptic.GetNumLongitudes:cardinal;
{------------------------------------------------------------------------------}
{ This function returns the number of longitudes in each data grid. If the     }
{ data grids are not all of exactly the same spatial structure, the function   }
{ raises an exception.                                                         }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).NumLong;
end; {End of T4DSynoptic.GetNumLongitudes.}

function T4DSynoptic.GetNumLatitudes:cardinal;
{------------------------------------------------------------------------------}
{ This function returns the number of latitudes in each data grid. If the data }
{ grids are not all of exactly the same spatial structure, the function raises }
{ an exception.                                                                }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).NumLat;
end; {End of T4DSynoptic.GetNumLatitudes.}

function T4DSynoptic.GetStartLat:double;
{------------------------------------------------------------------------------}
{ This function returns the start latitude in each data grid. If the data      }
{ grids are not all of exactly the same spatial structure, the function raises }
{ an exception.                                                                }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).StartLat;
end; {End of T4DSynoptic.GetStartLat.}

function T4DSynoptic.GetStartLong:double;
{------------------------------------------------------------------------------}
{ This function returns the start longitude in each data grid. If the data     }
{ grids are not all of exactly the same spatial structure, the function raises }
{ an exception.                                                                }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).StartLong;
end; {End of T4DSynoptic.GetStartLong.}

function T4DSynoptic.GetEndLat:double;
{------------------------------------------------------------------------------}
{ This function returns the end latitude in each data grid. If the data grids  }
{ are not all of exactly the same spatial structure, the function raises an    }
{ exception.                                                                   }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).EndLatitude;
end; {End of T4DSynoptic.GetEndLat.}

function T4DSynoptic.GetEndLong:double;
{------------------------------------------------------------------------------}
{ This function returns the end longitude in each data grid. If the data grids }
{ are not all of exactly the same spatial structure, the function raises an    }
{ exception.                                                                   }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).EndLongitude;
end; {End of T4DSynoptic.GetEndLong.}

function T4DSynoptic.GetDeltaLat:double;
{------------------------------------------------------------------------------}
{ This function returns the delta latitude in each data grid. If the data      }
{ grids are not all of exactly the same spatial structure, the function raises }
{ an exception.                                                                }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).DeltaLat;
end; {End of T4DSynoptic.GetDeltaLat.}

function T4DSynoptic.GetDeltaLong:double;
{------------------------------------------------------------------------------}
{ This function returns the delta longitude in each data grid. If the data     }
{ grids are not all of exactly the same spatial structure, the function raises }
{ an exception.                                                                }
{------------------------------------------------------------------------------}
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).DeltaLong;
end; {End of T4DSynoptic.GetDeltaLong.}

function T4DSynoptic.GetValueByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double;Interpolate:boolean=false):single;
var
  IntTime,
  Value:array[1..2] of double;
  HoldIndex,
  I:longint;
begin
  Result:=NullValue;
  if Interpolate then
  begin
    if ConfirmSynopticGlobalBuffer(ADateTime,ALevel) then
    begin
      Result:=FSynopticGlobalBuffer.BiLinear(ALatitude,ALongitude);
      exit;
    end;
    if Count=0 then raise E4DSynoptic.Create('there are no intervals in this data set.');
    if ADateTime<GetDateTimeForInterval(1) then raise E4DSynoptic.Create(
      'the passed date and time occurred before the first available interval.');
    if ADateTime>GetDateTimeForInterval(Count) then raise E4DSynoptic.Create(
      'the passed date and time occurred after the last available interval.');

    if Count=1 then {There is only one interval and we are there.}
    begin
      Result:=GetStackObject(Count).InterpolateValue(ALevel,ALongitude,ALatitude,false);
      if ((Result>=NullValue) and (FInterpolationFailPlan=raiseexception)) then
        raise E4DSynoptic.Create('a valid value could not be calculated.');
      exit;
    end;

    {Find the intervals than span the prescribed data and time.}
    if not DateTimeBetween(ADateTime,GetDateTimeForInterval(Int[1]),GetDateTimeForInterval(Int[2])) then
    begin
      Int[2]:=Count;
      Int[1]:=1;
    end;
    while (Int[2]-Int[1])>1 do
    begin
      HoldIndex:=(Int[1]+Int[2]) div 2;
      if GetDateTimeForInterval(HoldIndex)<ADateTime then Int[1]:=HoldIndex else Int[2]:=HoldIndex;
    end;
    if Int[1]=Int[2] then raise E4DSynoptic.Create('Int[1] was equal to Int[2].');

    {Now see if we are exactly at one of these two intervals. We do this because}
    {if this is the case we need only retrieve one value rather that two values }
    {from the Global SynopticStack and it doesn't matter if one of the values   }
    {retrieved from one of the GlobalSynopticStacks is null.}
    for I:=1 to 2 do
    if SameDateTime(GetDateTimeForInterval(Int[I]),ADateTime) then
    begin
      Result:=GetStackObject(Int[I]).InterpolateValue(ALevel,ALongitude,ALatitude,false);
      if ((Result>=NullValue) and (FInterpolationFailPlan=raiseexception)) then
        raise E4DSynoptic.Create('the value was null.');
      exit;
    end;

    {Now we know that we will need to interpolate between two value. Let's get them.}
    for I:=1 to 2 do
    begin
      Value[I]:=GetStackObject(Int[I]).InterpolateValue(ALevel,ALongitude,ALatitude,false);
      if Value[I]>=NullValue then
      case FInterpolationFailPlan of
        raiseexception:raise E4DSynoptic.Create('value number '+inttostr(I)+
          ' couldn''t be retrieved from the stack of synoptic grids.');
        returnnull:exit;
      end;
      IntTime[I]:=GetDateTimeForInterval(Int[I]);
    end;

    Result:=(ADateTime-IntTime[1])*(Value[2]-Value[1])/(IntTime[2]-IntTime[1])+Value[1];
    if Int[2]>Count then raise E4DSynoptic.Create(
      'Int[2] was greater than the number of records. This should never happen.');
    if Int[1]<1 then raise E4DSynoptic.Create(
      'Int[1] was less than zero. This should never happen.');
  end
  else
  begin
    try
      Result:=GetSynopticGlobal(ADateTime,ALevel,false).GetValueByCoords(ALongitude,ALatitude,false);
    except
      on E:exception do
      begin
        case FInterpolationFailPlan of
          raiseexception:raise;
          returnnull:exit;
        end;
      end;
    end;
  end;
end; {End of T4DSynoptic.GetValueByCoords.}

function T4DSynoptic.GetValueByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal):single;
begin
  Result:=NullValue;
  try
    Result:=GetSynopticGlobal(ADateTime,ALevel,false).GetValue(LongIndex,LatIndex);
  except
    on E:exception do
    begin
      case FInterpolationFailPlan of
        raiseexception:raise;
        returnnull:exit;
      end;
    end;
  end;
end; {End of T4DSynoptic.GetValueByIndex.}

function T4DSynoptic.InterpolateSynopticGlobal(ADateTime:TDateTime;ALevel:double;Owner:boolean=false):TSynopticGlobal;
var
  Interval,
  Level,
  I,J:cardinal;
begin
  if not DateTimeSpanned(ADateTime) then raise E4DSynoptic.Create('the data do not span the data and time specified.');
  Interval:=GetIntervalOnOrBeforeDateTime(ADateTime);
  Level:=GetStackObject(Interval).GetLevelNumAtOrBelow(ALevel);

  {We get the SynopticGlobal here just to get the grid info. The values themselves}
  {are fully interpolated.}
  Result:=GetStackObject(Interval).GetSynopticGlobal(Level,Owner);
  for I:=1 to Result.NumLong do
    for J:=1 to Result.NumLat do
      Result[I,J]:=GetValueByCoords(ADateTime,ALevel,Result.GetLongitude(I),Result.GetLatitude(J),true);
  Result.DateTime:=ADateTime;
end; {End if T4DSynoptic.InterpolateSynopticGlobal.}

function T4DSynoptic.IsNullByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double):boolean;
begin
  try Result:=GetSynopticGlobal(ADateTime,ALevel,false).IsNullByCoords(ALongitude,ALatitude);
  except
    on E:exception do Result:=true;
  end;
end; {End of T4DSynoptic.IsNullByCoords.}

function T4DSynoptic.IsNullByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal):boolean;
begin
  try Result:=GetSynopticGlobal(ADateTime,ALevel,false).IsNull(LongIndex,LatIndex);
  except
    on E:exception do Result:=true;
  end;
end; {End of T4DSynoptic.IsNullByIndex.}

function T4DSynoptic.IsNotNullByCoords(ADateTime:TDateTime;ALevel,ALongitude,ALatitude:double):boolean;
begin
  Result:=not IsNullByCoords(ADateTime,ALevel,ALongitude,ALatitude);
end; {End of T4DSynoptic.IsNotNullByCoords.}

function T4DSynoptic.IsNotNullByIndex(ADateTime:TDateTime;ALevel:double;LongIndex,LatIndex:cardinal):boolean;
begin
  Result:=not IsNullByIndex(ADateTime,ALevel,LongIndex,LatIndex);
end; {End of T4DSynoptic.IsNotNullByIndex.}

procedure T4DSynoptic.Clear;
{------------------------------------------------------------------------------}
{ Clear frees the memory and removes all of the objects on the TList from      }
{ which T4DSynoptic descends.                                                  }
{------------------------------------------------------------------------------}
var
  I:longint;
  Hold:TGlobalSynopticStack;
begin
  for I:=0 to Count-1 do
  if Items[I]<>nil then
  begin
    Hold:=TGlobalSynopticStack(Items[I]);
    FreeAndNil(Hold);
    Items[I]:=nil; {This line is vital since Notify checks whether the items are nil.}
  end;
  inherited Clear;
  FSynopticGlobalBuffer:=nil;
  Int[1]:=1;
  Int[2]:=2;
end; {End of T4DSynoptic.Clear.}

function T4DSynoptic.GetCellArea(Lat:cardinal):single;
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).GetCellArea(Lat);
end; {End of T4DSynoptic.GetCellArea.}

function T4DSynoptic.GetMemoryUsed:cardinal;
var
  I:integer;
begin
  Result:=0;
  for I:=1 to Count do
    if Items[I-1]<>nil then Result:=Result+GetStackObject(I).GetMemoryUsed;
end; {End of T4DSynoptic.GetMemoryUsed.}

procedure T4DSynoptic.ReplaceSynopticGlobal(ADateTime:TDateTime;ALevel:double;Value:TSynopticGlobal);
var
  IntervalNum,
  LevelNum:cardinal;
begin
  GetIntervalAndLevel(ADateTime,ALevel,false,IntervalNum,LevelNum);
  ReplaceSynopticGlobalByIndices(IntervalNum,LevelNum,Value);
end; {End of T4DSynoptic.ReplaceSynopticGlobal.}

function Compare(Item1,Item2:pointer):integer;
begin
  Result:=0;
  if TGlobalSynopticStack(Item1).GetDateStamp<TGlobalSynopticStack(Item2).GetDateStamp then Result:=-1;
  if TGlobalSynopticStack(Item1).GetDateStamp>TGlobalSynopticStack(Item2).GetDateStamp then Result:=1;
end;

procedure T4DSynoptic.SortByInterval;
var
  I:integer;
begin
  Sort(Compare);
  for I:=1 to Count-1 do
    if GetDateTimeForInterval(I)=GetDateTimeForInterval(I+1) then raise E4DSynoptic.Create(
       'date stamps for two of the intervals were the same. This is not a good idea.');
end; {End of T4DSynoptic.SortByInterval.}

procedure T4DSynoptic.AddNewStack(Stack:TGlobalSynopticStack);
begin
  {In the line below we pass FMaxMemory since later, when we call Add(Stack)  }
  {the notify procedure is called and the memory used is reassessed. So mainly}
  {we call RestrictMemoryUsed to pass the temporary file name.}
  if FMemRestriction then Stack.RestrictMemoryUsed(FMaxMemory,TempFilename+'.'+inttostrf(Count,5,true));
  Stack.SetNewSynopticGlobalDefaults(NewSynopticGlobalDefaults);
  Stack.ConfineToLevels:=FConfineToLevels;
  Stack.ExtendToPole:=FExtendToPole;
  Stack.InterpolationFailPlan:=FInterpolationFailPlan;
  Stack.Owner:=self;
  Add(Stack);
  SortByInterval;
end; {End of T4DSynoptic.AddNewStack.}

function T4DSynoptic.AllGridsSameStructure:boolean;
{------------------------------------------------------------------------------}
{ This function returns true if all of the SynopticGlobals in the object have  }
{ the same spatial grid structure. If there are no data in the object, the     }
{ function returns true.                                                       }
{------------------------------------------------------------------------------}
var
  Hold:TGlobalSynopticStack;
  FirstGrid:TSynopticGlobal;
  I,J:integer;
begin
  Result:=true;
  FirstGrid:=GetStackObject(1).GetSynopticGlobal(1);
  for I:=1 to Count do
  begin
    Hold:=GetStackObject(I);
    for J:=1 to Hold.Count do
    if not FirstGrid.SameGridStructure(Hold.GetSynopticGlobal(J)) then
    begin
      Result:=false;
      exit;
    end;
  end;
end; {End of T4DSynoptic.AllGridsSameStructure.}

procedure T4DSynoptic.ConfirmLatLongAttributesSet;
begin
  with NewSynopticGlobalDefaults.Header do
  begin
    if NumLat=0 then raise E4DSynoptic.Create('the default number of latitudes has not been set');
    if NumLong=0 then raise E4DSynoptic.Create('the default number of longitude has not been set');
    if StartLat=-999.99 then raise E4DSynoptic.Create('the default start latitude has not been set');
    if StartLong=-999.99 then raise E4DSynoptic.Create('the default start longitude has not been set');
    if DeltaLat=-999.99 then raise E4DSynoptic.Create('the default delta latitude has not been set');
    if DeltaLong=-999.99 then raise E4DSynoptic.Create('the default delta longitude has not been set');
  end;
end; {End of T4DSynoptic.ConfirmLatLongAttributesSet.}

procedure T4DSynoptic.ConfirmOffsetAndScalingSet;
begin
  with NewSynopticGlobalDefaults do
  begin
    if EquivDouble(Scaling,-999.99) then raise E4DSynoptic.Create('the default scaling value has not been set');
    if EquivDouble(Offset,-999.99) then raise E4DSynoptic.Create('the default offset value has not been set');
  end;
end;

function T4DSynoptic.ConfirmSynopticGlobalBuffer(ADateTime:TDateTime;ALevel:double):boolean;
var
  Test:boolean;
begin
  Result:=true;

  {Sometimes FSynopticGlobalBuffer becomes undefined for some reason and so on}
  {10 July 2010 I added the code below to trap and fix that. I did some tests }
  {and confirmed that it works.}
  try
    if not (FSynopticGlobalBuffer=nil) then
      Test:=SameDateTime(ADateTime,FSynopticGlobalBuffer.DateTime);
  except
    on E:Exception do FSynopticGlobalBuffer:=nil;
  end;

  if FSynopticGlobalBuffer=nil then Result:=false else
  begin
    if not Test then Result:=false else
      if not EquivDouble(ALevel,FSynopticGlobalBuffer.Level) then Result:=false;
  end;
end; {End of T4DSynoptic.ConfirmSynopticGlobalBuffer.}

function T4DSynoptic.FindTemporaryStorageFilename:string;
{------------------------------------------------------------------------------}
{ This function finds an unused temporary storage file name. The temporary     }
{ files have extensions added according to the period number of the stack of   }
{ data that they hold, so for now we look to see if any such files exist.      }
{------------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variabile through file names.}
  DirInfo:TSearchRec; {Used to see if any temporary storage files exist.}
  AlreadyUsed:boolean; {Used to see if any temporary storage files exist.}
begin
  I:=0;
  repeat
    inc(I);
    Result:='Synoptic4D_junk_'+inttostrf(I,4,true);
    if findfirst(Result+'.*',faAnyFile,DirInfo)=0 then AlreadyUsed:=true else AlreadyUsed:=false;
    findclose(DirInfo);
  until ((not AlreadyUsed) or (I=9999));
  if ((I=9999) and AlreadyUsed) then raise E4DSynoptic.Create('No temporary storage file names left');
end; {End of T4DSynoptic.FindTemporaryStorageFilename.}

function T4DSynoptic.GetDateTimeForInterval(INum:integer):TDateTime;
begin
  Result:=GetStackObject(INum).GetDateStamp;
end; {End of T4DSynoptic.GetDateTimeForInterval.}

procedure T4DSynoptic.GetIntervalAndLevel(ADateTime:TDateTime;ALevel:double;AllowAddition:boolean;out INum,LNum:cardinal);
begin
  if QueryDataAvailableAtDateTime(ADateTime) then INum:=GetIntervalForDateTime(ADateTime) else
    if AllowAddition then INum:=AddInterval(ADateTime,ALevel) else
    raise E4DSynoptic.Create('there were no data available at '+DateTimeToStr(ADateTime,'HH:TT:SS DD/MM/YYYY'));
  if GetStackObject(INum).DataAvailableForLevel(ALevel) then LNum:=GetLevelNumForLevel(INum,ALevel) else
    if AllowAddition then LNum:=GetStackObject(INum).AddLevel(ALevel) else
    raise E4DSynoptic.Create('there were no data available at Level='+floattostrf(ALevel,ffFixed,10,5));
end; {End of T4DSynoptic.GetIntervalAndLevel.}

function T4DSynoptic.GetIntervalForDateTime(ADateTime:TDateTime):cardinal;
begin
  Result:=GetStackNumForDateTime(ADateTime);
end; {End of T4DSynoptic.GetIntervalForDateTime.}

function T4DSynoptic.GetIntervalOnOrBeforeDateTime(ADateTime:TDateTime):cardinal;
begin
  if Count=0 then Result:=0 else
  begin
    Result:=Count+1;
    repeat dec(Result) until ((ADateTime>=GetDateTimeForInterval(Result)) or (Result=1));
    if ADateTime<GetDateTimeForInterval(Result) then Result:=0;
  end;
end; {End of T4DSynoptic.GetIntervalOnOrBeforeDateTime.}

function T4DSynoptic.GetLatitude(I:word):double;
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).GetLatitude(I);
end; {End of T4DSynoptic.GetLatitude.}

function T4DSynoptic.GetLevelNumForLevel(INum:cardinal;ALevel:double):cardinal;
begin
  Result:=GetStackObject(INum).GetLevelNum(ALevel);
end; {End of T4DSynoptic.GetLevelNumForLevel.}

function T4DSynoptic.GetLongitude(I:word):double;
begin
  if not AllGridsSameStructure then raise E4DSynoptic.Create(
    'all of the data grids do not have the same spatial structure.' );
  Result:=GetStackObject(1).GetSynopticGlobal(1).GetLongitude(I);
end; {End of T4DSynoptic.GetLongitude.}

function T4DSynoptic.GetStackNumForDateTime(ADateTime:TDateTime):longint;
{------------------------------------------------------------------------------}
{ This function returns the stack number for the passed date and time. If no   }
{ stack with a corresponding date and time exists, the function raises an      }
{ exception.                                                                   }
{------------------------------------------------------------------------------}
var
  I:longint;
begin
  Result:=0;
  for I:=1 to Count do
    if SameDateTime(GetDateTimeForInterval(I),ADateTime) then Result:=I;
  if Result=0 then raise E4DSynoptic.Create('there were no data available at '+DateTimeToStr(ADateTime,'HH:TT:SS DD/MM/YYYY'));
end; {End of T4DSynoptic.GetStackNumForDateTime.}

function T4DSynoptic.GetFirstStackNumOnDate(ADateTime:TDateTime):longint;
{------------------------------------------------------------------------------}
{ This function returns the stack number for the first stack occurring on the  }
{ date passed via ADateTime. If no stack exists on this date, the function     }
{ raises an exception.                                                         }
{------------------------------------------------------------------------------}
var
  I:longint;
begin
  Result:=0;
  for I:=1 to Count do
    if SameDate(GetDateTimeForInterval(I),ADateTime) then Result:=I;
  if Result=0 then raise E4DSynoptic.Create('there were no data available on '+DateTimeToStr(ADateTime,'DD/MM/YYYY'));
end; {End of T4DSynoptic.GetFirstStackNumOnDate.}

function T4DSynoptic.GetStackObject(INum:integer):TGlobalSynopticStack;
{------------------------------------------------------------------------------}
{ This function returns the TGlobalSynopticStack object corresponding to the   }
{ time interval specified by INum. This stack object may or may not be in      }
{ DataOnDisk mode.                                                             }
{------------------------------------------------------------------------------}
begin
  if ((INum<1) or (INum>Count)) then raise E4DSynoptic.Create('the interval number was out of range.');
  Result:=TGlobalSynopticStack(Items[INum-1]);
end; {End of T4DSynoptic.GetStackObject.}

procedure T4DSynoptic.Notify(Ptr:Pointer;Action:TListNotification);
{------------------------------------------------------------------------------}
{ This procedure is called when items are added, removed, or extracted from    }
{ the list. The main job here is to perform memory management.                 }
{------------------------------------------------------------------------------}
var
  INum,
  Delta,
  I:longint;
  StartIndex:cardinal;
begin
  if (FMemRestriction and (GetMemoryUsed>FMaxMemory)) then
  begin
    INum:=IndexOf(Ptr);
    if INum=0 then StartIndex:=Count-1 else StartIndex:=0;
    if INum=0 then Delta:=-1 else Delta:=1;
    I:=StartIndex;
    repeat
      if not TGlobalSynopticStack(Items[I]).DataOnDisk then
        TGlobalSynopticStack(Items[I]).DataOnDisk:=true;
      I:=I+Delta;
    until GetMemoryUsed<=FMaxMemory;
  end;
  inherited Notify(Ptr,Action);
end; {End of T4DSynoptic.Notify.}

procedure T4DSynoptic.ReplaceSynopticGlobalByIndices(INum,LNum:cardinal;Value:TSynopticGlobal);
begin
  GetStackObject(INum).PutSynopticGlobal(LNum,Value);
end; {End of T4DSynoptic.ReplaceSynopticGlobalByIndices.}

procedure T4DSynoptic.RestrictMemoryTo(MaxMem:cardinal);
begin
  if MaxMem<100 then raise E4DSynoptic.Create('maxmem is less than 100. Don''t be an idiot.');
  FMaxMemory:=MaxMem;
  MemoryRestricted:=true;
end; {End of T4DSynoptic.RestrictMemoryTo.}

procedure T4DSynoptic.RestrictMemory(Value:boolean);
{------------------------------------------------------------------------------}
{ This procedure switches memory restriction within the object on or off. The  }
{ procedure will most likely be called to switch memory restriction off since  }
{ setting the MaxMemory property is used to switch the memory restriction on.  }
{ However, because this routine does not change the value of MaxMemory, the    }
{ routine can be called sequentially to toggle memory restriction (to the set  }
{ value of MaxMemory) on an off. If MaxMemory has not been set and this        }
{ procedure is called, an exception is raised.                                 }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  if FMaxMemory=0 then raise E4DSynoptic.Create('a maximum memory value has not been set.');
  FMemRestriction:=Value;
  if FMemRestriction then
  begin
    if TempFilename='' then TempFilename:=FindTemporaryStorageFilename;
    for I:=1 to Count do GetStackObject(I).StorageFile:=TempFilename+'.'+inttostrf(I,5,true);
    I:=0;
    while GetMemoryUsed>FMaxMemory do
    begin
      inc(I);
      GetStackObject(I).DataOnDisk:=true;
    end;
  end
  else for I:=1 to Count do GetStackObject(I).DataOnDisk:=false;
end; {End of T4DSynoptic.SwitchMemoryRestriction.}

{$WARNINGS OFF}
procedure T4DSynoptic.WriteExtendToPole(Value:boolean);
var
  I:integer;
begin
  FExtendToPole:=Value;
  try
    for I:=1 to Count do GetStackObject(I).ExtendToPole:=Value;
  except
    on E:Exception do raise E4DSynoptic.Create('The call to WriteExtendToPole failed at I='+
    inttostr(I)+' and the error message was '+E.Message);
  end;
end; {End of T4DSynoptic.WriteExtendToPole.}
{$WARNINGS ON}

procedure T4DSynoptic.WriteConfineToLevels(Value:boolean);
var
  I:integer;
begin
  FConfineToLevels:=Value;
  for I:=1 to Count do GetStackObject(I).ConfineToLevels:=Value;
end; {End of T4DSynoptic.WriteConfineToLevels.}

procedure T4DSynoptic.SetNetCDFLevelIDStr(Value:string);
begin
  FNetCDFLevelIDStr:=LowerCase(Value);
end;

procedure T4DSynoptic.SetNetCDFTimeIDStr(Value:string);
begin
  FNetCDFTimeIDStr:=LowerCase(Value);
end;

procedure T4DSynoptic.SetNetCDFVarStr(Value:string);
begin
  FNetCDFVarStr:=Value;
end;

end.
