//@***h* BDBP/gb_AutocorrelationModel.pas
// 
// NAME
//   gb_AutocorrelationModel.pas
//
// PURPOSE
//  Deals with all aspects of an autocorrelation model that is needed for a multiple
//  linear regression model.
//
// DESCRIPTION
//  This unit deals with all aspects of an autocorrelation model that is needed for 
//  a multiple linear regression model.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_AutocorrelationModel;

{$I UNITDIREC.INC}

interface

uses
  SysUtils,
  gb_array,
  gb_TripletArray;

type
  TAutoCorrelationModel=class
  private
    {This field holds the autocorrelation coefficients.}
    FCoeffs:T1DDoubleArray;

    {This object holds all of the data to which the basis functions will be}
    {fitted. The object is created in the constructor and destroyed in the }
    {destructor.}
    FData:TRegressionDataArray;

    {FOrder specifies the order of the autocorrelation model. If FOrder=2 this}
    {means that the Nth residual is modelled as being some function of the    }
    {N-1th and N-2th residuals.}
    FOrder:word;

    {The number of coefficients in the autocorrelation model.}
    FNumCoeff:word;
  public
    constructor Create(Measurements:TRegressionDataArray);virtual;
    function  GenerateTransformedDataSeries(Data:T1DDoubleArray;FreeInputData:boolean):T1DDoubleArray;virtual;abstract;
    function  GenerateRevisedMeasurementErrors(OrigMeasurementErrors,MonthlyStdv:T1DDoubleArray):T1DDoubleArray;virtual;abstract;
    function  GetBasisFuncVector(I:integer):TDoubleArray;virtual;abstract;
    procedure Run;
    destructor Destroy;override;
  end;

  TAutoCorrelationModelClass=class of TAutoCorrelationModel;

  EAutoCorrelationModel=class(exception);

  TTwoTermAutoCorrelation=class(TAutoCorrelationModel)
  public
    constructor Create(Measurements:TRegressionDataArray);override;
    function GenerateTransformedDataSeries(Data:T1DDoubleArray;FreeInputData:boolean):T1DDoubleArray;override;
    function GenerateRevisedMeasurementErrors(OrigMeasurementErrors,MonthlyStdv:T1DDoubleArray):T1DDoubleArray;override;
    function GetBasisFuncVector(I:integer):TDoubleArray;override;
  end;

implementation

uses
  DateUtils,
  gb_math;

constructor TAutoCorrelationModel.Create(Measurements:TRegressionDataArray);
var
  I:integer;
begin
  FData:=TRegressionDataArray.Create;

  {We replace the measurements with the residuals, assuming that the errors on }
  {the measurements will also be the errors on the residuals, and that constit-}
  {utes all of the data that the autocorrelation model requires.}
  FData.DeepCopy(Measurements);
  for I:=1 to FData.Count do FData.PutMeasurement(I,FData.GetResidual(I));

  FNumCoeff:=0; {This value MUST be overwritten by the descendent constructor.}
  FOrder:=0; {This value MUST be overwritten by the descendent constructor.}
end;

procedure TAutocorrelationModel.Run;
var
  Beta:T1DDoubleArray;
  Covar:T2DDoubleArray;
  Sigma2i,
  Weight:double;
  I,J,K:integer;
  BasisFunc:TDoubleArray; {The basis function vector evaluated at a specific time.}
begin
  if FNumCoeff=0 then raise EAutocorrelationModel.Create(
    'the descendent object has not set the number of coefficients in its constructor.');
  if FOrder=0 then raise EAutocorrelationModel.Create(
    'the descendent object has not set the order for the autocorrelation model in its constructor.');
  Beta:=T1DDoubleArray.Create(FNumCoeff);
  Covar:=T2DDoubleArray.Create(FNumCoeff,FNumCoeff);
  FreeAndNil(FCoeffs);
  FCoeffs:=T1DDoubleArray.Create(FNumCoeff);
  try
    for I:=1 to FNumCoeff do
    begin
      for J:=1 to FNumCoeff do Covar[I,J]:=0.0;
      Beta[I]:=0.0;
    end;
    for I:=1+FOrder to FData.Count do
    begin
      BasisFunc:=GetBasisFuncVector(I);
      Sigma2i:=1.0/sqr(FData.GetMeasurementError(I));
      for J:=1 to FNumCoeff do
      begin
        Weight:=BasisFunc[J-1]*Sigma2i;
        for K:=1 to J do Covar[J,K]:=Covar[J,K]+(Weight*BasisFunc[K-1]);
        Beta[J]:=Beta[J]+(Weight*FData.GetMeasurement(I));
      end;
    end;
    for I:=2 to FNumCoeff do
      for J:=1 to I-1 do
        Covar[J,I]:=Covar[I,J];
    GaussJordan(Covar,Beta);
    for I:=1 to FNumCoeff do FCoeffs[I]:=Beta[I];
  finally
    FreeAndNil(Beta);
    FreeAndNil(Covar);
  end;
end;

destructor TAutoCorrelationModel.Destroy;
begin
  FreeAndNil(FData);
  FreeAndNil(FCoeffs);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for the TTwoTermAutoCorrelation class.      }
{------------------------------------------------------------------------------}
constructor TTwoTermAutoCorrelation.Create(Measurements:TRegressionDataArray);
begin
  inherited Create(Measurements);
  FNumCoeff:=2;
  FOrder:=2;
end;

function TTwoTermAutoCorrelation.GenerateTransformedDataSeries(Data:T1DDoubleArray;FreeInputData:boolean):T1DDoubleArray;
var
  I:integer;
begin
  Result:=T1DDoubleArray.Create(Data.NumData);
  for I:=1 to Data.NumData do
  begin
    if I=1 then Result[1]:=Data[1]*(1-FCoeffs[2]-FCoeffs[1]);
    if I=2 then Result[2]:=Data[2]-Data[1]*FCoeffs[2]-Data[1]*FCoeffs[1];
    if I>2 then Result[I]:=Data[I]-Data[I-1]*FCoeffs[2]-Data[I-2]*FCoeffs[1];
  end;
  if FreeInputData then FreeAndNil(Data);
end;

function TTwoTermAutoCorrelation.GenerateRevisedMeasurementErrors(OrigMeasurementErrors,MonthlyStdv:T1DDoubleArray):T1DDoubleArray;
var
  TermA,
  TermB:double;
  Month:word;
  I:integer;
begin
  Result:=T1DDoubleArray.Create(FData.Count);
  for I:=1 to OrigMeasurementErrors.NumData do
  try
    Month:=MonthOf(FData.GetDateTime(I));
    TermA:=(1-FCoeffs[1])*sqr(MonthlyStdv[Month])/((1+FCoeffs[1])*(sqr(1-FCoeffs[1])-sqr(FCoeffs[2])));
    if TermA<0.0 then TermA:=0.0; {We assume that the inclusion of the data variance whould never reduce the error on the measurements.}
    TermB:=sqr(OrigMeasurementErrors[I]);
    Result[I]:=sqrt(TermA+TermB);
  except
    on E:Exception do raise EAutoCorrelationModel.Create(
      'an error occurred in GenerateRevisedMeasurementErrors at I='+inttostr(I));
  end;
end;

function TTwoTermAutoCorrelation.GetBasisFuncVector(I:integer):TDoubleArray;
begin
  setlength(Result,2);
  Result[0]:=FData.GetMeasurement(I-2);
  Result[1]:=FData.GetMeasurement(I-1);
end;

end.
