//@***h* BDBP/gb_BasisFunctions.pas
// 
// NAME
//   gb_BasisFunctions.pas
//
// PURPOSE
//  Contains all subroutines that deal with preparing, reading and handling the basis 
//  functions for the multiple linear regression model.
//
// DESCRIPTION
//  The unit contains all subroutines that deal with preparing, reading and handling 
//  the basis functions for the multiple linear regression model.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_BasisFunctions;

{$I UNITDIREC.INC}

interface

uses
  Classes,
  gb_RegressionBasisFunction,
  gb_array,
  gb_EESCServer,
  gb_g1binary,
  gb_g1means,
  gb_g2binary,
  gb_TripletArray,
  gb_SynopticGlobal,
  gb_XYSignal;


type
  TYearTriplet=record
    StartYear,
    MiddleYear,
    EndYear:double;
  end;

  TOffsetBasisFunction=class(TBasisFunction)
  public
    constructor Create;overload;
    constructor Create(ANumFourier:word);overload;
    function GetValue(ADateTime:TDateTime):double;override;
  end;

  TTrendBasisFunction=class(TBasisFunction)
  public
    constructor Create;overload;
    constructor Create(ANumFourier:word);overload;
    function GetValue(ADateTime:TDateTime):double;override;
  end;

  TTruncatedTrendBasisFunction=class(TBasisFunction)
  strict private
    FStartDateTime,
    FEndDateTime:TDateTime;
  public
    constructor Create(AStartDateTime,AEndDateTime:TDateTime);overload;
    constructor Create(AStartDateTime,AEndDateTime:TDateTime;ANumFourier:word);overload;
    function GetValue(ADateTime:TDateTime):double;override;
    property StartDateTime:TDateTime read FStartDateTime write FStartDateTime;
    property EndDateTime:TDateTime read FEndDateTime write FEndDateTime;
  end;

  TTriangleBasisFunction=class(TBasisFunction)
  strict private
    FYearTriplet:TYearTriplet;
  public
    constructor Create(AYearTriplet:TYearTriplet);overload;
    constructor Create(AYearTriplet:TYearTriplet;ANumFourier:word);overload;
    function GetValue(ADateTime:TDateTime):double;override;
  end;

  TCO2BasisFunction=class(TBasisFunction)
  strict private
    Data:array[1850..2500] of double;
    FileData:TStringList;
    function DetermineFileType:byte;
  public
    constructor Create(Filename:string);overload;
    constructor Create(ANumFourier:word;Filename:string);overload;
    function  GetValue(ADateTime:TDateTime):double;override;
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
    destructor Destroy;
  end;

  TTruncatedCO2BasisFunction=class(TCO2BasisFunction)
  strict private
    FStartDateTime,
    FEndDateTime:TDateTime;
  public
    constructor Create(AStartDateTime,AEndDateTime:TDateTime;Filename:string);overload;
    constructor Create(AStartDateTime,AEndDateTime:TDateTime;ANumFourier:word;Filename:string);overload;
    function  GetValue(ADateTime:TDateTime):double;override;
  end;

  TEESCBasisFunction=class(TBasisFunction)
  strict private
    Data:TG1Binary;
    XYData:TXYSignal;
  public
    constructor Create(SourceTG1Binary:string);overload;
    constructor CreateFromNewman(ANumFourier:word;TextFilename:string);
    constructor Create(ANumFourier:word;SourceTG1Binary:string);overload;
    function  GetValue(ADateTime:TDateTime):double;override;
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
    destructor Destroy;override;
  end;

  TEESCBasisFunctionWithAoA_SpecifyLatAlt=class(TBasisFunction)
  strict private
    FLatitude,
    FAltitude:double;
    FPathForEESCFiles:string;
    EESCServer:TEESCServer;
    procedure LoadTheData;
  private
    FStartYear,
    FEndYear:word;
    Data:T2DSingleArray;
  public
    constructor Create(PathForEESCFiles:string);overload;
    constructor Create(ANumFourier:word;PathForEESCFiles:string);overload;
    procedure SetLatAndAlt(Latitude,Altitude:double);
    procedure SetApproximateStartAndEndYear(StartYear,EndYear:word);
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
    function  GetValue(ADateTime:TDateTime):double;override;
    destructor Destroy;override;
  end;

  TEESCBasisFunctionWithAoA_SpecifyAlt=class(TBasisFunction)
  strict private
    FAltitude:double;
    FPathForEESCFiles:string;
    EESCServer:TEESCServer;
    Header:TSynopticGlobalHeader;
    FWithRespectTo1980:boolean;
    procedure LoadTheData;
    procedure WriteWRT1980(Value:boolean);
  private
    FStartYear,
    FEndYear:word;

    {I use a TSynopticGlobal here to hold the EESC data, substituting longitude}
    {for time to facilitate interpolation in latitude and time.}
    Data:TSynopticGlobal;
  public
    constructor Create(PathForEESCFiles:string);overload;
    constructor Create(ANumFourier:word;PathForEESCFiles:string);overload;
    function GetBasisVector(ADateTime:TDateTime;AInfoObj:TInfoObject):TDoubleArray;
    function GetTreatedValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;override;
    procedure SetAlt(Altitude:double);
    procedure SetApproximateStartAndEndYear(StartYear,EndYear:word);
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
    function  GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;override;
    destructor Destroy;override;
    property WithRespectTo1980:boolean read FWithRespectTo1980 write WriteWRT1980;
  end;

  TEESCSquaredBasisFunctionWithAoA_SpecifyLatAlt=class(TEESCBasisFunctionWithAoA_SpecifyLatAlt)
  public
    constructor Create(ANumFourier:word;PathForEESCFiles:string);overload;
    function GetValue(ADateTime:TDateTime):double;override;
  end;

  TPinatuboBasisFunction=class(TBasisFunction)
  strict private
    FSeedYear,
    FStartYear:double;
  public
    constructor Create;overload;
    constructor Create(ANumFourier,ANumSteps:word);overload;
    constructor Create(ANumFourier,ANumSteps:word;ASeedYear:double);overload;
    function  GetValue(ADateTime:TDateTime):double;override;
    procedure SetStep(N:word);override;
  end;

  TPinatuboBasisFunctionWithPresOffset=class(TBasisFunction)
  strict private
    FStartYear:double;
    FLevel:word;

    {I use a TSynopticGlobal here to hold the offset data, substituting longitude}
    {for level to facilitate interpolation in latitude.}
    FOffset:TSynopticGlobal;
  public
    constructor Create;overload;
    constructor Create(OffsetFilename:string);overload;
    constructor Create(ANumFourier:word;OffsetFilename:string);overload;
    procedure SetLevel(ALevel:word);
    function  GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;override;
    destructor Destroy;override;
  end;

  TElChichonBasisFunction=class(TBasisFunction)
  strict private
    FSeedYear,
    FStartYear:double;
  public
    constructor Create;overload;
    constructor Create(ANumFourier,ANumSteps:word);overload;
    constructor Create(ANumFourier,ANumSteps:word;ASeedYear:double);overload;
    function  GetValue(ADateTime:TDateTime):double;override;
    procedure SetStep(N:word);override;
    property StartYear:double read FStartYear write FStartYear;
  end;

  TQBOandSOIBasisFunction=class(TBasisFunction)
  protected
    IndicesFilename:string;
    Data:TG2Binary;
  public
    procedure LoadTheDataFromFile;virtual;abstract;
    procedure DownloadNewData;virtual;abstract;
    procedure GetStartAndEndTimes(var StartDateTime,EndDateTime:TDateTime);
    function  TestResolution:boolean;override;
    function  GetValue(ADateTime:TDateTime):double;override;
    function  InsufficientCoverage(ADateTimeList:TDoubleArray):boolean;
    procedure ClimatologicalExtension(FirstYear,LastYear:word);
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
    destructor Destroy;override;
  end;

  TQBOBasisFunction=class(TQBOandSOIBasisFunction)
  public
    constructor Create(QBOIndicesFile:string);overload;
    constructor Create(ANumFourier:word;QBOIndicesFile:string);overload;
    procedure WriteDataGridToTextfile(Filename:string);
    procedure LoadTheDataFromFile;override;
    procedure DownloadNewData;override;
  end;

  TQBOOrthogBasisFunction=class(TQBOBasisFunction)
  public
    constructor Create(QBOIndicesFile:string);overload;
    constructor Create(ANumFourier:word;QBOIndicesFile:string);overload;
    procedure OrthogTheData;
  end;

  TSOIBasisFunction=class(TQBOandSOIBasisFunction)
  public
    constructor Create(SOIIndicesFile:string);overload;
    constructor Create(ANumFourier:word;SOIIndicesFile:string);overload;
    procedure LoadTheDataFromFile;override;
    procedure DownloadNewData;override;
  end;

  TSolarBasisFunction=class(TBasisFunction)
  strict private
    Data:TG1Mean;
    FilePath:string;
    procedure LoadTheDataFromFile;
    procedure DownloadNewData;
  public
    constructor Create(PathForSolarFiles:string);overload;
    constructor Create(ANumFourier:word;PathForSolarFiles:string);overload;
    function  TestResolution:boolean;override;
    function  IsNull(ADateTime:TDateTime):boolean;
    function  GetValue(ADateTime:TDateTime):double;override;
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
    destructor Destroy;override;
  end;

  TGlobalSurfTempType=(global,nh_ocean,nh_land,sh_ocean,sh_land);

  TGlobalSurfTemp=class(TBasisFunction)
  type
    TDataRec=record
      Year:word;
      Value:double;
    end;
  private
    FBaseLineStartYear,
    FBaseLineEndYear,
    FirstYear,
    LastYear:word;
    Data:array of TDataRec;
  public
    constructor Create(Filename:string;AnomStartYear,AnomEndYear:word;AType:TGlobalSurfTempType);overload;
    constructor Create(ANumFourier:word;Filename:string;AnomStartYear,AnomEndYear:word;AType:TGlobalSurfTempType);overload;
    function  GetValue(ADateTime:TDateTime):double;override;
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
  end;

  TColumnSet=set of 1..10;

  TCH4AndN20BasisFunction=class(TBasisFunction)
  type
    TDataRec=record
      Year:word;
      Value:double;
    end;
  private
    FirstYear,
    LastYear:word;
    Data:array of TDataRec;
  public
    constructor Create(FileList:TStringList;Columns:TColumnSet);overload;
    constructor Create(ANumFourier:word;FileList:TStringList;Columns:TColumnSet);overload;
    function  GetValue(ADateTime:TDateTime):double;override;
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);override;
  end;

implementation

//{$DEFINE ALWAYSCLIM}

uses
  Controls,
  DateUtils,
  Dialogs,
  Math,
  SysUtils,
  IdHTTP,
  idftp,
  idsocks,
  gb_datetime,
  gb_files,
  gb_GeneralConstantsAndTypes,
  gb_strings;

type
  TQBOSOIPhase=class
  strict private
    Data:TXYSignalArray;
    PhaseNum:word;
    MeanPhase:TDoubleArray;
    function GetMeanPhaseLength:word;
  public
    constructor Create;
    procedure NextPhase;
    procedure PutValue(Value:double);
    procedure DropLastPhase;
    procedure CalculateAveragePhase;
    function  GetMeanPhaseValue(Index:integer):double;
    destructor Destroy;override;

    property MeanPhaseLength:word read GetMeanPhaseLength;
  end;

{------------------------------------------------------------------------------}
{ TQBOSOIPhase                                                                 }
{------------------------------------------------------------------------------}
constructor TQBOSOIPhase.Create;
begin
  PhaseNum:=0;
  setlength(Data,0);
  setlength(MeanPhase,0);
end;

procedure TQBOSOIPhase.NextPhase;
begin
  inc(PhaseNum);
  setlength(Data,PhaseNum);
  Data[PhaseNum-1]:=TXYSignal.Create(-999.99,-999.99);
end;

procedure TQBOSOIPhase.PutValue(Value:double);
var
  I:integer;
begin
  I:=Data[PhaseNum-1].GetNumData+1;
  Data[PhaseNum-1].AddCoord(I,Value);
end;

procedure TQBOSOIPhase.DropLastPhase;
begin
  dec(PhaseNum);
  FreeAndNil(Data[PhaseNum]);
  setlength(Data,PhaseNum);
end;

procedure TQBOSOIPhase.CalculateAveragePhase;
var
  MeanPhaseLength,
  NumPhases,
  I,J:integer;
  Num:gb_array.TWordArray;
begin
  MeanPhaseLength:=0;
  NumPhases:=0;
  for I:=Low(Data) to High(Data) do
  if Data[I].GetNumData>0 then
  begin
    MeanPhaseLength:=MeanPhaseLength+Data[I].GetNumData;
    inc(NumPhases);
  end;
  if NumPhases=0 then raise Exception.Create('there were insufficients phases to'+
    ' calculate the mean phase length.');
  MeanPhaseLength:=round(MeanPhaseLength/NumPhases);
  setlength(MeanPhase,MeanPhaseLength);
  setlength(Num,MeanPhaseLength);
  for I:=1 to MeanPhaseLength do
  begin
    MeanPhase[I-1]:=0.0;
    Num[I-1]:=0;
  end;
  for I:=Low(Data) to High(Data) do
    for J:=0 to MeanPhaseLength-1 do
    if J<Data[I].GetNumData then
    begin
      MeanPhase[J]:=MeanPhase[J]+Data[I].GetYValue(J+1);
      inc(Num[J]);
    end;
  for J:=0 to MeanPhaseLength-1 do MeanPhase[J]:=MeanPhase[J]/Num[J];
end; {End of TQBOSOIPhase.GetAveragePhase.}

function TQBOSOIPhase.GetMeanPhaseValue(Index:integer):double;
begin
  if length(MeanPhase)=0 then CalculateAveragePhase;
  Result:=MeanPhase[Index-1];
end;

function TQBOSOIPhase.GetMeanPhaseLength:word;
begin
  Result:=length(MeanPhase);
end;

destructor TQBOSOIPhase.Destroy;
var
  I:integer;
begin
  for I:=Low(Data) to High(Data) do FreeAndNil(Data[I]);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TOffsetBasisFunction                                                         }
{------------------------------------------------------------------------------}
constructor TOffsetBasisFunction.Create;
begin
  Create(0);
end;

constructor TOffsetBasisFunction.Create(ANumFourier:word);
begin
  inherited Create(FillInitRec('Offset',ANumFourier,DontChange));
end;

function TOffsetBasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  Result:=1.0;
end;

{------------------------------------------------------------------------------}
{ TTrendBasisFunction                                                          }
{------------------------------------------------------------------------------}
constructor TTrendBasisFunction.Create;
begin
  Create(0);
end;

constructor TTrendBasisFunction.Create(ANumFourier:word);
begin
  inherited Create(FillInitRec('Trend',ANumFourier,RemoveOffset));
end;

function TTrendBasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  Result:=ConvertDateTimeToDecYear(ADateTime)-1980;
end;

{------------------------------------------------------------------------------}
{ TTruncatedTrendBasisFunction                                                 }
{------------------------------------------------------------------------------}
constructor TTruncatedTrendBasisFunction.Create(AStartDateTime,AEndDateTime:TDateTime);
begin
  Create(AStartDateTime,AEndDateTime,0);
end;

constructor TTruncatedTrendBasisFunction.Create(AStartDateTime,AEndDateTime:TDateTime;ANumFourier:word);
begin
  inherited Create(FillInitRec('TruncatedTrend',ANumFourier,RemoveOffset));
  FStartDateTime:=AStartDateTime;
  FEndDateTime:=AEndDateTime;
end;

function TTruncatedTrendBasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  if ((ADateTime<FStartDateTime) or (ADateTime>FEndDateTime)) then Result:=0.0
  else Result:=ConvertDateTimeToDecYear(ADateTime)-ConvertDateTimeToDecYear(FStartDateTime);
end;

{------------------------------------------------------------------------------}
{ TTriangleBasisFunction.                                                      }
{------------------------------------------------------------------------------}
constructor TTriangleBasisFunction.Create(AYearTriplet:TYearTriplet);
begin
  Create(AYearTriplet,0);
end;

constructor TTriangleBasisFunction.Create(AYearTriplet:TYearTriplet;ANumFourier:word);
begin
  inherited Create(FillInitRec('Triangle',ANumFourier,DontChange));
  FYearTriplet:=AYearTriplet;
end;

function TTriangleBasisFunction.GetValue(ADateTime:TDateTime):double;
var
  DecYear:double;
begin
  DecYear:=ConvertDateTimeToDecYear(ADateTime)+0.5;
  with FYearTriplet do
    if ((DecYear>=StartYear) and (DecYear<=MiddleYear)) then Result:=1-(MiddleYear-DecYear)/(MiddleYear-StartYear) else
      if ((DecYear>MiddleYear) and (DecYear<=EndYear)) then Result:=1-(DecYear-MiddleYear)/(EndYear-MiddleYear) else
         Result:=0.0;
end;

{------------------------------------------------------------------------------}
{ TCO2BasisFunction basis                                                      }
{------------------------------------------------------------------------------}
function TCO2BasisFunction.DetermineFileType:byte;
begin
  Result:=0;
  if ((FileData[0]='') and (GetWord(FileData[2],' ',1)='CONTENT:') and
    (GetWord(FileData[3],' ',1)='RUN:')) then Result:=1;
end;

constructor TCO2BasisFunction.Create(Filename:string);
begin
  Create(0,Filename);
end;

constructor TCO2BasisFunction.Create(ANumFourier:word;Filename:string);
var
  DecYear,Value:double;
  FirstDataRow,
  I:integer;
  Test:string;
begin
  inherited Create(FillInitRec('CO2',ANumFourier,RemoveOffset));
  FileData:=TStringList.Create;
  FileData.LoadFromFile(Filename);
  case DetermineFileType of
    1:begin {CMIP5 standard output file.}
        I:=0;
        repeat inc(I) until pos('THISFILE_FIRSTDATAROW',FileData[I])>0;
        FirstDataRow:=strtoint(EraseEndChar(GetWord(FileData[I],' ',3),','));
        for I:=FirstDataRow to FileData.Count do
        if FileData[I-1]<>'' then
        begin
          DecYear:=StrToFloat(GetWord(FileData[I-1],' ',1));
          Value:=StrToFloat(GetWord(FileData[I-1],' ',4));
          if ((trunc(DecYear)>=1850) and (trunc(DecYear)<=2500)) then
            Data[trunc(DecYear)]:=Value;
        end;
      end;
  else {Assume generic file of Year and CO2.}
    for I:=1 to FileData.Count do
    begin
      DecYear:=StrToFloat(GetWord(FileData[I-1],' ',1));
      Value:=StrToFloat(GetWord(FileData[I-1],' ',2));
      if ((trunc(DecYear)>=1850) and (trunc(DecYear)<=2500)) then
        Data[trunc(DecYear)]:=Value;
    end;
  end;
end; {End of TCO2BasisFunction.Create.}

function TCO2BasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  Result:=Data[YearOf(ADateTime)];
end;

procedure TCO2BasisFunction.TestDateTimeList(ADateTimeList:TDoubleArray);
begin
  if YearOf(ADateTimeList[0])<1850 then raise EBasisFunction.Create(
    'the ESC basis function start date doesn''t cover the start of the measurements.');
  if YearOf(ADateTimeList[High(ADateTimeList)])>2500 then raise EBasisFunction.Create(
    'the ESC basis function end date doesn''t cover the end of the measurements.');
end;

destructor TCO2BasisFunction.Destroy;
begin
  FreeAndNil(FileData);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TTruncatedCO2BasisFunction                                                   }
{------------------------------------------------------------------------------}
constructor TTruncatedCO2BasisFunction.Create(AStartDateTime,AEndDateTime:TDateTime;Filename:string);
begin
  Create(AStartDateTime,AEndDateTime,0,Filename);
end; {End of TTruncatedCO2BasisFunction.Create.}

constructor TTruncatedCO2BasisFunction.Create(AStartDateTime,AEndDateTime:TDateTime;ANumFourier:word;Filename:string);
begin
  inherited Create(ANumFourier,Filename);
  FStartDateTime:=AStartDateTime;
  FEndDateTime:=AEndDateTime;
end; {End of TTruncatedCO2BasisFunction.Create.}

function TTruncatedCO2BasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  if ((ADateTime<FStartDateTime) or (ADateTime>FEndDateTime)) then Result:=0.0
  else Result:=inherited GetValue(ADateTime)-inherited GetValue(FStartDateTime);
end; {End of TTruncatedCO2BasisFunction.GetValue.}

{------------------------------------------------------------------------------}
{ TEESCBasisFunction                                                           }
{------------------------------------------------------------------------------}
constructor TEESCBasisFunction.Create(SourceTG1Binary:string);
begin
  Create(0,SourceTG1Binary);
end;

constructor TEESCBasisFunction.CreateFromNewman(ANumFourier:word;TextFilename:string);
var
  Source:textfile;
  I:integer;
  DecYear,Value:double;
begin
  inherited Create(FillInitRec('EESC',ANumFourier,RemoveOffset));
  if not FileExists(textFilename) then raise EBasisFunction(
    'the file named '+TextFilename+' couldn''t be found.');
  Data:=nil;
  XYData:=TXYSignal.Create(-999.99,-999.99);
  assignfile(Source,TextFilename);
  reset(Source);
  for I:=1 to 8 do readln(Source);
  repeat
    readln(Source,DecYear,Value);
    XYData.AddCoord(DecYear,Value/1000);
  until eof(Source);
  closefile(Source);
end; {End of TEESCBasisFunction.CreateFromNewman.}

constructor TEESCBasisFunction.Create(ANumFourier:word;SourceTG1Binary:string);
begin
  inherited Create(FillInitRec('EESC',ANumFourier,RemoveOffset));
  if not FileExists(SourceTG1Binary) then raise EBasisFunction(
    'the file named '+SourceTG1Binary+' couldn''t be found.');
  Data:=TG1Binary.Create(SourceTG1Binary);
  XYData:=nil;
end;

function TEESCBasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  if Data<>nil then Result:=Data.GetInterpolValue(ADateTime);
  if XYData<>nil then Result:=XYData.InterpolateYValue(ConvertDateTimeToDecYear(ADateTime),false,false);
end;

procedure TEESCBasisFunction.TestDateTimeList(ADateTimeList:TDoubleArray);
begin
  if Data<>nil then
  begin
    if ADateTimeList[0]<Data.GetStartDateTime then raise EBasisFunction.Create(
      'the EESC basis function start date doesn''t cover the start of the measurements.');
    if ADateTimeList[High(ADateTimeList)]>Data.GetEndDateTime then raise EBasisFunction.Create(
      'the EESC basis function end date doesn''t cover the end of the measurements.');
  end;
  if XYData<>nil then
  begin
    if ConvertDateTimeToDecYear(ADateTimeList[0])<XYData.GetFirstCoord.XValue then
      raise EBasisFunction.Create('the EESC basis function start date doesn''t cover the start of the measurements.');
    if ConvertDateTimeToDecYear(ADateTimeList[High(ADateTimeList)])>XYData.GetLastCoord.XValue then
      raise EBasisFunction.Create('the EESC basis function end date doesn''t cover the end of the measurements.');
  end;
end;

destructor TEESCBasisFunction.Destroy;
begin
  FreeAndNil(Data);
  FreeAndNil(XYData);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TEESCBasisFunctionWithAoA_SpecifyLatAlt                                                  }
{------------------------------------------------------------------------------}
constructor TEESCBasisFunctionWithAoA_SpecifyLatAlt.Create(PathForEESCFiles:string);
begin
  Create(0,PathForEESCFiles);
end;

constructor TEESCBasisFunctionWithAoA_SpecifyLatAlt.Create(ANumFourier:word;PathForEESCFiles:string);
begin
  inherited Create(FillInitRec('EESC',ANumFourier,RemoveOffset));
  FLatitude:=-999.99;
  FAltitude:=-999.99;
  FStartYear:=1950;
  FEndYear:=2099;
  FPathForEESCFiles:=PathForEESCFiles;
  EESCServer:=TEESCServer.Create(FPathForEESCFiles);
  Data:=T2DSingleArray.Create(FEndYear-FStartYear+1,366);
end;

procedure TEESCBasisFunctionWithAoA_SpecifyLatAlt.LoadTheData;
var
  Year,
  Day:word;
  DecYear:double;
begin
  if ((FLatitude<-999) or (FAltitude<-999)) then raise EBasisFunction.Create(
    'the latitude and altitude values have not been set');
  for Year:=FStartYear to FEndYear do
    for Day:=1 to DaysInAYear(Year) do
    begin
      DecYear:=Year+(Day-0.5)/DaysInAYear(Year);
      Data[Year-FStartYear+1,Day]:=EESCServer.GetEESC(FLatitude,FAltitude,DecYear);
    end;
end;

procedure TEESCBasisFunctionWithAoA_SpecifyLatAlt.SetLatAndAlt(Latitude,Altitude:double);
begin
  if abs(Latitude)>90 then EBasisFunction.Create('the latitude was out of range.');
  if ((Altitude<10) or (Altitude>70)) then EBasisFunction.Create('the altitude was out of range.');
  if not ((FLatitude=Latitude) and (FAltitude=Altitude)) then
  begin
    FLatitude:=Latitude;
    FAltitude:=Altitude;
    LoadTheData;
  end;
end;

procedure TEESCBasisFunctionWithAoA_SpecifyLatAlt.SetApproximateStartAndEndYear(StartYear,EndYear:word);
begin
  if ((StartYear<1950) or (EndYear>2099)) then raise EBasisFunction.Create(
    'the start or end year were out of range.');
  FStartYear:=StartYear;
  FEndYear:=EndYear;
  if Data<>nil then FreeAndNil(Data);
  Data:=T2DSingleArray.Create(FEndYear-FStartYear+1,366);
end;

procedure TEESCBasisFunctionWithAoA_SpecifyLatAlt.TestDateTimeList(ADateTimeList:TDoubleArray);
begin
  if YearOf(ADateTimeList[0])<FStartYear then raise EBasisFunction.Create(
    'the EESC basis function start date doesn''t cover the start of the measurements.');
  if YearOf(ADateTimeList[High(ADateTimeList)])>FEndYear then raise EBasisFunction.Create(
    'the EESC basis function end date doesn''t cover the end of the measurements.');
end;

function TEESCBasisFunctionWithAoA_SpecifyLatAlt.GetValue(ADateTime:TDateTime):double;
begin
  Result:=Data[YearOf(ADateTime)-FStartYear+1,DayOfTheYear(ADateTime)];
end;

destructor TEESCBasisFunctionWithAoA_SpecifyLatAlt.Destroy;
begin
  FreeAndNil(EESCServer);
  FreeAndNil(Data);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TEESCBasisFunctionWithAoA_SpecifyAlt                                         }
{------------------------------------------------------------------------------}
constructor TEESCBasisFunctionWithAoA_SpecifyAlt.Create(PathForEESCFiles:string);
begin
  Create(0,PathForEESCFiles);
end;

constructor TEESCBasisFunctionWithAoA_SpecifyAlt.Create(ANumFourier:word;PathForEESCFiles:string);
begin
  inherited Create(FillInitRec('EESC',ANumFourier,RemoveOffset));
  FAltitude:=-999.99;
  FStartYear:=1950;
  FEndYear:=2099;
  WithRespectTo1980:=false;
  FPathForEESCFiles:=PathForEESCFiles;
  EESCServer:=TEESCServer.Create(FPathForEESCFiles);
  with Header do
  begin
    {Evaluate the EESC function 10 times each year and once at the end.}
    NumLong:=(FEndYear-FStartYear+1)*10+1;
    NumLat:=91;
    DeltaLong:=0.1;
    DeltaLat:=2;
    StartLong:=0;
    StartLat:=-90;
  end;
  Data:=TSynopticGlobal.Create(Header);
end;

function TEESCBasisFunctionWithAoA_SpecifyAlt.GetBasisVector(ADateTime:TDateTime;AInfoObj:TInfoObject):TDoubleArray;
var
  I:integer;
  DecYear:double;
begin
  {We don't call GetNumCoefficients in the next line because descendent objects}
  {may overwrite that method and we don't want to call the descendent object's }
  {method.}
  setlength(Result,1+NumFourier*2);

  Result[0]:=GetTreatedValue(ADateTime,AInfoObj);
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  for I:=1 to NumFourier do
  begin
    Result[I*2-1]:=Result[0]*sin(2*pi*I*DecYear);
    Result[I*2]:=Result[0]*cos(2*pi*I*DecYear);
  end;
  if Extender<>nil then Result:=Extender.GetBasisVector(Result,AInfoObj);
end;

function TEESCBasisFunctionWithAoA_SpecifyAlt.GetTreatedValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
var
  ReturnZero:boolean;
  Latitude:double;
  Month,
  Date:word;
begin
  ReturnZero:=false;
  if AInfoObj is TLatitudeInfo then
  begin
    Latitude:=abs((AInfoObj as TLatitudeInfo).Latitude);
    {I include the code below specifically for the generating of the Tier 1.x database.}
    if ((Latitude<25) and (FAltitude<16)) then ReturnZero:=true;
    if ((Latitude<27.5) and (FAltitude<15)) then ReturnZero:=true;
    if ((Latitude<30) and (FAltitude<14)) then ReturnZero:=true;
    if ((Latitude<35) and (FAltitude<13)) then ReturnZero:=true;
    if ((Latitude<40) and (FAltitude<12)) then ReturnZero:=true;
    if ((Latitude<50) and (FAltitude<11)) then ReturnZero:=true;
  end;
  if ReturnZero then Result:=0.0 else
    if FWithRespectTo1980 then
      Result:=GetValue(ADateTime,AInfoObj)-GetValue(RecodeYear(ADateTime,1980),AInfoObj)
    else Result:=inherited GetTreatedValue(ADateTime,AInfoObj);
end;

procedure TEESCBasisFunctionWithAoA_SpecifyAlt.LoadTheData;
var
  Year,
  YearStep,
  LatStep:word;
  Latitude,
  DecYear:double;
begin
  if FAltitude<-999 then raise EBasisFunction.Create('the altitude value has not been set');
  for LatStep:=1 to 91 do
  begin
    Latitude:=-92+LatStep*2;
    for Year:=FStartYear to FEndYear do
      for YearStep:=1 to 10 do
      begin
        DecYear:=Year+(YearStep-1)*0.1;
        Data.PutValue((Year-FStartYear)*10+YearStep,LatStep,EESCServer.GetEESC(Latitude,FAltitude,DecYear));
      end;
    DecYear:=FEndYear+1;
    Data.PutValue(Data.NumLong,LatStep,EESCServer.GetEESC(Latitude,FAltitude,DecYear));
  end;
end;

procedure TEESCBasisFunctionWithAoA_SpecifyAlt.WriteWRT1980(Value:boolean);
begin
  FWithRespectTo1980:=Value;
  if Value then Treatment:=DontChange else Treatment:=RemoveOffset;
end;

procedure TEESCBasisFunctionWithAoA_SpecifyAlt.SetAlt(Altitude:double);
begin
  if ((Altitude<10) or (Altitude>70)) then raise EBasisFunction.Create('the altitude was out of range.');
  if FAltitude<>Altitude then
  begin
    FAltitude:=Altitude;
    LoadTheData;
  end;
end;

procedure TEESCBasisFunctionWithAoA_SpecifyAlt.SetApproximateStartAndEndYear(StartYear,EndYear:word);
begin
  if ((StartYear<1950) or (EndYear>2099)) then raise EBasisFunction.Create(
    'the start or end year were out of range.');
  FStartYear:=StartYear;
  FEndYear:=EndYear;
  if Data<>nil then FreeAndNil(Data);
  Header.NumLong:=(FEndYear-FStartYear+1)*10+1;
  Data:=TSynopticGlobal.Create(Header);
end;

procedure TEESCBasisFunctionWithAoA_SpecifyAlt.TestDateTimeList(ADateTimeList:TDoubleArray);
begin
  if YearOf(ADateTimeList[0])<FStartYear then raise EBasisFunction.Create(
    'the EESC basis function start date doesn''t cover the start of the measurements.');
  if YearOf(ADateTimeList[High(ADateTimeList)])>FEndYear then raise EBasisFunction.Create(
    'the EESC basis function end date doesn''t cover the end of the measurements.');
end;

function TEESCBasisFunctionWithAoA_SpecifyAlt.GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
var
  DecYear:double;
begin
  if AInfoObj=nil then raise EBasisFunction.Create('InfoObj was nil.');
  if not (AInfoObj is TLatitudeInfo) then raise EBasisFunction.Create(
    'InfoObj was not of the correct type.');
  DecYear:=ConvertDateTimeToDecYear(ADateTime)-FStartYear;
  Result:=Data.BiLinear(TLatitudeInfo(AInfoObj).Latitude,DecYear);
end;

destructor TEESCBasisFunctionWithAoA_SpecifyAlt.Destroy;
begin
  FreeAndNil(EESCServer);
  FreeAndNil(Data);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TEESCSquaredBasisFunctionWithAoA_SpecifyLatAlt                               }
{------------------------------------------------------------------------------}
constructor TEESCSquaredBasisFunctionWithAoA_SpecifyLatAlt.Create(ANumFourier:word;PathForEESCFiles:string);
begin
  inherited Create(ANumFourier,PathForEESCFiles);
  Name:='EESC squared';
end;

function TEESCSquaredBasisFunctionWithAoA_SpecifyLatAlt.GetValue(ADateTime:TDateTime):double;
begin
  Result:=sqr(Data[YearOf(ADateTime)-FStartYear+1,DayOfTheYear(ADateTime)]);
end;

{------------------------------------------------------------------------------}
{ TPinatuboBasisFunction                                                       }
{------------------------------------------------------------------------------}
constructor TPinatuboBasisFunction.Create;
begin
  Create(0,1);
end;

constructor TPinatuboBasisFunction.Create(ANumFourier,ANumSteps:word);
begin
  inherited Create(FillInitRec('Pinatubo',ANumFourier,DontChange));
  NumSteps:=ANumSteps;
  FSeedYear:=1991.4547;
end;

constructor TPinatuboBasisFunction.Create(ANumFourier,ANumSteps:word;ASeedYear:double);
begin
  Create(ANumFourier,ANumSteps);
  FSeedYear:=ASeedYear;
end;

function TPinatuboBasisFunction.GetValue(ADateTime:TDateTime):double;
var
  DecYear:double;
begin
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  if DecYear<FStartYear then Result:=0.0 else Result:=exp(FStartYear-DecYear)*(1-exp(6*(FStartYear-DecYear)));
end;

procedure TPinatuboBasisFunction.SetStep(N:word);
begin
  FStartYear:=FSeedYear+(N-1)*0.0833333;
end;

{------------------------------------------------------------------------------}
{ TPinatuboBasisFunctionWithPresOffset                                         }
{------------------------------------------------------------------------------}
constructor TPinatuboBasisFunctionWithPresOffset.Create;
begin
  raise EBasisFunction.Create('calling the contructor with no parameters is not permitted.');
end;

constructor TPinatuboBasisFunctionWithPresOffset.Create(OffsetFilename:string);
begin
  Create(0,OffsetFilename);
end;

constructor TPinatuboBasisFunctionWithPresOffset.Create(ANumFourier:word;OffsetFilename:string);
var
  Source:textfile;
  Value,
  I,J:integer;
  Header:TSynopticGlobalHeader;
begin
  with Header do
  begin
    NumLong:=70;
    NumLat:=36;
    DeltaLong:=1;
    DeltaLat:=5;
    StartLong:=1;
    StartLat:=-87.5;
  end;
  FOffset:=TSynopticGlobal.Create(Header);
  assignfile(Source,OffsetFilename);
  reset(Source);
  DeleteHeader(Source);
  repeat
    readln(Source,I,J,Value);
    FOffset[J,I]:=Value;
  until eof(Source);
  closefile(Source);
  inherited Create(FillInitRec('Pinatubo',ANumFourier,DontChange));
end;

procedure TPinatuboBasisFunctionWithPresOffset.SetLevel(ALevel:word);
begin
  FLevel:=ALevel;
end;

function TPinatuboBasisFunctionWithPresOffset.GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
var
  DecYear:double;
begin
  if AInfoObj=nil then raise EBasisFunction.Create('InfoObj was nil.');
  if not (AInfoObj is TLatitudeInfo) then raise EBasisFunction.Create(
    'InfoObj was not of the correct type.');
  FStartYear:=1991.4547+(FOffset.BiLinear(TLatitudeInfo(AInfoObj).Latitude,FLevel)-1)*0.0833333;
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  if DecYear<FStartYear then Result:=0.0 else Result:=exp(FStartYear-DecYear)*(1-exp(6*(FStartYear-DecYear)));
end;

destructor TPinatuboBasisFunctionWithPresOffset.Destroy;
begin
  FreeAndNil(FOffset);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TElChichonBasisFunction                                                      }
{------------------------------------------------------------------------------}
constructor TElChichonBasisFunction.Create;
begin
  Create(0,1);
end;

constructor TElChichonBasisFunction.Create(ANumFourier,ANumSteps:word);
begin
  inherited Create(FillInitRec('ElChichon',ANumFourier,DontChange));
  NumSteps:=ANumSteps;
  FSeedYear:=1982.2384;
end;

constructor TElChichonBasisFunction.Create(ANumFourier,ANumSteps:word;ASeedYear:double);
begin
  Create(ANumFourier,ANumSteps);
  FStartYear:=ASeedYear;
end;

function TElChichonBasisFunction.GetValue(ADateTime:TDateTime):double;
var
  DecYear:double;
begin
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  if DecYear<FStartYear then Result:=0.0 else Result:=exp(FStartYear-DecYear)*(1-exp(6*(FStartYear-DecYear)));
end;

procedure TElChichonBasisFunction.SetStep(N:word);
begin
  FStartYear:=FSeedYear+(N-1)*0.0833333;
end;

{------------------------------------------------------------------------------}
{ TQBOandSOIBasisFunction                                                           }
{------------------------------------------------------------------------------}
procedure TQBOandSOIBasisFunction.GetStartAndEndTimes(var StartDateTime,EndDateTime:TDateTime);
begin
  StartDateTime:=Data.GetFirstDateTime;
  EndDateTime:=Data.GetLastDateTime;
end;

function TQBOandSOIBasisFunction.TestResolution:boolean;
begin
//  if FResolution<ResMonth then raise EBasisFunction.Create('the QBO and SOI basis '+
//    'functions are currently not designed for higher than monthly resolution.');
  Result:=true;
end;

function TQBOandSOIBasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  Result:=Data.GetValue(ADateTime);
end;

function TQBOandSOIBasisFunction.InsufficientCoverage(ADateTimeList:TDoubleArray):boolean;
var
  StartDateTime,
  EndDateTime:TDateTime;
begin
  GetStartAndEndTimes(StartDateTime,EndDateTime);
  Result:=((ADateTimeList[High(ADateTimeList)]>EndDateTime) or (ADateTimeList[0]<StartDateTime));
end; {End of TQBOandSOIBasisFunction.InsufficientCoverage}

{$WARNINGS OFF}
procedure TQBOandSOIBasisFunction.ClimatologicalExtension(FirstYear,LastYear:word);
type
  TPhaseArray=array of array of double;
var
  PositivePhases,
  NegativePhases,
  Hold:TQBOSOIPhase;
  FirstPhaseChange,
  StartDateTime,
  EndDateTime,
  DateTime:TDateTime;
  Ind:integer;
  HoldInterpolate,
  Done:boolean;
begin
  PositivePhases:=TQBOSOIPhase.Create;
  NegativePhases:=TQBOSOIPhase.Create;

  try
    {First generate the mean positive and negative phases.}
    {Find the first phase change if there is one.}
    HoldInterpolate:=Data.Interpolate;
    Data.Interpolate:=false;
    Done:=false;
    StartDateTime:=Data.GetFirstDateTime;
    EndDateTime:=Data.GetLastDateTime;
    FirstPhaseChange:=StartDateTime;
    repeat
      if (Data.IsNotNull(FirstPhaseChange) and Data.IsNotNull(incmonth(FirstPhaseChange)) and
        (sign(Data.GetValue(FirstPhaseChange))*sign(Data.GetValue(incmonth(FirstPhaseChange)))=-1)) then Done:=true;
      FirstPhaseChange:=incmonth(FirstPhaseChange);
    until (Done or (incmonth(FirstPhaseChange)>EndDateTime));

    {Now accumulate the positive and negative phases.}
    DateTime:=FirstPhaseChange;
    if Data.GetValue(DateTime)>0 then Hold:=PositivePhases else Hold:=NegativePhases;
    Hold.NextPhase;
    Hold.PutValue(Data.GetValue(DateTime));
    repeat
      if (Data.IsNotNull(DateTime) and Data.IsNotNull(incmonth(DateTime)) and
        (sign(Data.GetValue(DateTime))*sign(Data.GetValue(incmonth(DateTime)))=-1)) then {Phase has changed.}
      begin
        if Hold=PositivePhases then Hold:=NegativePhases else Hold:=PositivePhases;
        Hold.NextPhase;
      end;
      Hold.PutValue(Data.GetValue(incmonth(DateTime)));
      DateTime:=incmonth(DateTime);
    until incmonth(DateTime)>EndDateTime;
    Hold.DropLastPhase;
    PositivePhases.CalculateAveragePhase;
    NegativePhases.CalculateAveragePhase;

    {Now fill backward from the first phase change.}
    DateTime:=incmonth(FirstPhaseChange,-1);
    if Data.GetValue(DateTime)>0 then Hold:=PositivePhases else Hold:=NegativePhases;
    Ind:=Hold.MeanPhaseLength;
    repeat
      DateTime:=incmonth(DateTime,-1);
      dec(Ind);
      if Ind=0 then
      begin
        if Hold=PositivePhases then Hold:=NegativePhases else Hold:=PositivePhases;
        Ind:=Hold.MeanPhaseLength;
      end;
      if Data.IsNull(DateTime) then Data.PutValue(DateTime,Hold.GetMeanPhaseValue(Ind));
    until (YearOf(incmonth(DateTime,-1))<=(FirstYear-1));

    {Now fill forward from the first phase change.}
    DateTime:=FirstPhaseChange;
    if Data.GetValue(DateTime)>0 then Hold:=PositivePhases else Hold:=NegativePhases;
    Ind:=1;
    repeat
      DateTime:=incmonth(DateTime);
      inc(Ind);
      if Ind>Hold.MeanPhaseLength then
      begin
        if Hold=PositivePhases then Hold:=NegativePhases else Hold:=PositivePhases;
        Ind:=1;
      end;
      if Data.IsNull(DateTime) then Data.PutValue(DateTime,Hold.GetMeanPhaseValue(Ind));
    until (YearOf(incmonth(DateTime))>=(LastYear+1));
  finally
    Data.Interpolate:=HoldInterpolate;
    FreeAndNil(PositivePhases);
    FreeAndNil(NegativePhases);
  end;
end;
{$WARNINGS ON}

procedure TQBOandSOIBasisFunction.TestDateTimeList(ADateTimeList:TDoubleArray);
var
  FirstYear,
  LastYear:word;
begin
  FirstYear:=YearOf(ADateTimeList[Low(ADateTimeList)]);
  LastYear:=YearOf(ADateTimeList[High(ADateTimeList)]);
  if InsufficientCoverage(ADateTimeList) then
  {$IFDEF ALWAYSCLIM}
  ClimatologicalExtension(FirstYear,LastYear);
  {$ELSE}
  case MessageDlg('The '+Name+' basis function does not span the full time period.'+
    ' Try to download a newer version of the data file?',mtConfirmation,[mbYes,mbNo],0) of
    mrYes:begin
      try
        DownloadNewData;
        LoadTheDataFromFile;
      except
        on E:Exception do ShowMessage('There was a problem when trying to download '+
          'a new version of the '+Name+' data file. The error message is "'+E.Message+
          '". But don''t worry, there are other options.');
      end;
      if InsufficientCoverage(ADateTimeList) then
      case MessageDlg('There was still insufficient data to span the full period '+
        'of regression data. Use climatological extension to fill the missing data?'+
        ' If you say no, the program will terminate.',mtConfirmation,[mbYes,mbNo],0) of
        mrYes:ClimatologicalExtension(FirstYear,LastYear);
        mrNo:raise EBasisFunction.Create('There were insufficient data for the '+Name+' basis function.');
      end;
    end;
    mrNo:begin
      case MessageDlg('Use climatological extension to fill the missing data?'+
        ' If you say no, the program will terminate.',mtConfirmation,[mbYes,mbNo],0) of
        mrYes:ClimatologicalExtension(FirstYear,LastYear);
        mrNo:raise EBasisFunction.Create('There were insufficient data for the '+Name+' basis function.');
      end;
    end;
  end;
  {$ENDIF}
end; {End of TQBOandSOIBasisFunction.TestDateTimeList.}

destructor TQBOandSOIBasisFunction.Destroy;
begin
  FreeAndNil(Data);
  inherited Destroy;
end;


{------------------------------------------------------------------------------}
{ TQBOBasisFunction                                                           }
{------------------------------------------------------------------------------}
constructor TQBOBasisFunction.Create(QBOIndicesFile:string);
begin
  Create(0,QBOIndicesFile);
end;

constructor TQBOBasisFunction.Create(ANumFourier:word;QBOIndicesFile:string);
begin
  inherited Create(FillInitRec('QBO',ANumFourier,RemoveOffsetAndTrend));
  if not FileExists(QBOIndicesFile) then raise EBasisFunction.Create(
    'the file named '+QBOIndicesFile+' couldn''t be found.');
  IndicesFilename:=QBOIndicesFile;
  LoadTheDataFromFile;
end;

procedure TQBOBasisFunction.WriteDataGridToTextfile(Filename:string);
var
  Dest:textfile;
  I:integer;
  Hold:TXYSignal;
begin
  Hold:=Data.GetXYSignalOfValues;
  try
    assignfile(Dest,Filename);
    rewrite(Dest);
    writeln(Dest,'''Column A) The year.');
    writeln(Dest,'''Column B) The month.');
    writeln(Dest,'''Column C) The decimal year.');
    writeln(Dest,'''Column D) The QBO value.');
    for I:=1 to Hold.GetNumData do
      writeln(Dest,YearOf(Hold.GetXValue(I)):4,' ',MonthOf(Hold.GetXValue(I)):2,
      ' ',ConvertDateTimeToDecYear(Hold.GetXValue(I)):8:3,' ',Hold.GetYValue(I):8:3);
    closefile(Dest);
  finally
    FreeAndNil(Hold);
  end;
end;

procedure TQBOBasisFunction.LoadTheDataFromFile;
var
  BaseYear,
  LineCount,
  Year,Month:word;
  WorkStr:string;
  QBOValue:double;
  DataReadIn:boolean;
  FileData:TStringList; {We use this to avoid EOL problems.}
begin
  {First clear the data that may already be in the data buffer. The Data}
  {object created here is freed in the destructor.}
  FreeAndNil(Data);
  Data:=TG2Binary.Create(ResMonth);

  if not FileExists(IndicesFilename) then raise EBasisFunction.Create(
    'the file named '+IndicesFilename+' couldn''t be found.');
  FileData:=TStringList.Create;
  try
    FileData.LoadFromFile(IndicesFilename);
    DataReadIn:=false;
//    if copy(FileData[0],1,46)='Monthly mean zonal wind components u (0.1 m/s)' then
    if copy(FileData[1],1,46)='Monthly mean zonal wind components u (0.1 m/s)' then
    begin
      LineCount:=0;
      repeat inc(LineCount) until copy(FileData[LineCount],1,10)='IIIII YYMM';
      BaseYear:=1900;
      repeat
        inc(LineCount);
        if FileData[LineCount]<>'' then
        begin
          WorkStr:=GetWord(FileData[LineCount],' ',2);
          if WorkStr='0001' then BaseYear:=BaseYear+100;
          Year:=BaseYear+strtoint(copy(WorkStr,1,2));
          Month:=strtoint(copy(WorkStr,3,2));
          QBOValue:=strtofloat(copy(FileData[LineCount],20,4))/10;
          Data.PutValue(Year,Month,QBOValue);
        end;
      until LineCount=FileData.Count-1;
      DataReadIn:=true;
    end;
    if copy(FileData[0],1,20)=' 50 mb zonal wind in' then
    begin
      LineCount:=0;
      repeat inc(LineCount) until copy(FileData[LineCount],1,10)='YEAR   JAN';
      repeat
        inc(LineCount);
        if copy(FileData[LineCount],1,10)<>' 50 mb zon' then
        begin
          Year:=strtoint(copy(FileData[LineCount],1,4));
          for Month:=1 to 12 do
          begin
            QBOValue:=strtofloat(Trim(copy(FileData[LineCount],5+(Month-1)*7,7)));
            if QBOValue>-999 then Data.PutValue(Year,Month,QBOValue);
          end;
        end;
      until copy(FileData[LineCount],1,10)=' 50 mb zon';
      DataReadIn:=true;
    end;
    if not DataReadIn then raise EBasisFunction.Create('the file specified as the source'+
      ' for the QBO data was not a valid format.');
  finally
    FreeAndNil(FileData);
  end;
end; {End of TQBOBasisFunction.LoadTheDataFromFile.}

procedure TQBOBasisFunction.DownloadNewData;
{$DEFINE FUB}
var
  {$IFDEF CPC}
  FTP:TIdFTP; {The instance of the FTP session.}
  {$ELSE}
  Http: TIdHTTP;
  MS: TMemoryStream;
  {$ENDIF}
begin
  {$IFDEF CPC}
  FTP:=TIdFTP.Create(nil);
  FTP.Passive:=true;
  with FTP do
  begin
    {Now available only thru anonymous ftp at: ftp.cpc.ncep.noaa.gov/wd52dg/data/indices}
    Host:='ftp.cpc.ncep.noaa.gov';
    UserName:='';
    Password:='';
    try Connect except
      on E:Exception do Abort;
    end;
    Noop;
    ChangeDir('wd52dg/data/indices');
    Noop;
    Get('qbo.u50.index',IndicesFilename,true);
  end;
  FTP.Disconnect;
  FTP.Free;
  {$ELSE}
  Http:=TIdHTTP.Create(nil);
  try
    MS:=TMemoryStream.Create;
    try
      Http.Get('http://www.geo.fu-berlin.de/met/ag/strat/produkte/qbo/qbo.dat',MS);
      MS.SaveToFile(IndicesFilename);
    finally
      MS.Free;
    end;
  finally
    Http.Free;
  end;
  {$ENDIF}
end; {End of TQBOBasisFunction.DownloadNewData.}

{------------------------------------------------------------------------------}
{ TQBOOrthogBasisFunction                                                      }
{------------------------------------------------------------------------------}
constructor TQBOOrthogBasisFunction.Create(QBOIndicesFile:string);
begin
  Create(0,QBOIndicesFile);
end;

constructor TQBOOrthogBasisFunction.Create(ANumFourier:word;QBOIndicesFile:string);
begin
  inherited Create(ANumFourier,QBOIndicesFile);
  OrthogTheData;
  Name:='Orthogonal QBO';
end;

procedure TQBOOrthogBasisFunction.OrthogTheData;
var
  Done:boolean; {Set to true when the shift has been completed.}
  IntProduct,     {The integrated product of Data and ShiftedData.}
  PrevIntProduct, {The previous integrated product value.}
  Shift:double;   {The current shift in decimal days.}
  ShiftedDateTime,    {The shifted date and time.}
  StartDateTime,      {The first date and time in the QBO time series.}
  EndDateTime,        {The last date and time in the QBO time series.}
  DateTime:TDateTime; {The current date and time.}
  ShiftedData:TG2Binary; {The shifted QBO time series.}
begin
  ShiftedData:=TG2Binary.Create(ResMonth); {Freed below.}
  Data.OffsetBy(Data.GetMeanValue);
  StartDateTime:=Data.GetFirstDateTime;
  EndDateTime:=Data.GetLastDateTime;

  {Now start trying various shifts.}
  Shift:=0.0; {This is the shift in decimal days.}
  IntProduct:=0.0;
  Done:=false;
  repeat
    ShiftedData.Clear;
    Shift:=Shift+1.0;
    PrevIntProduct:=IntProduct;
    IntProduct:=0.0;

    {Fill shifted data.}
    DateTime:=StartDateTime;
    repeat
      ShiftedDateTime:=DateTime-Shift;
      if ((ShiftedDateTime>=StartDateTime) and (ShiftedDateTime<=EndDateTime)) then
        ShiftedData.PutValue(DateTime,Data.GetValue(ShiftedDateTime));
      DateTime:=incmonth(DateTime);
    until ShiftedDateTime>EndDateTime;
    ShiftedData.OffsetBy(ShiftedData.GetMeanValue);

    {Calculate the integral of the product of the two signals.}
    DateTime:=StartDateTime;
    repeat
      if (Data.IsNotNull(DateTime) and ShiftedData.IsNotNull(DateTime)) then
        IntProduct:=IntProduct+(Data.GetValue(DateTime)*ShiftedData.GetValue(DateTime));
      DateTime:=incmonth(DateTime);
    until DateTime>EndDateTime;

    if ((IntProduct<=0.0) and (PrevIntProduct>0.0)) then Done:=true;
    if Shift>730.0 then Done:=true;
  until Done;

  if Shift>729 then
  begin
    Shift:=168.0;
    ShiftedData.Clear;
    DateTime:=StartDateTime;
    repeat
      ShiftedDateTime:=DateTime-Shift;
      if ((ShiftedDateTime>=StartDateTime) and (ShiftedDateTime<=EndDateTime)) then
        ShiftedData.PutValue(DateTime,Data.GetValue(ShiftedDateTime));
      DateTime:=incmonth(DateTime);
    until ShiftedDateTime>EndDateTime;
  end;

  {Overwrite the original QBO data with the shifted values.}
  FreeAndNil(Data);
  Data:=ShiftedData;
end;

{------------------------------------------------------------------------------}
{ TSOIBasisFunction                                                            }
{------------------------------------------------------------------------------}
constructor TSOIBasisFunction.Create(SOIIndicesFile:string);
begin
  Create(0,SOIIndicesFile);
end;

constructor TSOIBasisFunction.Create(ANumFourier:word;SOIIndicesFile:string);
begin
  inherited Create(FillInitRec('SOI',ANumFourier,RemoveOffsetAndTrend));
  if not FileExists(SOIIndicesFile) then raise EBasisFunction.Create(
    'the file named '+SOIIndicesFile+' couldn''t be found.');
  IndicesFilename:=SOIIndicesFile;
  LoadTheDataFromFile;
end;

procedure TSOIBasisFunction.LoadTheDataFromFile;
var
  FileData:TStringList; {Use this to circumvent EOL problems.}
  LineCount,
  Year,Month:word;
  Linein:string;
  SOIValue:double;
begin
  {First clear the data that may already be in the data buffer. The Data}
  {object created here is freed in the destructor.}
  FreeAndNil(Data);
  Data:=TG2Binary.Create(ResMonth);

  if not FileExists(IndicesFilename) then raise EBasisFunction.Create(
    'the file named '+IndicesFilename+' couldn''t be found.');
  FileData:=TStringList.Create;
  try
    FileData.LoadFromFile(IndicesFilename);
    LineCount:=0;
    repeat inc(LineCount) until copy(FileData[LineCount],1,10)='YEAR   JAN';
    repeat
      inc(LineCount);
      if copy(FileData[LineCount],1,10)<>'(STAND TAH' then
      begin
        Year:=strtoint(copy(FileData[LineCount],1,4));
        for Month:=1 to 12 do
        begin
          SOIValue:=strtofloat(Trim(copy(FileData[LineCount],5+(Month-1)*6,6)));
          if SOIValue>-999 then Data.PutValue(Year,Month,SOIValue);
        end;
      end;
    until copy(FileData[LineCount],1,10)='(STAND TAH';
  finally
    FreeAndNil(FileData);
  end;
end; {End of SIOBasisFunction.LoadTheDataFromFile.}

procedure TSOIBasisFunction.DownloadNewData;
var
  FTP:TIdFTP; {The instance of the FTP session.}
begin
  FTP:=TIdFTP.Create(nil);

  {We need to figure out how to deal with the proxy settings here.}
//  FTP.SocksInfo.Host:='socks.niwa.co.nz';
//  FTP.SocksInfo.Port:=27559;
//  FTP.SocksInfo.Version:=svSocks5;

  FTP.Passive:=true;
  with FTP do
  begin
    {Now available only thru anonymous ftp at: ftp.cpc.ncep.noaa.gov/wd52dg/data/indices}
    Host:='ftp.cpc.ncep.noaa.gov';
    UserName:='anonymous';
    Password:='g.bodeker@niwa.co.nz';
    try Connect except
      on E:Exception do Abort;
    end;
    Noop;
    ChangeDir('wd52dg/data/indices');
    Noop;
    Get('soi',IndicesFilename,true);
  end;
  FTP.Disconnect;
  FTP.Free;
end;

{------------------------------------------------------------------------------}
{ TSolarBasisFunction                                                          }
{------------------------------------------------------------------------------}
constructor TSolarBasisFunction.Create(PathForSolarFiles:string);
begin
  Create(0,PathForSolarFiles);
end;

constructor TSolarBasisFunction.Create(ANumFourier:word;PathForSolarFiles:string);
begin
  inherited Create(FillInitRec('Solar',ANumFourier,RemoveOffsetAndTrend));
  if not DirectoryExists(PathForSolarFiles) then raise EBasisFunction.Create(
    'the '+PathForSolarFiles+' doesn''t exist.');
  FilePath:=PathForSolarFiles;
  LoadTheDataFromFile;
end; {End of TSolarBasisFunction.Create.}

procedure TSolarBasisFunction.LoadTheDataFromFile;
var
  Header:TG1Header;
  FileData,
  FileList:TStringList;
  LineCount,
  FileCount,
  I:integer;
  Source:textfile;
  Month,
  Date,
  TestDate,
  Year:word;
  Linein,
  WorkStr:string;
begin
  {First clear the data that may already be in the data buffer. The Data}
  {object created here is freed in the destructor.}
  FreeAndNil(Data);
  with Header do
  begin
    FileType:=1;
    FullName:='Ottawa';
    AbbrName:='ott';
    DataDesc:='Measured 10.7 cm solar flux with 11 year lagged interpolation.';
    Units:='10^-23 J/s/m^2/Hz ';
    Latitude:=0.0;
    Longitude:=0.0;
    NullData:=-999.9;
    Altitude:=65535;
    NumPerDay:=1;
  end;
  Data:=TG1Mean.Create('',Header);
  Data.SetScalingAndOffset(6553.5,0.1);
  Data.SetRestrictions(1,20,12);
  FileList:=GetFileList(FilePath+'\????.obs',true);
//  FileList:=GetFileList(FilePath+'\????',true);
  try
    for FileCount:=0 to FileList.Count-1 do
    begin
      {Using a TStringList get us around the EOL problem in UNIX files.}
      FileData:=TStringList.Create;
      try
        FileData.LoadFromFile(FileList[FileCount]);
        Year:=strtoint(GetWord(FileData[0],' ',1));
        if Year<>strtoint(copy(ExtractFilename(FileList[FileCount]),1,4)) then
          raise Exception.Create('year mismatch in reading from '+FileList[FileCount]);
        LineCount:=4;
        TestDate:=0;
        repeat
          inc(LineCount);
          Linein:=Trim(FileData[LineCount]);
          if ((length(Linein)>10) and (copy(Linein,1,10)<>'----------')) then
          begin
            inc(TestDate);
            try
              Date:=strtoint(GetWord(Linein,' ',1));
              if Date<>TestDate then raise Exception.Create('date<>test data in '+FileList[FileCount]);
            except
              on E:Exception do raise Exception.Create('the date could not be decoded from '+FileList[FileCount]);
            end;
            for Month:=1 to 12 do
            begin
              WorkStr:=GetWord(Linein,' ',Month+1);
              for I:=1 to length(WorkStr) do if not (WorkStr[I] in ['0'..'9']) then WorkStr[I]:=' ';
              WorkStr:=Trim(WorkStr);
              WorkStr:=EraseFrontChar(WorkStr,'0');
              if WorkStr<>'' then
              try
                Data.PutValue(Year,CalcDay(Year,Month,Date),1,strtofloat(WorkStr));
              except
                on E:Exception do raise Exception.Create('could not decode '+WorkStr+'. Error message was: '+E.Message);
              end
            end;
          end;
        until copy(Linein,1,10)='----------';
      finally
        FreeAndNil(FileData);
      end;
    end;
  finally
    FreeAndNil(FileList);
  end;
  Data.FillMeans;
end;

procedure TSolarBasisFunction.DownloadNewData;
var
  StartYearFiles,
  EndYearFiles,
  FileYear,
  Year,    {Loop control variable through years.}
  I:word;
  FTP:TIdFTP; {The instance of the FTP session.}
  FileList:TStringList;
begin
  {This is Birgits code and I need to confirm that it works.}
  StartYearFiles:=3000;
  EndYearFiles:=1000;
  FTP:=TIdFTP.Create(nil);

  {We need to figure out how to deal with the proxy settings here.}
//  FTP.SocksInfo.Host:='socks.niwa.co.nz';
//  FTP.SocksInfo.Port:=27559;
//  FTP.SocksInfo.Version:=svSocks5;

  FTP.Passive:=true;
  with FTP do
  begin
    {Now available only thru anonymous ftp at: ftp.cpc.ncep.noaa.gov/wd52dg/data/indices}
    Host:='ftp.ngdc.noaa.gov';
    UserName:='anonymous';
    Password:='g.bodeker@niwa.co.nz';
    try Connect except
      on E:Exception do Abort;
    end;
//    Noop;
    ChangeDir('STP/SOLAR_DATA/SOLAR_RADIO/FLUX');
    {get the range of available data files}
    FileList:=GetFileList(Host+'/STP/SOLAR_DATA/SOLAR_RADIO/FLUX/'+'*.obs');
    for I:=0 to FileList.Count-1 do
    begin
      try
        FileYear:=StrToInt(copy(FileList[I],1,4));
        if FileYear<StartYearFiles then StartYearFiles:=FileYear;
        if FileYear>EndYearFiles then EndYearFiles:=FileYear;
      except
        on exception do;
      end;
    end;
    {get all available data files}
    for Year:=StartYearFiles to EndYearFiles do
      try Get(inttostr(Year)+'.OBS',FilePath+inttostr(Year)+'.obs',true);
      except
        on E:Exception do Get(inttostr(Year)+'.obs',FilePath+inttostr(Year)+'.obs',true);
      end;
  end;
  FTP.Disconnect;
  FTP.Free;
end;

function TSolarBasisFunction.TestResolution:boolean;
begin
  if not ((FResolution=ResDay) or (FResolution=ResMonth) or (FResolution=ResYear)) then
    raise EBasisFunction.Create('the solar cycle basis function is currently not designed for this resolution.');
  Result:=true;
end;

function TSolarBasisFunction.IsNull(ADateTime:TDateTime):boolean;
begin
  Result:=true; {The default return value.}
  if FResolution=ResDay then Result:=Data.IsNull(ADateTime);
  if FResolution=ResMonth then Result:=(Data.GetMonthlyMean(YearOf(ADateTime),MonthOf(ADateTime))<=0.0);
  if FResolution=ResYear then Result:=(Data.GetAnnualMean(YearOf(ADateTime))<=0.0);
end;

function TSolarBasisFunction.GetValue(ADateTime:TDateTime):double;
begin
  Result:=NullValue;
  if FResolution=ResDay then Result:=Data[YearOf(ADateTime),DayOfTheYear(ADateTime),1];
  if FResolution=ResMonth then Result:=Data.GetMonthlyMean(YearOf(ADateTime),MonthOf(ADateTime));
  if FResolution=ResYear then Result:=Data.GetAnnualMean(YearOf(ADateTime));
end;

procedure TSolarBasisFunction.TestDateTimeList(ADateTimeList:TDoubleArray);
var
  FillFromClimatology,
  MissingValues:boolean;
  DT1,DT2:array[-2..2] of TDateTime;
  DateTime:TDateTime;
  Factor,
  I:integer;
  Done:boolean;
begin
  MissingValues:=false;
  if ADateTimeList[0]<Data.GetStartDateTime then MissingValues:=true else
    if ADateTimeList[High(ADateTimeList)]>Data.GetEndDateTime then MissingValues:=true else
      for I:=Low(ADateTimeList) to High(ADateTimeList) do if IsNull(ADateTimeList[I]) then MissingValues:=true;
  if MissingValues then
  begin
    {$IFDEF ALWAYSCLIM}
    FillFromClimatology:=true;
    {$ELSE}
    case MessageDlg('The solar cycle basis function does not contain values for'+
      ' each measurement in the regression. Do you want to try and automatically'+
      ' download newer versions of the input files?',mtConfirmation,[mbYes,mbNo],0) of
      mrYes:begin
        try
          DownloadNewData;
          LoadTheDataFromFile;
          MissingValues:=false;
          if ADateTimeList[0]<Data.GetStartDateTime then MissingValues:=true else
            if ADateTimeList[High(ADateTimeList)]>Data.GetEndDateTime then MissingValues:=true else
              for I:=Low(ADateTimeList) to High(ADateTimeList) do if IsNull(ADateTimeList[I]) then MissingValues:=true;
          if MissingValues then
            case MessageDlg('Even after downloading the latest version of the'+
              ' input files the solar cycle basis cycle still has missing values.'+
              ' Use a climatology to fill the gaps?',mtConfirmation,[mbYes,mbNo],0) of
            mrYes:FillFromClimatology:=true;
            mrNo:FillFromClimatology:=false;
          end;
        except
          on Exception do
          case MessageDlg('There was a problem with downloading the newest data files.'+
            ' Use a climatology to fill the gaps?',mtConfirmation,[mbYes,mbNo],0) of
            mrYes:FillFromClimatology:=true;
            mrNo:FillFromClimatology:=false;
          end;
        end;
      end;
      mrNo:begin
        case MessageDlg('Use a climatology to fill the gaps?',mtConfirmation,[mbYes,mbNo],0) of
          mrYes:FillFromClimatology:=true;
          mrNo:FillFromClimatology:=false;
        end;
      end;
    end;
    {$ENDIF}
    if FillFromClimatology then
    begin
      {Fill by year from the last year to the first year.}
      DateTime:=EncodeDate(YearOf(ADateTimeList[0]),1,1);
      repeat
        if IsNull(DateTime) then
        begin
          Factor:=1;
          Done:=false;
          repeat
            for I:=-2 to 2 do DT1[I]:=incyear(DateTime,11*Factor)+I;
            for I:=-2 to 2 do DT2[I]:=incyear(DateTime,-11*Factor)+I;
            for I:=-2 to 2 do
            begin
              if Data.IsNotNull(DT1[I]) then
              begin
                Data.PutValue(DateTime,Data.GetValue(DT1[I]));
                Done:=true;
              end
              else
              if Data.IsNotNull(DT2[I]) then
              begin
                Data.PutValue(DateTime,Data.GetValue(DT2[I]));
                Done:=true;
              end;
            end;
            if ((DT2[0]<Data.GetStartDateTime) and (DT1[0]>Data.GetEndDateTime)) then Done:=true;
            inc(Factor);
          until Done;
        end;
        DateTime:=incday(DateTime);
      until (YearOf(DateTime)>=YearOf(ADateTimeList[High(ADateTimeList)])+1);
      Data.FillMeans;
      for I:=Low(ADateTimeList) to High(ADateTimeList) do if IsNull(ADateTimeList[I]) then
        raise EBasisFunction.Create('it was not possible to fill the missing values'+
          ' in the solar cycle basis function.');
    end;
  end;
end;

destructor TSolarBasisFunction.Destroy;
begin
  FreeAndNil(Data);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TGlobalSurfTemp                                                              }
{------------------------------------------------------------------------------}
constructor TGlobalSurfTemp.Create(Filename:string;AnomStartYear,AnomEndYear:word;AType:TGlobalSurfTempType);
begin
  Create(0,Filename,AnomStartYear,AnomEndYear,AType);
end;

constructor TGlobalSurfTemp.Create(ANumFourier:word;Filename:string;AnomStartYear,AnomEndYear:word;AType:TGlobalSurfTempType);
const
  DescStr:array[0..4] of string=('Global','NH ocean','NH land','SH ocean','SH land');
var
  Source:textfile;
  Linein:string;
  Year:word;
  Total:double;
  Num,I:integer;
begin
  inherited Create(FillInitRec(DescStr[ord(AType)],ANumFourier,DontChange));
  FBaseLineStartYear:=AnomStartYear;
  FBaseLineEndYear:=AnomEndYear;
  FirstYear:=65535;
  LastYear:=0;
  Total:=0.0;
  Num:=0;
  assignfile(Source,Filename);
  reset(Source);
  DeleteHeader(Source);
  repeat
    readln(Source,Linein);
    if not (Linein='') then
    begin
      Year:=strtoint(GetWord(Linein,' ',1));
      if Year<FirstYear then FirstYear:=Year;
      if Year>LastYear then LastYear:=Year;
      setlength(Data,length(Data)+1);
      Data[High(Data)].Year:=Year;
      Data[High(Data)].Value:=strtofloat(GetWord(Linein,' ',2+ord(AType)));
      if ((Year>=FBaseLineStartYear) and (Year<=FBaseLineEndYear)) then
      begin
        Total:=Total+Data[High(Data)].Value;
        inc(Num);
      end;
    end;
  until eof(Source);
  closefile(Source);
  dec(LastYear); {Need this since it does temporal interpolation.}

  if Num=0 then raise EBasisFunction.Create('the climatology could not be calculated.');

  {Now subtract off the climatology.}
  Total:=Total/Num;
  for I:=Low(Data) to High(Data) do Data[I].Value:=Data[I].Value-Total;
end;

function TGlobalSurfTemp.GetValue(ADateTime:TDateTime):double;
var
  DecYear:double;
  I:integer;
begin
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  I:=-1;
  repeat
    inc(I);
  until ((Data[I].Year<=DecYear) and (Data[I+1].Year>DecYear));
  Result:=Data[I].Value+(Data[I+1].Value-Data[I].Value)*frac(DecYear);
end;

procedure TGlobalSurfTemp.TestDateTimeList(ADateTimeList: TDoubleArray);
var
  I:integer;
begin
  for I:=Low(ADateTimeList) to High(ADateTimeList) do
  begin
    if YearOf(ADateTimeList[I])<FirstYear then raise Exception.Create('too bad');
    if YearOf(ADateTimeList[I])>LastYear then raise Exception.Create('too bad');
  end;
end;

{------------------------------------------------------------------------------}
{ TCH4AndN20BasisFunction                                                      }
{------------------------------------------------------------------------------}
constructor TCH4AndN20BasisFunction.Create(FileList:TStringList;Columns:TColumnSet);
begin
  Create(0,FileList,Columns);
end;

constructor TCH4AndN20BasisFunction.Create(ANumFourier:word;FileList:TStringList;Columns:TColumnSet);
var
  Source:textfile;
  Linein:string;
  Year:word;
  I,J:integer;
  DescStr:string;
  Max,Min,
  Value:double;
begin
  DescStr:='';
  for I in Columns do DescStr:=DescStr+'Column'+inttostr(I)+'+';
  DescStr:=EraseEndChar(DescStr,'+');
  inherited Create(FillInitRec(DescStr,ANumFourier,DontChange));
  Year:=0;
  Max:=-1E10;
  Min:=1E10;
  for I:=1 to FileList.Count do
  begin
    assignfile(Source,FileList[I-1]);
    reset(Source);
    readln(Source,Linein);
    reset(Source);
    for J:=1 to strtoint(GetWord(Linein,',',1)) do readln(Source);
    repeat
      readln(Source,Linein);
      if not (Linein='') then
      begin
        inc(Year);
        setlength(Data,length(Data)+1);
        Data[High(Data)].Year:=Year;
        Value:=1.0;
        for J in Columns do Value:=Value*strtofloat(GetWord(Linein,' ',J));
        Data[High(Data)].Value:=Value;
        if Value>Max then Max:=Value;
        if Value<Min then Min:=Value;
      end;
    until eof(Source);
    closefile(Source);
  end;
  FirstYear:=1;
  LastYear:=Year;
end;

function TCH4AndN20BasisFunction.GetValue(ADateTime:TDateTime):double;
var
  DecYear:double;
  I:integer;
begin
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  if trunc(DecYear)>=LastYear then
    Result:=Data[High(Data)].Value
  else
  begin
    I:=-1;
    repeat
      inc(I);
    until ((Data[I].Year<=DecYear) and (Data[I+1].Year>DecYear));
    Result:=Data[I].Value+(Data[I+1].Value-Data[I].Value)*frac(DecYear);
  end;
end;

procedure TCH4AndN20BasisFunction.TestDateTimeList(ADateTimeList: TDoubleArray);
var
  I:integer;
begin
  for I:=Low(ADateTimeList) to High(ADateTimeList) do
  begin
    if YearOf(ADateTimeList[I])<FirstYear then raise Exception.Create('too bad');
    if YearOf(ADateTimeList[I])>LastYear then raise Exception.Create('too bad');
  end;
end;

end.
