//@***h* BDBP/gb_GlobalSynopticStack.pas
// 
// NAME
//   gb_GlobalSynopticStack.pas
//
// PURPOSE
//  Defines and handles a four dimensional synoptic data object.
//
// DESCRIPTION
//  This unit defines and handles a four dimensional synoptic data object.
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_GlobalSynopticStack;

{$I UNITDIREC.INC}

interface

uses
  {These are standard delphi units}
  Classes,
  SysUtils,

  {These are specific units.}
  gb_array,
  gb_SynopticGlobal;

type
  TStackType=(compressed,uncompressed);

  TNewSynopticGlobalDefaults=record
    Header:TSynopticGlobalHeader;
    Scaling,
    Offset:double;
    FLatitudes:TDoubleArray;
  end;

  TGlobalSynopticStack=class(TList)
  private
    {The NewSynopticGlobalDefaults record specifies the default values to be   }
    {used for any new SynopticGlobals that are created.                        }
    NewSynopticGlobalDefaults:TNewSynopticGlobalDefaults;

    StackType:TStackType;
    DateStamp:TDateTime;
    FOwner:TList;
    MemoryUsed,
    MaxMemory:cardinal;
    TempStorageFile:string;
    FInterpolationFailPlan:TInterpolationFailPlan;
    FExtendToPole,
    FConfineToLevels,
    FDataOnDisk:boolean;
    FVertInterpolDist,
    FDistToLevel:double;
    Int:array[1..2] of longint;
    function  AddSynopticGlobalToStack(var Data:TSynopticGlobal):integer;
    procedure SetDefaultValues;
    procedure WriteDataOnDisk(Value:boolean);
    procedure WriteExtendToPole(Value:boolean);
    function  GetSynopticObject(LNum:integer):TSynopticGlobal;
  protected
    procedure Notify(Ptr:Pointer;Action:TListNotification);override;
  public
    {Constructors}
    constructor Create(AStackType:TStackType;ADateTime:TDateTime);overload;
    constructor Create(var Source:file);overload;
    constructor Create(Filename:string);overload;

    {Other methods listed in alphabetical order.}
    function  AddLevel(ALevel:double):integer;
    procedure Clear;override;
    function  DataAvailableForLevel(ALevel:double):boolean;
    function  GetDateStamp:TDateTime;
    function  GetLevel(LNum:cardinal):double;
    function  GetLevelNum(ALevel:double):cardinal;
    function  GetLevelNumAtOrBelow(ALevel:double):cardinal;
    function  GetMemoryUsed:cardinal;
    function  GetTempStorageFile:string;
    function  GetValue(LNum,Long,Lat:cardinal):single;
    function  GetSynopticGlobal(LNum:cardinal;Owner:boolean=false):TSynopticGlobal;
    function  InterpolateValue(Level,Longitude,Latitude:double;LogVertInterpol:boolean=false):single;
    procedure PutValue(LNum,Long,Lat:cardinal;Value:single);
    function  IsNull(LNum,Long,Lat:cardinal):boolean;
    function  IsNotNull(LNum,Long,Lat:cardinal):boolean;
    procedure PutSynopticGlobal(LNum:cardinal;Value:TSynopticGlobal);
    procedure ReadFromFile(Filename:string);overload;
    procedure ReadFromFile(var Source:file);overload;
    procedure RestrictMemoryUsed(MaxMem:cardinal;TFilename:string);
    procedure SaveToFile(Filename:string);overload;
    procedure SaveToFile(var Dest:file);overload;
    procedure SetNewSynopticGlobalDefaults(Defaults:TNewSynopticGlobalDefaults);
    procedure SetTempStorageFile(Value:string);
    procedure SortByLevel;

    {Destructor}
    destructor Destroy;override;

    {Properties}
    property ConfineToLevels:boolean read FConfineToLevels write FConfineToLevels;
    property Data[LNum,Long,Lat:cardinal]:single read GetValue write PutValue; default;
    property DataOnDisk:boolean read FDataOnDisk write WriteDataOnDisk;
    property InterpolationFailPlan:TInterpolationFailPlan read FInterpolationFailPlan write FInterpolationFailPlan;
    property StorageFile:string read GetTempStorageFile write SetTempStorageFile;
    property ExtendToPole:boolean read FExtendToPole write WriteExtendToPole;
    property Owner:TList read FOwner write FOwner;
  end;

  EGlobalSynopticStack=class(exception);

implementation

uses
  Dialogs,

  {These are specific units.}
  gb_4DSynoptic,
  gb_files,
  gb_math;

constructor TGlobalSynopticStack.Create(AStackType:TStackType;ADateTime:TDateTime);
begin
  inherited Create;
  StackType:=AStackType;
  DateStamp:=ADateTime;
  SetDefaultValues;
end; {End of TGlobalSynopticStack.Create.}

constructor TGlobalSynopticStack.Create(var Source:file);
begin
  inherited Create;
  SetDefaultValues;
  ReadFromFile(Source);
end; {End of TGlobalSynopticStack.Create.}

constructor TGlobalSynopticStack.Create(Filename:string);
var
  Source:file;
begin
  assignfile(Source,Filename);
  reset(Source,1);
  Create(Source);
  closefile(Source);
end; {End of TGlobalSynopticStack.Create.}

procedure TGlobalSynopticStack.SetDefaultValues;
begin
  Int[1]:=0;
  Int[2]:=1;
  MaxMemory:=high(cardinal);
  MemoryUsed:=0;
  TempStorageFile:='';
  FConfineToLevels:=true;
  FInterpolationFailPlan:=raiseexception;
  FDataOnDisk:=false;
  FExtendToPole:=false;
end; {End of TGlobalSynopticStack.SetDefaultValues.}

function TGlobalSynopticStack.AddSynopticGlobalToStack(var Data:TSynopticGlobal):integer;
var
  FHandle:file;
  MemoryAvailable:longint;
begin
  {Check if memory restrictions are in place and if so whether there is enough memory.}
  if ((not DataOnDisk) and (TempStorageFile<>'')) then
  begin
    MemoryAvailable:=MaxMemory-MemoryUsed;
    {Note: the possible call of DataOnDisk:=true results in a call to WriteDataOnDisk(true).}
    if Data.GetMemoryUsed>cardinal(MemoryAvailable) then DataOnDisk:=true;
  end;
  if DataOnDisk then
  begin
    if TempStorageFile='' then raise EGlobalSynopticStack.Create(
      'the temporary storage file name has not been set');
    assignfile(FHandle,TempStorageFile);
    if FileExists(TempStorageFile) then
    begin
      reset(FHandle,1);
      seek(FHandle,filesize(FHandle));
    end
    else rewrite(FHandle);
    Data.WriteToFile(FHandle);
    closefile(FHandle);
    FreeAndNil(Data);
  end
  else MemoryUsed:=MemoryUsed+Data.GetMemoryUsed;
  Add(Data);
  if not DataOnDisk then SortByLevel;
  Result:=IndexOf(Data)+1;
end; {End of TGlobalSynopticStack.AddSynopticGlobalToStack.}

procedure TGlobalSynopticStack.WriteDataOnDisk(Value:boolean);
var
  I:integer;
  FHandle:file;
begin
  if Value then
  begin
    if FDataOnDisk then
    begin
      {Data requested to be put on disk and data are already on disk.}
      for I:=0 to Count-1 do
        if Items[I]<>nil then raise EGlobalSynopticStack.Create('Items['+inttostr(I)+'] was not nil');
    end
    else
    begin
      {Data requested to be put on disk and data are not already on disk.}
      if TempStorageFile='' then raise EGlobalSynopticStack.Create(
        'the temporary storage file name has not been set');
      assignfile(FHandle,TempStorageFile);
      rewrite(FHandle,1);
      for I:=0 to Count-1 do
      begin
        if Items[I]=nil then raise EGlobalSynopticStack.Create('Items['+inttostr(I)+'] was nil');
        TSynopticGlobal(Items[I]).WriteToFile(FHandle);
        MemoryUsed:=MemoryUsed-TSynopticGlobal(Items[I]).GetMemoryUsed;
        TSynopticGlobal(Items[I]).Free;
        Items[I]:=nil;
      end;
      closefile(FHandle);
      FDataOnDisk:=true;
    end;
  end
  else
  begin
    if FDataOnDisk then
    begin
      {Data requested to be taken off disk and data are on disk.}
      if TempStorageFile='' then raise EGlobalSynopticStack.Create(
        'the temporary storage file name has not been set');
      assignfile(FHandle,TempStorageFile);
      reset(FHandle,1);
      for I:=0 to Count-1 do
      begin
        if Items[I]<>nil then raise EGlobalSynopticStack.Create('Items['+inttostr(I)+'] was not nil');
        Items[I]:=TSynopticGlobal.Create(FHandle);
        MemoryUsed:=MemoryUsed+TSynopticGlobal(Items[I]).GetMemoryUsed;
      end;
      closefile(FHandle);
      FDataOnDisk:=false;
      SortByLevel;
    end
    else
    begin
      {Data requested to be taken off disk and data are already off disk.}
      for I:=0 to Count-1 do
        if Items[I]=nil then raise EGlobalSynopticStack.Create('Items['+inttostr(I)+'] was nil');
    end;
  end;
end; {End of TGlobalSynopticStack.ReconcileDataOnDisk.}

procedure TGlobalSynopticStack.WriteExtendToPole(Value:boolean);
var
  I:longint;
begin
  FExtendToPole:=Value;
  for I:=1 to Count do GetSynopticObject(I).ExtendToPole:=Value;
end; {End of TGlobalSynopticStack.WriteExtendToPole.}

function TGlobalSynopticStack.GetSynopticObject(LNum:integer):TSynopticGlobal;
begin
  if ((LNum<1) or (LNum>Count)) then raise EGlobalSynopticStack.Create('the level number was out of range.');
  if DataOnDisk then DataOnDisk:=false;
  Result:=TSynopticGlobal(Items[LNum-1]);
end; {End of TGlobalSynopticStack.GetSynopticObject.}

procedure TGlobalSynopticStack.Notify(Ptr:Pointer;Action:TListNotification);
begin
  {Still need to write this code.}
end; {End of TGlobalSynopticStack.Notify.}

function TGlobalSynopticStack.AddLevel(ALevel:double):integer;
{------------------------------------------------------------------------------}
{ This procedure adds a data level to the stack. The stack object takes        }
{ responsibility for freeing the TSynopticGlobal object that is added to the   }
{ stack.                                                                       }
{------------------------------------------------------------------------------}
var
  Hold:TSynopticGlobal;
begin
  with NewSynopticGlobalDefaults do
  begin
    if StackType=compressed then
    begin
      if Scaling=-999.99 then raise EGlobalSynopticStack.Create('the default scaling value has not been set');
      if Offset=-999.99 then raise EGlobalSynopticStack.Create('the default offset value has not been set');
    end;
    Hold:=TSynopticGlobal.Create(Header);
    Hold.DateTime:=DateStamp;
    Hold.Level:=ALevel;
    if StackType=compressed then Hold.SwitchCompressionOn(65535*Scaling+Offset,Scaling+Offset);
    if length(FLatitudes)>0 then Hold.SpecifyNonuniformLatitudes(FLatitudes);
    Hold.InterpolationFailPlan:=FInterpolationFailPlan;
    Hold.ExtendToPole:=FExtendToPole;
  end;
  Result:=AddSynopticGlobalToStack(Hold);
end; {End of TGlobalSynopticStack.AddLevel.}

procedure TGlobalSynopticStack.Clear;
var
  I:longint;
begin
  for I:=0 to Count-1 do
    if Items[I]<>nil then TSynopticGlobal(Items[I]).Free;
  inherited Clear;
end; {End of TGlobalSynopticStack.Clear.}

function TGlobalSynopticStack.DataAvailableForLevel(ALevel:double):boolean;
var
  I:cardinal;
begin
  Result:=false;
  for I:=1 to Count do if EquivDouble(GetSynopticObject(I).Level,ALevel) then Result:=true;
end; {End of TGlobalSynopticStack.DataAvailableForLevel.}

function TGlobalSynopticStack.GetDateStamp:TDateTime;
begin
  Result:=DateStamp;
end; {End of TGlobalSynopticStack.GetDateStamp.}

function TGlobalSynopticStack.GetLevel(LNum:cardinal):double;
begin
  if LNum>cardinal(Count) then raise EGlobalSynopticStack.Create(
    'the value passed as the level number exceeded the number of levels');
  Result:=GetSynopticObject(LNum).Level;
end; {End of TGlobalSynopticStack.GetLevel.}

function TGlobalSynopticStack.GetLevelNum(ALevel:double):cardinal;
var
  I:cardinal;
begin
  Result:=0;
  for I:=1 to Count do
    if EquivDouble(GetSynopticObject(I).Level,ALevel) then Result:=I;
  if Result=0 then raise EGlobalSynopticStack.Create('there were no data available at '+floattostrf(ALevel,ffFixed,10,3));
end; {End of TGlobalSynopticStack.GetLevelNum.}

function TGlobalSynopticStack.GetLevelNumAtOrBelow(ALevel:double):cardinal;
begin
  Result:=Count+1;
  repeat dec(Result) until ((GetSynopticObject(Result).Level<=ALevel) or (Result=1));
  if ALevel<GetSynopticObject(Result).Level then
    raise EGlobalSynopticStack.Create('there were no levels at or below '+floattostrf(ALevel,ffFixed,10,3));
end; {End of TGlobalSynopticStack.GetLevelNumAtOrBelow.}

function TGlobalSynopticStack.GetMemoryUsed:cardinal;
begin
  Result:=MemoryUsed;
end; {End of TGlobalSynopticStack.GetMemoryUsed.}

function TGlobalSynopticStack.GetTempStorageFile:string;
begin
  Result:=TempStorageFile;
end; {End of TGlobalSynopticStack.GetTempStorageFile.}

function TGlobalSynopticStack.GetValue(LNum,Long,Lat:cardinal):single;
begin
  Result:=GetSynopticObject(LNum)[Long,Lat];
end; {End of TGlobalSynopticStack.GetValue.}

function TGlobalSynopticStack.GetSynopticGlobal(LNum:cardinal;Owner:boolean=false):TSynopticGlobal;
{------------------------------------------------------------------------------}
{ This function returns the SynopticGlobal object at level 'LNum'. If 'Owner'  }
{ is true then a copy of the object is returned and it is the responsibility   }
{ of the calling program to free this object. If 'Owner' is false, a pointer   }
{ to the SynopticGlobal owned by the Stack, is passed.                         }
{------------------------------------------------------------------------------}
begin
  if Owner then Result:=TSynopticGlobal.Create(GetSynopticObject(LNum))
  else Result:=GetSynopticObject(LNum);
end; {End of TGlobalSynopticStack.GetSynopticGlobal.}

function TGlobalSynopticStack.InterpolateValue(Level,Longitude,Latitude:double;
  LogVertInterpol:boolean=false):single;
{------------------------------------------------------------------------------}
{ We assume that the levels have already been sorted at this stage and so we   }
{ don't sort them again.                                                       }
{------------------------------------------------------------------------------}
var
  I:byte;
  HoldIndex:longint;
  StillLooking:boolean;
  LowestLevel,
  HighestLevel:double;
  Hold:TSynopticGlobal;
  InterpolLevel,
  Value:array[1..2] of double;
begin
  if Count=0 then raise EGlobalSynopticStack.Create('there are no data levels in this stack.');
  Result:=NullValue;
  LowestLevel:=GetSynopticObject(1).Level;
  HighestLevel:=GetSynopticObject(Count).Level;

  {First deal with the level being out of range.}
  if ((Level<LowestLevel) or (Level>HighestLevel)) then
  begin
    if FConfineToLevels then raise EGlobalSynopticStack.Create('the prescribed '+
      'level was not spanned by the available levels: '+floattostrf(LowestLevel,
      ffFixed,5,0)+' to '+floattostrf(HighestLevel,ffFixed,5,0))
    else
    begin
      if Level<LowestLevel then HoldIndex:=-1 else HoldIndex:=Count;
      repeat
        if Level<LowestLevel then inc(HoldIndex);
        if Level>HighestLevel then dec(HoldIndex);
        Result:=TSynopticGlobal(HoldIndex).BiLinear(Latitude,Longitude);
      until ((Result<NullValue) or (HoldIndex=(Count-1)) or (HoldIndex=0));
      if ((Result>=NullValue) and (FInterpolationFailPlan=raiseexception)) then
        raise EGlobalSynopticStack.Create('a valid value could not be found at any level.');
      FDistToLevel:=Level-TSynopticGlobal(HoldIndex).Level;
      FVertInterpolDist:=1E30;
      exit;
    end;
  end;

  {Now if there is only one level, we must be at that level.}
  if Count=1 then
  begin
    Hold:=TSynopticGlobal(First);
    if FOwner is T4DSynoptic then (FOwner as T4DSynoptic).SynopticGlobalBuffer:=Hold;
    Result:=Hold.BiLinear(Latitude,Longitude);
    if ((Result>=NullValue) and (FInterpolationFailPlan=raiseexception)) then
      raise EGlobalSynopticStack.Create('a valid value could not be found at any level.');
    FDistToLevel:=0.0;
    FVertInterpolDist:=0.0;
    exit;
  end;

  {First find the data levels than span the prescribed level.}
  if ((Level<TSynopticGlobal(Items[Int[1]]).Level) or (Level>TSynopticGlobal(Items[Int[2]]).Level)) then
  begin
    Int[2]:=Count-1;
    Int[1]:=0;
  end;
  while (Int[2]-Int[1])>1 do
  begin
    HoldIndex:=(Int[1]+Int[2]) div 2;
    if TSynopticGlobal(Items[HoldIndex]).Level<Level then Int[1]:=HoldIndex else Int[2]:=HoldIndex;
  end;
  if Int[1]=Int[2] then raise EGlobalSynopticStack.Create('Int[1] was equal to Int[2].');

  {Now see if we are exactly at one of these two levels.}
  for I:=1 to 2 do
  if TSynopticGlobal(Items[Int[I]]).Level=Level then
  begin
    Hold:=TSynopticGlobal(Items[Int[I]]);
    if FOwner is T4DSynoptic then (FOwner as T4DSynoptic).SynopticGlobalBuffer:=Hold;
    Result:=Hold.BiLinear(Latitude,Longitude);
    if Result<NullValue then exit else
      if FConfineToLevels then
      case FInterpolationFailPlan of
        raiseexception:raise EGlobalSynopticStack.Create('a valid value could not be found at any level.');
        returnnull:exit;
      end;
  end;

  {Then get the values and levels that span the prescribed level.}
  StillLooking:=true;
  while StillLooking do
  begin
    Value[1]:=TSynopticGlobal(Items[Int[1]]).BiLinear(Latitude,Longitude);
    if ((Value[1]>=NullValue) and (not FConfineToLevels) and (Int[1]>=0)) then dec(Int[1]);
    if (FConfineToLevels or (Value[1]<1E29) or (Int[1]<0)) then StillLooking:=false;
  end;
  if Int[1]<0 then Int[1]:=0;
  StillLooking:=true;
  while StillLooking do
  begin
    Value[2]:=TSynopticGlobal(Items[Int[2]]).BiLinear(Latitude,Longitude);
    if ((Value[2]>=NullValue) and (not FConfineToLevels) and (Int[2]<Count)) then inc(Int[2]);
    if (FConfineToLevels or (Value[2]<1E29) or (Int[2]=Count)) then StillLooking:=false;
  end;
  if Int[2]=Count then Int[2]:=Count-1;
  for I:=1 to 2 do
  begin
    if Value[I]>=NullValue then
    case FInterpolationFailPlan of
      raiseexception:raise EGlobalSynopticStack.Create('a valid value could not be found at any level.');
      returnnull:exit;
    end;
    InterpolLevel[I]:=TSynopticGlobal(Items[Int[I]]).Level;
  end;
  FVertInterpolDist:=InterpolLevel[2]-InterpolLevel[1];

  {Now do the vertical and temporal interpolation.}
  if LogVertInterpol then
  begin
    if ((Value[1]<=0.0) or (Value[2]<=0.0)) then raise EGlobalSynopticStack.Create(
      'logarithmic interpolation cannot be performed on negative or zero data.');
    Result:=exp((Level-InterpolLevel[1])*(ln(Value[2])-ln(Value[1]))/
      FVertInterpolDist+ln(Value[1]));
  end
  else Result:=(Level-InterpolLevel[1])*(Value[2]-Value[1])/FVertInterpolDist+Value[1];
  if (Level-InterpolLevel[1])<(InterpolLevel[2]-Level) then FDistToLevel:=
    Level-InterpolLevel[1] else FDistToLevel:=InterpolLevel[2]-Level;
  if Int[2]>(Count-1) then raise EGlobalSynopticStack.Create(
    'Int[2] was greater than the number of records. This should never happen.');
  if Int[1]<0 then raise EGlobalSynopticStack.Create(
    'Int[1] was less than zero. This should never happen.');
end; {End of TGlobalSynopticStack.InterpolateValue.}

procedure TGlobalSynopticStack.PutValue(LNum,Long,Lat:cardinal;Value:single);
begin
  if LNum>cardinal(Count) then raise EGlobalSynopticStack.Create(
    'the value passed as the level number exceeded the number of levels');
  GetSynopticObject(LNum)[Long,Lat]:=Value;
end; {End of TGlobalSynopticStack.PutValue.}

function TGlobalSynopticStack.IsNull(LNum,Long,Lat:cardinal):boolean;
begin
  if LNum>cardinal(Count) then raise EGlobalSynopticStack.Create(
    'the value passed as the level number exceeded the number of levels');
  Result:=GetSynopticObject(LNum).IsNull(Long,Lat);
end; {End of TGlobalSynopticStack.IsNull.}

function TGlobalSynopticStack.IsNotNull(LNum,Long,Lat:cardinal):boolean;
begin
  if LNum>cardinal(Count) then raise EGlobalSynopticStack.Create(
    'the value passed as the level number exceeded the number of levels');
  Result:=GetSynopticObject(LNum).IsNotNull(Long,Lat);
end; {End of TGlobalSynopticStack.IsNotNull.}

procedure TGlobalSynopticStack.PutSynopticGlobal(LNum:cardinal;Value:TSynopticGlobal);
var
  Hold:TSynopticGlobal;
begin
  Hold:=GetSynopticObject(LNum);
  FreeAndNil(Hold);
  Items[LNum-1]:=Value;
end; {End of TGlobalSynopticStack.PutSynopticGlobal.}

procedure TGlobalSynopticStack.ReadFromFile(Filename:string);
var
  Source:file;
begin
  assignfile(Source,Filename);
  reset(Source,1);
  ReadFromFile(Source);
  closefile(Source);
end; {End of TGlobalSynopticStack.ReadFromFile.}

procedure TGlobalSynopticStack.ReadFromFile(var Source:file);
var
  Hold:TSynopticGlobal;
  NumBytes:byte;
  NumLevels,
  I:cardinal;
begin
  Clear;
  blockread(Source,NumBytes,sizeof(byte));
  if NumBytes=2 then StackType:=compressed;
  if NumBytes=4 then StackType:=uncompressed;
  blockread(Source,DateStamp,sizeof(TDateTime));
  blockread(Source,NumLevels,sizeof(cardinal));
  MemoryUsed:=0;
  for I:=1 to NumLevels do
  begin
    Hold:=TSynopticGlobal.Create(Source);
    MemoryUsed:=MemoryUsed+Hold.GetMemoryUsed;
    Add(Hold);
  end;
  if ((TempStorageFile<>'') and (MemoryUsed>MaxMemory)) then DataOnDisk:=true;
end; {End of TGlobalSynopticStack.ReadFromFile.}

procedure TGlobalSynopticStack.RestrictMemoryUsed(MaxMem:cardinal;TFilename:string);
{------------------------------------------------------------------------------}
{ This procedure is called to restrict the memory used by the global synoptic  }
{ stack. In addition to specifying the maxmimum memory to be used, the calling }
{ program must also pass a temporary storage file name that can be used to     }
{ swap data to the disk. If the data stack is on the disk and the new maximum  }
{ memory is greater than the current memory used, the data are first taken off }
{ the disk. This results in a recalculation of the memory used. The test for   }
{ whether the memory used exceeds the maximum memory allowed is then repeated. }
{------------------------------------------------------------------------------}
begin
  MaxMemory:=MaxMem;
  StorageFile:=TFilename;
  if (DataOnDisk and (MemoryUsed<MaxMemory)) then DataOnDisk:=false;
  if ((not DataOnDisk) and (MemoryUsed>MaxMemory)) then DataOnDisk:=true;
end; {End of TGlobalSynopticStack.RestrictMemoryUsed.}

procedure TGlobalSynopticStack.SaveToFile(Filename:string);
var
  Dest:file;
begin
  assignfile(Dest,Filename);
  rewrite(Dest,1);
  SaveToFile(Dest);
  closefile(Dest);
end; {End of TGlobalSynopticStack.SaveToFile.}

procedure TGlobalSynopticStack.SaveToFile(var Dest:file);
var
  I:longint;
  Source:file;
  NumBytes:byte;
  Numread,Numwrite:integer;      {The number of values transferred.}
  Buffer:array[1..4096] of byte; {The data buffer.}
begin
  if StackType=compressed then NumBytes:=2;
  if StackType=uncompressed then NumBytes:=4;
  blockwrite(Dest,NumBytes,sizeof(byte));
  blockwrite(Dest,DateStamp,sizeof(TDateTime));
  blockwrite(Dest,Count,sizeof(cardinal));
  if DataOnDisk then
  begin
    assignfile(Source,TempStorageFile);
    reset(Source,1);
    repeat
      blockread(Source,Buffer,sizeof(Buffer),Numread);
      blockwrite(Dest,Buffer,Numread,Numwrite);
    until (Numread=0) or (Numwrite<>Numread);
    closefile(Source);
    if (Numwrite<>Numread) then raise EGlobalSynopticStack.Create(
      'something went wrong with the copying of files in SaveToFile.');
  end
  else
  for I:=0 to Count-1 do
  begin
    if Items[I]=nil then raise EGlobalSynopticStack.Create('Items['+inttostr(I)+'] was nil');
    TSynopticGlobal(Items[I]).WriteToFile(Dest);
  end;
end; {End of TGlobalSynopticStack.SaveToFile.}

procedure TGlobalSynopticStack.SetNewSynopticGlobalDefaults(Defaults:TNewSynopticGlobalDefaults);
begin
  NewSynopticGlobalDefaults:=Defaults;
//  NewSynopticGlobalDefaults.Header:=Defaults.Header;
//  NewSynopticGlobalDefaults.Scaling:=Defaults.Scaling;
//  NewSynopticGlobalDefaults.Offset:=Defaults.Offset;
//  NewSynopticGlobalDefaults.FLatitudes:=Copy(Defaults.FLatitudes);
end; {End of TGlobalSynopticStack.SetNewSynopticGlobalDefaults.}

procedure TGlobalSynopticStack.SetTempStorageFile(Value:string);
begin
  if DataOnDisk then
  begin
    copyfile(TempStorageFile,Value,nil);
    DeleteFile(TempStorageFile);
  end;
  TempStorageFile:=Value;
end; {End of TGlobalSynopticStack.SetTempStorageFile.}

function CompareLevels(Item1,Item2:pointer):integer;
begin
  Result:=0;
  if TSynopticGlobal(Item1).Level<TSynopticGlobal(Item2).Level then Result:=-1;
  if TSynopticGlobal(Item1).Level>TSynopticGlobal(Item2).Level then Result:=1;
end; {End of comparelevels.}

procedure TGlobalSynopticStack.SortByLevel;
var
  I:integer;
begin
  if DataOnDisk then DataOnDisk:=false;
  if Count>1 then Sort(CompareLevels);
  for I:=0 to Count-2 do
    if TSynopticGlobal(Items[I]).Level=TSynopticGlobal(Items[I+1]).Level then
    raise EGlobalSynopticStack.Create('there are two levels in the stack with'+
      ' the same level value. This is not a good idea.');
end; {End of TGlobalSynopticStack.SortByLevel.}

destructor TGlobalSynopticStack.Destroy;
begin
  if TempStorageFile<>'' then DeleteFile(TempStorageFile);
  inherited Destroy; {This calls Clear anyway which frees the memory of each object in the TList.}
end; {End of TGlobalSynopticStack.Destroy.}

end.
