//@***h* BDBP/gb_LeastSquaresRegression.pas
// 
// NAME
//   gb_LeastSquaresRegression.pas
//
// PURPOSE
//  Defines a least squares regression analysis class.
//
// DESCRIPTION
//  This unit defines a least squares regression analysis class.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_LeastSquaresRegression;

{$I UNITDIREC.INC}

interface

uses
  Classes,
  Dialogs,
  Forms,
  SysUtils,
  gb_array,
  gb_AutocorrelationModel,
  gb_GeneralConstantsAndTypes,
  gb_RegressionBasisFunction,
  gb_TripletArray;

type
  TBlankPeriod=record
    StartDateTime,
    EndDateTime:TDateTime;
  end;
  TBlankPeriodArray=array of TBlankPeriod;

  TLeastSquaresRegression=class
  strict private
    {This field specifies the temporal resolution of the data to which the}
    {basis functions will be fitted.}
    FResolution:TResolution;

    {This object holds all of the data to which the basis functions will be}
    {fitted. The object is created in the constructor and destroyed in the }
    {destructor.}
    FData:TRegressionDataArray;

    {This field holds a list of dates and times that were passed to the basis}
    {functions. It is used to ensure that identical specifications of the    }
    {date and time list are not repeatedly sent to the basis functions.}
    DateTimeListPassedToBFs:TDoubleArray;

    {This object holds all of the basis functions registered with the}
    {regression model.}
    BasisFunctionObjects:TBasisFunctionObjectList;

    {This array holds the period during which the measurements should not be}
    {used for the regression fit.}
    BlankPeriod:TBlankPeriodArray;

    {The total number of coefficients that will be fitted.}
    NumCoeff:word;

    {The Covariance and Beta matrices.}
    Covar:T2DimDoubleArray;
    Beta:TDoubleArray;

    {This field holds the autocorrelation model to be used to account for}
    {autocorrelation in the residuals.}
    AutoCorrClass:TAutocorrelationModelClass;

    {This field holds the default number of Fourier pairs for each basis function.}
    DefaultNumFourier:TWordArray;

    {This array is filled with the optimal steps for each basis function when}
    {the .Run method is called.}
    OptimalSteps:TCardinalArray;

    {This field, controlled by a property, will determine whether the data}
    {provided to the regression are pre-sorted by time.}
    FSortByTime:boolean;

    function  CountRegressionSteps:cardinal;
    procedure PassIndependentVariablesToBasisFunctions;
    procedure PassRegCoeffsToBasisFunctions(ACoeffType:TCoeffType);
    function  RecalculateMeasurementUncertainties(AutoCorr:TAutocorrelationModel):T1DDoubleArray;
    function  RegressionFit(FillResidualArray:boolean;NewErrors:T1DDoubleArray=nil):double;
    procedure RegressionFitTransformed(AutoCorr:TAutocorrelationModel;RevisedMeasurementUncertainties:T1DDoubleArray);
    function  ScreenAndCountData:cardinal;
    procedure SetupStorageArrays;
  protected

  public
    constructor Create;overload;
    constructor Create(AResolution:TResolution;AModel:TAutocorrelationModelClass;AutoAddBFs:boolean);overload;
    function  GetNumCoefficients:word;

    {These are all methods associated with passing the measurements to the regression object.}
    procedure PassMeasurement(ADateTime:TDateTime;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure PassMeasurement(Year,Month,Date,Hour,Minute,Second:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure PassMeasurement(Year,Month,Date,Hour,Minute:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure PassMeasurement(Year,Month,Date,Hour:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure PassMeasurement(Year,Month,Date:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure PassMeasurement(Year,Month:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure PassMeasurement(Year:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure ClearAllMeasurements;

    {These are the methods used to set periods of measurements to ignore.}
    procedure SpecifyBlankPeriod(AStartDateTime,AEndDateTime:TDateTime);
    procedure ClearAllBlankPeriods;

    {These are the methods used to manipulate the basis functions.}
    procedure AddBasisFunction(ABasisFunction:TBasisFunction);
    function  GetBasisFuncVector(I:integer):TDoubleArray;
    function  NumBasisFunctions:integer;
    procedure IgnoreBasisFunction(BasisFunc:TClass);
    procedure RestoreBasisFunction(BasisFunc:TClass);
    procedure ChangeNumStepsFor(BasisFunc:TClass;NewNumSteps:word);
    procedure ChangeNumFourierFor(BasisFunc:TClass;NewNumFourier:word);
    procedure SetSeasonalResolutionForBasisFunctions(StepsPerYear:word);
    procedure MaxFourierForBasisFunctions(MaxFourier:word);
    procedure RestoreNumFourierToOriginal;
    procedure RemoveAllExtenders;

    function Run:double;

    {This is a special method for excluding the measurements with the highest residuals.}
    procedure RemoveXPercentOfDataByResiduals(Percent:double);

    {These are methods for extracting the results from the regression model run.}
    function  GetBasisFunctionsCoefficients(BasisFunc:TClass):T3DDoubleArray;overload;
    function  GetBasisFunctionsCoefficients(BasisFunc:TClass;ACoeffType:TCoeffType):T2DDoubleArray;overload;
    function  GetResolvedCoefficientArray(BasisFunc:TClass;ACoeffType:TCoeffType):T2DDoubleArray;overload;
    function  GetResolvedCoefficientArray(ABasisFunction:TBasisFunction;ACoeffType:TCoeffType):T2DDoubleArray;overload;
    function  GetBasisFunctionObject(BasisFunc:TClass):TBasisFunction;overload;
    function  GetBasisFunctionObject(I:integer):TBasisFunction;overload;
    function  GetRegressionModelTimeSeries(ACoeffType:TCoeffType;ADateTimeList:TDoubleArray=nil):T2DDoubleArray;
    function  GetRegressionResiduals:T1DDoubleArray;
    function  GetRegressionTimeSeries:T1DDoubleArray;
    function  GetDateTimeList:T1DDoubleArray;
    function  GetMeasurementTimeSeries:T1DDoubleArray;
    function  GetBasisFunctionContribution(ACoeffType:TCoeffType;BasisFunc:TClass;ADateTimeList:TDoubleArray=nil):T1DDoubleArray;
    function  GetOptimalSteps:TCardinalArray;
    procedure WriteBasisFunctionTimeSeries(BasisFunc:TClass;Filename:string);
    procedure WriteAllBasisFunctionTimeSeries(Filename:string;Original:boolean=true);
    procedure WriteMeasModelAndRes(Filename:string;ACoeffType:TCoeffType);
    procedure WriteContributionFile(Filename:string;ACoeffType:TCoeffType);overload;
    procedure WriteContributionFile(Filename:string;ACoeffType:TCoeffType;ADateTimeList:TDoubleArray);overload;
    procedure WriteBasisFunctionCoefficients(Filename:string);
    procedure WriteResolvedCoefficientArray(Filename:string;BasisFunc:TClass;ACoeffType:TCoeffType);

    destructor Destroy;override;

    property SortByTime:boolean read FSortByTime write FSortByTime;
  end;

  ELeastSquaresRegression=class(Exception);

implementation

uses
  DateUtils,
  gb_datetime,
  gb_excelutils,
  gb_math,
  gb_strings;

{------------------------------------------------------------------------------}
{ These are the method definitions for TLeastSquaresRegression.                }
{------------------------------------------------------------------------------}
constructor TLeastSquaresRegression.Create;
begin
  raise ELeastSquaresRegression.Create('the call to a TLeastSquaresRegression constructor with no parameters is not allowed.');
end;

constructor TLeastSquaresRegression.Create(AResolution:TResolution;AModel:TAutoCorrelationModelClass;AutoAddBFs:boolean);
var
  I:integer;
begin
  FResolution:=AResolution;
  FData:=TRegressionDataArray.Create;
  BasisFunctionObjects:=TBasisFunctionObjectList.Create;
  BasisFunctionObjects.OwnsObjects:=false;
  DateTimeListPassedToBFs:=nil;
  BlankPeriod:=nil;
  AutoCorrClass:=AModel;
  FSortByTime:=true;

  {Add the basis functions from the BasisFunctionList object defined in the}
  {gb_RegressionBasisFunction unit if that has been requested.}
  if AutoAddBFs then
    for I:=0 to BasisFunctionList.Count-1 do
    begin
      TBasisFunction(BasisFunctionList[I]).SetResolution(AResolution);
      BasisFunctionObjects.Add(TBasisFunction(BasisFunctionList[I]));
    end;
end;

function TLeastSquaresRegression.CountRegressionSteps:cardinal;
var
  I:integer;
begin
  Result:=1;
  for I:=0 to BasisFunctionObjects.Count-1 do
    Result:=Result*BasisFunctionObjects[I].NumSteps;
end; {End of TLeastSquaresRegression.CountRegressionSteps.}

function TLeastSquaresRegression.GetNumCoefficients:word;
var
  I:integer;
begin
  Result:=0;
  for I:=0 to BasisFunctionObjects.Count-1 do
  if BasisFunctionObjects[I].Active then Result:=Result+BasisFunctionObjects[I].GetNumCoefficients;
end; {End of TLeastSquaresRegression.GetNumCoefficients.}

procedure TLeastSquaresRegression.PassIndependentVariablesToBasisFunctions;
var
  I:integer;
  DateTimeList:TDoubleArray;
begin
  if FSortByTime then FData.SortByTime;

  {Note that we pass all of the dates and times to the basis functions}
  {irrespective of the blank periods since we still want to calculate }
  {residuals through the blank periods.}
  DateTimeList:=FData.GetDateTimeList;

  {Previously I did a lot of checks here and only called TestDateTimeList and  }
  {SpecifyIndependentVars if I thought it was absolutely necessary. But that's }
  {just too risky. It might be that the dates and times haven't changed but the}
  {latitudes or longitudes (if the ozone data are being sampled thus) may have }
  {changed which would affect the spatial sampling of those basis functions. So}
  {just to be safe I always call TestDateTimeList and SpecifyIndpendentVars.   }
  for I:=0 to BasisFunctionObjects.Count-1 do
  with BasisFunctionObjects[I] do
  if Active then
  begin
    TestDateTimeList(DateTimeList);
    SpecifyIndependentVars(FData);
  end;
end; {End of TLeastSquaresRegression.PassIndependentVariablesToBasisFunctions.}

procedure TLeastSquaresRegression.PassRegCoeffsToBasisFunctions(ACoeffType:TCoeffType);
var
  I,J,Index:integer;
  Hold:double;
begin
  Index:=-1;
  for I:=0 to BasisFunctionObjects.Count-1 do
  begin
    for J:=1 to BasisFunctionObjects[I].GetNumCoefficients do
      if BasisFunctionObjects[I].Active then
      begin
        inc(Index);
        Hold:=Covar[Index,Index];
        if Hold<0.0 then Hold:=0.0;
        BasisFunctionObjects[I].SetRegCoeff(ACoeffType,J,Beta[Index],sqrt(Hold));
      end
      else BasisFunctionObjects[I].SetRegCoeff(ACoeffType,J,0.0,0.0);
  end;
end; {End of TLeastSquaresRegression.PassRegCoeffsToBasisFunctions.}

function TLeastSquaresRegression.RecalculateMeasurementUncertainties(AutoCorr:TAutocorrelationModel):T1DDoubleArray;
type
  StdvRec=record
    Mean,
    Stdv:double;
    Num:longint;
  end;
var
  ResidualsAfterAutoCorrelation,
  MonthlyStdv,
  OrigMeasurementErrors,
  ResidualsBeforeAutoCorrelation:T1DDoubleArray;
  I:integer;
  Month:word;
  MonthlyData:array[1..12] of StdvRec;
begin
  {Calculate the monthly standard deviations of the residuals which have been}
  {transformed to account for their autocorrelation.}
  try {Start of exception trapping block.}
    try {Start of resource protection block.}
      ResidualsBeforeAutoCorrelation:=FData.Get1DResidualsArray;
      ResidualsAfterAutoCorrelation:=AutoCorr.GenerateTransformedDataSeries(ResidualsBeforeAutoCorrelation,false);
      OrigMeasurementErrors:=T1DDoubleArray.Create(FData.Count);
      MonthlyStdv:=T1DDoubleArray.Create(12);
      for I:=1 to 12 do
      begin
        MonthlyData[I].Mean:=0.0;
        MonthlyData[I].Stdv:=0.0;
        MonthlyData[I].Num:=0;
      end;
      for I:=1 to FData.Count do
      begin
        Month:=MonthOf(FData.GetDateTime(I));
        MonthlyData[Month].Mean:=MonthlyData[Month].Mean+ResidualsAfterAutoCorrelation[I];
        inc(MonthlyData[Month].Num);
      end;
      for I:=1 to 12 do
        if MonthlyData[I].Num>0 then MonthlyData[I].Mean:=MonthlyData[I].Mean/MonthlyData[I].Num else MonthlyData[I].Mean:=0.0;
      for I:=1 to FData.Count do
      begin
        Month:=MonthOf(FData.GetDateTime(I));
        MonthlyData[Month].Stdv:=MonthlyData[Month].Stdv+sqr(ResidualsAfterAutoCorrelation[I]-MonthlyData[Month].Mean);
      end;
      for I:=1 to 12 do
        if MonthlyData[I].Num>1 then MonthlyStdv[I]:=sqrt(MonthlyData[I].Stdv/(MonthlyData[I].Num-1)) else MonthlyStdv[I]:=0.0;
      for I:=1 to FData.Count do OrigMeasurementErrors[I]:=FData.GetMeasurementError(I);
      Result:=AutoCorr.GenerateRevisedMeasurementErrors(OrigMeasurementErrors,MonthlyStdv);
    finally
      FreeAndNil(ResidualsBeforeAutoCorrelation);
      FreeAndNil(ResidualsAfterAutoCorrelation);
      FreeAndNil(MonthlyStdv);
      FreeAndNil(OrigMeasurementErrors);
    end;
  except
    on E:EAutoCorrelationModel do raise ELeastSquaresRegression.Create(
      'RecalculateMeasurementUncertainties failed because of a known error. The error message was: '+E.Message);
    on E:Exception do raise ELeastSquaresRegression.Create(
      'RecalculateMeasurementUncertainties failed because of an unknown error. The error message was: '+E.Message);
  end;
end;

function TLeastSquaresRegression.GetBasisFuncVector(I:integer):TDoubleArray;
var
  Index,
  J,K:integer;
  Hold:TDoubleArray;
begin
  setlength(Result,NumCoeff);
  Index:=0;
  for J:=0 to BasisFunctionObjects.Count-1 do
  if BasisFunctionObjects[J].Active then
  begin
    Hold:=BasisFunctionObjects[J].GetBasisVector(FData.GetDateTime(I),FData.GetInfoObject(I));
    for K:=Low(Hold) to High(Hold) do
    begin
      Result[Index]:=Hold[K];
      inc(Index);
    end;
  end;
end;

function TLeastSquaresRegression.NumBasisFunctions:integer;
begin
  Result:=BasisFunctionObjects.Count;
end;

procedure TLeastSquaresRegression.IgnoreBasisFunction(BasisFunc:TClass);
begin
  BasisFunctionObjects.GetBasisObj(BasisFunc).Active:=false;
end;

procedure TLeastSquaresRegression.RestoreBasisFunction(BasisFunc:TClass);
begin
  BasisFunctionObjects.GetBasisObj(BasisFunc).Active:=true;
end;

procedure TLeastSquaresRegression.ChangeNumStepsFor(BasisFunc:TClass;NewNumSteps:word);
begin
  BasisFunctionObjects.GetBasisObj(BasisFunc).NumSteps:=NewNumSteps;
end;

procedure TLeastSquaresRegression.ChangeNumFourierFor(BasisFunc:TClass;NewNumFourier:word);
begin
  BasisFunctionObjects.GetBasisObj(BasisFunc).NumFourier:=NewNumFourier;
end;

procedure TLeastSquaresRegression.SetSeasonalResolutionForBasisFunctions(StepsPerYear:word);
var
  I:integer;
begin
  for I:=0 to BasisFunctionObjects.Count-1 do
    BasisFunctionObjects[I].SeasonalResolution:=StepsPerYear;
end;

procedure TLeastSquaresRegression.MaxFourierForBasisFunctions(MaxFourier:word);
var
  I:integer;
begin
  setlength(DefaultNumFourier,BasisFunctionObjects.Count);
  for I:=0 to BasisFunctionObjects.Count-1 do
  begin
    DefaultNumFourier[I]:=BasisFunctionObjects[I].NumFourier;
    if BasisFunctionObjects[I].NumFourier>MaxFourier then BasisFunctionObjects[I].NumFourier:=MaxFourier;
  end;
end;

procedure TLeastSquaresRegression.RestoreNumFourierToOriginal;
var
  I:integer;
begin
  if length(DefaultNumFourier)<>0 then
  begin
    for I:=0 to BasisFunctionObjects.Count-1 do BasisFunctionObjects[I].NumFourier:=DefaultNumFourier[I];
    setlength(DefaultNumFourier,0);
  end;
end;

procedure TLeastSquaresRegression.RemoveAllExtenders;
var
  I:integer;
begin
  for I:=0 to BasisFunctionObjects.Count-1 do BasisFunctionObjects[I].RemoveExtender;
end;

function TLeastSquaresRegression.RegressionFit(FillResidualArray:boolean;NewErrors:T1DDoubleArray=nil):double;
var
  WorkStr:string;
  Dest:textfile; {For reporting errors.}
  I,J,K:integer; {Loop control variables.}
  MeasMinusModel, {The residual value.}
  Weight,         {The weighting parameter.}
  Sum,            {Used in the final chi squared summation.}
  Sigma2i:double; {1 divided by sigma squared.}
  BasisFunc:TDoubleArray; {The basis function vector evaluated at a specific time.}
begin
  for I:=0 to NumCoeff-1 do
  begin
    for J:=0 to NumCoeff-1 do Covar[I,J]:=0.0;
    Beta[I]:=0.0;
  end;
  for I:=1 to FData.Count do
  if FData.UseInRegression(I) then
  begin
    BasisFunc:=GetBasisFuncVector(I);
    if NewErrors=nil then Sigma2i:=1.0/sqr(FData.GetMeasurementError(I)) else Sigma2i:=1.0/sqr(NewErrors[I]);
    for J:=0 to NumCoeff-1 do
    begin
      Weight:=BasisFunc[J]*Sigma2i;
      for K:=0 to J do Covar[J,K]:=Covar[J,K]+(Weight*BasisFunc[K]);
      Beta[J]:=Beta[J]+(Weight*FData.GetMeasurement(I));
    end;
  end;
  for I:=1 to NumCoeff-1 do
    for J:=0 to I-1 do
      Covar[J,I]:=Covar[I,J];
  Application.ProcessMessages;
  try
    GaussJordan(Covar,Beta);
  except
    on E:EGBMath do
    begin
      assignfile(Dest,'b:\junk.txt');
      rewrite(Dest);
      write(Dest,'      ');
      for I:=0 to BasisFunctionObjects.Count-1 do
      if BasisFunctionObjects[I].Active then
      begin
        WorkStr:=BasisFunctionObjects[I].Name;
        BackBuffer(WorkStr,' ',BasisFunctionObjects[I].GetNumCoefficients*9);
        write(Dest,WorkStr);
      end;
      writeln(Dest);
      for I:=1 to FData.Count do
      if FData.UseInRegression(I) then
      begin
        BasisFunc:=GetBasisFuncVector(I);
        write(Dest,I:5,' ');
        for J:=0 to NumCoeff-1 do write(Dest,BasisFunc[J]:8:3,' ');
        writeln(Dest);
      end;
      for I:=Low(Beta) to High(Beta) do Beta[I]:=0.0;
      ShowMessage('GaussJordan failed. See b:\junk.txt. All regression fit coefficients are set to zero.');
    end;
  end;
  {Beta now holds the regression coefficients.}
  Result:=0.0;
  for I:=1 to FData.Count do
  begin
    BasisFunc:=GetBasisFuncVector(I);
    Sum:=0.0;
    for J:=0 to NumCoeff-1 do Sum:=Sum+(Beta[J]*BasisFunc[J]);
    MeasMinusModel:=FData.GetMeasurement(I)-Sum;
    if FillResidualArray then FData.PutResidual(I,MeasMinusModel);
    if FData.UseInRegression(I) then
    begin
      if NewErrors=nil then Result:=Result+sqr(MeasMinusModel/FData.GetMeasurementError(I))
      else Result:=Result+sqr(MeasMinusModel/NewErrors[I]);
    end;
  end;
end;

procedure TLeastSquaresRegression.RegressionFitTransformed(AutoCorr:TAutocorrelationModel;RevisedMeasurementUncertainties:T1DDoubleArray);
var
  I,J,K:integer;
  TransformedMeasurements:T1DDoubleArray;
  Sigma2i,
  Weight:double;
  TransformedBasisFunctions:array of T1DDoubleArray;
  BasisFunc:TDoubleArray; {The basis function vector evaluated at a specific time.}
begin
  {First calculate the transformed measurements.}
  TransformedMeasurements:=AutoCorr.GenerateTransformedDataSeries(FData.Get1DMeasurementArray,true);

  {Now calculate the transformed basis functions.}
  setlength(TransformedBasisFunctions,NumCoeff);
  for I:=1 to NumCoeff do TransformedBasisFunctions[I-1]:=T1DDoubleArray.Create(FData.Count);
  for I:=1 to FData.Count do
  begin
    BasisFunc:=GetBasisFuncVector(I);
    for J:=1 to NumCoeff do TransformedBasisFunctions[J-1][I]:=BasisFunc[J-1];
  end;
  for I:=1 to NumCoeff do TransformedBasisFunctions[I-1]:=AutoCorr.GenerateTransformedDataSeries(TransformedBasisFunctions[I-1],true);

  {Now we are ready to go.}
  for I:=0 to NumCoeff-1 do
  begin
    for J:=0 to NumCoeff-1 do Covar[I,J]:=0.0;
    Beta[I]:=0.0;
  end;
  try
    for I:=1 to FData.Count do
    if FData.UseInRegression(I) then
    begin
      Sigma2i:=1.0/sqr(RevisedMeasurementUncertainties[I]);
      for J:=0 to NumCoeff-1 do
      begin
        Weight:=TransformedBasisFunctions[J][I]*Sigma2i;
        for K:=0 to J do Covar[J,K]:=Covar[J,K]+(Weight*TransformedBasisFunctions[K][I]);
        Beta[J]:=Beta[J]+(Weight*TransformedMeasurements[I]);
      end;
    end;
    for I:=1 to NumCoeff-1 do
      for J:=0 to I-1 do
        Covar[J,I]:=Covar[I,J];
    GaussJordan(Covar,Beta);
    {Beta now holds the regression coefficients.}
  finally
    for I:=0 to NumCoeff-1 do FreeAndNil(TransformedBasisFunctions[I]);
    FreeAndNil(TransformedMeasurements);
  end;
end;

function TLeastSquaresRegression.ScreenAndCountData:cardinal;
var
  I,J:integer;
  TestDateTime:TDateTime;
begin
  Result:=0;
  for I:=1 to FData.Count do
  begin
    TestDateTime:=FData.GetDateTime(I);
    for J:=Low(BlankPeriod) to High(BlankPeriod) do
      if ((TestDateTime>=BlankPeriod[J].StartDateTime) and
          (TestDateTime<=BlankPeriod[J].EndDateTime)) then FData.PutUseInRegression(I,false);
    if FData.UseInRegression(I) then inc(Result);
  end;
end; {End of TLeastSquaresRegression.ScreenAndCountData.}

procedure TLeastSquaresRegression.SetupStorageArrays;
begin
  setlength(Covar,NumCoeff,NumCoeff);
  setlength(Beta,NumCoeff);
end; {End of TLeastSquaresRegression.SetupStorageArrays.}

procedure TLeastSquaresRegression.PassMeasurement(ADateTime:TDateTime;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject);
begin
  if AMeasurementError<=0.0 then raise ELeastSquaresRegression.Create('negative or zero measurement error values are not permitted.');
  FData.AddEntry(ADateTime,AMeasurement,AMeasurementError,-1E30,true,AInfoObj);
end;

procedure TLeastSquaresRegression.PassMeasurement(Year,Month,Date,Hour,Minute,Second:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);
begin
  if AMeasurementError<=0.0 then raise ELeastSquaresRegression.Create('negative or zero measurement error values are not permitted.');
  {Note that the convention for the DateTime stamp is the middle of the time period.}
  if FResolution in [ResRandom,ResSecond] then FData.AddEntry(EncodeDateTime(Year,
    Month,Date,Hour,Minute,Second,500),AMeasurement,AMeasurementError,-1E30,true,AInfoObj)
  else raise ELeastSquaresRegression.Create('this method should only be used when the resolution is ResRandom or ResSecond');
end;

procedure TLeastSquaresRegression.PassMeasurement(Year,Month,Date,Hour,Minute:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);
begin
  if AMeasurementError<=0.0 then raise ELeastSquaresRegression.Create('negative or zero measurement error values are not permitted.');
  {Note that the convention for the DateTime stamp is the middle of the time period.}
  if FResolution in [ResRandom,ResMinute] then FData.AddEntry(EncodeDateTime(Year,
    Month,Date,Hour,Minute,30,0),AMeasurement,AMeasurementError,-1E30,true,AInfoObj)
  else raise ELeastSquaresRegression.Create('this method should only be used when the resolution is ResRandom or ResMinute');
end;

procedure TLeastSquaresRegression.PassMeasurement(Year,Month,Date,Hour:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);
begin
  if AMeasurementError<=0.0 then raise ELeastSquaresRegression.Create('negative or zero measurement error values are not permitted.');
  {Note that the convention for the DateTime stamp is the middle of the time period.}
  if FResolution in [ResRandom,ResHour] then FData.AddEntry(EncodeDateTime(Year,
    Month,Date,Hour,30,0,0),AMeasurement,AMeasurementError,-1E30,true,AInfoObj)
  else raise ELeastSquaresRegression.Create('this method should only be used when the resolution is ResRandom or ResHour');
end;

procedure TLeastSquaresRegression.PassMeasurement(Year,Month,Date:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);
begin
  if AMeasurementError<=0.0 then raise ELeastSquaresRegression.Create('negative or zero measurement error values are not permitted.');
  {Note that the convention for the DateTime stamp is the middle of the time period.}
  if FResolution in [ResRandom,ResDay] then FData.AddEntry(EncodeDate(Year,Month,Date)+0.5,
    AMeasurement,AMeasurementError,-1E30,true,AInfoObj)
  else raise ELeastSquaresRegression.Create('this method should only be used when the resolution is ResRandom or ResDay');
end;

procedure TLeastSquaresRegression.PassMeasurement(Year,Month:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);
begin
  if AMeasurementError<=0.0 then raise ELeastSquaresRegression.Create('negative or zero measurement error values are not permitted.');
  {Note that the convention for the DateTime stamp is the middle of the time period.}
  if FResolution in [ResRandom,ResMonth] then FData.AddEntry(EncodeDate(Year,Month,1)+
    DaysInAMonth(Year,Month)/2,AMeasurement,AMeasurementError,-1E30,true,AInfoObj)
  else raise ELeastSquaresRegression.Create('this method should only be used when the resolution is ResRandom or ResMonth');
end;

procedure TLeastSquaresRegression.PassMeasurement(Year:word;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);
begin
  if AMeasurementError<=0.0 then raise ELeastSquaresRegression.Create('negative or zero measurement error values are not permitted.');
  {Note that the convention for the DateTime stamp is the middle of the time period.}
  if FResolution in [ResRandom,ResYear] then FData.AddEntry(EncodeDate(Year,1,1)+
    DaysInAYear(Year)/2,AMeasurement,AMeasurementError,-1E30,true,AInfoObj)
  else raise ELeastSquaresRegression.Create('this method should only be used when the resolution is ResRandom or ResYear');
end;

procedure TLeastSquaresRegression.ClearAllMeasurements;
begin
  FreeAndNil(FData);
  FData:=TRegressionDataArray.Create;
end;

procedure TLeastSquaresRegression.SpecifyBlankPeriod(AStartDateTime,AEndDateTime:TDateTime);
var
  I:integer;
begin
  I:=length(BlankPeriod);
  setlength(BlankPeriod,I+1);
  BlankPeriod[I].StartDateTime:=AStartDateTime;
  BlankPeriod[I].EndDateTime:=AEndDateTime;
end;

procedure TLeastSquaresRegression.ClearAllBlankPeriods;
begin
  setlength(BlankPeriod,0);
end;

procedure TLeastSquaresRegression.AddBasisFunction(ABasisFunction:TBasisFunction);
begin
  {The only information we can pass to the basis function at this time is the}
  {resolution of the data.}
  if not ABasisFunction.SetResolution(FResolution) then raise ELeastSquaresRegression.Create(
    'The basis function couldn''t be added to the regression model because it has '+
    'not been designed to cope with the temporal resolution set for the regression model.');
  BasisFunctionObjects.Add(ABasisFunction);
end;

function TLeastSquaresRegression.Run:double;
var
  I:integer;
  NumSteps,
  StepCount,
  NumData:cardinal; {The total number of measurements to be used in the regression.}
  MinRes:double;    {The current minimum of the sum of the squares of the residuals.}
  StepIndex:TCardinalArray;
  RevisedMeasurementUncertainties:T1DDoubleArray;
  AutoCorr:TAutocorrelationModel;
begin
  {Determine which measurements will be used in the regression model.}
  NumData:=ScreenAndCountData;
  NumCoeff:=GetNumCoefficients;

  if NumData>=(2*NumCoeff) then
  begin
    {Tell all of the basis functions what dates and times they will have to deal}
    {with. This routine also sorts measurements into ascending time order.}
    PassIndependentVariablesToBasisFunctions;

    {Setup the various storage arrays ie. the Covariance matrix and the Beta matrix.}
    SetupStorageArrays;

    setlength(StepIndex,BasisFunctionObjects.Count);
    for I:=1 to BasisFunctionObjects.Count do StepIndex[I-1]:=1;
    NumSteps:=CountRegressionSteps;
    MinRes:=1E30;
    for StepCount:=1 to NumSteps do
    begin
      {Set the step indices for each basis function.}
      for I:=0 to BasisFunctionObjects.Count-1 do
        BasisFunctionObjects[I].SetStep(StepIndex[I]);

      {Run the model and get the sum squared residuals.}
      Result:=RegressionFit(true);

      if Result<MinRes then
      begin
        OptimalSteps:=copy(StepIndex);
        MinRes:=Result;
      end;

      {Step through the step indices.}
      if StepCount<>NumSteps then
      begin
        inc(StepIndex[0]);
        for I:=0 to BasisFunctionObjects.Count-1 do
        if StepIndex[I]>BasisFunctionObjects[I].NumSteps then
        begin
          StepIndex[I]:=1;
          inc(StepIndex[I+1]);
        end;
      end;
    end;

    {If necessary, run the model again with the optimal steps.}
    if not CompareTCardinalArrays(OptimalSteps,StepIndex) then
    begin
      for I:=0 to BasisFunctionObjects.Count-1 do BasisFunctionObjects[I].SetStep(OptimalSteps[I]);
      RegressionFit(true);
    end;

    {Transfer the model regression coefficients to the basis functions.}
    PassRegCoeffsToBasisFunctions(ResInitial);

    {Run the autocorrelation model if necessary.}
    if AutoCorrClass<>nil then
    begin
      AutoCorr:=AutoCorrClass.Create(FData);
      try
        AutoCorr.Run;
        RevisedMeasurementUncertainties:=nil; {Need to do this in case RecalculateMeasurementUncertainties raises an exception.}
        RevisedMeasurementUncertainties:=RecalculateMeasurementUncertainties(AutoCorr);
        RegressionFit(false,RevisedMeasurementUncertainties);
        PassRegCoeffsToBasisFunctions(ResIntermediate);
        RegressionFitTransformed(AutoCorr,RevisedMeasurementUncertainties);
        PassRegCoeffsToBasisFunctions(ResFinal);
      finally
        FreeAndNil(RevisedMeasurementUncertainties);
        FreeAndNil(AutoCorr);
      end;
    end;
  end
  else raise ELeastSquaresRegression.Create('there were insufficient measurements to run the regression model.');
end; {End of TLeastSquaresRegression.Run.}

procedure TLeastSquaresRegression.RemoveXPercentOfDataByResiduals(Percent:double);
{------------------------------------------------------------------------------}
{ This procedure sorts the FData array by the residuals and then deletes the   }
{ top and bottom Percent/2 percent of the data.                                }
{------------------------------------------------------------------------------}
var
  NumToDelete,
  I:integer;
begin
  FData.SortByResiduals;
  NumToDelete:=trunc(FData.Count*Percent/200);
  for I:=1 to NumToDelete do
  begin
    {Note that because deleting indexing changes the record numbers}
    {we just delete the first and last entries.}
    FData.DeleteEntry(1);
    FData.DeleteEntry(FData.Count);
  end;
  FData.SortByTime;
end;

function TLeastSquaresRegression.GetBasisFunctionsCoefficients(BasisFunc:TClass):T3DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
begin
  Result:=BasisFunctionObjects.GetBasisObj(BasisFunc).GetRegCoeffArray;
end;

function TLeastSquaresRegression.GetBasisFunctionsCoefficients(BasisFunc:TClass;ACoeffType:TCoeffType):T2DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
begin
  Result:=BasisFunctionObjects.GetBasisObj(BasisFunc).GetRegCoeffArray(ACoeffType);
end;

function TLeastSquaresRegression.GetResolvedCoefficientArray(BasisFunc:TClass;ACoeffType:TCoeffType):T2DDoubleArray;
begin
  Result:=BasisFunctionObjects.GetBasisObj(BasisFunc).GetResolvedCoeffArray(ACoeffType);
end;

function TLeastSquaresRegression.GetResolvedCoefficientArray(ABasisFunction:TBasisFunction;ACoeffType:TCoeffType):T2DDoubleArray;
begin
  Result:=ABasisFunction.GetResolvedCoeffArray(ACoeffType);
end;

function TLeastSquaresRegression.GetBasisFunctionObject(BasisFunc:TClass):TBasisFunction;
begin
  Result:=BasisFunctionObjects.GetBasisObj(BasisFunc);
end;

function TLeastSquaresRegression.GetBasisFunctionObject(I:integer):TBasisFunction;
begin
  if ((I<1) or (I>BasisFunctionObjects.Count)) then raise ELeastSquaresRegression.Create(
    'the index passed to GetBasisFunctionObject was out of range.');
  Result:=BasisFunctionObjects[I-1];
end;

function TLeastSquaresRegression.GetRegressionModelTimeSeries(ACoeffType:TCoeffType;ADateTimeList:TDoubleArray=nil):T2DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
var
  I,J:integer;
  Hold:TDoubleArray;
  Total:double;
begin
  if ADateTimeList=nil then Hold:=FData.GetDateTimeList else Hold:=ADateTimeList;
  Result:=T2DDoubleArray.Create(2,length(Hold));
  for I:=Low(Hold) to High(Hold) do
  begin
    Total:=0.0;
    for J:=1 to BasisFunctionObjects.Count do
      Total:=Total+BasisFunctionObjects[J-1].GetContribution(ACoeffType,Hold[I]);
    Result[1,I+1]:=Hold[I];
    Result[2,I+1]:=Total;
  end;
end;

function TLeastSquaresRegression.GetRegressionResiduals:T1DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
begin
  Result:=FData.Get1DResidualsArray;
end;

function TLeastSquaresRegression.GetRegressionTimeSeries:T1DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  Result:=T1DDoubleArray.Create(FData.Count);
  for I:=1 to FData.Count do Result[I]:=FData.GetMeasurement(I)-FData.GetResidual(I);
end;

function TLeastSquaresRegression.GetDateTimeList:T1DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
var
  Hold:TDoubleArray;
  I:integer;
begin
  Hold:=FData.GetDateTimeList;
  Result:=T1DDoubleArray.Create(length(Hold));
  for I:=Low(Hold) to High(Hold) do Result[I+1]:=Hold[I];
  setlength(Hold,0);
end; {End of TLeastSquaresRegression.GetDateTimeList.}

function TLeastSquaresRegression.GetMeasurementTimeSeries:T1DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  Result:=T1DDoubleArray.Create(FData.Count);
  for I:=1 to FData.Count do Result[I]:=FData.GetMeasurement(I);
end; {End of TLeastSquaresRegression.GetMeasurementTimeSeries.}

function TLeastSquaresRegression.GetBasisFunctionContribution(ACoeffType:TCoeffType;BasisFunc:TClass;ADateTimeList:TDoubleArray=nil):T1DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling programme to free the returned       }
{ object.                                                                      }
{------------------------------------------------------------------------------}
var
  Hold:TDoubleArray;
  HoldObj:TBasisFunction;
  I:integer;
begin
  HoldObj:=BasisFunctionObjects.GetBasisObj(BasisFunc);
  if ADateTimeList=nil then Hold:=FData.GetDateTimeList else Hold:=ADateTimeList;
  Result:=T1DDoubleArray.Create(length(Hold));
  for I:=Low(Hold) to High(Hold) do
    Result[I+1]:=HoldObj.GetContribution(ACoeffType,Hold[I]);
end;

function TLeastSquaresRegression.GetOptimalSteps:TCardinalArray;
begin
  Result:=copy(OptimalSteps);
end; {End of TLeastSquaresRegression.GetOptimalSteps.}

procedure TLeastSquaresRegression.WriteBasisFunctionTimeSeries(BasisFunc:TClass;Filename:string);
begin
  BasisFunctionObjects.GetBasisObj(BasisFunc).WriteTimeSeriesToFile(FData.GetDateTimeList,Filename);
end;

procedure TLeastSquaresRegression.WriteAllBasisFunctionTimeSeries(Filename:string;Original:boolean=true);
var
  Dest:textfile;
  Hold:TDoubleArray;
  I,J:integer;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  write(Dest,'''These are the ');
  if Original then write(Dest,'original') else write(Dest,'treated');
  writeln(Dest,' values from all of the basis functions used in the regression.');
  writeln(Dest,'''Column A) The decimal year.');
  for I:=0 to BasisFunctionObjects.Count-1 do
    writeln(Dest,'''Column '+ConvertExcelNumToCol(I+2)+') The values for the '+
    BasisFunctionObjects[I].Name+' basis function.');
  Hold:=FData.GetDateTimeList;
  for I:=Low(Hold) to High(Hold) do
  begin
    write(Dest,ConvertDateTimeToDecYear(Hold[I]):9:4,' ');
    for J:=0 to BasisFunctionObjects.Count-1 do
      if Original then write(Dest,BasisFunctionObjects[J].GetValue(Hold[I]):10:4)
      else write(Dest,BasisFunctionObjects[J].GetTreatedValue(Hold[I]):10:4);
    writeln(Dest);
  end;
  closefile(Dest);
end;

procedure TLeastSquaresRegression.WriteMeasModelAndRes(Filename:string;ACoeffType:TCoeffType);
var
  I,J:integer;
  Dest:textfile;
  Total:double;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  if FData.GetInfoObject(1)=nil then {I haven't written the code for the alternative case yet.}
  begin
    writeln(Dest,'''Column 1) The Delphi Date/Time stamp.');
    writeln(Dest,'''Column 2) The decimal year.');
    writeln(Dest,'''Column 3) The measurement.');
    writeln(Dest,'''Column 4) The regression modelled value.');
    writeln(Dest,'''Column 5) The residual.');
  end;
  for I:=1 to FData.Count do
  begin
    write(Dest,FData.GetDateTime(I):9:3,' ');
    write(Dest,ConvertDateTimeToDecYear(FData.GetDateTime(I)):9:4,' ');
    if FData.GetInfoObject(I)<>nil then FData.GetInfoObject(I).WriteToFile(Dest);
    write(Dest,FData.GetMeasurement(I):10:5,' ');
    Total:=0.0;
    for J:=1 to BasisFunctionObjects.Count do Total:=Total+
      BasisFunctionObjects[J-1].GetContribution(ACoeffType,FData.GetDateTime(I),FData.GetInfoObject(I));
    write(Dest,Total:10:5,' ');
    writeln(Dest,FData.GetMeasurement(I)-Total:10:5);
  end;
  closefile(Dest);
end;

procedure TLeastSquaresRegression.WriteContributionFile(Filename:string;ACoeffType:TCoeffType);
var
  I,J,Col:integer;
  Dest:textfile;
  BFContribution,
  Total:double;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  writeln(Dest,'''Column A) The delphi date/time value.');
  writeln(Dest,'''Column B) The decimal year.');
  Col:=2;
  if FData.GetInfoObject(1)<>nil then
    for I:=1 to FData.GetInfoObject(1).NumFields do
    begin
      inc(Col);
      writeln(Dest,'''Column '+ConvertExcelNumToCol(Col)+') '+FData.GetInfoObject(1).GetFieldDescriptor(I));
    end;
  inc(Col);
  writeln(Dest,'''Column '+ConvertExcelNumToCol(Col)+') The measurements passed to the regression model.');
  for I:=1 to BasisFunctionObjects.Count do
  begin
    inc(Col);
    writeln(Dest,'''Column '+ConvertExcelNumToCol(Col)+') The contribution from the '+BasisFunctionObjects[I-1].Name+' basis function.');
  end;
  inc(Col);
  writeln(Dest,'''Column '+ConvertExcelNumToCol(Col)+') The sum of the contributions from all of the basis functions.');
  inc(Col);
  writeln(Dest,'''Column '+ConvertExcelNumToCol(Col)+') The measurements minus the regression model i.e. the residuals.');
  write(Dest,'''DateTime DecYear     ');
  if FData.GetInfoObject(1)<>nil then write(Dest,FData.GetInfoObject(1).GetColumnsDescriptor);
  write(Dest,'Measments ');
  for I:=1 to BasisFunctionObjects.Count do write(Dest,copy(BasisFunctionObjects[I-1].Name+'         ',1,11));
  writeln(Dest,'Total_Reg  Residuals');
  for I:=1 to FData.Count do
  begin
    write(Dest,FData.GetDateTime(I):9:3,' ');
    write(Dest,ConvertDateTimeToDecYear(FData.GetDateTime(I)):10:5,' ');
    if FData.GetInfoObject(I)<>nil then FData.GetInfoObject(I).WriteToFile(Dest);
    write(Dest,FData.GetMeasurement(I):10:5,' ');
    Total:=0.0;
    for J:=1 to BasisFunctionObjects.Count do
    begin
      BFContribution:=BasisFunctionObjects[J-1].GetContribution(ACoeffType,FData.GetDateTime(I),FData.GetInfoObject(I));
      Total:=Total+BFContribution;
      write(Dest,BFContribution:10:5,' ');
    end;
    write(Dest,Total:10:5,' ');
    writeln(Dest,FData.GetMeasurement(I)-Total:10:5);
  end;
  closefile(Dest);
end;

procedure TLeastSquaresRegression.WriteContributionFile(Filename:string;ACoeffType:TCoeffType;ADateTimeList:TDoubleArray);
var
  I,J,Col:integer;
  Dest:textfile;
  BFContribution,
  Total:double;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  writeln(Dest,'''Column A) The delphi date/time value.');
  writeln(Dest,'''Column B) The decimal year.');
  Col:=2;
  for I:=1 to BasisFunctionObjects.Count do
  begin
    inc(Col);
    writeln(Dest,'''Column '+ConvertExcelNumToCol(Col)+') The contribution from the '+BasisFunctionObjects[I-1].Name+' basis function.');
  end;
  inc(Col);
  writeln(Dest,'''Column '+ConvertExcelNumToCol(Col)+') The sum of the contributions from all of the basis functions.');
  write(Dest,'''DateTime DecYear     ');
  for I:=1 to BasisFunctionObjects.Count do write(Dest,copy(BasisFunctionObjects[I-1].Name+'         ',1,11));
  writeln(Dest,'Total_Reg');
  for I:=Low(ADateTimeList) to High(ADateTimeList) do
  begin
    write(Dest,ADateTimeList[I]:9:3,' ');
    write(Dest,ConvertDateTimeToDecYear(ADateTimeList[I]):10:5,' ');
    Total:=0.0;
    for J:=1 to BasisFunctionObjects.Count do
    begin
      BFContribution:=BasisFunctionObjects[J-1].GetContribution(ACoeffType,ADateTimeList[I]);
      Total:=Total+BFContribution;
      write(Dest,BFContribution:10:5,' ');
    end;
    writeln(Dest,Total:10:5);
  end;
  closefile(Dest);
end;

procedure TLeastSquaresRegression.WriteBasisFunctionCoefficients(Filename:string);
var
  Dest:textfile;
  I,J,K:integer;
  Coeffs:T3DDoubleArray;
  StringExp:TStringList;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  for I:=1 to BasisFunctionObjects.Count do
  begin
    Coeffs:=BasisFunctionObjects[I-1].GetRegCoeffArray;
    StringExp:=BasisFunctionObjects[I-1].GetStringListOfExpansions;
    try
      writeln(Dest,'These are the coefficients for the '+BasisFunctionObjects[I-1].Name+' basis function.');
      for J:=1 to 3 do
      begin
        writeln(Dest);
        case J of
          1:writeln(Dest,'  These are the coefficients before auto-correlation was accounted for:');
          2:writeln(Dest,'  These are the coefficients after the measurement uncertainties were revised:');
          3:writeln(Dest,'  These are the coefficients after auto-correlation in the residuals was fully accounted for:');
        end;
        for K:=1 to Coeffs.GetNumZ do
          writeln(Dest,'  ',Coeffs[1,J,K]:8:4,' +- ',Coeffs[2,J,K]:8:4,' --> ',StringExp[K-1]);
      end;
      writeln(Dest);
      writeln(Dest);
    finally
      FreeAndNil(Coeffs);
      FreeAndNil(StringExp);
    end;
  end;
  closefile(Dest);
end;

procedure TLeastSquaresRegression.WriteResolvedCoefficientArray(Filename:string;BasisFunc:TClass;ACoeffType:TCoeffType);
var
  Hold:T2DDoubleArray;
begin
  Hold:=GetResolvedCoefficientArray(BasisFunc,ACoeffType);
  try
    Hold.WriteToTextFile(Filename);
  finally
    FreeAndNil(Hold);
  end;
end;



destructor TLeastSquaresRegression.Destroy;
begin
  FreeAndNil(BasisFunctionObjects);
  FreeAndNil(FData);
  inherited Destroy;
end;

end.
