//@***h* BDBP/gb_RegressionBasisFunction.pas
// 
// NAME
//   gb_RegressionBasisFunction.pas
//
// PURPOSE
//  Deals with all the technical aspects of handling and providing basis functions 
//  for a regression model.
//
// DESCRIPTION
//  This unit deals with all the technical aspects of handling and providing basis 
//  functions for a regression model.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_RegressionBasisFunction;

{$I UNITDIREC.INC}

interface

uses
  Classes,
  Contnrs,
  SysUtils,
  gb_4DSynoptic,
  gb_Array,
  gb_AutocorrelationModel,
  gb_GeneralConstantsAndTypes,
  gb_Math,
  gb_SynopticGlobal,
  gb_tripletarray;

type
  TCoeffType=(ResInitial,ResIntermediate,ResFinal);
  TBasisFunctionTreatment=(DontChange,RemoveOffset,RemoveOffsetAndTrend);
  TPType=T2DimDoubleArray;
  TPMType=array of array of array of array of double;

  TCoeffRec=record
    Coeff,
    Error:double;
  end;

  TBasisFuncInitRec=record
    Name:string;
    NumFourier:word;
    Treatment:TBasisFunctionTreatment;
  end;

  TBasisFunctionExtender=class
  private
    FOwner:TObject;
  public
    function GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;virtual;abstract;

    {Descendents of this virtual method must return the number of additional}
    {regression model fit coefficients introduced by the extender.}
    function GetNumCoefficients:word;virtual;abstract;

    function GetStringListOfExpansions(InputList:TStringList):TStringList;virtual;abstract;

    {Given the array of fit coefficients from the regression model, descendents}
    {of this virtual method must return a two dimensional array of 'evaluated  }
    {coefficients'. An evaluated coefficient is Coeffs[I]*ExtenderBF[I] where  }
    {ExtenderBF[I] is the Ith basis function of the extender. The returned     }
    {array will have N+1 columns where N is the number of dimensions intro-    }
    {duced by the extender, and has many rows as is necessary to evaluated the }
    {extender basis functions at their respective resolutions.                 }
    function GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;virtual;abstract;

    property Owner:TObject read FOwner write FOwner;
  end;

  TPreloadedPattern=class
  public
    function GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;overload;virtual;abstract;
    function GetValue(ADateTime:TDateTime;Latitude:double):double;overload;virtual;abstract;
  end;

  TBasisFunction=class
  strict private
    M,C:double; {These are the coefficients used to remove offset and/or trend from the basis function.}
    FNumSteps,
    FNumFourier:word;
    FName:string;
    FFitOnlyPreloaded,
    FActive:boolean;
    FTreatment:TBasisFunctionTreatment;

    {FRegCoeffs holds the basis function fit coefficients. The first index (1..2)}
    {differentiates between the fit coefficient value and its error. The second  }
    {index (1..3) differentiates between the initial, intermediate and final     }
    {coefficient estimates. The third index runs through the total number of fit }
    {coefficients for the basis function.}
    FRegCoeffs:T3DDoubleArray;

    FExtender:TBasisFunctionExtender;
    FPreloadedPattern:TPreloadedPattern;
    FSeasonalResolution:word;
    procedure SetNumFourier(AValue:word);
    function CheckIfHasPreload:boolean;
  protected
    FResolution:TResolution;
  public
    constructor Create(ABasisFuncInitRec:TBasisFuncInitrec);
    function  FillInitRec(AName:string;ANumFourier:word;ATreatment:TBasisFunctionTreatment):TBasisFuncInitRec;
    function  GetBasisVector(ADateTime:TDateTime;AInfoObj:TInfoObject=nil):TDoubleArray;
    function  GetCoefficientErrorVector(ADateTime:TDateTime;ACoeffType:TCoeffType;AInfoObj:TInfoObject=nil):TDoubleArray;
    function  GetNumCoefficients:word;
    function  GetStringListOfExpansions:TStringList;
    function  GetRegCoeff(ACoeffType:TCoeffType;I:word):TCoeffRec;
    function  GetRegCoeffForPreload(ACoeffType:TCoeffType):TCoeffRec;
    function  GetRegCoeffArray:T3DDoubleArray;overload;
    function  GetRegCoeffArray(ACoeffType:TCoeffType):T2DDoubleArray;overload;
    function  GetResolvedCoeffArray(ACoeffType:TCoeffType):T2DDoubleArray;
    function  GetTreatedValue(ADateTime:TDateTime;AInfoObj:TInfoObject=nil):double;virtual;
    function  GetValue(ADateTime:TDateTime):double;overload;virtual;abstract;
    function  GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;overload;virtual;
    function  GetPreloadValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
    function  GetArrayOfValues(ADateTimeList:TDoubleArray):TDoubleArray;
    procedure ExtendWith(AnExtender:TBasisFunctionExtender);
    procedure PreloadWith(PreloadedPattern:TPreloadedPattern);
    procedure RemoveExtender;
    function  HasExtender:boolean;
    function  SetResolution(AResolution:TResolution):boolean;
    procedure SetStep(N:word);virtual;
    procedure SetRegCoeff(ACoeffType:TCoeffType;I:word;Coeff,CoeffError:double);
    procedure SpecifyIndependentVars(Data:TRegressionDataArray);virtual;final;
    procedure TestDateTimeList(ADateTimeList:TDoubleArray);virtual;
    procedure WriteTimeSeriesToFile(ADateTimeList:TDoubleArray;Filename:string);virtual;
    function  TestResolution:boolean;virtual;
    function  GetContribution(ACoeffType:TCoeffType;ADateTime:TDateTime;AInfoObj:TInfoObject=nil):double;
    destructor Destroy;override;

    property NumSteps:word read FNumSteps write FNumSteps;
    property Treatment:TBasisFunctionTreatment read FTreatment write FTreatment;
    property Name:string read FName write FName;
    property FitOnlyPreloaded:boolean read FFitOnlyPreloaded write FFitOnlyPreloaded;
    property Active:boolean read FActive write FActive;
    property HasPreload:boolean read CheckIfHasPreload;
    property NumFourier:word read FNumFourier write SetNumFourier;
    property Extender:TBasisFunctionExtender read FExtender;
    property SeasonalResolution:word read FSeasonalResolution write FSeasonalResolution;
  end;

  TBasisFunctionClass=class of TBasisFunction;

  EBasisFunction=class(exception);

  TBasisFunctionObjectList=class(TObjectList)
  strict private
    function GetBasisObj(I:word):TBasisFunction;overload;
    procedure PutBasisObj(I:word;AValue:TBasisFunction);
  public
    function Add(AObject:TBasisFunction):integer;
    function GetBasisObj(BasisFunc:TClass):TBasisFunction;overload;
    property Hold[I:word]:TBasisFunction read GetBasisObj write PutBasisObj;default;
  end;

  {------------------------------------------------------------------------------}
  { Here in an object that defines a preloaded regression pattern as a function  }
  { of latitude and season.                                                      }
  {------------------------------------------------------------------------------}
  TPreloadLatAndSeason=class(TPreloadedPattern)
  private
    Data:TSynopticGlobal;
  public
    constructor Create(Filename:string;Caveat:string='');
    function GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;override;
    function GetValue(ADateTime:TDateTime;Latitude:double):double;override;
    destructor Destroy;override;
  end;

  {------------------------------------------------------------------------------}
  { Here in an object that defines a preloaded regression pattern as a function  }
  { of latitude, longitude, and season.                                          }
  {------------------------------------------------------------------------------}
  TPreloadLatLongAndSeason=class(TPreloadedPattern)
  private
    Data:T4DSynoptic;
  public
    constructor Create(Filename:string);
    function GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;override;
    destructor Destroy;override;
  end;

  {------------------------------------------------------------------------------}
  { Here are two objects that facilitate time dependence     in basis functions. }
  {------------------------------------------------------------------------------}
  TTimeInfo=class(TInfoObject)
  public
    DateTime:double;
    constructor Create;
    procedure WriteToFile(var Dest:textfile);override;
    function  GetFieldDescriptor(FieldNum:word):string;override;
    function  GetColumnsDescriptor:string;override;
    function  SpawnCopy:TInfoObject;override;
  end;

  TTimeDependenceExtender=class(TBasisFunctionExtender)
  strict private
    FStartDateTime,
    FEndDateTime:TDateTime;
    FDayStep:integer;
  public
    constructor Create(StartDateTime,EndDateTime:TDateTime;DayStep:integer);
    function GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;override;
    function GetNumCoefficients:word;override;
    function GetStringListOfExpansions(InputList:TStringList):TStringList;override;
    function GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;override;
  end;


  {------------------------------------------------------------------------------}
  { Here are two objects that facilitate longitude expansion in basis functions. }
  {------------------------------------------------------------------------------}
  TLongitudeInfo=class(TInfoObject)
  public
    Longitude:double;
    constructor Create;
    procedure WriteToFile(var Dest:textfile);override;
    function  GetFieldDescriptor(FieldNum:word):string;override;
    function  GetColumnsDescriptor:string;override;
    function  SpawnCopy:TInfoObject;override;
  end;

  TLongitudeExtender=class(TBasisFunctionExtender)
  strict private
    FNumLongFourier:word;
    FLongResolution:word;
  public
    constructor Create(ANumLongFourier:word);
    function GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;override;
    function GetNumCoefficients:word;override;
    function GetStringListOfExpansions(InputList:TStringList):TStringList;override;
    function GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;override;
    property LongResolution:word read FLongResolution write FLongResolution;
  end;

  {------------------------------------------------------------------------------}
  { Here are two objects that facilitate latitude expansion in spherical         }
  { harmonics for the basis functions.                                           }
  {------------------------------------------------------------------------------}
  TLatitudeInfo=class(TInfoObject)
  public
    Latitude:double;
    constructor Create;
    procedure WriteToFile(var Dest:textfile);override;
    function  GetFieldDescriptor(FieldNum:word):string;override;
    function  GetColumnsDescriptor:string;override;
    function  SpawnCopy:TInfoObject;override;
  end;

  TLatitudeSphericalHarmonicExtender=class(TBasisFunctionExtender)
  strict private
    FDegree:word;
    FLatResolution:word;
  public
    constructor Create(ADegree:word);
    function GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;override;
    function GetNumCoefficients:word;override;
    function GetStringListOfExpansions(InputList:TStringList):TStringList;override;
    function GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;override;
    property Degree:word read FDegree write FDegree;
  end;

  {------------------------------------------------------------------------------}
  { Here is an object that facilitates latitude and longitude expansion in       }
  { spherical harmonics for the basis functions.                                 }
  {------------------------------------------------------------------------------}
  TLatAndLonInfo=class(TLatitudeInfo)
  public
    Longitude:double;
    constructor Create;
    procedure WriteToFile(var Dest:textfile);override;
    function  GetFieldDescriptor(FieldNum:word):string;override;
    function  GetColumnsDescriptor:string;override;
    function  SpawnCopy:TInfoObject;override;
  end;

  TSphericalHarmonicExtender=class(TBasisFunctionExtender)
  strict private
    FDegree,
    FMaxOrder,
    FLatResolution,
    FLongResolution:word;
    MemLat,
    MemLong:double;
    MemBasisFV:TDoubleArray;
  public
    constructor Create(ADegree,AOrder:word);
    function GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;override;
    function GetNumCoefficients:word;override;
    function GetStringListOfExpansions(InputList:TStringList):TStringList;override;
    function GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;override;
    property LatResolution:word read FLatResolution write FLatResolution;
    property LongResolution:word read FLongResolution write FLongResolution;
  end;


var
  BasisFunctionList:TObjectList;

implementation

uses
  DateUtils,
  Dialogs,
  Math,
  gb_datetime,
  gb_files;

{------------------------------------------------------------------------------}
{ TBasisFunction                                                               }
{------------------------------------------------------------------------------}
constructor TBasisFunction.Create(ABasisFuncInitRec:TBasisFuncInitrec);
begin
  FName:=ABasisFuncInitRec.Name;
  FNumFourier:=ABasisFuncInitRec.NumFourier;
  FTreatment:=ABasisFuncInitRec.Treatment;
  FFitOnlyPreloaded:=false;
  FRegCoeffs:=T3DDoubleArray.Create(2,3,GetNumCoefficients);
  FNumSteps:=1;
  FSeasonalResolution:=36;
  FActive:=true;
  FExtender:=nil;
  BasisFunctionList.Add(self);
end;

function TBasisFunction.FillInitRec(AName:string;ANumFourier:word;ATreatment:TBasisFunctionTreatment):TBasisFuncInitRec;
begin
  with Result do
  begin
    Name:=AName;
    NumFourier:=ANumFourier;
    Treatment:=ATreatment;
  end;
end;

function TBasisFunction.GetRegCoeff(ACoeffType:TCoeffType;I:word):TCoeffRec;
begin
  Result.Coeff:=FRegCoeffs[1,ord(ACoeffType)+1,I];
  Result.Error:=FRegCoeffs[2,ord(ACoeffType)+1,I];
end; {End of TBasisFunction.GetRegCoeff.}

function TBasisFunction.GetRegCoeffForPreload(ACoeffType:TCoeffType):TCoeffRec;
begin
  if FPreloadedPattern=nil then raise EBasisFunction.Create('no pattern had been preloaded.');
  Result.Coeff:=FRegCoeffs[1,ord(ACoeffType)+1,FRegCoeffs.GetNumZ];
  Result.Error:=FRegCoeffs[2,ord(ACoeffType)+1,FRegCoeffs.GetNumZ];
end;

function TBasisFunction.GetRegCoeffArray:T3DDoubleArray;
begin
  Result:=T3DDoubleArray.Create(FRegCoeffs);
end;

function TBasisFunction.GetRegCoeffArray(ACoeffType:TCoeffType):T2DDoubleArray;
var
  I,J:integer;
begin
  Result:=T2DDoubleArray.Create(FRegCoeffs.GetNumX,FRegCoeffs.GetNumZ);
  for I:=1 to FRegCoeffs.GetNumX do
    for J:=1 to FRegCoeffs.GetNumZ do
      Result[I,J]:=FRegCoeffs[I,ord(ACoeffType)+1,J];
end;

function GetSingleCoeff(Coeff:TDoubleArray):T2DDoubleArray;
begin
  Result:=T2DDoubleArray.Create(1,1);
  Result[1,1]:=Coeff[0];
end;

function TBasisFunction.GetResolvedCoeffArray(ACoeffType:TCoeffType):T2DDoubleArray;
{------------------------------------------------------------------------------}
{ The purpose of this function is to return a two dimensional array containing }
{ the fully resolved regression model coefficients. The array has N+1 columns  }
{ where N is the number of 'dimensions' in which the coefficient has been      }
{ expanded. For example, if we have expanded in Fourier pairs to account for   }
{ seasonality, then N is 1. If we have expanded in Fouriers for seasonality    }
{ and in Legendres for latitudes, then N is 2. The number of rows is determin- }
{ ed by the resolution at which the respective dimensions are evaluated. NOTE: }
{ the output array returned by this function does not include any contribution }
{ from any preloaded fit coefficient pattern. It is left to the calling        }
{ program to do that. The reason for this is that if no extender is specified, }
{ or if the specified extender doesn't cover the same dimensions as the pre-   }
{ loaded pattern, then the preloaded pattern won't be evaluated at the correct }
{ vertices.                                                                    }
{------------------------------------------------------------------------------}
var
  {NumExpansionCoeffs is the number of terms in which each of the Fourier fit }
  {coefficients is expanded. This will be 1 if no extender has been specified.}
  NumExpansionCoeffs,

  NumExtenderEvaluatedPoints, {The total number of vertices at which the extender is resolved.}
  NumExtenderDimensions,      {The number of dimensions for any extender.}

  DayStep,
  I,J,K,
  Ind:integer;
  RadiansThroughYear:double;

  {CoeffArray is an array of 2D arrays with one element for each Fourier fit   }
  {coefficient. The 2D arrays hold the evaluated fit coefficients from any     }
  {extender. If there is no extender, the 2D array collapses to a single cell. }
  CoeffArray:array of T2DDoubleArray;

  {FitCoeffsForExpansion holds the fit coefficients associated with each Fourier}
  {term. If the Fourier term fit coefficients are not expanded, this array will }
  {have only 1 entry.}
  FitCoeffsForExpansion:TDoubleArray;
begin
  if HasExtender then NumExpansionCoeffs:=FExtender.GetNumCoefficients else NumExpansionCoeffs:=1;
  setlength(FitCoeffsForExpansion,NumExpansionCoeffs);

  {First fill the CoeffArray.}
  setlength(CoeffArray,2*NumFourier+1);
  for I:=Low(CoeffArray) to High(CoeffArray) do
  begin
    for J:=1 to NumExpansionCoeffs do FitCoeffsForExpansion[J-1]:=
      FRegCoeffs[1,ord(ACoeffType)+1,I*NumExpansionCoeffs+J];
    if HasExtender then CoeffArray[I]:=FExtender.GetEvaluatedCoefficients(FitCoeffsForExpansion)
    else CoeffArray[I]:=GetSingleCoeff(FitCoeffsForExpansion);
    if FFitOnlyPreloaded then
      for J:=2 to CoeffArray[I].GetNumX do
        for K:=1 to CoeffArray[I].GetNumY do
          CoeffArray[I][J,K]:=0.0;
  end;

  {Setup the result array. Because all of the 2D arrays in CoeffArray are the}
  {same size, we just use the first entry to get the necessary dimensions.   }
  NumExtenderEvaluatedPoints:=CoeffArray[0].GetNumY;
  NumExtenderDimensions:=CoeffArray[0].GetNumX;
  Result:=T2DDoubleArray.Create(NumExtenderDimensions+1,NumExtenderEvaluatedPoints*FSeasonalResolution);

  {Evaluate at the seasonal resolution and accumulate in the Result array.}
  for DayStep:=1 to FSeasonalResolution do
  begin
    RadiansThroughYear:=(DayStep-0.5)/FSeasonalResolution*2*Pi;
    for I:=1 to NumExtenderEvaluatedPoints do
    begin
      Ind:=(DayStep-1)*NumExtenderEvaluatedPoints+I;

      {Do the zeroth term.}
      Result[1,Ind]:=(DayStep-0.5)*365.25/FSeasonalResolution;
      for J:=1 to NumExtenderDimensions do Result[J+1,Ind]:=CoeffArray[0][J,I];

      {Do the sine and cosine terms.}
      for J:=1 to NumFourier do
      begin
        Result[NumExtenderDimensions+1,Ind]:=Result[NumExtenderDimensions+1,Ind]+
          CoeffArray[J*2-1][NumExtenderDimensions,I]*sin(J*RadiansThroughYear);
        Result[NumExtenderDimensions+1,Ind]:=Result[NumExtenderDimensions+1,Ind]+
          CoeffArray[J*2][NumExtenderDimensions,I]*cos(J*RadiansThroughYear);
      end;
    end;
  end;

  {Free the CoeffArray objects.}
  for I:=Low(CoeffArray) to High(CoeffArray) do FreeAndNil(CoeffArray[I]);
end;

{I have excluded the method below from being compiled since I haven't finished}
{working on it. I started working on it but then realised that I didn't       }
{actually need it. But I may want to resurrect it some time in the future.    }

{$IFDEF XX}
function TBasisFunction.GetResolvedCoeffErrorSquaredArray(ACoeffType:TCoeffType):T2DDoubleArray;
{------------------------------------------------------------------------------}
{ This function returns a two dimensional array containing the fully resolved  }
{ errors on the regression model fit coefficients. The array has N+1 columns   }
{ where N is the number of 'dimensions' in which the coefficient has been      }
{ expanded. For example, if we have expanded in Fourier pairs to account for   }
{ seasonality, then N is 1. If we have expanded in Fouriers for seasonality    }
{ and in Legendres for latitudes, then N is 2. The number of rows is determin- }
{ ed by the resolution at which the respective dimensions are evaluated. NOTE: }
{ the output array returned by this function does not include any contribution }
{ from any preloaded fit coefficient pattern. It is left to the calling        }
{ program to do that. The reason for this is that if no extender is specified, }
{ or if the specified extender doesn't cover the same dimensions as the pre-   }
{ loaded pattern, then the preloaded pattern won't be evaluated at the correct }
{ vertices.                                                                    }
{------------------------------------------------------------------------------}
var
  {NumExpansionCoeffs is the number of terms in which each of the Fourier fit }
  {coefficients is expanded. This will be 1 if no extender has been specified.}
  NumExpansionCoeffs,
  NumExtenderEvaluatedPoints, {The total number of vertices at which the extender is resolved.}
  NumExtenderDimensions,      {The number of dimensions for any extender.}

  DayStep,
  I,J,
  Ind:integer;
  RadiansThroughYear:double;

  {CoeffErrorArray is an array of 2D arrays with one element for each Fourier fit}
  {coefficient. The 2D arrays hold the evaluated fit coefficient errors from any }
  {extender. If there is no extender, the 2D array collapses to a single cell.   }
  CoeffErrorArray:array of T2DDoubleArray;

  {FitCoeffsErrorsForExpansion holds the fit coefficient errors associated with}
  {each Fourier term. If the Fourier term fit coefficients are not expanded,   }
  {this array will have only 1 entry.}
  FitCoeffsErrorsForExpansion:TDoubleArray;
begin
  if HasExtender then NumExpansionCoeffs:=FExtender.GetNumCoefficients else NumExpansionCoeffs:=1;
  setlength(FitCoeffErrorsForExpansion,NumExpansionCoeffs);

  {First fill the CoeffArray.}
  setlength(CoeffErrorArray,2*NumFourier+1);
  for I:=Low(CoeffErrorArray) to High(CoeffErrorArray) do
  begin
    {Recall that FRegCoeffs[1..] refers to the fit coefficient while FRegCoeffs[2..] refers to }
    {the error on the fit coefficient. So, let's say the extender is a Lengendre expansion in  }
    {latitudes. So then say A=A0 + A1*L1 + A2*L2 + A3*L3... the uncertainty on A0 is then given}
    {by Sigma_A0^2 = Sigma_A0^2 + Sigma_A1^2*L1^2 + Sigma_A2^2*L2^2 + ... which is mathematic- }
    {ally the same as Sigma_A0^2 = Sigma_A0^2 + (Sigma_A1*L1)^2 + (Sigma_A2*L2)^2 + ... and so }
    {we can pass the fit coefficient errors for the expansion to the GetEvaluatedCoefficients  }
    {method of }





    for J:=1 to NumExpansionCoeffs do FitCoeffErrorsForExpansion[J-1]:=
      FRegCoeffs[2,ord(ACoeffType)+1,I*NumExpansionCoeffs+J];



    if HasExtender then CoeffArray[I]:=FExtender.GetEvaluatedCoefficients(FitCoeffErrorsForExpansion)
    else CoeffArray[I]:=GetSingleCoeff(FitCoeffsForExpansion);
  end;

  {Setup the result array. Because all of the 2D arrays in CoeffArray are the}
  {same size, we just use the first entry to get the necessary dimensions.   }
  NumExtenderEvaluatedPoints:=CoeffArray[0].GetNumY;
  NumExtenderDimensions:=CoeffArray[0].GetNumX;
  Result:=T2DDoubleArray.Create(NumExtenderDimensions+1,NumExtenderEvaluatedPoints*FSeasonalResolution);

  {Evaluate at the seasonal resolution and accumulate in the Result array.}
  for DayStep:=1 to FSeasonalResolution do
  begin
    RadiansThroughYear:=(DayStep-0.5)/FSeasonalResolution*2*Pi;
    for I:=1 to NumExtenderEvaluatedPoints do
    begin
      Ind:=(DayStep-1)*NumExtenderEvaluatedPoints+I;

      {Do the zeroth term.}
      Result[1,Ind]:=(DayStep-0.5)*365.25/FSeasonalResolution;
      for J:=1 to NumExtenderDimensions do Result[J+1,Ind]:=CoeffArray[0][J,I];

      {Do the sine and cosine terms.}
      for J:=1 to NumFourier do
      begin
        Result[NumExtenderDimensions+1,Ind]:=Result[NumExtenderDimensions+1,Ind]+
          CoeffArray[J*2-1][NumExtenderDimensions,I]*sin(J*RadiansThroughYear);
        Result[NumExtenderDimensions+1,Ind]:=Result[NumExtenderDimensions+1,Ind]+
          CoeffArray[J*2][NumExtenderDimensions,I]*cos(J*RadiansThroughYear);
      end;
    end;
  end;

  {Free the CoeffArray objects.}
  for I:=Low(CoeffArray) to High(CoeffArray) do FreeAndNil(CoeffArray[I]);
end;
{$ENDIF}

function TBasisFunction.GetTreatedValue(ADateTime:TDateTime;AInfoObj:TInfoObject=nil):double;
begin
  case FTreatment of
    DontChange:Result:=GetValue(ADateTime,AInfoObj);
    RemoveOffset:Result:=GetValue(ADateTime,AInfoObj)-C;
    RemoveOffsetAndTrend:Result:=GetValue(ADateTime,AInfoObj)-(M*ADateTime+C);
  end;
end;

function TBasisFunction.GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
{------------------------------------------------------------------------------}
{ If this method is not overwritten, calling this routine is the same as       }
{ calling GetValue and not passing the information object.                     }
{------------------------------------------------------------------------------}
begin
  Result:=GetValue(ADateTime);
end;

function TBasisFunction.GetPreloadValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
begin
  if FPreloadedPattern=nil then raise EBasisFunction.Create('no pattern had been preloaded.');
  Result:=FPreloadedPattern.GetValue(ADateTime,AInfoObj);
end;

function TBasisFunction.GetArrayOfValues(ADateTimeList:TDoubleArray):TDoubleArray;
var
  I:integer;
begin
  setlength(Result,length(ADateTimeList));
  for I:=Low(ADateTimeList) to High(ADateTimeList) do Result[I]:=GetValue(ADateTimeList[I]);
end;

procedure TBasisFunction.ExtendWith(AnExtender:TBasisFunctionExtender);
begin
  if FExtender<>nil then FreeAndNil(FExtender);
  FExtender:=AnExtender;
  FExtender.Owner:=self;
  {Now need to resize the FRegCoeffs array.}
  FreeAndNil(FRegCoeffs);
  FRegCoeffs:=T3DDoubleArray.Create(2,3,GetNumCoefficients);
end;

procedure TBasisFunction.PreloadWith(PreloadedPattern:TPreloadedPattern);
begin
  if FPreloadedPattern<>nil then FreeAndNil(FPreloadedPattern);
  FPreloadedPattern:=PreloadedPattern;
  {Now need to resize the FRegCoeffs array.}
  FreeAndNil(FRegCoeffs);
  FRegCoeffs:=T3DDoubleArray.Create(2,3,GetNumCoefficients);
end;

procedure TBasisFunction.RemoveExtender;
begin
  FreeAndNil(FExtender);
  {Now need to resize the FRegCoeffs array.}
  FreeAndNil(FRegCoeffs);
  FRegCoeffs:=T3DDoubleArray.Create(2,3,GetNumCoefficients);
end;

function TBasisFunction.HasExtender:boolean;
begin
  Result:=FExtender<>nil;
end;

procedure TBasisFunction.SetRegCoeff(ACoeffType:TCoeffType;I:word;Coeff,CoeffError:double);
begin
  FRegCoeffs[1,ord(ACoeffType)+1,I]:=Coeff;
  FRegCoeffs[2,ord(ACoeffType)+1,I]:=CoeffError;
end; {End of TBasisFunction.SetRegCoeff.}

procedure TBasisFunction.SetStep(N:word);
begin
  if ((N<1) or (N>FNumSteps)) then raise EBasisFunction.Create(
    'the value passed to SetStep was out of range.');
end;

procedure TBasisFunction.SetNumFourier(AValue:word);
begin
  FNumFourier:=AValue;
  {Now need to resize the FRegCoeffs array.}
  FreeAndNil(FRegCoeffs);
  FRegCoeffs:=T3DDoubleArray.Create(2,3,GetNumCoefficients);
end;

function TBasisFunction.CheckIfHasPreload:boolean;
begin
  Result:=FPreloadedPattern<>nil;
end;

function TBasisFunction.GetNumCoefficients:word;
begin
  if FFitOnlyPreloaded then Result:=0 else Result:=1+FNumFourier*2;
  if FExtender<>nil then Result:=Result*FExtender.GetNumCoefficients;
  if FPreloadedPattern<>nil then inc(Result);
end;

function TBasisFunction.GetStringListOfExpansions:TStringList;
var
  I:integer;
begin
  Result:=TStringList.Create;
  Result.Add('['+FName+']');
  for I:=1 to FNumFourier do
  begin
    Result.Add(Result[0]+'*sin('+inttostr(2*I)+'*Pi*Decimal Year)');
    Result.Add(Result[0]+'*cos('+inttostr(2*I)+'*Pi*Decimal Year)');
  end;
  if FExtender<>nil then Result:=FExtender.GetStringListOfExpansions(Result);
end;

function TBasisFunction.SetResolution(AResolution:TResolution):boolean;
begin
  FResolution:=AResolution;
  Result:=TestResolution;
end;

procedure TBasisFunction.SpecifyIndependentVars(Data:TRegressionDataArray);
var
  N,         {The number of entries in the date and time list.}
  I:integer; {Loop control variable.}
  SumX,      {The sum of all the X values.}
  SumY,      {The sum of all the Y values.}
  SumXY,     {The sum of the product of the X and Y values.}
  SumXX,     {The sum of all the X values squared.}
  SumYY,     {The sum of all the Y values squared.}
  Hold:double; {Temporary storage variable.}
  DateTime:TDateTime; {The date and time.}
begin
  N:=Data.Count;
  case FTreatment of
    DontChange:begin
      M:=1E30;
      C:=1E30;
    end;
    RemoveOffset:begin
      M:=1E30;
      C:=0.0;
      for I:=1 to Data.Count do C:=C+GetValue(Data.GetDateTime(I),Data.GetInfoObject(I))/N;
    end;
    RemoveOffsetAndTrend:begin
      {The code in this section may require some further thinking since the trend}
      {in the data may be a function of latitude, altitude, longitude or some    }
      {other independent variable.}
      if N<2 then raise EBasisFunction.Create('there were less than 2 entries in the date and time list.');
      SumX:=0.0;
      SumY:=0.0;
      SumXY:=0.0;
      SumXX:=0.0;
      SumYY:=0.0;
      for I:=1 to Data.Count do
      begin
        DateTime:=Data.GetDateTime(I);
        Hold:=GetValue(DateTime,Data.GetInfoObject(I));
        SumX:=SumX+DateTime;
        SumY:=SumY+Hold;
        SumXY:=SumXY+(DateTime*Hold);
        SumXX:=SumXX+sqr(DateTime);
        SumYY:=SumYY+sqr(Hold);
      end;
      Hold:=(N*SumXX)-sqr(SumX);
      if Hold=0 then raise EBasisFunction.Create('the trend could not be determined for the basis function.');
      M:=((N*SumXY)-(SumX*SumY))/Hold;
      C:=(SumY-M*SumX)/N;
    end;
  end;
end; {End of TBasisFunction.SpecifyDateTimeList.}

procedure TBasisFunction.TestDateTimeList(ADateTimeList:TDoubleArray);
begin
  {Empty code so that this doesn't have to be defined as an abstract method and}
  {thereby force descendenst to have to override it.}
end;

procedure TBasisFunction.WriteTimeSeriesToFile(ADateTimeList:TDoubleArray;Filename:string);
var
  Dest:textfile;
  I:integer;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  writeln(Dest,'''These are the values from the '+Name+' basis function.');
  writeln(Dest,'''Column A) The decimal year.');
  writeln(Dest,'''Column B) The original basis function value.');
  writeln(Dest,'''Column C) The treated basis function value.');
  for I:=Low(ADateTimeList) to High(ADateTimeList) do
    writeln(Dest,ConvertDateTimeToDecYear(ADateTimeList[I]):9:4,' ',GetValue(ADateTimeList[I]):10:5,
    ' ',GetTreatedValue(ADateTimeList[I]):10:5);
  closefile(Dest);
end;

function TBasisFunction.TestResolution:boolean;
begin
  Result:=true;
  if ((FResolution=ResYear) and (FNumFourier>0)) then Result:=false;
  if ((FResolution=ResSemiAnnual) and (FNumFourier>1)) then Result:=false;
  if ((FResolution=ResQuarterly) and (FNumFourier>2)) then Result:=false;
  if ((FResolution=ResMonth) and (FNumFourier>5)) then Result:=false;
end;

function TBasisFunction.GetContribution(ACoeffType:TCoeffType;ADateTime:TDateTime;AInfoObj:TInfoObject=nil):double;
var
  BasisVector:TDoubleArray;
  I:integer;
begin
  Result:=0.0;
  BasisVector:=GetBasisVector(ADateTime,AInfoObj);
  for I:=Low(BasisVector) to High(BasisVector) do
    Result:=Result+(BasisVector[I]*FRegCoeffs[1,ord(ACoeffType)+1,I+1]);
end;

function TBasisFunction.GetBasisVector(ADateTime:TDateTime;AInfoObj:TInfoObject):TDoubleArray;
var
  I:integer;
  DecYear:double;
begin
  {We don't call GetNumCoefficients in the next line because descendent objects}
  {may overwrite that method and we don't want to call the descendent object's }
  {method.}
  setlength(Result,1+FNumFourier*2); {This is modified below if FPreloadedPattern.}

  Result[0]:=GetTreatedValue(ADateTime,AInfoObj);
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  for I:=1 to FNumFourier do
  begin
    Result[I*2-1]:=Result[0]*sin(2*pi*I*DecYear);
    Result[I*2]:=Result[0]*cos(2*pi*I*DecYear);
  end;

  {The length of Result gets extended by 1 in the call to FExtender.GetBasisFector}
  {below which calls GetNumCoefficients.}
  if FExtender<>nil then Result:=FExtender.GetBasisVector(Result,AInfoObj);
  if FPreloadedPattern<>nil then
  begin
    if FExtender=nil then setlength(Result,length(Result)+1);
    Result[High(Result)]:=Result[0]*FPreloadedPattern.GetValue(ADateTime,AInfoObj);
  end;
end;

function TBasisFunction.GetCoefficientErrorVector(ADateTime:TDateTime;ACoeffType:TCoeffType;AInfoObj:TInfoObject=nil):TDoubleArray;
var
  I:integer;
  DecYear:double;
begin
  {We don't call GetNumCoefficients in the next line because descendent objects}
  {may overwrite that method and we don't want to call the descendent object's }
  {method.}
  setlength(Result,1+FNumFourier*2); {This is modified below if FPreloadedPattern.}

  Result[0]:=1;
  DecYear:=ConvertDateTimeToDecYear(ADateTime);
  for I:=1 to FNumFourier do
  begin
    Result[I*2-1]:=sin(2*pi*I*DecYear);
    Result[I*2]:=cos(2*pi*I*DecYear);
  end;
  if FExtender<>nil then Result:=FExtender.GetBasisVector(Result,AInfoObj);
  if FPreloadedPattern<>nil then
  begin
    if FExtender=nil then setlength(Result,length(Result)+1);
    Result[High(Result)]:=FPreloadedPattern.GetValue(ADateTime,AInfoObj);
  end;
  for I:=Low(Result) to High(Result) do
    Result[I]:=sqr(Result[I])*sqr(FRegCoeffs[2,ord(ACoeffType)+1,I+1]);
end;

destructor TBasisFunction.Destroy;
begin
  FreeAndNil(FRegCoeffs);
  FreeAndNil(FExtender);
  FreeAndNil(FPreloadedPattern);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TBasisFunctionObjectList                                                     }
{------------------------------------------------------------------------------}
function TBasisFunctionObjectList.GetBasisObj(I:word):TBasisFunction;
begin
  Result:=Items[I] as TBasisFunction;
end;

procedure TBasisFunctionObjectList.PutBasisObj(I:word;AValue:TBasisFunction);
begin
  Items[I]:=AValue;
end;

function TBasisFunctionObjectList.Add(AObject:TBasisFunction):integer;
begin
  Result:=inherited Add(AObject);
end;

function TBasisFunctionObjectList.GetBasisObj(BasisFunc:TClass):TBasisFunction;
var
  I:integer;
  TestStr1,
  TestStr2:string;
begin
  Result:=nil;
  TestStr1:=BasisFunc.ClassName;
  for I:=0 to Count-1 do
  begin
    TestStr2:=(Items[I] as TBasisFunction).ClassName;
    if (TestStr1=TestStr2) then Result:=GetBasisObj(I);
  end;
  if Result=nil then raise EBasisFunction.Create('there were no '+BasisFunc.ClassName+
    ' basis function in the regression mode when GetBasisObj was called.');
end;

{------------------------------------------------------------------------------}
{ TPreloadLatAndSeason                                                         }
{------------------------------------------------------------------------------}
{$HINTS OFF}
constructor TPreloadLatAndSeason.Create(Filename:string;Caveat:string='');
var
  Header:TSynopticGlobalHeader;
  Source:textfile;
  NumLats,
  NumDays,
  DataLines,
  I,J:integer;
  Lat,
  PrevLat,
  DecDay,
  Junk,Value:double;
begin
  if not FileExists(Filename) then raise EBasisFunction.Create('there was no file named '+
    Filename+' from which to load the preloaded pattern.');

  {First we need to understand the dimensions of the file.}
  DataLines:=CountDataLines(Filename);
  assignfile(Source,Filename);
  reset(Source);
  DeleteHeader(Source);
  Lat:=-999.99;
  repeat
    PrevLat:=Lat;
    readln(Source,DecDay,Lat);
  until (not EquivDouble(PrevLat,-999.99) and (Lat<>PrevLat));
  NumLats:=round(180/(abs(Lat-PrevLat)));
  NumDays:=DataLines div NumLats;

  with Header do
  begin
    NumLong:=NumDays;
    NumLat:=NumLats;
    TagValue:=0;
    DeltaLong:=360/NumDays;
    DeltaLat:=abs(Lat-PrevLat);
    StartLong:=0.0;
    StartLat:=-90+(DeltaLat/2);
  end;
  Data:=TSynopticGlobal.Create(Header);
  Data.ExtendToPole:=true;
  reset(Source);
  DeleteHeader(Source);
  repeat
    readln(Source,DecDay,Lat,Value);
    I:=round((DecDay*NumDays/365.25)+0.5);
    J:=round(((Lat+90)*NumLats/180)+0.5);
    if ((Caveat='NegOnly') and (Value>0.0)) then Value:=0.0;
    Data[I,J]:=Value;
  until eof(Source);
  closefile(Source);
end; {End of TPreloadLatAndSeason.Create.}
{$HINTS ON}

function TPreloadLatAndSeason.GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
var
  Latitude,
  DayScale:double;
begin
  Latitude:=TLatitudeInfo(AInfoObj).Latitude;
  DayScale:=(DayOfTheYear(ADateTime)-1)*360/365;
  Result:=Data.BiLinear(Latitude,DayScale);
end; {End of TPreloadLatAndSeason.GetValue.}

function TPreloadLatAndSeason.GetValue(ADateTime:TDateTime;Latitude:double):double;
var
  DayScale:double;
begin
  DayScale:=(DayOfTheYear(ADateTime)-1)*360/365;
  Result:=Data.BiLinear(Latitude,DayScale);
end;

destructor TPreloadLatAndSeason.Destroy;
begin
  FreeAndNil(Data);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TPreloadLatLongAndSeason                                                     }
{------------------------------------------------------------------------------}
{$HINTS OFF}
constructor TPreloadLatLongAndSeason.Create(Filename:string);
begin
  if not FileExists(Filename) then raise EBasisFunction.Create('there was no file named '+
    Filename+' from which to load the preloaded pattern.');
  Data:=T4DSynoptic.Create;
  Data.ReadFromFile(Filename);
  Data.ExtendToPole:=true;
end; {End of TPreloadLatLongAndSeason.Create.}
{$HINTS ON}

function TPreloadLatLongAndSeason.GetValue(ADateTime:TDateTime;AInfoObj:TInfoObject):double;
var
  Latitude,
  Longitude:double;
  Month,
  Date:word;
begin
  Latitude:=TLatAndLonInfo(AInfoObj).Latitude;
  Longitude:=TLatAndLonInfo(AInfoObj).Longitude;
  Month:=MonthOf(ADateTime);
  Date:=DayOf(ADateTime);
  ADateTime:=EncodeDate(2000,Month,Date);
  Result:=Data.GetValueByCoords(ADateTime,0.0,Longitude,Latitude,true);
end; {End of TPreloadLatLongAndSeason.GetValue.}

destructor TPreloadLatLongAndSeason.Destroy;
begin
  FreeAndNil(Data);
  inherited Destroy;
end;

{------------------------------------------------------------------------------}
{ TTimeInfo                                                                    }
{------------------------------------------------------------------------------}
constructor TTimeInfo.Create;
begin
  NumFields:=1;
end;

procedure TTimeInfo.WriteToFile(var Dest:textfile);
begin
  write(Dest,ConvertDateTimeToDecYear(DateTime):9:4,' ');
end;

function TTimeInfo.GetFieldDescriptor(FieldNum:word):string;
begin
  Result:='The decimal year.';
end;

function TTimeInfo.GetColumnsDescriptor:string;
begin
  Result:='DecimalYear ';
end;

function TTimeInfo.SpawnCopy:TInfoObject;
begin
  Result:=TTimeInfo.Create;
  (Result as TTimeInfo).DateTime:=DateTime;
end;

{------------------------------------------------------------------------------}
{ TTimeDependenceExtender                                                      }
{------------------------------------------------------------------------------}
constructor TTimeDependenceExtender.Create(StartDateTime,EndDateTime:TDateTime;DayStep:integer);
begin
  FStartDateTime:=StartDateTime;
  FEndDateTime:=EndDateTime;
  FDayStep:=DayStep;
end;

function TTimeDependenceExtender.GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;
var
  I,J:integer;
  Long:double;
begin
  if AInfoObj=nil then raise EBasisFunction.Create('InfoObj was nil.');
  if not (AInfoObj is TTimeInfo) then raise EBasisFunction.Create(
    'InfoObj was not of the correct type.');
  setlength(Result,TBasisFunction(FOwner).GetNumCoefficients);
  for I:=0 to length(InputBasisVector)-1 do
  begin
    Result[I*2]:=InputBasisVector[I];
    Result[I*2+1]:=(ConvertDateTimeToDecYear(TTimeInfo(AInfoObj).DateTime)-1980)*InputBasisVector[I];
  end;
end;

function TTimeDependenceExtender.GetNumCoefficients:word;
begin
  Result:=2;
end;

function TTimeDependenceExtender.GetStringListOfExpansions(InputList:TStringList):TStringList;
var
  I,J:integer;
begin
  Result:=TStringList.Create;
  for I:=0 to InputList.Count-1 do
  begin
    Result.Add(InputList[I]);
    Result.Add(InputList[I]+'*t');
  end;
  FreeAndNil(InputList);
end;

function TTimeDependenceExtender.GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;
var
  NumSteps,
  I:integer;
  DateTime:TDateTime;
begin
  NumSteps:=round(FEndDateTime-FStartDateTime) div FDayStep;
  Result:=T2DDoubleArray.Create(2,NumSteps);
  for I:=1 to NumSteps do
  begin
    DateTime:=FStartDateTime+(I-0.5)*FDayStep;
    Result[1,I]:=ConvertDateTimeToDecYear(DateTime);
    Result[2,I]:=Coeffs[0]+Coeffs[1]*(Result[1,I]-1980);
  end;
end;

{------------------------------------------------------------------------------}
{ TLongitudeInfo                                                               }
{------------------------------------------------------------------------------}
constructor TLongitudeInfo.Create;
begin
  NumFields:=1;
end;

procedure TLongitudeInfo.WriteToFile(var Dest:textfile);
begin
  write(Dest,Longitude:9:4,' ');
end;

function TLongitudeInfo.GetFieldDescriptor(FieldNum:word):string;
begin
  Result:='The longitude in degrees.';
end;

function TLongitudeInfo.GetColumnsDescriptor:string;
begin
  Result:='Longitude ';
end;

function TLongitudeInfo.SpawnCopy:TInfoObject;
begin
  Result:=TLongitudeInfo.Create;
  (Result as TLongitudeInfo).Longitude:=Longitude;
end;

{------------------------------------------------------------------------------}
{ TLongitudeExtender                                                           }
{------------------------------------------------------------------------------}
constructor TLongitudeExtender.Create(ANumLongFourier:word);
begin
  FNumLongFourier:=ANumLongFourier;
  FLongResolution:=18;
end;

function TLongitudeExtender.GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;
var
  I,J:integer;
  Long:double;
begin
  if AInfoObj=nil then raise EBasisFunction.Create('InfoObj was nil.');
  if not (AInfoObj is TLongitudeInfo) then raise EBasisFunction.Create(
    'InfoObj was not of the correct type.');
  Long:=TLongitudeInfo(AInfoObj).Longitude;
  setlength(Result,TBasisFunction(FOwner).GetNumCoefficients);
  for I:=0 to length(InputBasisVector)-1 do
  begin
    Result[I*(FNumLongFourier*2+1)]:=InputBasisVector[I];
    for J:=1 to FNumLongFourier do
    begin
      Result[I*(FNumLongFourier*2+1)+(2*J)-1]:=InputBasisVector[I]*sin(2*J*Pi*Long/360);
      Result[I*(FNumLongFourier*2+1)+(2*J)]:=InputBasisVector[I]*cos(2*J*Pi*Long/360);
    end;
  end;
end;

function TLongitudeExtender.GetNumCoefficients:word;
begin
  Result:=FNumLongFourier*2+1;
end;

function TLongitudeExtender.GetStringListOfExpansions(InputList:TStringList):TStringList;
var
  I,J:integer;
begin
  Result:=TStringList.Create;
  for I:=0 to InputList.Count-1 do
  begin
    Result.Add(InputList[I]);
    for J:=1 to FNumLongFourier do
    begin
      Result.Add(InputList[I]+'*sin('+inttostr(2*J)+'*Pi*Longitude/360)');
      Result.Add(InputList[I]+'*cos('+inttostr(2*J)+'*Pi*Longitude/360)');
    end;
  end;
  FreeAndNil(InputList);
end;

function TLongitudeExtender.GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;
var
  I,J:integer;
  Total,
  Long:double;
begin
  Result:=T2DDoubleArray.Create(2,FLongResolution);
  for I:=1 to FLongResolution do
  begin
    Long:=-180+(I-0.5)*360/FLongResolution;
    Total:=Coeffs[0];
    for J:=1 to FNumLongFourier do
    begin
      Total:=Total+Coeffs[J*2-1]*sin(J*DegToRad(Long));
      Total:=Total+Coeffs[J*2]*cos(J*DegToRad(Long));
    end;
    Result[1,I]:=Long;
    Result[2,I]:=Total;
  end;
end;

{------------------------------------------------------------------------------}
{ TLatitudeInfo                                                                }
{------------------------------------------------------------------------------}
constructor TLatitudeInfo.Create;
begin
  NumFields:=1;
end;

procedure TLatitudeInfo.WriteToFile(var Dest:textfile);
begin
  write(Dest,Latitude:9:4,' ');
end;

function TLatitudeInfo.GetFieldDescriptor(FieldNum:word):string;
begin
  Result:='The latitude in degrees.';
end;

function TLatitudeInfo.GetColumnsDescriptor:string;
begin
  Result:='Latitude ';
end;

function TLatitudeInfo.SpawnCopy:TInfoObject;
begin
  Result:=TLatitudeInfo.Create;
  (Result as TLatitudeInfo).Latitude:=Latitude;
end;

{------------------------------------------------------------------------------}
{ TLatitudeSphericalHarmonicExtender                                           }
{------------------------------------------------------------------------------}
constructor TLatitudeSphericalHarmonicExtender.Create(ADegree:word);
begin
  FDegree:=ADegree;
  FLatResolution:=36;
end;

function TLatitudeSphericalHarmonicExtender.GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;
var
  I,J:integer;
  Lat:double;
  Hold:TDoubleArray;
begin
  if AInfoObj=nil then raise EBasisFunction.Create('InfoObj was nil.');
  if not (AInfoObj is TLatitudeInfo) then raise EBasisFunction.Create(
    'InfoObj was not of the correct type.');
  Lat:=DegToRad(90.0+TLatitudeInfo(AInfoObj).Latitude);
  setlength(Hold,FDegree);
  Hold[0]:=1;
  if FDegree>1 then Hold[1]:=cos(Lat);
  for I:=2 to FDegree-1 do
    Hold[I]:=((2*I-1)*cos(Lat)*Hold[I-1]-(I-1)*Hold[I-2])/I;
  setlength(Result,TBasisFunction(FOwner).GetNumCoefficients);
  for I:=0 to length(InputBasisVector)-1 do
    for J:=1 to FDegree do Result[(I*FDegree)+J-1]:=InputBasisVector[I]*Hold[J-1];
end;

function TLatitudeSphericalHarmonicExtender.GetNumCoefficients:word;
begin
  Result:=FDegree;
end;

function TLatitudeSphericalHarmonicExtender.GetStringListOfExpansions(InputList:TStringList):TStringList;
var
  I,J:integer;
begin
  Result:=TStringList.Create;
  for I:=0 to InputList.Count-1 do
    for J:=1 to FDegree do Result.Add(InputList[I]+'*P_'+inttostr(J)+'(cos(Latitude))');
  FreeAndNil(InputList);
end;

function TLatitudeSphericalHarmonicExtender.GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;
var
  I,J:integer;
  Total,
  Lat:double;
  Hold:TDoubleArray;
begin
  setlength(Hold,FDegree);
  Hold[0]:=1;
  Result:=T2DDoubleArray.Create(2,FLatResolution);
  for I:=1 to FLatResolution do
  begin
    Lat:=DegToRad((I-0.5)*180/FLatResolution);
    if FDegree>1 then Hold[1]:=cos(Lat);
    for J:=2 to FDegree-1 do Hold[J]:=((2*J-1)*cos(Lat)*Hold[J-1]-(J-1)*Hold[J-2])/J;
    Total:=Coeffs[0];
    for J:=1 to FDegree-1 do Total:=Total+Coeffs[J]*Hold[J];
    Result[1,I]:=-90+(I-0.5)*180/FLatResolution;
    Result[2,I]:=Total;
  end;
end;

{------------------------------------------------------------------------------}
{ TLatAndLonInfo                                                                }
{------------------------------------------------------------------------------}
constructor TLatAndLonInfo.Create;
begin
  NumFields:=2;
end;

procedure TLatAndLonInfo.WriteToFile(var Dest:textfile);
begin
  write(Dest,Latitude:9:4,' ');
  write(Dest,Longitude:9:4,' ');
end;

function TLatAndLonInfo.GetFieldDescriptor(FieldNum:word):string;
begin
  if FieldNum=1 then
    Result:='The latitude in degrees.';
  if FieldNum=2 then
    Result:='The longitude in degrees.';
end;

function TLatAndLonInfo.GetColumnsDescriptor:string;
var
  I:word;
begin
  for I:=1 to NumFields do
  begin
    if I=1 then
      Result:='Latitude ';
    if I=2 then
      Result:='Longitude ';
  end;
end;

function TLatAndLonInfo.SpawnCopy:TInfoObject;
begin
  Result:=TLatAndLonInfo.Create;
  (Result as TLatAndLonInfo).Latitude:=Latitude;
  (Result as TLatAndLonInfo).Longitude:=Longitude;
end;

{------------------------------------------------------------------------------}
{ TSphericalHarmonicExtender                                                   }
{------------------------------------------------------------------------------}
constructor TSphericalHarmonicExtender.Create(ADegree,AOrder:word);
begin
  FDegree:=ADegree;
  FMaxOrder:=AOrder;
  MemLat:=-999.99;
  MemLong:=-999.99;
  setlength(MemBasisFV,0);
end;

procedure GetSphericalHarmonicArrays(Degree,MaxOrder:word;var P:TPType;var PM:TPMType);
var
  I,J,M:integer;
  Coeff:double;
begin
  setlength(P,Degree);
  setlength(PM,Degree);
  {Set up the first two Ps and PMs}
  if (Degree>=1) then
  begin
    {Set up the first P}
    setlength(P[0],1);
    P[0,0]:=1;
    {get the associated Legendre function}
    setlength(PM[0],1,2,1);
    PM[0,0,0,0]:=0;
    PM[0,0,1,0]:=0;
  end;
  if (Degree>=2) then
  begin
    setlength(P[1],2);
    P[1,0]:=0;
    P[1,1]:=1;
    {get the associated Legendre function}
    setlength(PM[1],2,2,2);
    PM[1,0,0,0]:=0;
    PM[1,0,1,0]:=0;
    PM[1,1,0,0]:=-1;
    PM[1,1,1,1]:=1;
  end;
  for I:=2 to Degree-1 do
  begin
    {get the P's right}
    setlength(P[I],I+1);
    Coeff:=(2*I-1)/I;
    P[I,0]:=0;
    for J:=1 to I do P[I,J]:=Coeff*P[I-1,J-1];
    Coeff:=(I-1)/I;
    for J:=0 to I-2 do P[I,J]:=P[I,J]-Coeff*P[I-2,J];
    {now get the PM's}
    setlength(PM[I],I+1,2,I+1);
    for J:=0 to 1 do PM[I,0,J,0]:=0;
    for M:=1 to I do                        {loop over number of associated Legendre functions}
    begin
      PM[I,M,1,M]:=1;
      PM[I,M,1,M-1]:=0;
      for J:=1 to I do                      {loop over terms}
      begin
        if M=1 then
          Coeff:=P[I,J]*(J)
        else
          Coeff:=PM[I,M-1,0,J]*(J);
        if abs(Coeff)>0 then
          PM[I,M,0,J-1]:=Coeff*(-1);
      end;
    end;
  end;
end;

function TSphericalHarmonicExtender.GetBasisVector(InputBasisVector:TDoubleArray;AInfoObj:TInfoObject):TDoubleArray;
var
  boundary,
  OldArrayLength,
  Counter,
  I,J,K:integer;
  Lat,
  Lon,
  Hold_Value:double;
  ValueArray:TDoubleArray;
  P:TPType;
  PM:TPMType;
begin
  if AInfoObj=nil then raise EBasisFunction.Create('InfoObj was nil.');
  if not (AInfoObj is TLatAndLonInfo) then raise EBasisFunction.Create(
    'InfoObj was not of the correct type.');
  Lat:=DegToRad(90.0+TLatAndLonInfo(AInfoObj).Latitude);
  {Longitude starts with 0degrees (Greenwich), then goes east (positive numbers)}
  {and from there back to Greenwich (negative numbers)}
  if TLatAndLonInfo(AInfoObj).Longitude>0.0 then
    Lon:=DegToRad(TLatAndLonInfo(AInfoObj).Longitude)
  else
    Lon:=DegToRad(360.0+TLatAndLonInfo(AInfoObj).Longitude);
  {check if the Lat and Lon are the same as before}
  if (abs(Lat-MemLat)>1e-06) or (abs(Lon-MemLong)>1e-06) then
  begin
    GetSphericalHarmonicArrays(FDegree,FMaxOrder,P,PM);
    {now get the spherical harmonics}
    {and get rid of the old MemBasisFV}
    setlength(ValueArray,0);
    setlength(MemBasisFV,0);
    for I:=0 to FDegree-1 do
    begin
      if I<FMaxOrder-1 then
        boundary:=I
      else
        boundary:=FMaxOrder-1;
      Counter:=0;
      OldArrayLength:=length(ValueArray);
      setlength(ValueArray,length(ValueArray)+(2*boundary+1));
      setlength(MemBasisFV,length(MemBasisFV)+(2*boundary+1));
      for J:=0 to boundary do
      begin
        inc(Counter);
        if J=0 then
          for K:=0 to I do
            ValueArray[OldArrayLength+Counter-1]:=ValueArray[OldArrayLength+Counter-1]+(P[I,K]*power(cos(Lat),K));
        if J<>0 then
        begin
          for K:=0 to I do
            ValueArray[OldArrayLength+Counter-1]:=ValueArray[OldArrayLength+Counter-1]+(PM[I,abs(J),0,K]*power(cos(Lat),K));
          ValueArray[OldArrayLength+Counter-1]:=ValueArray[OldArrayLength+Counter-1]*power(sin(Lat),abs(J));
          Hold_Value:=ValueArray[OldArrayLength+Counter-1];
          {first get the sin part}
          ValueArray[OldArrayLength+Counter-1]:=Hold_Value*sin(J*Lon);
          {then get the cos part - don't forget to increase the counter!}
          inc(Counter);
          ValueArray[OldArrayLength+Counter-1]:=Hold_Value*cos(J*Lon);
        end;
      end;
    end;
    MemLat:=Lat;
    MemLong:=Lon;
    for J:=0 to length(ValueArray)-1 do
      MemBasisFV[J]:=ValueArray[J];
  end;
  setlength(Result,TBasisFunction(FOwner).GetNumCoefficients);
  for I:=0 to length(InputBasisVector)-1 do
  begin
    if abs(InputBasisVector[I])>10000 then
      ShowMessage('the problem is in the basis vector');
  end;
  for I:=0 to length(InputBasisVector)-1 do
    for J:=1 to length(MemBasisFV) do Result[(I*length(MemBasisFV))+J-1]:=InputBasisVector[I]*MemBasisFV[J-1];
end;

function TSphericalHarmonicExtender.GetNumCoefficients:word;
var
  boundary,
  NumCoeffs,
  I:integer;
begin
  NumCoeffs:=0;
  for I:=0 to FDegree-1 do
  begin
    if I<FMaxOrder-1 then boundary:=I
    else boundary:=FMaxOrder-1;
    NumCoeffs:=NumCoeffs+(2*boundary+1);
  end;
  Result:=NumCoeffs;
end;

function TSphericalHarmonicExtender.GetStringListOfExpansions(InputList:TStringList):TStringList;
var
  boundary,
  I,J,K:integer;
begin
  Result:=TStringList.Create;
  for I:=0 to InputList.Count-1 do
    for J:=0 to FDegree-1 do
    begin
      if I<FMaxOrder-1 then
        boundary:=I
      else
        boundary:=FMaxOrder-1;
      for K:=0 to boundary do
      begin
        if K=0 then
          Result.Add(InputList[I]+'*P_L'+inttostr(J)+'M'+inttostr(abs(K))+'(cos(Latitude))');
        if K>0 then
        begin
          Result.Add(InputList[I]+'*P_L'+inttostr(J)+'M'+inttostr(abs(K))+'(cos(Latitude),cos(Longitude))*sin('+inttostr(abs(K))+'*Longitude)');
          Result.Add(InputList[I]+'*P_L'+inttostr(J)+'M'+inttostr(abs(K))+'(cos(Latitude),cos(Longitude))*cos('+inttostr(abs(K))+'*Longitude)');;
        end;
      end;
    end;
  FreeAndNil(InputList);
end;

function TSphericalHarmonicExtender.GetEvaluatedCoefficients(Coeffs:TDoubleArray):T2DDoubleArray;
var
  LatCount,
  LonCount,
  boundary,
  Counter,
  Eq_Counter,
  I,J,K:integer;
  Total,
  Lat,
  Long,
  RadLat,
  RadLong,
  Hold_Value:double;
  P:TPType;
  PM:TPMType;
begin
  Result:=T2DDoubleArray.Create(3,FLongResolution*FLatResolution);
  GetSphericalHarmonicArrays(FDegree,FMaxOrder,P,PM);
  Counter:=1;
  for LatCount:=1 to FLatResolution do
  begin
    Lat:=-90+(LatCount-0.5)*(180/FLatResolution);
    for LonCount:=1 to FLongResolution do
    begin
      if LonCount<=round(360/(360/FLongResolution))/2 then
        Long:=(LonCount-0.5)*(360/FLongResolution)
      else
        Long:=-360+((LonCount-0.5)*(360/FLongResolution));
      RadLat:=DegToRad(90.0+Lat);
      if Long>0.0 then RadLong:=DegToRad(Long)
        else RadLong:=DegToRad(360.0+Long);
      Eq_Counter:=0;
      Total:=0.0;

      {now get the spherical harmonics}
      for I:=0 to FDegree-1 do
      begin
        if I<FMaxOrder-1 then
          boundary:=I
        else
          boundary:=FMaxOrder-1;
        for J:=0 to boundary do
        begin
          if J=0 then
          begin
            Hold_Value:=0.0;
            for K:=0 to I do
              Hold_Value:=Hold_Value+(P[I,K]*power(cos(RadLat),K));
            Total:=Total+Hold_Value*Coeffs[Eq_Counter];
            inc(Eq_Counter);
          end;
          if J<>0 then
          begin
            Hold_Value:=0.0;
            for K:=0 to I do
              Hold_Value:=Hold_Value+(PM[I,abs(J),0,K]*power(cos(RadLat),K));
            Hold_Value:=Hold_Value*power(sin(RadLat),abs(J));
            {first get the sin part}
            Total:=Total+Hold_Value*sin(J*RadLong)*Coeffs[Eq_Counter];
            inc(Eq_Counter);
            {then get the cos part}
            Total:=Total+Hold_Value*cos(J*RadLong)*Coeffs[Eq_Counter];
            inc(Eq_Counter);
          end;
        end;
      end;
      {then write the result in the array}
      Result[1,Counter]:=Lat;
      Result[2,Counter]:=Long;
      Result[3,Counter]:=Total;
      inc(Counter);
    end;
  end;
end;

initialization
  BasisFunctionList:=TObjectList.Create(true);

finalization
  FreeAndNil(BasisFunctionList);

end.
