//@***h* BDBP/gb_SpatioTemporalBiasCorrector.pas
// 
// NAME
//   gb_SpatioTemporalBiasCorrector.pas
//
// PURPOSE
//  Deals with the spatial and temporal correction of the individual data points 
//  while calculating the monthly mean zonal mean.
//
// DESCRIPTION
//  This unit deals with the spatial and temporal correction of the individual data 
//  points while calculating the monthly mean zonal mean.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_SpatioTemporalBiasCorrector;

{$I UNITDIREC.INC}

interface

uses
  gb_4DSynoptic,
  gb_SynopticGlobal;

type
  TSpatioTemporalBiasCorrector=class
  strict private
    FDailyDataPath,
    FMonthlyDataPath,
    FSetDesc:string;
    FDailyData:T4DSynoptic;
    FMonthlyData:array of double;
    FDeltaLat:double;
    FMonthlyYear,
    FMonthlyMonth,
    FDailyYear:word;
    procedure LoadNewDailyData(Year:word);
    procedure LoadNewMonthlyData(Year,Month:word);
  public
    constructor Create(DailyDataPath,MonthlyDataPath,SetDesc:string;DeltaLat:double);
    function GetCorrection(ADateTime:TDateTime;ALatitude,ALongitude:double):double;
    destructor Destroy;override;
  end;

implementation

uses
  DateUtils,
  SysUtils,
  gb_Strings;

constructor TSpatioTemporalBiasCorrector.Create(DailyDataPath,MonthlyDataPath,SetDesc:string;DeltaLat:double);
begin
  if not DirectoryExists(DailyDataPath) then raise Exception.Create('the path named '+
    DailyDataPath+' doesn''t exist.');
  FDailyDataPath:=DailyDataPath;
  if not DirectoryExists(MonthlyDataPath) then raise Exception.Create('the path named '+
    MonthlyDataPath+' doesn''t exist.');
  FMonthlyDataPath:=MonthlyDataPath;
  FSetDesc:=SetDesc;
  setlength(FMonthlyData,round(180/DeltaLat));
  FDeltaLat:=DeltaLat;
  FDailyYear:=65535;
  FMonthlyYear:=65535;
  FMonthlyMonth:=65535;
end;

function TSpatioTemporalBiasCorrector.GetCorrection(ADateTime:TDateTime;ALatitude,ALongitude:double):double;
var
  MonthOzone,
  SpotOzone:double;
begin
  Result:=1.0;
  if YearOf(ADateTime)<>FDailyYear then LoadNewDailyData(YearOf(ADateTime));
  if not ((YearOf(ADateTime)=FMonthlyYear) and (MonthOf(ADateTime)=FMonthlyMonth)) then
    LoadNewMonthlyData(YearOf(ADateTime),MonthOf(ADateTime));
  MonthOzone:=FMonthlyData[trunc((ALatitude+90)/FDeltaLat)];
  if MonthOzone>0.0 then
  try
    SpotOzone:=FDailyData.GetValueByCoords(ADateTime,0.0,ALongitude,ALatitude,true);
    Result:=SpotOzone/MonthOzone;
  except
    on Exception do;
  end;
end;

procedure TSpatioTemporalBiasCorrector.LoadNewDailyData(Year:word);
var
  Filename:string;
begin
  FileMode:=fmOpenRead;
  Filename:=FDailyDataPath+'\NIWAAssimOzoneV2.7_'+inttostr(Year)+'_'+FSetDesc+'.4dsyn';
  if not FileExists(Filename) then raise Exception.Create('The file named '+
    Filename+' doesn''t exist.');
  FreeAndNil(FDailyData);
  FDailyData:=T4DSynoptic.Create;
  FDailyData.ReadFromFile(Filename);
  FDailyYear:=Year;
  FileMode:=fmOpenReadWrite;
end;

procedure TSpatioTemporalBiasCorrector.LoadNewMonthlyData(Year,Month:word);
var
  Filename:string;
  Hold:TSynopticGlobal;
  I:integer;
  Value:double;
begin
  FileMode:=fmOpenRead;
  Filename:=FMonthlyDataPath+'\NIWAAssimV2.7_'+inttostr(Year)+inttostrf(Month,2,true)+'_'+FSetDesc+'.syn';
  if not FileExists(Filename) then raise Exception.Create('The file named '+
    Filename+' doesn''t exist.');
  Hold:=TSynopticGlobal.Create(Filename);
  try
    for I:=1 to round(180/FDeltaLat) do
    begin
      Value:=Hold.GetAreaWeightedMean((I-1)*FDeltaLat-90,I*FDeltaLat-90,-180,180);
      if Hold.FracMissing>0.1 then Value:=-999.99;
      FMonthlyData[I-1]:=Value;
    end;
  finally
    FreeAndNil(Hold);
  end;
  FMonthlyYear:=Year;
  FMonthlyMonth:=Month;
  FileMode:=fmOpenReadWrite;
end;

destructor TSpatioTemporalBiasCorrector.Destroy;
begin
  FreeAndNil(FDailyData);
end;


end.
