//@***h* BDBP/gb_array.pas
// 
// NAME
//   gb_array.pas
//
// PURPOSE
//  Defines a data storage class which can be used much like an array.
//
// DESCRIPTION
//  This unit defines a data storage class which can be used much like an 
//  array. The size of the array is limited only by the free disk space. The 
//  class defined in this unit implements a single dimensional array, leaving 
//  multidimension implementation to descendent objects. Any data type can be 
//  stored in the array object. The maximum amount of heap memory that can be 
//  used by the object is specified by the calling program. If the object's 
//  data array(s) can be stored on the heap without exceeding this amount, the 
//  object uses this approach otherwise a temporary storage file is created 
//  and pages of memory are swapped to and from disk. The size of the data 
//  which the object manages is specified in the initialisation. The object 
//  also defines two methods for rapidly writing and reading data to and from files. 
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_array;

{$I UNITDIREC.INC}

interface

uses
  {This is a standard delphi unit.}
  SysUtils;

const
  {The maximim memory page size is set to 2^22 bytes = 4Mb}
  MaxPageSize=4194303;

type
  TByteArray=array of byte;
  TCardinalArray=array of cardinal;
  TIntegerArray=array of integer;
  TWordArray=array of word;
  TSingleArray=array of single;
  TDoubleArray=array of double;
  PDoubleArray=^TDoubleArray;
  T2DimDoubleArray=array of array of double;
  TBooleanArray=array of boolean;

  Page=array[0..MaxPageSize] of byte;
  PPage=^Page;
  Book=array[0..1023] of PPage;
  PBook=^Book;
  TLargeArray=class
  private
    DiskData:PPage;   {The pointer to the data array used for disk storage.}
    HeapData:PBook;   {The pointer to the data array used for heap storage.}
    StoreFile:file;   {The file handle for the storage data file, if needed.}
    CurrentPage,      {The current page number.}
    EntrySize,        {The size in bytes of each entry.}
    PageSize,         {The size in bytes of each of the first N-1 pages.}
    LastPageSize,     {The size of the last page in bytes.}
    NumPerPage,       {The number of entries per page.}
    NumPages:cardinal;{The total number of pages.}
    Shift:word;       {The byte shift parameter.}
    CloseFiles,       {True if files must be closed after read/write.}
    Disk:boolean;     {True if a disk storage file is to be used.}
  public
    NumData:cardinal; {The number of entries in the object data array(s).}
    constructor Create(MaxHeap,NData,ESize:cardinal);
    procedure Clear;virtual;
    function  Index(Entry:cardinal):pointer;
    function  SaveData(var Dest:file):byte;virtual;
    function  LoadData(var Source:file):byte;virtual;
    function  GetNumData:cardinal;
    function  GetSizeOfData:cardinal;
    function  GetMemoryUsed:cardinal;
    procedure GetMemoryStats(var TDisk:boolean;var TNumPages:cardinal);
    destructor Destroy;override;
  end;

  TLargeArrayArray=array[1..50] of TLargeArray;

  TByteType=byte;
  TBytePtr=^TByteType;
  TIntegerType=Integer;
  TIntegerPtr=^TIntegerType;
  TWordType=word;
  TWordPtr=^TWordType;
  TSingleType=single;
  TSinglePtr=^TSingleType;
  TRealType=real;
  TRealPtr=^TRealType;
  TDoubleType=double;
  TDoublePtr=^TDoubleType;
  TCardinalType=cardinal;
  TCardinalPtr=^TCardinalType;

  T1DByteArray=class(TLargeArray)
  private
    NumX:cardinal;
    function Index(X:cardinal):TBytePtr;
  public
    constructor Create(X:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X:cardinal):byte;
    procedure PutValue(X:cardinal;Value:byte);
    property Data[X:cardinal]:byte read GetValue write PutValue; default;
  end;

  T2DByteArray=class(TLargeArray)
  private
    NumX,
    NumY:cardinal;
    function Index(X,Y:cardinal):TBytePtr;
  public
    constructor Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X,Y:cardinal):byte;
    procedure PutValue(X,Y:cardinal;Value:byte);
    property Data[X,Y:cardinal]:byte read GetValue write PutValue; default;
  end;

  T3DByteArray=class(TLargeArray)
  private
    NumX,
    NumY,
    NumZ:cardinal;
    function Index(X,Y,Z:cardinal):TBytePtr;
  public
    constructor Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X,Y,Z:cardinal):byte;
    procedure PutValue(X,Y,Z:cardinal;Value:byte);
    property Data[X,Y,Z:cardinal]:byte read GetValue write PutValue; default;
  end;

  T2DRealArray=class(TLargeArray)
  private
    NumX,
    NumY:cardinal;
    function Index(X,Y:cardinal):TRealPtr;
  public
    constructor Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X,Y:cardinal):real;
    procedure PutValue(X,Y:cardinal;Value:real);
    property Data[X,Y:cardinal]:real read GetValue write PutValue; default;
  end;

  T2DSingleArray=class(TLargeArray)
  private
    NumX,
    NumY:cardinal;
    function Index(X,Y:cardinal):TSinglePtr;
  public
    constructor Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
    procedure DeepCopy(Source:T2DSingleArray);
    function GetValue(X,Y:cardinal):single;
    procedure PutValue(X,Y:cardinal;Value:single);
    property Data[X,Y:cardinal]:single read GetValue write PutValue; default;
    property GetNumX:cardinal read NumX;
    property GetNumY:cardinal read NumY;
  end;

  T3DSingleArray=class(TLargeArray)
  private
    NumX,
    NumY,
    NumZ:cardinal;
    function Index(X,Y,Z:cardinal):TSinglePtr;
  public
    constructor Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X,Y,Z:cardinal):single;
    procedure PutValue(X,Y,Z:cardinal;Value:single);
    property Data[X,Y,Z:cardinal]:single read GetValue write PutValue; default;
  end;

  T4DSingleArray=class(TLargeArray)
  private
    FWOffset,
    FXOffset,
    FYOffset,
    FZOffset,
    NumW,
    NumX,
    NumY,
    NumZ:cardinal;
    function Index(W,X,Y,Z:cardinal):TSinglePtr;
  public
    constructor Create(W,X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(W,X,Y,Z:cardinal):single;
    procedure PutValue(W,X,Y,Z:cardinal;Value:single);
    property Data[W,X,Y,Z:cardinal]:single read GetValue write PutValue; default;
    property WOffset:cardinal read FWOffset write FWOffset;
    property XOffset:cardinal read FXOffset write FXOffset;
    property YOffset:cardinal read FYOffset write FYOffset;
    property ZOffset:cardinal read FZOffset write FZOffset;
  end;

  T2DIntegerArray=class(TLargeArray)
  private
    NumX,
    NumY:cardinal;
    function Index(X,Y:cardinal):TIntegerPtr;
  public
    constructor Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X,Y:cardinal):integer;
    procedure PutValue(X,Y:cardinal;Value:integer);
    property Data[X,Y:cardinal]:integer read GetValue write PutValue; default;
    property GetNumX:cardinal read NumX;
    property GetNumY:cardinal read NumY;
  end;

  T1DWordArray=class(TLargeArray)
  private
    NumX:cardinal;
    function Index(X:cardinal):TWordPtr;
  public
    constructor Create(X:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X:cardinal):Word;
    procedure PutValue(X:cardinal;Value:Word);
    property Data[X:cardinal]:Word read GetValue write PutValue; default;
  end;

  T2DWordArray=class(TLargeArray)
  private
    NumX,
    NumY:cardinal;
    function Index(X,Y:cardinal):TWordPtr;
  public
    constructor Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
    procedure DeepCopy(Source:T2DWordArray);
    function GetValue(X,Y:cardinal):word;
    procedure PutValue(X,Y:cardinal;Value:word);
    property Data[X,Y:cardinal]:word read GetValue write PutValue; default;
    property GetNumX:cardinal read NumX;
    property GetNumY:cardinal read NumY;
  end;

  T3DWordArray=class(TLargeArray)
  private
    NumX,
    NumY,
    NumZ:cardinal;
    function Index(X,Y,Z:cardinal):TWordPtr;
  public
    constructor Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X,Y,Z:cardinal):word;
    procedure PutValue(X,Y,Z:cardinal;Value:word);
    property Data[X,Y,Z:cardinal]:word read GetValue write PutValue; default;
  end;

  T4DWordArray=class(TLargeArray)
  private
    NumW,
    NumX,
    NumY,
    NumZ:cardinal;
    function Index(W,X,Y,Z:cardinal):TWordPtr;
  public
    constructor Create(W,X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(W,X,Y,Z:cardinal):word;
    procedure PutValue(W,X,Y,Z:cardinal;Value:word);
    property Data[W,X,Y,Z:cardinal]:word read GetValue write PutValue; default;
  end;

  T1DDoubleArray=class(TLargeArray)
  private
    FXOffset,
    NumX:cardinal;
    function Index(X:cardinal):TDoublePtr;
  public
    constructor Create(X:cardinal;MaxMem:cardinal=high(cardinal));overload;
    constructor Create(Source:T1DDoubleArray);overload;
    procedure DeepCopy(Source:T1DDoubleArray);
    function GetValue(X:cardinal):double;
    procedure PutValue(X:cardinal;Value:double);
    property Data[X:cardinal]:double read GetValue write PutValue; default;
    property GetNumX:cardinal read NumX;
    property XOffset:cardinal read FXOffset write FXOffset;
  end;

  T2DDoubleArray=class(TLargeArray)
  private
    FXOffset,
    FYOffset,
    NumX,
    NumY:cardinal;
    function Index(X,Y:cardinal):TDoublePtr;
  public
    constructor Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));overload;
    constructor Create(Source:T2DDoubleArray);overload;
    procedure DeepCopy(Source:T2DDoubleArray);
    function GetValue(X,Y:cardinal):double;
    procedure PutValue(X,Y:cardinal;Value:double);
    procedure FillWith(Value:double);
    procedure WriteToTextFile(Filename:string);
    property Data[X,Y:cardinal]:double read GetValue write PutValue; default;
    property GetNumX:cardinal read NumX;
    property GetNumY:cardinal read NumY;
    property XOffset:cardinal read FXOffset write FXOffset;
    property YOffset:cardinal read FYOffset write FYOffset;
  end;

  T3DDoubleArray=class(TLargeArray)
  private
    FXOffset,
    FYOffset,
    FZOffset,
    NumX,
    NumY,
    NumZ:cardinal;
    function Index(X,Y,Z:cardinal):TDoublePtr;
  public
    constructor Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));overload;
    constructor Create(Source:T3DDoubleArray);overload;
    procedure DeepCopy(Source:T3DDoubleArray);
    function GetValue(X,Y,Z:cardinal):double;
    procedure PutValue(X,Y,Z:cardinal;Value:double);
    property Data[X,Y,Z:cardinal]:double read GetValue write PutValue; default;
    property GetNumX:cardinal read NumX;
    property GetNumY:cardinal read NumY;
    property GetNumZ:cardinal read NumZ;
    property XOffset:cardinal read FXOffset write FXOffset;
    property YOffset:cardinal read FYOffset write FYOffset;
    property ZOffset:cardinal read FZOffset write FZOffset;
  end;

  T2DCardinalArray=class(TLargeArray)
  private
    NumX,
    NumY:cardinal;
    function Index(X,Y:cardinal):TCardinalPtr;
  public
    constructor Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
    function GetValue(X,Y:cardinal):cardinal;
    procedure PutValue(X,Y:cardinal;Value:cardinal);
    property Data[X,Y:cardinal]:cardinal read GetValue write PutValue; default;
    property GetNumX:cardinal read NumX;
    property GetNumY:cardinal read NumY;
  end;

  EArray=class(Exception);

function AddTDoubleArrays(A1,A2:TDoubleArray):TDoubleArray;
function AddT1DDoubleArrays(A1,A2:T1DDoubleArray):T1DDoubleArray;
function Calc1DDoubleArrayMean(X:T1DDoubleArray):double;
function CompareTCardinalArrays(X,Y:TCardinalArray):boolean;
function CompareTDoubleArrays(X,Y:TDoubleArray):boolean;
function CompareTBooleanArrays(X,Y:TBooleanArray):boolean;
function EqualSpacing(X:TDoubleArray):boolean;
procedure FillTDoubleArrayWithValue(var X:TDoubleArray;Value:double);

implementation

uses
  gb_files, {The unit defining file manipulation routines.}
  gb_math;

const
  Mask:array[0..22] of cardinal=(0,1,3,7,15,31,63,127,255,511,1023,2047,4095,8191,
    16383,32767,65535,131071,262143,524287,1048575,2097151,4194303);

constructor TLargeArray.Create(MaxHeap,NData,ESize:cardinal);
{------------------------------------------------------------------------------}
{ The parameters passed to this constructor are:                               }
{   MaxHeap: The maximum heap memory that can be used by the object in bytes.  }
{   NData: The number of entries in object's data array(s).                    }
{   ESize: The size of each entry in bytes.                                    }
{ The constructor first checks whether it is possible to hold the entire data  }
{ array on the heap without exceeding `MaxHeap'. If this is possible, the      }
{ object uses the HeapData pointer to store the data array on the heap. If     }
{ this is not possible, as much memory as possible, but still not exceeding    }
{ `MaxHeap', is assigned to the DiskData pointer which is then used to swap    }
{ one page of memory to and from a temporary storage file. The name of the     }
{ temporary storage data file is determined through a call to `FindJunkFile'   }
{ which is defined in the files unit.                                          }
{------------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variable.}
  ErrStr:string;
begin
  try
    if MaxHeap<100 then MaxHeap:=100;
    NumData:=NData;
    EntrySize:=ESize;
    NumPerPage:=MaxPageSize div ESize;
    ErrStr:='stage 1';
    if NumPerPage>NData then NumPerPage:=NData;
    I:=4194304;
    Shift:=22;
    while ((I>NumPerPage) or (I*ESize>MaxHeap)) do
    begin
      I:=I div 2;
      dec(Shift);
    end;
    ErrStr:='stage 2';
    NumPerPage:=I;
    PageSize:=I*ESize;
    NumPages:=NData shr Shift;
    LastPageSize:=(NData and Mask[Shift])*ESize;
    ErrStr:='stage 3';
    if LastPageSize<>0 then inc(NumPages) else LastPageSize:=PageSize;
    if (NData*ESize)>MaxHeap then Disk:=true
    else
    begin
      Disk:=false;
      if (PageSize*(NumPages-1)+LastPageSize+NumPages*sizeof(pointer))>MaxHeap
      then Disk:=true;
    end;
    ErrStr:='stage 4';
  except
    on E:Exception do raise EArray.Create('an error occurred in the TLargeArray '+
      'constructor after '+ErrStr+' with the error message '+E.Message);
  end;

  if Disk then
  begin
    getmem(DiskData,PageSize);
    fillchar(DiskData^,PageSize,0);
    assignfile(StoreFile,FindJunkFile);
    {$I-}rewrite(StoreFile,PageSize);{$I+}
    if ioresult<>0 then raise EArray.Create('couldn''t write to the temporary storage file.');
    for I:=1 to NumPages do
    begin
      {$I-}blockwrite(StoreFile,DiskData^,1);{$I+}
      if ioresult<>0 then raise EArray.Create('couldn''t write to the temporary storage file.');
    end;
    CurrentPage:=0;
    CloseFiles:=false;
  end
  else
  try
    getmem(HeapData,NumPages*sizeof(pointer));
    for I:=1 to NumPages do
      if I<NumPages then getmem(HeapData^[I-1],PageSize) else getmem(HeapData^[I-1],LastPageSize);
  except
    on E:Exception do raise EArray.Create('an error occurred at I='+inttostr(I)+
      ' and the error message was '+E.Message);
  end;
end; {End of TLargeArray.Create.}

procedure TLargeArray.Clear;
{------------------------------------------------------------------------------}
{ This procedure clears the entire object data array by filling it with zeros. }
{------------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variable.}
begin
  if Disk then
  begin
    if CloseFiles then reset(StoreFile,PageSize);
    fillchar(DiskData^,PageSize,0);
    seek(StoreFile,0);
    for I:=1 to NumPages do blockwrite(StoreFile,DiskData^,1);
    if CloseFiles then closefile(StoreFile);
  end
  else
  begin
    I:=NumPages-1;
    fillchar(HeapData^[I]^,LastPageSize,0);
    if I>0 then
    repeat
      dec(I);
      fillchar(HeapData^[I]^,PageSize,0);
    until I=0;
  end;
end; {End of TLargeArray.Clear.}

function TLargeArray.Index(Entry:cardinal):pointer;
{------------------------------------------------------------------------------}
{ This routine permits both writing to and reading from the array. If the      }
{ array index `Entry' is out of range, an EArray exception is raised.          }
{------------------------------------------------------------------------------}
var
  Page:word; {The page where the array element is held.}
begin
  if Entry>=NumData then raise EArray.Create('array index out of bounds'); {Entry cannot be <0 since defined as cardinal.}
  Page:=Entry shr Shift;
  if Disk then
  begin
    if Page<>CurrentPage then
    begin
      if CloseFiles then reset(StoreFile,PageSize);
      seek(StoreFile,CurrentPage);
      blockwrite(StoreFile,DiskData^,1);
      CurrentPage:=Page;
      seek(StoreFile,CurrentPage);
      blockread(StoreFile,DiskData^,1);
      if CloseFiles then closefile(StoreFile);
    end;
    Index:=@DiskData^[(Entry and Mask[Shift])*EntrySize];
  end
  else Index:=@HeapData^[Page]^[(Entry and Mask[Shift])*EntrySize];
end; {End of TLargeArray.Index.}

function TLargeArray.SaveData(var Dest:file):byte;
{------------------------------------------------------------------------------}
{ This method saves the entire data set to the file specified by `Dest'. It is }
{ assumed that the file has already been opened and is ready for being written }
{ to. The returned value reports error codes as follows:                       }
{ 0: The data saving was successful.                                           }
{ 1: The data saving was not successful.                                       }
{------------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variable.}
  ErrorCode:byte; {The error code returned by the function.}
begin
  ErrorCode:=0;
  if Disk then
  begin
    if CloseFiles then reset(StoreFile,PageSize);
    seek(StoreFile,CurrentPage);
    blockwrite(StoreFile,DiskData^,1);
    seek(StoreFile,0);
    for I:=1 to NumPages do
    begin
      blockread(StoreFile,DiskData^,1);
      {$I-}
      if I<NumPages then blockwrite(Dest,DiskData^,PageSize)
      else blockwrite(Dest,DiskData^,LastPageSize);
      {$I+}
      if ioresult<>0 then ErrorCode:=1;
    end;
    if CloseFiles then closefile(StoreFile);
    CurrentPage:=NumPages-1;
  end
  else
  for I:=1 to NumPages do
  begin
    {$I-}
    if I<NumPages then blockwrite(Dest,HeapData^[I-1]^,PageSize)
    else blockwrite(Dest,HeapData^[I-1]^,LastPageSize);
    {$I+}
    if ioresult<>0 then ErrorCode:=1;
   end;
  SaveData:=ErrorCode;
end; {End of TLargeArray.SaveData.}

function TLargeArray.LoadData(var Source:file):byte;
{------------------------------------------------------------------------------}
{ This method loads the data set from the file specified by `Source'. It is    }
{ assumed that the file has already been opened and is ready for being read    }
{ from. The value returned by the function is used to report error codes as    }
{ follows:                                                                     }
{ 0: The data loading was successful.                                          }
{ 1: The data loading was not successful.                                      }
{------------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variable.}
  ErrorCode:byte; {The error code returned by the function.}
begin
  ErrorCode:=0;
  if Disk then
  begin
    if CloseFiles then reset(StoreFile,PageSize);
    seek(StoreFile,CurrentPage);
    blockwrite(StoreFile,DiskData^,1);
    seek(StoreFile,0);
    for I:=1 to NumPages do
    begin
      {$I-}
      if I<NumPages then blockread(Source,DiskData^,PageSize)
      else blockread(Source,DiskData^,LastPageSize);
      {$I+}
      if ioresult=0 then blockwrite(StoreFile,DiskData^,1)
      else ErrorCode:=1
    end;
    if CloseFiles then closefile(StoreFile);
    CurrentPage:=NumPages-1;
  end
  else
  for I:=1 to NumPages do
  begin
    {$I-}
    if I<NumPages then blockread(Source,HeapData^[I-1]^,PageSize)
    else blockread(Source,HeapData^[I-1]^,LastPageSize);
    {$I+}
    if ioresult<>0 then ErrorCode:=1;
  end;
  LoadData:=ErrorCode;
end; {End of TLargeArray.LoadData.}

function TLargeArray.GetNumdata:cardinal;
begin
  Result:=NumData;
end;

function TLargeArray.GetSizeOfData:cardinal;
begin
  Result:=NumData*EntrySize;
end;

function TLargeArray.GetMemoryUsed:cardinal;
begin
  if Disk then Result:=PageSize else
  Result:=GetSizeOfData+NumPages*sizeof(pointer);
end;

procedure TLargeArray.GetMemoryStats(var TDisk:boolean;var TNumPages:cardinal);
{------------------------------------------------------------------------------}
{ This routine returns some memory and disk usage statistics that can be used  }
{ for diagnostic purposes.                                                     }
{------------------------------------------------------------------------------}
begin
  TDisk:=Disk;
  TNumPages:=NumPages;
end; {End of TLargeArray.GetMemoryStats.}

destructor TLargeArray.Destroy;
{------------------------------------------------------------------------------}
{ This destructor is responsible for freeing all of the assigned memory.       }
{------------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variable.}
begin
  if Disk then
  begin
    if CloseFiles then reset(StoreFile,PageSize);
    seek(StoreFile,CurrentPage);
    blockwrite(StoreFile,DiskData^,1);
    closefile(StoreFile);
    erase(StoreFile);
    freemem(DiskData,PageSize);
  end
  else
  begin
    I:=NumPages-1;
    freemem(HeapData^[I],LastPageSize);
    if I>0 then
    repeat
      dec(I);
      freemem(HeapData^[I],PageSize);
    until I=0;
    freemem(HeapData,NumPages*sizeof(pointer));
  end;
  inherited Destroy;
end; {End of TLargeArray.Destroy.}

{------------------------------------------------------------------------------}
{ These are the method definitions for T1DByteArray.                           }
{------------------------------------------------------------------------------}
constructor T1DByteArray.Create(X:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  inherited Create(MaxMem,X,sizeof(TByteType));
end;

function T1DByteArray.Index(X:cardinal):TBytePtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if X<1 then raise EArray.Create('X was less than 1');
    Index:=TBytePtr(inherited Index(X-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T1DByteArray.Index');
  end;
end;

function T1DByteArray.GetValue(X:cardinal):byte;
begin
  Result:=Index(X)^;
end;

procedure T1DByteArray.PutValue(X:cardinal;Value:byte);
begin
  Index(X)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T2DByteArray.                           }
{------------------------------------------------------------------------------}
constructor T2DByteArray.Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  inherited Create(MaxMem,X*Y,sizeof(TByteType));
end;

function T2DByteArray.Index(X,Y:cardinal):TBytePtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    Index:=TBytePtr(inherited Index((X-1)*NumY+Y-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T2DByteArray.Index');
  end;
end;

function T2DByteArray.GetValue(X,Y:cardinal):byte;
begin
  Result:=Index(X,Y)^;
end;

procedure T2DByteArray.PutValue(X,Y:cardinal;Value:byte);
begin
  Index(X,Y)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T3DByteArray.                           }
{------------------------------------------------------------------------------}
constructor T3DByteArray.Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  NumZ:=Z;
  inherited Create(MaxMem,X*Y*Z,sizeof(TByteType));
end;

function T3DByteArray.Index(X,Y,Z:cardinal):TBytePtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if Z>NumZ then raise EArray.Create('Z was greater than NumZ');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    if Z<1 then raise EArray.Create('Z was less than 1');
    Index:=TBytePtr(inherited Index((X-1)*NumY*NumZ+(Y-1)*NumZ+Z-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T3DByteArray.Index');
  end;
end;

function T3DByteArray.GetValue(X,Y,Z:cardinal):byte;
begin
  Result:=Index(X,Y,Z)^;
end;

procedure T3DByteArray.PutValue(X,Y,Z:cardinal;Value:byte);
begin
  Index(X,Y,Z)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T2DRealArray.                           }
{------------------------------------------------------------------------------}
constructor T2DRealArray.Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  inherited Create(MaxMem,X*Y,sizeof(TRealType));
end;

function T2DRealArray.Index(X,Y:cardinal):TRealPtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    Index:=TRealPtr(inherited Index((X-1)*NumY+Y-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T2DRealArray.Index');
  end;
end;

function T2DRealArray.GetValue(X,Y:cardinal):real;
begin
  Result:=Index(X,Y)^;
end;

procedure T2DRealArray.PutValue(X,Y:cardinal;Value:real);
begin
  Index(X,Y)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T2DSingleArray.                         }
{------------------------------------------------------------------------------}
constructor T2DSingleArray.Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  inherited Create(MaxMem,X*Y,sizeof(TSingleType));
end;

procedure T2DSingleArray.DeepCopy(Source:T2DSingleArray);
var
  I,J:cardinal;
begin
  if Source.NumX<>NumX then raise EArray.Create('NumX of source did not match NumX.');
  if Source.NumY<>NumY then raise EArray.Create('NumY of source did not match NumY.');
  for I:=1 to NumX do
    for J:=1 to NumY do
      Data[I,J]:=Source[I,J];
end; {End of T2DSingleArray.DeepCopy.}

function T2DSingleArray.Index(X,Y:cardinal):TSinglePtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    Index:=TSinglePtr(inherited Index((X-1)*NumY+Y-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T2DSingleArray.Index');
  end;
end;

function T2DSingleArray.GetValue(X,Y:cardinal):single;
begin
  Result:=Index(X,Y)^;
end;

procedure T2DSingleArray.PutValue(X,Y:cardinal;Value:single);
begin
  Index(X,Y)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T3DSingleArray.                         }
{------------------------------------------------------------------------------}
constructor T3DSingleArray.Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  NumZ:=Z;
  inherited Create(MaxMem,X*Y*Z,sizeof(TSingleType));
end;

function T3DSingleArray.Index(X,Y,Z:cardinal):TSinglePtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if Z>NumZ then raise EArray.Create('Z was greater than NumZ');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    if Z<1 then raise EArray.Create('Z was less than 1');
    Index:=TSinglePtr(inherited Index((X-1)*NumY*NumZ+(Y-1)*NumZ+Z-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T3DSingleArray.Index');
  end;
end;

function T3DSingleArray.GetValue(X,Y,Z:cardinal):single;
begin
  Result:=Index(X,Y,Z)^;
end;

procedure T3DSingleArray.PutValue(X,Y,Z:cardinal;Value:single);
begin
  Index(X,Y,Z)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T4DSingleArray.                         }
{------------------------------------------------------------------------------}
constructor T4DSingleArray.Create(W,X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumW:=W;
  NumX:=X;
  NumY:=Y;
  NumZ:=Z;
  FWOffset:=0;
  FXOffset:=0;
  FYOffset:=0;
  FXOffset:=0;
  inherited Create(MaxMem,W*X*Y*Z,sizeof(TSingleType));
end;

function T4DSingleArray.Index(W,X,Y,Z:cardinal):TSinglePtr;
begin
  W:=W-FWOffset;
  X:=X-FXOffset;
  Y:=Y-FYOffset;
  Z:=Z-FZOffset;
  try
    if W>NumW then raise EArray.Create('W was greater than NumW');
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if Z>NumZ then raise EArray.Create('Z was greater than NumZ');
    if W<1 then raise EArray.Create('W was less than 1');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    if Z<1 then raise EArray.Create('Z was less than 1');
    Index:=TSinglePtr(inherited Index((W-1)*NumX*NumY*NumZ+(X-1)*NumY*NumZ+(Y-1)*NumZ+Z-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T4DSingleArray.Index');
  end;
end;

function T4DSingleArray.GetValue(W,X,Y,Z:cardinal):single;
begin
  Result:=Index(W,X,Y,Z)^;
end;

procedure T4DSingleArray.PutValue(W,X,Y,Z:cardinal;Value:single);
begin
  Index(W,X,Y,Z)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T2DIntegerArray.                        }
{------------------------------------------------------------------------------}
constructor T2DIntegerArray.Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  inherited Create(MaxMem,X*Y,sizeof(TIntegerType));
end;

function T2DIntegerArray.Index(X,Y:cardinal):TIntegerPtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    Index:=TIntegerPtr(inherited Index((X-1)*NumY+Y-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T2DIntegerArray.Index');
  end;
end;

function T2DIntegerArray.GetValue(X,Y:cardinal):integer;
begin
  Result:=Index(X,Y)^;
end;

procedure T2DIntegerArray.PutValue(X,Y:cardinal;Value:integer);
begin
  Index(X,Y)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T1DWordArray.                           }
{------------------------------------------------------------------------------}
constructor T1DWordArray.Create(X:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  inherited Create(MaxMem,X,sizeof(TWordType));
end;

function T1DWordArray.Index(X:cardinal):TWordPtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if X<1 then raise EArray.Create('X was less than 1');
    Index:=TWordPtr(inherited Index(X-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T1DWordArray.Index');
  end;
end;

function T1DWordArray.GetValue(X:cardinal):word;
begin
  Result:=Index(X)^;
end;

procedure T1DWordArray.PutValue(X:cardinal;Value:word);
begin
  Index(X)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T2DWordArray.                           }
{------------------------------------------------------------------------------}
constructor T2DWordArray.Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  try
    inherited Create(MaxMem,X*Y,sizeof(TWordType));
  except
    on E:Exception do raise EArray.Create('the call to inherited Create in T2DWordArray.Create '+
      'failed with the error message: '+E.Message);
  end;
end;

procedure T2DWordArray.DeepCopy(Source:T2DWordArray);
var
  I,J:cardinal;
begin
  if Source.NumX<>NumX then raise EArray.Create('NumX of source did not match NumX.');
  if Source.NumY<>NumY then raise EArray.Create('NumY of source did not match NumY.');
  for I:=1 to NumX do
    for J:=1 to NumY do
      Data[I,J]:=Source[I,J];
end;

function T2DWordArray.Index(X,Y:cardinal):TWordPtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    Index:=TWordPtr(inherited Index((X-1)*NumY+Y-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T2DWordArray.Index');
  end;
end;

function T2DWordArray.GetValue(X,Y:cardinal):word;
begin
  Result:=Index(X,Y)^;
end;

procedure T2DWordArray.PutValue(X,Y:cardinal;Value:word);
begin
  Index(X,Y)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T3DWordArray.                           }
{------------------------------------------------------------------------------}
constructor T3DWordArray.Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  NumZ:=Z;
  inherited Create(MaxMem,X*Y*Z,sizeof(TWordType));
end;

function T3DWordArray.Index(X,Y,Z:cardinal):TWordPtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if Z>NumZ then raise EArray.Create('Z was greater than NumZ');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    if Z<1 then raise EArray.Create('Z was less than 1');
    Index:=TWordPtr(inherited Index((X-1)*NumY*NumZ+(Y-1)*NumZ+Z-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T3DWordArray.Index');
  end;
end;

function T3DWordArray.GetValue(X,Y,Z:cardinal):word;
begin
  Result:=Index(X,Y,Z)^;
end;

procedure T3DWordArray.PutValue(X,Y,Z:cardinal;Value:word);
begin
  Index(X,Y,Z)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T4DWordArray.                           }
{------------------------------------------------------------------------------}
constructor T4DWordArray.Create(W,X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumW:=W;
  NumX:=X;
  NumY:=Y;
  NumZ:=Z;
  inherited Create(MaxMem,W*X*Y*Z,sizeof(TWordType));
end;

function T4DWordArray.Index(W,X,Y,Z:cardinal):TWordPtr;
begin
  try
    if W>NumW then raise EArray.Create('W was greater than NumW');
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if Z>NumZ then raise EArray.Create('Z was greater than NumZ');
    if W<1 then raise EArray.Create('W was less than 1');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    if Z<1 then raise EArray.Create('Z was less than 1');
    Index:=TWordPtr(inherited Index((W-1)*NumX*NumY*NumZ+(X-1)*NumY*NumZ+(Y-1)*NumZ+Z-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T4DWordArray.Index');
  end;
end;

function T4DWordArray.GetValue(W,X,Y,Z:cardinal):word;
begin
  Result:=Index(W,X,Y,Z)^;
end;

procedure T4DWordArray.PutValue(W,X,Y,Z:cardinal;Value:word);
begin
  Index(W,X,Y,Z)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T1DDoubleArray.                         }
{------------------------------------------------------------------------------}
constructor T1DDoubleArray.Create(X:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  inherited Create(MaxMem,X,sizeof(TDoubleType));
  FXOffset:=0;
end;

constructor T1DDoubleArray.Create(Source:T1DDoubleArray);
begin
  Create(Source.NumX);
  DeepCopy(Source);
end;

procedure T1DDoubleArray.DeepCopy(Source:T1DDoubleArray);
var
  I:integer;
begin
  if Source.NumX<>NumX then raise EArray.Create('NumX of source did not match NumX.');
  for I:=1 to NumX do Data[I]:=Source[I];
end;

function T1DDoubleArray.Index(X:cardinal):TDoublePtr;
begin
  X:=X-FXOffset;
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if X<1 then raise EArray.Create('X was less than 1');
    Index:=TDoublePtr(inherited Index(X-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T1DDoubleArray.Index');
  end;
end;

function T1DDoubleArray.GetValue(X:cardinal):double;
begin
  Result:=Index(X)^;
end;

procedure T1DDoubleArray.PutValue(X:cardinal;Value:double);
begin
  Index(X)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T2DDoubleArray.                         }
{------------------------------------------------------------------------------}
constructor T2DDoubleArray.Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  FXOffset:=0;
  FYOffset:=0;
  inherited Create(MaxMem,X*Y,sizeof(TDoubleType));
end;

constructor T2DDoubleArray.Create(Source:T2DDoubleArray);
begin
  Create(Source.NumX,Source.NumY);
  DeepCopy(Source);
end;

procedure T2DDoubleArray.DeepCopy(Source:T2DDoubleArray);
var
  I,J:cardinal;
begin
  if Source.NumX<>NumX then raise EArray.Create('NumX of source did not match NumX.');
  if Source.NumY<>NumY then raise EArray.Create('NumY of source did not match NumY.');
  for I:=1 to NumX do
    for J:=1 to NumY do
      Data[I,J]:=Source[I,J];
end;

function T2DDoubleArray.Index(X,Y:cardinal):TDoublePtr;
begin
  X:=X-FXOffset;
  Y:=Y-FYOffset;
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    Index:=TDoublePtr(inherited Index((X-1)*NumY+Y-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T2DDoubleArray.Index');
  end;
end;

function T2DDoubleArray.GetValue(X,Y:cardinal):double;
begin
  Result:=Index(X,Y)^;
end;

procedure T2DDoubleArray.PutValue(X,Y:cardinal;Value:double);
begin
  Index(X,Y)^:=Value;
end;

procedure T2DDoubleArray.FillWith(Value:double);
var
  I,J:cardinal;
begin
  for I:=1 to NumX do
    for J:=1 to NumY do Data[I,J]:=Value;
end;

procedure T2DDoubleArray.WriteToTextFile(Filename:string);
var
  I,J:cardinal;
  Dest:textfile;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  for J:=1 to NumY do
  begin
    for I:=1 to NumX do write(Dest,Data[I,J]:10:5,' ');
    writeln(Dest);
  end;
  closefile(Dest);
end;



{------------------------------------------------------------------------------}
{ These are the method definitions for T3DDoubleArray.                         }
{------------------------------------------------------------------------------}
constructor T3DDoubleArray.Create(X,Y,Z:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  NumZ:=Z;
  FXOffset:=0;
  FYOffset:=0;
  FZOffset:=0;
  inherited Create(MaxMem,X*Y*Z,sizeof(TDoubleType));
end;

constructor T3DDoubleArray.Create(Source:T3DDoubleArray);
begin
  Create(Source.NumX,Source.NumY,Source.NumZ);
  DeepCopy(Source);
end;

procedure T3DDoubleArray.DeepCopy(Source:T3DDoubleArray);
var
  I,J,K:cardinal;
begin
  if Source.NumX<>NumX then raise EArray.Create('NumX of source did not match NumX.');
  if Source.NumY<>NumY then raise EArray.Create('NumY of source did not match NumY.');
  if Source.NumZ<>NumZ then raise EArray.Create('NumZ of source did not match NumZ.');
  for I:=1 to NumX do
    for J:=1 to NumY do
      for K:=1 to NumZ do
        Data[I,J,K]:=Source[I,J,K];
end;

function T3DDoubleArray.Index(X,Y,Z:cardinal):TDoublePtr;
begin
  X:=X-FXOffset;
  Y:=Y-FYOffset;
  Z:=Z-FZOffset;
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if Z>NumZ then raise EArray.Create('Z was greater than NumZ');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    if Z<1 then raise EArray.Create('Z was less than 1');
    Index:=TDoublePtr(inherited Index((X-1)*NumY*NumZ+(Y-1)*NumZ+Z-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T3DDoubleArray.Index');
  end;
end;

function T3DDoubleArray.GetValue(X,Y,Z:cardinal):double;
begin
  Result:=Index(X,Y,Z)^;
end;

procedure T3DDoubleArray.PutValue(X,Y,Z:cardinal;Value:double);
begin
  Index(X,Y,Z)^:=Value;
end;

{------------------------------------------------------------------------------}
{ These are the method definitions for T2DCardinalArray.                       }
{------------------------------------------------------------------------------}
constructor T2DCardinalArray.Create(X,Y:cardinal;MaxMem:cardinal=high(cardinal));
begin
  NumX:=X;
  NumY:=Y;
  inherited Create(MaxMem,X*Y,sizeof(TCardinalType));
end;

function T2DCardinalArray.Index(X,Y:cardinal):TCardinalPtr;
begin
  try
    if X>NumX then raise EArray.Create('X was greater than NumX');
    if Y>NumY then raise EArray.Create('Y was greater than NumY');
    if X<1 then raise EArray.Create('X was less than 1');
    if Y<1 then raise EArray.Create('Y was less than 1');
    Index:=TCardinalPtr(inherited Index((X-1)*NumY+Y-1));
  except
    on E:EArray do raise EArray.Create(E.Message+' in call to T2DCardinalArray.Index');
  end;
end;

function T2DCardinalArray.GetValue(X,Y:cardinal):Cardinal;
begin
  Result:=Index(X,Y)^;
end;

procedure T2DCardinalArray.PutValue(X,Y:cardinal;Value:Cardinal);
begin
  Index(X,Y)^:=Value;
end;

function AddTDoubleArrays(A1,A2:TDoubleArray):TDoubleArray;
var
  I:cardinal;
begin
  if length(A1)<>length(A2) then raise EArray.Create('the arrays passed to AddTDoubleArrays were not the same length.');
  setlength(Result,length(A1));
  for I:=Low(A1) to High(A2) do Result[I]:=A1[I]+A2[I];
end;

function AddT1DDoubleArrays(A1,A2:T1DDoubleArray):T1DDoubleArray;
{------------------------------------------------------------------------------}
{ It is the responsibility of the calling program to free the object returned  }
{ by this function.                                                            }
{------------------------------------------------------------------------------}
var
  I:cardinal;
begin
  if A1.NumX<>A2.NumX then raise EArray.Create('the arrays passed to AddT1DDoubleArrays were not the same length.');
  Result:=T1DDoubleArray.Create(A1.NumX);
  for I:=1 to A1.NumX do Result[I]:=A1[I]+A2[I];
end;

function Calc1DDoubleArrayMean(X:T1DDoubleArray):double;
var
  I:cardinal;
begin
  Result:=0;
  for I:=1 to X.NumX do Result:=Result+X.Index(I)^/X.NumX;
end; {End of Calc1DDoubleArrayMean.}

function CompareTCardinalArrays(X,Y:TCardinalArray):boolean;
var
  I:cardinal;
begin
  Result:=true;
  if length(X)<>length(Y) then Result:=false else
    for I:=1 to length(X) do
      if X[I-1]<>Y[I-1] then Result:=false;
end;

function CompareTDoubleArrays(X,Y:TDoubleArray):boolean;
var
  I:cardinal;
begin
  Result:=true;
  if length(X)<>length(Y) then Result:=false else
    for I:=1 to length(X) do
      if not EquivDouble(X[I-1],Y[I-1]) then Result:=false;
end;

function CompareTBooleanArrays(X,Y:TBooleanArray):boolean;
var
  I:cardinal;
begin
  Result:=true;
  if length(X)<>length(Y) then Result:=false else
    for I:=1 to length(X) do
      if not X[I-1]=Y[I-1] then Result:=false;
end;

function EqualSpacing(X:TDoubleArray):boolean;
var
  I:cardinal;
  TestDiff:double;
begin
  if length(X)<2 then Result:=true else
  begin
    Result:=true;
    TestDiff:=X[1]-X[0];
    for I:=1 to length(X)-2 do
      if abs((X[I+1]-X[I])-TestDiff)>0.01 then Result:=false;
  end;
end;

procedure FillTDoubleArrayWithValue(var X:TDoubleArray;Value:double);
var
  I:cardinal;
begin
  for I:=Low(X) to High(X) do X[I]:=Value;
end;

{$WARNINGS OFF}
end. {End of gb_array unit.}
{$WARNINGS ON}

