//@***h* BDBP/gb_g2binary.pas
// 
// NAME
//   gb_g2binary.pas
//
// PURPOSE
//  Defines a class for managing time series of data at a range of  
//  temporal resolutions, including random time steps.        
//
// DESCRIPTION
//  This unit defines a class for managing time series of data at a range of  
//  temporal resolutions, including random time steps.   
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_g2binary;

{$I UNITDIREC.INC}

interface

uses
  Contnrs,
  Dialogs,
  SysUtils,
  gb_array,
  gb_GeneralConstantsAndTypes,
  gb_xysignal;

type
  TVersionArray=array[1..4] of byte;
  TInterpolationFailPlan=(raiseexception,returnnull);
  TG2GridTimeStamp=(TSStart,TSMiddle,TSEnd);

  {By defining descendants from the TG2GridStore class, the precision with which}
  {the data are stored in gridded data versions of TG2Binary are defined.}
  TG2GridStoreClass=class of TG2GridStore;
  TG2Binary=class(TObjectList)
  strict private
    {$REGION 'strict private fields and methods'}
    {The first two fields are pretty much 'constants' that are used to describe }
    {the object when the object is saved to file. The values for both fields are}
    {set in the object constructor and cannot be changed through any method calls.}
    FFileVersion:TVersionArray;
    FFileDesc:st43;

    {These two fields specifiy the data resolution and the precision at which the data are stored.}
    FResolution:TResolution;
    FPrecision:TG2GridStoreClass;

    FFilename:string; {The name of the data file associated with this object.}
    FFullLocationName:st40; {The name of the location associated with this object - could be empty.}
    FLocationAbbreviation:st3; {The three letter abbreviation associated with the location.}
    FDataDescription:st200; {The 200 character description of the data.}
    FDataUnits:st40; {The 40 character description of the units of the data.}
    FNullValue, {The value used to denote the null value.}
    FLatitude,  {The latitude associated with this location. Negative for southern hemisphere.}
    FLongitude, {The longitude associated with this location. Negative for western hemisphere.}
    FAltitude:double; {The altitude associated with this location in meters.}
    FSortedByYear, {Set to true if the objects on the list have been sorted by year.}

    {FInterpolate is set to true if linearly interpolation must be performed when}
    {the GetValue method is called. It is set to true by default but its value   }
    {can be changed through the Interpolate property.}
    FInterpolate:boolean;

    FTimeStamp:TG2GridTimeStamp;
    FForceFileExt:boolean;
    FInterpolationFailPlan:TInterpolationFailPlan;

    procedure SetNullValue(Value:double);
    procedure SetFileName(Value:string);
    procedure SetTimeStamp(Value:TG2GridTimeStamp);
    procedure SwitchPrecision;
    {$ENDREGION}
  public
    {$REGION 'public methods'}
    constructor Create;overload;
    constructor Create(Resolution:TResolution);overload;
    constructor Create(Source:TG2Binary);overload;
    procedure   DeepCopy(Source:TG2Binary);
    procedure   Clear;override;
    function    GetYearIndex(Year:word):integer;overload;
    function    GetYearIndex(DateTime:TDateTime):integer;overload;
    function    AddYear(Year:word):integer;
    procedure   PutValue(DateTime:TDateTime;Value:double);overload;
    procedure   PutValue(Year:word;I:cardinal;Value:double);overload;
    function    GetValue(DateTime:TDateTime):double;overload;
    function    GetValue(Year:word;I:cardinal):double;overload;
    function    GetXYSignalOfValues:TXYSignal;
    function    GetMeanValue:double;
    function    IsNull(DateTime:TDateTime):boolean;overload;
    function    IsNull(Year:word;I:cardinal):boolean;overload;
    function    IsNotNull(DateTime:TDateTime):boolean;overload;
    function    IsNotNull(Year:word;I:cardinal):boolean;overload;
    function    GetStartYear:word;
    function    GetEndYear:word;
    function    GetDateTimeList(OnlyNonNull:boolean=true):TDoubleArray;
    function    GetFirstDateTime(OnlyNonNull:boolean=true):TDateTime;
    function    GetLastDateTime(OnlyNonNull:boolean=true):TDateTime;
    procedure   FillWithValue(AValue:double;OnlyNonNull:boolean=true);
    procedure   ScaleBy(AValue:double;OnlyNonNull:boolean=true);
    procedure   OffsetBy(AValue:double;OnlyNonNull:boolean=true);
    procedure   SwitchToSinglePrecision;
    procedure   SwitchToPackedPrecision;
    procedure   SwitchToDoublePrecision;
    procedure   SortByYear;
    procedure   Filter(FilterArray:TDoubleArray);
    procedure   SaveToFile(Filename:string);overload;
    procedure   SaveToFile;overload;
    procedure   WriteToTextFile(Filename:string;IncNull:boolean=false);
    procedure   LoadFromFile(Filename:string);overload;
    procedure   LoadFromFile;overload;
    {$ENDREGION}
    {$REGION 'properties'}
    property FileVersion:TVersionArray read FFileVersion;
    property Filename:string read FFilename write SetFilename;
    property FullLocationName:st40 read FFullLocationName write FFullLocationName;
    property NullValue:double read FNullValue write SetNullValue;
    property LocationAbbreviation:st3 read FLocationAbbreviation write FLocationAbbreviation;
    property DataDescription:st200 read FDataDescription write FDataDescription;
    property DataUnits:st40 read FDataUnits write FDataUnits;
    property Latitude:double read FLatitude write FLatitude;
    property Longitude:double read FLongitude write FLongitude;
    property Altitude:double read FAltitude write FAltitude;
    property Resolution:TResolution read FResolution;
    property Interpolate:boolean read FInterpolate write FInterpolate;
    property GridTimeStamp:TG2GridTimeStamp read FTimeStamp write SetTimeStamp;
    property InterpolationFailPlan:TInterpolationFailPlan read FInterpolationFailPlan write FInterpolationFailPlan;
    property ForceFileExt:boolean read FForceFileExt write FForceFileExt;
    property Data[DateTime:TDateTime]:double read GetValue write PutValue;
    property DefData[Year:word;I:cardinal]:double read GetValue write PutValue;default;
    {$ENDREGION}
  end;

  {This is the fundamental ancestor class for the object in which each year of }
  {data is stored. Two objects descend from this one - TG2GridStore which      }
  {stores a year of data where the temporal spacing between data points is     }
  {fixed, and TG2StoreDateTime which stores all of the data (i.e. for multiple }
  {years) where the temporal spacing between data points is not fixed.         }
  TG2Store=class
  private
    Owner:TG2Binary;
    procedure PutValue(DateTime:TDateTime;Value:double);virtual;abstract;
    function  GetValue(DateTime:TDateTime):double;virtual;abstract;

    {The value used to denote the null value in a TG2Binary is specified and   }
    {set in the TG2Binary. The procedure below simply puts that prescribed null}
    {value into all null locations in the data grid.}
    procedure SetNullValue(Value:double);virtual;abstract;

    procedure WriteToTextFile(var Dest:textfile;IncNull:boolean=false);virtual;abstract;
    procedure WriteToFile(var Dest:file);virtual;abstract;
    procedure LoadFromFile(var Source:file);virtual;abstract;
    function  IsNullValue(Value:double):boolean;
    function  GetNumData:cardinal;virtual;abstract;
  protected
    procedure PutIndexValue(I:cardinal;Value:double);virtual;abstract;
    function  GetIndexValue(I:cardinal):double;virtual;abstract;
  public
    {When loading data from a file the TG2Binary class does not read in all of}
    {the data in since it may not all need to be accessed. Rather it reads in }
    {just the file position where each data storage object is stored in the   }
    {file. If this FilePosition field has a non-zero value, it means that its }
    {data are still in the file and haven't yet been loaded.}
    FilePosition:cardinal;

    constructor Create;overload;
    constructor Create(AOwner:TG2Binary);overload;
    function  GetIndex(DateTime:TDateTime):cardinal;
    function  GetDateTime(Index:cardinal):TDateTime;virtual;abstract;
    function  GetBytesInFile:longint;virtual;abstract;
    function  IsNull(I:cardinal):boolean;overload;virtual;abstract;
    function  IsNull(DateTime:TDateTime):boolean;overload;virtual;abstract;
    function  IsNotNull(I:cardinal):boolean;overload;virtual;abstract;
    function  IsNotNull(DateTime:TDateTime):boolean;overload;virtual;abstract;
    {Properties}
    property  NumData:cardinal read GetNumData;
  end;

  {Because no instances of TG2Store are ever created (only instances of its }
  {descendents, it is not necessary that TG2Store defines methods for all of}
  {the abstract methods in its ancestor.                                    }
  TG2GridStore=class(TG2Store)
  strict private
    FYear:word;
  private
    procedure PutValue(DateTime:TDateTime;Value:double);override;
    function  GetValue(DateTime:TDateTime):double;override;
    procedure WriteToTextFile(var Dest:textfile;IncNull:boolean=false);override;
    function  GetNumData:cardinal;override;
  public
    constructor Create;overload;
    constructor Create(Year:word;AOwner:TG2Binary);overload;
    constructor Create(Source:TG2GridStore);overload;
    procedure DeepCopy(Source:TG2GridStore);virtual;
    procedure GetMinAndMax(var Min,Max:double);
    function  GetDateTime(Index:cardinal):TDateTime;override;
    function  GetLastNonNullValue(var DateTime:TDateTime):double;
    function  GetFirstNonNullValue(var DateTime:TDateTime):double;
    procedure SetNullValue(Value:double);override;
    function  IsNull(I:cardinal):boolean;override;
    function  IsNull(DateTime:TDateTime):boolean;override;
    function  IsNotNull(I:cardinal):boolean;override;
    function  IsNotNull(DateTime:TDateTime):boolean;override;
    {Properties}
    property Year:word read FYear;
    property DefData[DateTime:TDateTime]:double read GetValue write PutValue;default;
    property Data[I:cardinal]:double read GetIndexValue write PutIndexValue;
  end;

  TG2StorePacked=class(TG2GridStore)
  strict private
    FData:TWordArray;
    FScaling,
    FOffset:double;
    procedure PutIndexWordValue(I:cardinal;Value:word);
    function  GetIndexWordValue(I:cardinal):word;
    procedure InitializeFData;
    procedure SetScaling(Value:double);
    procedure SetOffset(Value:double);
    procedure WriteToFile(var Dest:file);override;
    procedure LoadFromFile(var Source:file);override;
  private
    procedure PutValue(DateTime:TDateTime;Value:double);override;
  protected
    procedure PutIndexValue(I:cardinal;Value:double);override;
    function  GetIndexValue(I:cardinal):double;override;
  public
    constructor Create;overload;
    procedure DeepCopy(Source:TG2GridStore);override;
    procedure Rescale(Value:double);
    function  GetBytesInFile:longint;override;
    function  IsNull(I:cardinal):boolean;override;
    function  IsNotNull(I:cardinal):boolean;override;
    procedure SetMinAndMax(Min,Max:double);
    {Properties}
    property Scaling:double read FScaling write SetScaling;
    property Offset:double read FOffset write SetOffset;
    property WordData[I:cardinal]:word read GetIndexWordValue write PutIndexWordValue;
  end;

  TG2StoreSingle=class(TG2GridStore)
  strict private
    FData:TSingleArray;
    procedure InitializeFData;
    procedure WriteToFile(var Dest:file);override;
    procedure LoadFromFile(var Source:file);override;
  protected
    procedure PutIndexValue(I:cardinal;Value:double);override;
    function  GetIndexValue(I:cardinal):double;override;
  public
    constructor Create;overload;
    procedure DeepCopy(Source:TG2GridStore);override;
//    function  GetBytesInFile:longint;override;
  end;

  TG2StoreDouble=class(TG2GridStore)
  strict private
    FData:TDoubleArray;
    procedure InitializeFData;
    procedure WriteToFile(var Dest:file);override;
    procedure LoadFromFile(var Source:file);override;
  protected
    procedure PutIndexValue(I:cardinal;Value:double);override;
    function  GetIndexValue(I:cardinal):double;override;
  public
    constructor Create;overload;
    procedure DeepCopy(Source:TG2GridStore);override;
//    function  GetBytesInFile:longint;override;
  end;

  TG2StoreDateTime=class(TG2Store)
  strict private
    FData:TXYSignal;
    procedure WriteToTextFile(var Dest:textfile;IncNull:boolean=false);override;
    procedure WriteToFile(var Dest:file);override;
    procedure LoadFromFile(var Source:file);override;
    function  GetNumData:cardinal;override;
  private
    procedure PutValue(DateTime:TDateTime;Value:double);override;
    function  GetValue(DateTime:TDateTime):double;override;
  protected
    procedure PutIndexValue(I:cardinal;Value:double);override;
    function  GetIndexValue(I:cardinal):double;override;
  public
    constructor Create;overload;
    constructor Create(AOwner:TG2Binary);overload;
    constructor Create(Source:TG2StoreDateTime);overload;
    destructor Destroy;override;
    procedure Clear;
    procedure DeepCopy(Source:TG2StoreDateTime);
//    function  GetBytesInFile:longint;override;
    procedure SortByTime;
    function  GetIndexXValue(I:cardinal):double;
    function  GetIndexYValue(I:cardinal):double;
    function  GetDateTime(Index:cardinal):TDateTime;override;
    function  GetFirstDateTime:TDateTime;
    function  GetLastDateTime:TDateTime;
    procedure SetNullValue(Value:double);override;
    function  IsNull(I:cardinal):boolean;override;
    function  IsNull(DateTime:TDateTime):boolean;override;
    function  IsNotNull(I:cardinal):boolean;override;
    function  IsNotNull(DateTime:TDateTime):boolean;override;
    {Properties}
    property NumData:cardinal read GetNumData;
    property DefData[DateTime:TDateTime]:double read GetValue write PutValue;default;
  end;

  EG2Binary=class(Exception);
  EG2Store=class(Exception);

implementation

uses
  DateUtils,
  gb_datetime,
  gb_dialogutils;

{------------------------------------------------------------------------------}
{ Here is where the class method, procedure and function defintions start.     }
{------------------------------------------------------------------------------}
function NumPerYear(Resolution:TResolution):cardinal;
begin
  case Resolution of
    ResSecond:Result:=31622400;
    ResMinute:Result:=527040;
    ResHour:Result:=8784;
    ResDay:Result:=366;
    ResMonth:Result:=12;
    ResQuarterly:Result:=4;
    ResSemiAnnual:Result:=2;
    ResYear:Result:=1;
  else
    Result:=0;
  end;
end; {End of NumPerYear.}

function CalcResolution(NumPerYear:cardinal):TResolution;
begin
  case NumPerYear of
    1:Result:=ResYear;
    2:Result:=ResSemiAnnual;
    4:Result:=ResQuarterly;
    12:Result:=ResMonth;
    366:Result:=ResDay;
    8784:Result:=ResHour;
    527040:Result:=ResMinute;
    31622400:Result:=ResSecond;
  else
    raise EG2Store.Create('there is no defined resolution for NumPerYear='+inttostr(NumPerYear)+'.');
  end;
end; {End of Resolution.}

{------------------------------------------------------------------------------}
{ Here are the method definitions for TG2Store.                                }
{------------------------------------------------------------------------------}
constructor TG2Store.Create;
begin
  raise EG2Store.Create('calling the constructor without any parameters is not permitted.');
end; {End of TG2Store.Create.}

constructor TG2Store.Create(AOwner:TG2Binary);
begin
  if AOwner=nil then raise EG2Store.Create('an owner must be specified');
  Owner:=AOwner;
end; {End of TG2Store.Create.}

function TG2Store.GetIndex(DateTime:TDateTime):cardinal;
begin
  case Owner.Resolution of
    ResSecond:Result:=SecondOfTheYear(DateTime)+1;
    ResMinute:Result:=MinuteOfTheYear(DateTime)+1;
    ResHour:Result:=HourOfTheYear(DateTime)+1;
    ResDay:Result:=DayOfTheYear(DateTime);
    ResMonth:Result:=MonthOfTheYear(DateTime);
    ResQuarterly:Result:=((MonthOfTheYear(DateTime)-1) div 3)+1;
    ResSemiAnnual:Result:=((MonthOfTheYear(DateTime)-1) div 6)+1;
    ResYear:Result:=1;
    ResRandom:raise EG2Store.Create('GetIndex cannot be called with random resolution.');
  else
    Result:=0;
  end;
end; {End of TG2Store.GetIndex.}

function TG2Store.IsNullValue(Value:double):boolean;
begin
  if Owner.NullValue=0 then
  begin
    if abs(Value)<1E-20 then Result:=true else Result:=false;
  end
  else if abs((Value-Owner.NullValue)/Owner.NullValue)<1E-6 then Result:=true else Result:=false;
end; {End of TG2Store.IsNullValue.}

{------------------------------------------------------------------------------}
{ Here are the method definitions for TG2GridStore.                            }
{------------------------------------------------------------------------------}
constructor TG2GridStore.Create;
begin
  raise EG2Store.Create('calling the constructor without any parameters is not permitted.');
end; {End of TG2GridStore.Create.}

constructor TG2GridStore.Create(Year:word;AOwner:TG2Binary);
begin
  inherited Create(AOwner);
  if Owner.Resolution=ResRandom then raise EG2Store.Create('random resolution is not a valid resolution for TG2GridStore.');
  FYear:=Year;
  FilePosition:=0;
end; {End of TG2GridStore.Create.}

constructor TG2GridStore.Create(Source:TG2GridStore);
begin
  if Assigned(Source) then DeepCopy(Source);
end; {End of TG2GridStore.Create.}

procedure TG2GridStore.DeepCopy(Source:TG2GridStore);
begin
  Owner:=Source.Owner;
  FYear:=Source.Year;
  {Descendent objects define the actual procedure for copying data across from}
  {Source. They must therefore call inherited DeepCopy(Source).}
end; {End of TG2GridStore.DeepCopy.}

procedure TG2GridStore.GetMinAndMax(var Min,Max:double);
var
  I:cardinal;
  Hold:double;
begin
  Max:=-1E30;
  Min:=1E30;
  for I:=1 to NumData do
  if IsNotNull(I) then
  begin
    Hold:=GetIndexValue(I); {This should speed this up a bit.}
    if Hold>Max then Max:=Hold;
    if Hold<Min then Min:=Hold;
  end;
end; {End of TG2GridStore.GetMinAndMax.}

function TG2GridStore.GetDateTime(Index:cardinal):TDateTime;
var
  Offset:double;
  DaysInPeriod:word;
  I:integer;
begin
  if Index>NumData then raise EG2Store.Create('the Index value was greater that '+inttostr(NumData));
  if Index<1 then raise EG2Store.Create('the Index value cannot be less than 1.');
  case Owner.GridTimeStamp of
    TSStart:Offset:=0.0;
    TSMiddle:Offset:=0.5;
    TSEnd:Offset:=1.0;
  else
    Offset:=0.0;
  end;
  case Owner.Resolution of
    ResSecond:Result:=EncodeDate(FYear,1,1)+(Index-1+Offset)/86400.0;
    ResMinute:Result:=EncodeDate(FYear,1,1)+(Index-1+Offset)/1440.0;
    ResHour:Result:=EncodeDate(FYear,1,1)+(Index-1+Offset)/24.0;
    ResDay:Result:=EncodeDate(FYear,1,1)+(Index-1+Offset);
    ResMonth:Result:=EncodeDate(FYear,Index,1)+DaysInAMonth(FYear,Index)*Offset;
    ResQuarterly:begin
      DaysInPeriod:=0;
      for I:=Index*3-2 to Index*3 do DaysInPeriod:=DaysInPeriod+DaysInAMonth(FYear,I);
      Result:=EncodeDate(FYear,Index*3-2,1)+DaysInPeriod*Offset;
    end;
    ResSemiAnnual:Result:=EncodeDate(FYear,1,1)+DaysInAYear(FYear)*(Index-1+Offset)*0.5;
    ResYear:Result:=EncodeDate(FYear,1,1)+DaysInAYear(FYear)*Offset;
  else
    Result:=EncodeDate(FYear,1,1)+DaysInAYear(FYear)*Offset;
  end;
end; {End of TG2GridStore.GetDateTime.}

function TG2GridStore.GetLastNonNullValue(var DateTime:TDateTime):double;
var
  I:cardinal;
  Done:boolean;
begin
  Result:=Owner.NullValue;
  I:=NumPerYear(Owner.Resolution); {This is to speed things up.}
  Done:=false;
  repeat
    if IsNotNull(I) then
    begin
      Result:=GetIndexValue(I);
      DateTime:=GetDateTime(I);
      Done:=true;
    end;
    dec(I);
  until (Done or (I=0));
  if not Done then raise EG2Store.Create('there were no non-null values in the '+
    'call to GetLastNonNullValue.');
end;

function TG2GridStore.GetFirstNonNullValue(var DateTime:TDateTime):double;
var
  Num,
  I:cardinal;
  Done:boolean;
begin
  Result:=Owner.NullValue;
  I:=0;
  Num:=NumPerYear(Owner.Resolution); {This is to speed things up.}
  Done:=false;
  repeat
    inc(I);
    if IsNotNull(I) then
    begin
      Result:=GetIndexValue(I);
      DateTime:=GetDateTime(I);
      Done:=true;
    end;
  until (Done or (I=Num));
  if not Done then raise EG2Store.Create('there were no non-null values in the '+
    'call to GetLastNonNullValue.');
end;


procedure TG2GridStore.SetNullValue(Value:double);
var
  I:cardinal; {I must be of type cardinal to ensure that the correct version of IsNull is called.}
begin
  if Value<>Owner.NullValue then
    for I:=1 to NumData do
      if IsNull(I) then PutIndexValue(I,Value);
end; {End of TG2GridStore.SetNullValue.}

function TG2GridStore.IsNull(I:cardinal):boolean;
begin
  Result:=IsNullValue(GetIndexValue(I));
end; {End of TG2GridStore.IsNull.}

function TG2GridStore.IsNull(DateTime:TDateTime):boolean;
begin
  Result:=IsNull(GetIndex(DateTime));
end; {End of TG2GridStore.IsNull.}

function TG2GridStore.IsNotNull(I:cardinal):boolean;
begin
  Result:=not IsNull(I);
end; {End of TG2GridStore.IsNotNull.}

function TG2GridStore.IsNotNull(DateTime:TDateTime):boolean;
begin
  Result:=not IsNull(GetIndex(DateTime));
end; {End of TG2GridStore.IsNotNull.}

procedure TG2GridStore.PutValue(DateTime:TDateTime;Value:double);
begin
  PutIndexValue(GetIndex(DateTime),Value);
end; {End of TG2GridStore.PutValue.}

function TG2GridStore.GetValue(DateTime:TDateTime):double;
var
  UpperIndex,
  LowerIndex:cardinal;
  UpperValue,
  LowerValue:double;
  I:cardinal;
  MiddleDateTime:TDateTime;
begin
  I:=GetIndex(DateTime);
  if Owner.Interpolate then
  begin
    MiddleDateTime:=GetDateTime(I);
    if SameDateTime(MiddleDateTime,DateTime) then Result:=GetIndexValue(I) else
    begin
      if DateTime<MiddleDateTime then LowerIndex:=I-1 else LowerIndex:=I;
      UpperIndex:=LowerIndex+1;
      while (IsNull(LowerIndex) and (LowerIndex>1)) do dec(LowerIndex);
      while (IsNull(UpperIndex) and (UpperIndex<NumData)) do inc(UpperIndex);
      LowerValue:=GetIndexValue(LowerIndex);
      UpperValue:=GetIndexValue(UpperIndex);
      if (IsNull(LowerIndex) or IsNull(UpperIndex)) then
        raise EG2Store.Create('no non-nulls were available to perform the interpolation.')
      else Result:=LowerValue+(DateTime-GetDateTime(LowerIndex))*(UpperValue-LowerValue)/
        (GetDateTime(UpperIndex)-GetDateTime(LowerIndex));
    end;
  end
  else Result:=GetIndexValue(I);
end; {End of TG2GridStore.GetValue.}

procedure TG2GridStore.WriteToTextFile(var Dest:textfile;IncNull:boolean=false);
var
  I:cardinal;
begin
  for I:=1 to NumData do
    if IncNull then writeln(Dest,ConvertDateTimeToDecYear(GetDateTime(I)):10:5,' ',GetIndexValue(I):10:5)
    else if IsNotNull(I) then writeln(Dest,ConvertDateTimeToDecYear(GetDateTime(I)):10:5,' ',GetIndexValue(I):10:5);
end; {End of TG2GridStore.WriteToTextFile.}

function TG2GridStore.GetNumData:cardinal;
begin
  Result:=NumPerYear(Owner.Resolution);
end; {End of TG2GridStore.GetNumData.}

{------------------------------------------------------------------------------}
{ Here are the method definitions for TG2StorePacked.                          }
{------------------------------------------------------------------------------}
constructor TG2StorePacked.Create;
begin
  raise EG2Store.Create('calling the constructor without any parameters is not permitted.');
end; {End of TG2StorePacked.Create.}

procedure TG2StorePacked.DeepCopy(Source:TG2GridStore);
var
  I:cardinal;
  Min,Max:double;
begin
  inherited DeepCopy(Source);
  if Source is TG2StorePacked then
  begin
    FScaling:=(Source as TG2StorePacked).Scaling;
    FOffset:=(Source as TG2StorePacked).Offset;
    for I:=1 to Source.NumData do PutIndexWordValue(I,(Source as TG2StorePacked).WordData[I]);
  end
  else
  begin
    Source.GetMinAndMax(Min,Max);
    SetMinAndMax(Min,Max);
    for I:=1 to Source.NumData do
      if Source.IsNull(I) then PutIndexWordValue(I,0) else PutIndexValue(I,Source.GetIndexValue(I));
  end;
end; {End of TG2StorePacked.DeepCopy.}

procedure TG2StorePacked.Rescale(Value:double);
var
  NewScaling,
  NewOffset,
  Min,
  Max:double;
  I:cardinal;
begin
  GetMinAndMax(Min,Max);
  if Value<Min then Min:=Value;
  if Value>Max then Max:=Value;
  if Max=Min then NewScaling:=1E-5 else NewScaling:=(Max-Min)/65534.0;
  NewOffset:=Min-NewScaling;
  for I:=1 to NumData do
    if IsNotNull(I) then PutIndexWordValue(I,round((GetIndexValue(I)-NewOffset)/NewScaling));
  FScaling:=NewScaling;
  FOffset:=NewOffset;
end; {End of TG2StorePacked.Rescale.}

function TG2StorePacked.GetBytesInFile:longint;
begin
  Result:=sizeof(FScaling)+sizeof(FOffset)+(NumData*sizeof(word));
end; {End of TG2StorePacked.GetBytesInFile.}

function TG2StorePacked.IsNull(I:cardinal):boolean;
begin
  Result:=GetIndexWordValue(I)=0;
end; {End of TG2StorePacked.IsNull.}

function TG2StorePacked.IsNotNull(I:cardinal):boolean;
begin
  Result:=GetIndexWordValue(I)<>0;
end; {End of TG2StorePacked.IsNull.}

procedure TG2StorePacked.SetMinAndMax(Min,Max:double);
begin
  if length(FData)<>0 then raise EG2Store.Create(
    'the SetMinAndMax method must not be called when there are already data in the array.');
  FScaling:=(Max-Min)/65534.0;
  FOffset:=Min-FScaling;
end; {End of TG2StorePacked.SetMinAndMax.}

procedure TG2StorePacked.SetScaling(Value:double);
begin
  if length(FData)<>0 then raise EG2Store.Create(
    'the SetScaling method must not be called when there are already data in the array.');
  FScaling:=Value;;
end; {End of TG2StorePacked.SetScaling.}

procedure TG2StorePacked.SetOffset(Value:double);
begin
  if length(FData)<>0 then raise EG2Store.Create(
    'the SetOffset method must not be called when there are already data in the array.');
  FOffset:=Value;;
end; {End of TG2StorePacked.SetScaling.}

procedure TG2StorePacked.WriteToFile(var Dest:file);
begin
  {If anything gets added to here, make sure to modify the GetBytesInFile function.}
  blockwrite(Dest,FScaling,sizeof(FScaling));
  blockwrite(Dest,FOffset,sizeof(FOffset));
  blockwrite(Dest,FData,length(FData)*sizeof(word));
end; {End of TG2StorePacked.WriteToFile.}

procedure TG2StorePacked.LoadFromFile(var Source:file);
begin
  blockread(Source,FScaling,sizeof(FScaling));
  blockread(Source,FOffset,sizeof(FOffset));
  blockread(Source,FData,length(FData)*sizeof(word));
end; {End of TG2StorePacked.LoadFromFile.}

procedure TG2StorePacked.PutValue(DateTime:TDateTime;Value:double);
begin
  PutIndexValue(GetIndex(DateTime),Value);
end; {End of TG2StorePacked.PutValue.}

procedure TG2StorePacked.PutIndexValue(I:cardinal;Value:double);
var
  TestValue:double;
begin
  if IsNullValue(Value) then PutIndexWordValue(I,0) else
  begin
    if FScaling=0.0 then TestValue:=0.0 else TestValue:=round((Value-FOffset)/FScaling);
    if ((TestValue<1.0) or (TestValue>65535.0)) then Rescale(Value);
    PutIndexWordValue(I,round((Value-FOffset)/FScaling));
  end;
end; {End of TG2StorePacked.PutIndexValue.}

function TG2StorePacked.GetIndexValue(I:cardinal):double;
begin
  if GetIndexWordValue(I)=0 then Result:=Owner.NullValue else Result:=GetIndexWordValue(I)*FScaling+FOffset;
end; {End of TG2StorePacked.GetIndexValue.}

procedure TG2StorePacked.PutIndexWordValue(I:cardinal;Value:word);
begin
  if length(FData)=0 then InitializeFData;
  if I<1 then raise EG2Store.Create('the index passed was less than 1.');
  if I>cardinal(length(FData)) then raise EG2Store.Create('the value passed was greater than '+inttostr(length(FData)));
  FData[I-1]:=Value;
end; {End of TG2StorePacked.PutIndexWordValue.}

function TG2StorePacked.GetIndexWordValue(I:cardinal):word;
begin
  if length(FData)=0 then InitializeFData;
  if I<1 then raise EG2Store.Create('the index passed was less than 1.');
  if I>cardinal(length(FData)) then raise EG2Store.Create('the value passed was greater than '+inttostr(length(FData)));
  Result:=FData[I-1];
end; {End of TG2StorePacked.GetIndexWordValue.}

procedure TG2StorePacked.InitializeFData;
var
  I:cardinal;
  Source:file;
begin
  setlength(FData,NumData);
  if FilePosition>0 then {This means that the data grid is available in a file.}
  begin
    assignfile(Source,Owner.Filename);
    reset(Source);
    seek(Source,FilePosition);
//    LoadFromFile



    closefile(Source);
  end
  else for I:=Low(FData) to High(FData) do FData[I]:=0;
end; {End of TG2StorePacked.InitializeFData.}

{------------------------------------------------------------------------------}
{ Here are the method definitions for TG2StoreSingle.                          }
{------------------------------------------------------------------------------}
constructor TG2StoreSingle.Create;
begin
  raise EG2Store.Create('calling the constructor without any parameters is not permitted.');
end; {End of TG2StoreSingle.Create.}

procedure TG2StoreSingle.DeepCopy(Source:TG2GridStore);
var
  I:cardinal;
begin
  inherited DeepCopy(Source);
  for I:=1 to Source.NumData do PutIndexValue(I,Source.Data[I]);
end; {End of TG2StoreSingle.DeepCopy.}

procedure TG2StoreSingle.PutIndexValue(I:cardinal;Value:double);
begin
  if length(FData)=0 then InitializeFData;
  if I<1 then raise EG2Store.Create('the index passed was less than 1.');
  if I>cardinal(length(FData)) then raise EG2Store.Create('the value passed was greater than '+inttostr(length(FData)));
  FData[I-1]:=Value;
end; {End of TG2StoreSingle.PutIndexValue.}

function TG2StoreSingle.GetIndexValue(I:cardinal):double;
begin
  if length(FData)=0 then InitializeFData;
  if I<1 then raise EG2Store.Create('the index passed was less than 1.');
  if I>cardinal(length(FData)) then raise EG2Store.Create('the value passed was greater than '+inttostr(length(FData)));
  Result:=FData[I-1];
end; {End of TG2StoreSingle.GetIndexValue.}

procedure TG2StoreSingle.InitializeFData;
var
  I:cardinal;
begin
  setlength(FData,NumData);
  if FilePosition>0 then
  begin



  end
  else for I:=Low(FData) to High(FData) do FData[I]:=Owner.NullValue;
end; {End of TG2StoreSingle.InitializeFData.}

procedure TG2StoreSingle.WriteToFile(var Dest:file);
begin
  blockwrite(Dest,FData,length(FData)*sizeof(single));
end;

procedure TG2StoreSingle.LoadFromFile(var Source:file);
begin
  blockread(Source,FData,length(FData)*sizeof(single));
end; {End of TG2StoreSingle.LoadFromFile.}

{------------------------------------------------------------------------------}
{ Here are the method definitions for TG2StoreDouble.                          }
{------------------------------------------------------------------------------}
constructor TG2StoreDouble.Create;
begin
  raise EG2Store.Create('calling the constructor without any parameters is not permitted.');
end; {End of TG2StoreDouble.Create.}

procedure TG2StoreDouble.DeepCopy(Source:TG2GridStore);
var
  I:cardinal;
begin
  inherited DeepCopy(Source);
  for I:=1 to Source.NumData do PutIndexValue(I,Source.Data[I]);
end; {End of TG2StoreDouble.DeepCopy.}

procedure TG2StoreDouble.PutIndexValue(I:cardinal;Value:double);
begin
  if length(FData)=0 then InitializeFData;
  if I<1 then raise EG2Store.Create('the index passed was less than 1.');
  if I>cardinal(length(FData)) then raise EG2Store.Create('the value passed was greater than '+inttostr(length(FData)));
  FData[I-1]:=Value;
end; {End of TG2StoreDouble.PutIndexValue.}

function TG2StoreDouble.GetIndexValue(I:cardinal):double;
begin
  if length(FData)=0 then InitializeFData;
  if I<1 then raise EG2Store.Create('the index passed was less than 1.');
  if I>cardinal(length(FData)) then raise EG2Store.Create('the value passed was greater than '+inttostr(length(FData)));
  Result:=FData[I-1];
end; {End of TG2StoreDouble.GetIndexValue.}

procedure TG2StoreDouble.InitializeFData;
var
  I:cardinal;
begin
  setlength(FData,NumData);
  if FilePosition>0 then
  begin


  end
  else for I:=Low(FData) to High(FData) do FData[I]:=Owner.NullValue;
end; {End of TG2StoreDouble.InitializeFData.}

procedure TG2StoreDouble.WriteToFile(var Dest:file);
begin
  blockwrite(Dest,FData,length(FData)*sizeof(double));
end; {End of TG2StoreDouble.WriteToFile.}

procedure TG2StoreDouble.LoadFromFile(var Source:file);
begin
  blockread(Source,FData,length(FData)*sizeof(double));
end; {End of TG2StoreDouble.LoadFromFile.}

{------------------------------------------------------------------------------}
{ Here are the method definitions for TG2StoreDateTime.                        }
{------------------------------------------------------------------------------}
constructor TG2StoreDateTime.Create;
begin
  raise EG2Store.Create('calling the constructor without any parameters is not permitted.');
end; {End of TG2StoreDateTime.Create.}

constructor TG2StoreDateTime.Create(AOwner:TG2Binary);
begin
  inherited Create(AOwner);
  FData:=TXYSignal.Create(-1E6,Owner.NullValue);
end; {End of TG2StoreDateTime.Create.}

constructor TG2StoreDateTime.Create(Source:TG2StoreDateTime);
begin
  if Assigned(Source) then DeepCopy(Source);
end; {End of TG2StoreDateTime.Create.}

destructor TG2StoreDateTime.Destroy;
begin
  if FData<>nil then FreeAndNil(FData);
  inherited Destroy;
end; {End of TG2StoreDateTime.Destroy.}

procedure TG2StoreDateTime.Clear;
begin
  FData.Clear;
end;

procedure TG2StoreDateTime.DeepCopy(Source:TG2StoreDateTime);
begin
  if FData<>nil then FreeAndNil(FData);
  FData:=TXYSignal.Create(Source.FData);
end; {End of TG2StoreDateTime.DeepCopy.}

procedure TG2StoreDateTime.SortByTime;
begin
  FData.SortByX;
end; {End of TG2StoreDateTime.SortByTime.}

function TG2StoreDateTime.GetFirstDateTime:TDateTime;
begin
  SortByTime;
  Result:=FData.GetFirstCoord.XValue;
end; {End of TG2StoreDateTime.GetFirstDateTime.}

function TG2StoreDateTime.GetLastDateTime:TDateTime;
begin
  SortByTime;
  Result:=FData.GetLastCoord.XValue;
end; {End of TG2StoreDateTime.GetLastDateTime.}

procedure TG2StoreDateTime.SetNullValue(Value:double);
var
  I:integer;
begin
  if Value<>Owner.NullValue then
    for I:=1 to FData.GetNumData do
      if FData.IsNullY(I) then FData.PutYValue(I,Value);
end; {End of TG2StoreDateTime.SetNullValue.}

function TG2StoreDateTime.IsNull(I:cardinal):boolean;
begin
  Result:=FData.IsNullY(I);
end; {End of TG2StoreDateTime.IsNull.}

function TG2StoreDateTime.IsNull(DateTime:TDateTime):boolean;
var
  I:integer;
begin
  Result:=true;
  for I:=1 to FData.GetNumData do
    if (SameDateTime(FData.GetXValue(I),DateTime) and (not FData.IsNullY(I))) then Result:=false;
end; {End of TG2StoreDateTime.IsNull.}

function TG2StoreDateTime.IsNotNull(I:cardinal):boolean;
begin
  Result:=not FData.IsNullY(I);
end; {End of TG2StoreDateTime.IsNotNull.}

function TG2StoreDateTime.IsNotNull(DateTime:TDateTime):boolean;
begin
  Result:=not IsNull(DateTime);
end; {End of TG2StoreDateTime.IsNotNull.}

function TG2StoreDateTime.GetNumData:cardinal;
begin
  Result:=FData.GetNumData;
end; {End of TG2StoreDateTime.GetNumData.}

procedure TG2StoreDateTime.WriteToTextFile(var Dest:textfile;IncNull:boolean=false);
var
  I:integer;
begin
  for I:=1 to FData.GetNumData do
    if IncNull then writeln(Dest,ConvertDateTimeToDecYear(FData.GetXValue(I)):10:5,' ',FData.GetYValue(I):10:5)
    else if IsNotNull(I) then writeln(Dest,ConvertDateTimeToDecYear(FData.GetXValue(I)):10:5,' ',FData.GetYValue(I):10:5);
end; {End of TG2StoreDateTime.WriteToTextFile.}

procedure TG2StoreDateTime.WriteToFile(var Dest:file);
begin

end;

procedure TG2StoreDateTime.LoadFromFile(var Source:file);
begin

end;

procedure TG2StoreDateTime.PutValue(DateTime:TDateTime;Value:double);
begin
  FData.AddCoord(DateTime,Value);
end; {End of TG2StoreDateTime.PutValue.}

function TG2StoreDateTime.GetValue(DateTime:TDateTime):double;
begin
  Result:=FData.InterpolateYValue(DateTime,false,false);
end; {End of TG2StoreDateTime.GetValue.}

procedure TG2StoreDateTime.PutIndexValue(I:cardinal;Value:double);
begin
  FData.PutYValue(I,Value);
end; {End of TG2StoreDateTime.PutIndexValue.}

function TG2StoreDateTime.GetIndexValue(I:cardinal):double;
begin
  Result:=FData.GetYValue(I);
end; {End of TG2StoreDateTime.GetIndexValue.}

function TG2StoreDateTime.GetIndexXValue(I:cardinal):double;
begin
  Result:=FData.GetXValue(I);
end; {End of TG2StoreDateTime.GetIndexXValue.}

function TG2StoreDateTime.GetIndexYValue(I:cardinal):double;
begin
  Result:=FData.GetYValue(I);
end; {End of TG2StoreDateTime.GetIndexYValue.}

function TG2StoreDateTime.GetDateTime(Index:cardinal):TDateTime;
begin
  Result:=FData.GetXValue(Index);
end; {End of TG2StoreDateTime.GetDateTime.}

{------------------------------------------------------------------------------}
{ Here are the method definitions for TG2Binary.                               }
{------------------------------------------------------------------------------}
constructor TG2Binary.Create;
begin
  raise EG2Binary.Create('the call to a TG2Binary constructor with no parameters is not allowed.');
end; {End of TG2Binary.Create.}

constructor TG2Binary.Create(Resolution:TResolution);
begin
  OwnsObjects:=true;
  FResolution:=Resolution;

  {This line must come here since the calls below, in particular the call to}
  {TG2StoreDateTime.Create uses this value.}
  FNullValue:=1E30;

  {The following line is only meaningful is FResolution is not ResRandom, but we}
  {still set a value so that it is not undefined. The default precision is double.}
  FPrecision:=TG2StoreDouble;

  {If the temporal resolution is random then we need only one TG2StoreDateTime object.}
  if FResolution=ResRandom then Add(TG2StoreDateTime.Create(self));

  FInterpolate:=true;
  FTimeStamp:=TSMiddle;
  FSortedByYear:=false;
  FForceFileExt:=true;
  FFilename:='';
  FFullLocationName:='';
  FLocationAbbreviation:='';
  FDataDescription:='';
  FDataUnits:='';
  FLatitude:=-999.99;
  FLongitude:=-999.99;
  FAltitude:=-999.99;

  {FFileVersion is a 4 byte array that details the version of the file. The first}
  {value is always 2 to denote that this is a version 2 generic binary location  }
  {data file. The next 3 values denote the sub-version.}
  FFileVersion[1]:=2;
  FFileVersion[2]:=0;
  FFileVersion[3]:=0;
  FFileVersion[4]:=1;

  FFileDesc:='Generic version 2 binary location data file';
end; {End of TG2Binary.Create.}

constructor TG2Binary.Create(Source:TG2Binary);
begin
  OwnsObjects:=true;
  if Assigned(Source) then DeepCopy(Source);
end; {End of TG2Binary.Create.}

procedure TG2Binary.DeepCopy(Source:TG2Binary);
var
  I:integer;
begin
  FResolution:=Source.FResolution;
  FPrecision:=Source.FPrecision;
  FFilename:=Source.FFilename;
  FFullLocationName:=Source.FFullLocationName;
  FLocationAbbreviation:=Source.FLocationAbbreviation;
  FDataDescription:=Source.FDataDescription;
  FDataUnits:=Source.FDataUnits;
  FNullValue:=Source.FNullValue;
  FLatitude:=Source.FLatitude;
  FLongitude:=Source.FLongitude;
  FAltitude:=Source.FAltitude;
  FSortedByYear:=Source.FSortedByYear;
  FInterpolate:=Source.FInterpolate;
  FTimeStamp:=Source.FTimeStamp;
  inherited Clear;
  for I:=0 to Source.Count-1 do Add(FPrecision.Create(Source.Items[I] as FPrecision));
end; {End of TG2Binary.DeepCopy.}

procedure TG2Binary.Clear;
begin
  if Resolution=ResRandom then TG2StoreDateTime(Items[0]).Clear
  else inherited Clear;
end;

function TG2Binary.GetYearIndex(Year:word):integer;
var
  I:integer;
begin
  Result:=-1;
  if Count=0 then Exit;
  if Resolution=ResRandom then Result:=0 else
  begin
    I:=-1;
    repeat
      inc(I);
      if (Items[I] as TG2GridStore).Year=Year then Result:=I;
    until ((I=Count-1) or (Result<>-1));
  end;
end; {End of TG2Binary.GetYearIndex.}

function TG2Binary.GetYearIndex(DateTime:TDateTime):integer;
begin
  if Resolution=ResRandom then Result:=0 else
    if Count=0 then Result:=-1 else Result:=GetYearIndex(YearOf(DateTime));
end; {End of  TG2Binary.GetYearIndex.}

function TG2Binary.AddYear(Year:word):integer;
var
  Hold:TG2GridStore;
begin
  if FResolution=ResRandom then raise EG2Binary.Create(
    'this method cannot be called for random temporal sampled data.');
  Hold:=FPrecision.Create(Year,self);
  Result:=Add(Hold);
  FSortedByYear:=false;
end; {End of TG2Binary.AddYear.}

procedure TG2Binary.PutValue(DateTime:TDateTime;Value:double);
var
  YearIndex:integer;
begin
  YearIndex:=GetYearIndex(DateTime);
  if YearIndex=-1 then YearIndex:=AddYear(YearOf(DateTime));
  (Items[YearIndex] as TG2Store).PutValue(DateTime,Value);
end; {End of TG2Binary.PutValue.}

procedure TG2Binary.PutValue(Year:word;I:cardinal;Value:double);
var
  YearIndex:integer;
begin
  if FResolution=ResRandom then raise EG2Binary.Create(
    'this method cannot be called for random temporal sampled data.');
  YearIndex:=GetYearIndex(Year);
  if YearIndex=-1 then YearIndex:=AddYear(Year);
  (Items[YearIndex] as TG2GridStore).Data[I]:=Value;
end; {End of TG2Binary.PutValue.}

function TG2Binary.GetValue(DateTime:TDateTime):double;
type
  DataRec=record
    Value:double;
    DateTime:TDateTime;
  end;
var
  YearIndex:integer;
  I:cardinal;
  CrossesYears:boolean;
  Year:word;
  Done:boolean;
  V:array[1..2] of DataRec;
begin
  Result:=FNullValue;

  {This is the start of an exception trapping block. Any exceptions raised in}
  {this function are dealt with at the end.}
  try
    YearIndex:=GetYearIndex(DateTime);

    {First deal with the simple case where we don't interpolate.}
    if not FInterpolate then
    begin
      if FResolution=ResRandom then Result:=(Items[0] as TG2StoreDateTime).GetValue(DateTime) else {This does actually interpolate.}
      begin
        if YearIndex=-1 then raise Exception.Create('');
        I:=(Items[YearIndex] as TG2GridStore).GetIndex(DateTime);
        if (Items[YearIndex] as TG2GridStore).IsNull(I) then raise Exception.Create('');
        Result:=(Items[YearIndex] as TG2GridStore).GetIndexValue(I);
      end;
      Exit;
    end;

    {Now deal with the case where we have to interpolate.}
    if ((DateTime<GetFirstDateTime) or (DateTime>GetLastDateTime)) then
      raise EG2Binary.Create('the DateTime passed to GetValue was out of range.');

    {We know now that the data are sorted by year because this call is made in GetFirstDateTime.}
    if YearIndex=-1 then
    begin
      {Prepare for the procedure below by moving to a year for which data are available.}
      Year:=YearOf(DateTime);
      repeat
        dec(Year);
        YearIndex:=GetYearIndex(Year);
      until YearIndex>-1; {This must eventually happen since we know the data span DateTime.}
      CrossesYears:=true;
    end
    else
    begin
      try
        {The call to GetValue below can raise an EG2Store exception if there}
        {are no data spanning DateTime. If Items[YearIndex] is a            }
        {TG2StoreDateTime object, rather than an EG2Store exception being   }
        {raised an EXYSignal exception is raised in which case the function }
        {exists with that exception.}
        Result:=(Items[YearIndex] as TG2Store).GetValue(DateTime);
        CrossesYears:=false;
      except
        on EG2Store do CrossesYears:=true;
      end;
    end;
    if CrossesYears then
    begin
      {Note that we should never come in here with the resolution being   }
      {ResRandom. If the resolution is ResRandom, either a valid result is}
      {obtained from the call above, or an EXYSignal exception is raised. }
      V[2].Value:=(Items[YearIndex] as TG2GridStore).GetFirstNonNullValue(V[2].DateTime);
      V[1].Value:=(Items[YearIndex] as TG2GridStore).GetLastNonNullValue(V[1].DateTime);
      if DateTime<V[2].DateTime then {Look for the value below and put in V[1].}
      begin
        Done:=false;
        Year:=YearOf(DateTime)-1;
        repeat
          YearIndex:=GetYearIndex(Year);
          if YearIndex=-1 then dec(Year) else
          try
            V[1].Value:=(Items[YearIndex] as TG2GridStore).GetLastNonNullValue(V[1].DateTime);
            Done:=true;
          except
            {This can happen if there are no non-null values in the year record.}
            on EG2Store do dec(Year);
          end;
        until Done;
      end
      else {Look for the value above and put in V[2].}
      begin
        Done:=false;
        Year:=YearOf(DateTime)+1;
        repeat
          YearIndex:=GetYearIndex(Year);
          if YearIndex=-1 then inc(Year) else
          begin
            try
              V[2].Value:=(Items[YearIndex] as TG2GridStore).GetFirstNonNullValue(V[2].DateTime);
              Done:=true;
            except
              {This can happen if there are no non-null values in the year record.}
              on EG2Store do inc(Year);
            end;
          end;
        until Done;
      end;
      Result:=(DateTime-V[1].DateTime)*(V[2].Value-V[1].Value)/(V[2].DateTime-V[1].DateTime)+V[1].Value;
    end;
  except
    on Exception do
    case FInterpolationFailPlan of
      raiseexception:raise EG2Binary.Create('it was not possible to retieve a value in GetValue.');
      returnnull:Exit;
    end;
  end;
end; {End of TG2Binary.GetValue.}

function TG2Binary.GetValue(Year:word;I:cardinal):double;
var
  YearIndex:integer;
begin
  Result:=FNullValue;
  if FResolution=ResRandom then raise EG2Binary.Create(
    'this method cannot be called for random temporal sampled data.');
  YearIndex:=GetYearIndex(Year);
  if YearIndex>-1 then Result:=(Items[YearIndex] as TG2GridStore).Data[I];
end; {End of TG2Binary.GetValue.}

function TG2Binary.GetXYSignalOfValues:TXYSignal;
var
  NumData,
  ObjCount,
  I:cardinal; {I must be of type cardinal to ensure that the correct version of IsNull is called.}
  Hold:TG2Store;
begin
  Result:=TXYSignal.Create(NullValue,NullValue);
  for ObjCount:=0 to Count-1 do
  begin
    Hold:=Items[ObjCount] as TG2Store;
    NumData:=Hold.GetNumData;
    for I:=1 to NumData do
      if not Hold.IsNull(I) then Result.AddCoord(Hold.GetDateTime(I),Hold.GetIndexValue(I));
  end;
end; {End of TG2Binary.GetXYSignalOfValues.}

function TG2Binary.GetMeanValue:double;
var
  Total:double;
  NumPerYear,
  NumData,
  ObjCount:integer;
  I:cardinal; {I must be of type cardinal to ensure that the correct version of IsNull is called.}
  Hold:TG2Store;
begin
  NumData:=0;
  Total:=0.0;
  for ObjCount:=0 to Count-1 do
  begin
    Hold:=Items[ObjCount] as TG2Store;
    NumPerYear:=Hold.GetNumData;
    for I:=1 to NumPerYear do
      if not Hold.IsNull(I) then
      begin
        inc(NumData);
        Total:=Total+Hold.GetIndexValue(I);
      end;
  end;
  if NumData=0 then raise EG2Binary.Create('there were no non-null data available to calculate the mean.');
  Result:=Total/NumData;
end; {End of TG2Binary.GetMeanValue.}

function TG2Binary.IsNull(DateTime:TDateTime):boolean;
var
  YearIndex:integer;
begin
  YearIndex:=GetYearIndex(DateTime);
  if YearIndex>-1 then Result:=(Items[YearIndex] as TG2Store).IsNull(DateTime)
  else Result:=true;
end; {End of TG2Binary.IsNull.}

function TG2Binary.IsNull(Year:word;I:cardinal):boolean;
var
  YearIndex:integer;
begin
  if FResolution=ResRandom then raise EG2Binary.Create(
    'this method cannot be called for random temporal sampled data.');
  YearIndex:=GetYearIndex(Year);
  if YearIndex>-1 then Result:=(Items[YearIndex] as TG2GridStore).IsNull(I) else Result:=true;
end; {End of TG2Binary.IsNull.}

function TG2Binary.IsNotNull(DateTime:TDateTime):boolean;
begin
  Result:=not IsNull(DateTime);
end; {End of TG2Binary.IsNotNull.}

function TG2Binary.IsNotNull(Year:word;I:cardinal):boolean;
begin
  Result:=not IsNull(Year,I);
end; {End of TG2Binary.IsNotNull.}

function TG2Binary.GetStartYear:word;
begin
  if FResolution=ResRandom then Result:=YearOf(TG2StoreDateTime(Items[0]).GetFirstDateTime) else
  begin
    SortByYear;
    Result:=TG2GridStore(Items[0]).Year
  end;
end; {End of TG2Binary.GetStartYear.}

function TG2Binary.GetEndYear:word;
begin
  if FResolution=ResRandom then Result:=YearOf(TG2StoreDateTime(Items[0]).GetLastDateTime) else
  begin
    SortByYear;
    Result:=TG2GridStore(Items[Count-1]).Year
  end;
end; {End of TG2Binary.GetEndYear.}

function TG2Binary.GetDateTimeList(OnlyNonNull:boolean=true):TDoubleArray;
var
  NumData,
  ObjCount:integer;
  I:cardinal; {I must be of type cardinal to ensure that the correct version of IsNull is called.}
  AddDateTime:boolean;
begin
  setlength(Result,0);
  for ObjCount:=0 to Count-1 do
  begin
    NumData:=(Items[ObjCount] as TG2Store).GetNumData;
    for I:=1 to NumData do
    begin
      AddDateTime:=true;
      if (OnlyNonNull and (Items[ObjCount] as TG2Store).IsNull(I)) then AddDateTime:=false;
      if AddDateTime then
      begin
        setlength(Result,length(Result)+1);
        Result[High(Result)]:=(Items[ObjCount] as TG2Store).GetDateTime(I);
      end;
    end;
  end;
end; {End of TG2Binary.GetDateTimeList.}

function TG2Binary.GetFirstDateTime(OnlyNonNull:boolean=true):TDateTime;
var
  ObjCount:integer;
  NumData,
  I:cardinal;
  Done:boolean;
begin
  Result:=-1E30;
  if Count=0 then exit;
  if FResolution=ResRandom then TG2StoreDateTime(Items[0]).SortByTime else SortByYear;
  if OnlyNonNull then
  begin
    ObjCount:=0;
    NumData:=(Items[ObjCount] as TG2Store).GetNumData;
    I:=1;
    Done:=false;
    repeat
      if (Items[ObjCount] as TG2Store).IsNotNull(I) then
      begin
        Result:=(Items[ObjCount] as TG2Store).GetDateTime(I);
        Done:=true;
      end;
      inc(I);
      if I>NumData then
      begin
        I:=1;
        inc(ObjCount);
        if ObjCount=Count then Done:=true;
      end;
    until Done;
  end
  else Result:=(Items[0] as TG2Store).GetIndexValue(1);
end;

function TG2Binary.GetLastDateTime(OnlyNonNull:boolean=true):TDateTime;
var
  NumData,
  ObjCount:integer;
  I:cardinal;
  Done:boolean;
begin
  Result:=-1E30;
  if Count=0 then exit;
  if FResolution=ResRandom then TG2StoreDateTime(Items[0]).SortByTime else SortByYear;
  NumData:=(Items[0] as TG2Store).GetNumData;
  if OnlyNonNull then
  begin
    ObjCount:=Count-1;
    I:=NumData;
    Done:=false;
    repeat
      if (Items[ObjCount] as TG2Store).IsNotNull(I) then
      begin
        Result:=(Items[ObjCount] as TG2Store).GetDateTime(I);
        Done:=true;
      end;
      dec(I);
      if I<1 then
      begin
        I:=NumData;
        dec(ObjCount);
        if ObjCount=-1 then Done:=true;
      end;
    until Done;
  end
  else Result:=(Items[Count-1] as TG2Store).GetIndexValue(NumData);
end;

procedure TG2Binary.FillWithValue(AValue:double;OnlyNonNull:boolean=true);
var
  ObjCount,
  NumData:integer;
  I:cardinal; {I must be of type cardinal to ensure that the correct version of IsNull is called.}
begin
  for ObjCount:=0 to Count-1 do
  begin
    NumData:=(Items[ObjCount] as TG2Store).GetNumData;
    for I:=1 to NumData do
      if not (OnlyNonNull and (Items[ObjCount] as TG2Store).IsNull(I)) then
        (Items[ObjCount] as TG2Store).PutIndexValue(I,AValue);
  end;
end;

procedure TG2Binary.ScaleBy(AValue:double;OnlyNonNull:boolean=true);
var
  ObjCount:integer;
  NumData,
  I:cardinal; {I must be of type cardinal to ensure that the correct version of IsNull is called.}
  Value:double;
begin
  for ObjCount:=0 to Count-1 do
  begin
    NumData:=(Items[ObjCount] as TG2Store).GetNumData;
    for I:=1 to NumData do
      if not (OnlyNonNull and (Items[ObjCount] as TG2Store).IsNull(I)) then
      begin
        Value:=(Items[ObjCount] as TG2Store).GetIndexValue(I)*AValue;
        (Items[ObjCount] as TG2Store).PutIndexValue(I,Value);
      end;
  end;
end;

procedure TG2Binary.OffsetBy(AValue:double;OnlyNonNull:boolean=true);
var
  ObjCount:integer;
  NumData,
  I:cardinal; {I must be of type cardinal to ensure that the correct version of IsNull is called.}
  Value:double;
begin
  for ObjCount:=0 to Count-1 do
  begin
    NumData:=(Items[ObjCount] as TG2Store).GetNumData;
    for I:=1 to NumData do
      if not (OnlyNonNull and (Items[ObjCount] as TG2Store).IsNull(cardinal(I))) then
      begin
        Value:=(Items[ObjCount] as TG2Store).GetIndexValue(I)-AValue;
        (Items[ObjCount] as TG2Store).PutIndexValue(I,Value);
      end;
  end;
end;

procedure TG2Binary.SwitchPrecision;
var
  I:integer;
  Hold:TG2GridStore;
begin
  if Resolution=ResRandom then raise EG2Store.Create(
    'precision switching is not allowed with random temporal resolution data.');
  for I:=0 to Count-1 do
  if not (Items[I] is FPrecision) then
  begin
    Hold:=Items[I] as TG2GridStore;
    Insert(I,FPrecision.Create(Hold));
    Remove(Hold);
  end;
end; {End of TG2Binary.SwitchPrecision.}

procedure TG2Binary.SwitchToSinglePrecision;
begin
  FPrecision:=TG2StoreSingle;
  SwitchPrecision;
end; {End of TG2Binary.SwitchToSinglePrecision.}

procedure TG2Binary.SwitchToPackedPrecision;
begin
  FPrecision:=TG2StorePacked;
  SwitchPrecision;
end; {End of TG2Binary.SwitchToPackedPrecision.}

procedure TG2Binary.SwitchToDoublePrecision;
begin
  FPrecision:=TG2StoreDouble;
  SwitchPrecision;
end; {End of TG2Binary.SwitchToDoublePrecision.}

function CompareYear(Item1,Item2:pointer):integer;
begin
  Result:=0;
  if TG2GridStore(Item1).Year<TG2GridStore(Item2).Year then Result:=-1;
  if TG2GridStore(Item1).Year>TG2GridStore(Item2).Year then Result:=1;
end;

procedure TG2Binary.SortByYear;
begin
  if Resolution=ResRandom then raise EG2Store.Create(
    'this procedure should not be called for randomly temporally spaced data.');
  if FSortedByYear then exit;
  Sort(CompareYear);
  FSortedByYear:=true;
end;

procedure TG2Binary.Filter(FilterArray:TDoubleArray);
var
  YearBelow,
  YearAbove,
  Year:word;
  StartYear,
  EndYear,
  NumInYear,
  HFW,          {The half width of the filter.}
  CurrentWidth,
  IBelow,
  IAbove,
  I:cardinal;
  Store:TDoubleArray;
  TotalWeight,
  TotalData:double;
  Done:boolean;
begin
  if Resolution=ResRandom then raise EG2Binary.Create('this filter method cannot be called for random '+
    'temporal resolution data. Suggest calling GaussianFilter.');
  if (length(FilterArray) mod 2)=0 then raise EG2Binary.Create(
    'the filter array must have an odd number of elements');
  HFW:=length(FilterArray) div 2;
  StartYear:=GetStartYear;
  EndYear:=GetEndYear;
  NumInYear:=NumPerYear(Resolution);
  setlength(Store,(EndYear-StartYear+1)*NumInYear);
  for Year:=StartYear to EndYear do
    for I:=1 to NumInYear do
    if IsNotNull(Year,I) then
    begin
      CurrentWidth:=0;
      TotalWeight:=FilterArray[HFW];
      TotalData:=GetValue(Year,I)*TotalWeight;
      Done:=false;
      repeat
        inc(CurrentWidth);
        YearBelow:=Year;
        IBelow:=I-CurrentWidth;
        while IBelow<1 do
        begin
          dec(YearBelow);
          IBelow:=IBelow+NumInYear;
        end;
        YearAbove:=Year;
        IAbove:=I+CurrentWidth;
        while IAbove>NumInYear do
        begin
          inc(YearAbove);
          IAbove:=IAbove-NumInYear;
        end;
        if (IsNotNull(YearBelow,IBelow) and IsNotNull(YearAbove,IAbove)) then
        begin
          TotalWeight:=TotalWeight+FilterArray[HFW-CurrentWidth]+FilterArray[HFW+CurrentWidth];
          TotalData:=TotalData+(GetValue(YearBelow,IBelow)*FilterArray[HFW-CurrentWidth])+
                               (GetValue(YearAbove,IAbove)*FilterArray[HFW+CurrentWidth]);
        end
        else Done:=true;
        if CurrentWidth=HFW then Done:=true;
      until Done;
      Store[(Year-StartYear)*NumInYear+I-1]:=TotalData/TotalWeight;
    end
    else Store[(Year-StartYear)*NumInYear+I-1]:=NullValue;

  {Now copy the data in the store back into the object.}
  for Year:=StartYear to EndYear do
    for I:=1 to NumInYear do
      if Store[(Year-StartYear)*NumInYear+I-1]<>NullValue then PutValue(Year,I,Store[(Year-StartYear)*NumInYear+I-1]);
end; {End of TG2Binary.Filter.}


procedure TG2Binary.SaveToFile(Filename:string);
begin
  FFilename:=Filename;
  SaveToFile;
end; {End of TG2Binary.SaveToFile.}

procedure TG2Binary.SaveToFile;
var
  Dest:file;
  Direc:string;
  I:integer;
begin
  if FFilename='' then raise EG2Binary.Create('the file name has not yet been set.');
  Direc:=ExtractFilePath(FFilename);
  ConfirmDirectory(Direc,true,true);
  if FForceFileExt then FFilename:=ChangeFileExt(FFilename,'.g2bin');
  assignfile(Dest,FFilename);
  rewrite(Dest,1);

  {First we write the general file description data.}
  blockwrite(Dest,FFileVersion,sizeof(TVersionArray));
  blockwrite(Dest,FFileDesc,sizeof(st43));

  {Then we write the next two most important pieces of information i.e.}
  {the precision with which the data are stored and the data resolution.}
  blockwrite(Dest,FPrecision,sizeof(TG2GridStoreClass));
  blockwrite(Dest,FResolution,sizeof(TResolution));

  {Now write the information describing the location.}
  blockwrite(Dest,FFullLocationName,sizeof(st40));
  blockwrite(Dest,FLocationAbbreviation,sizeof(st3));
  blockwrite(Dest,FDataDescription,sizeof(st200));
  blockwrite(Dest,FDataUnits,sizeof(st40));
  blockwrite(Dest,FLatitude,sizeof(double));
  blockwrite(Dest,FLongitude,sizeof(double));
  blockwrite(Dest,FNullValue,sizeof(double));
  blockwrite(Dest,FAltitude,sizeof(double));

  {Finally write the data objects to file.}

  {I think that special consideration needs to be given here to ResRandom.}

  for I:=0 to Count-1 do (Items[I] as FPrecision).WriteToFile(Dest);

  closefile(Dest);
end;

procedure TG2Binary.WriteToTextFile(Filename:string;IncNull:boolean=false);
var
  I:integer;
  Dest:textfile;
begin
  SortByYear;
  assignfile(Dest,Filename);
  rewrite(Dest);
  for I:=0 to Count-1 do
    (Items[I] as FPrecision).WriteToTextFile(Dest,IncNull);
  closefile(Dest);
end; {End of TG2Binary.WriteToTextFile.}

procedure TG2Binary.LoadFromFile(Filename:string);
begin
  FFilename:=Filename;
  LoadFromFile;
end; {End of TG2Binary.LoadFromFile.}

procedure TG2Binary.LoadFromFile;
type
  TFileDataRec=record
    Year:word;
    Source:st40;
  end;
var
  Source:file;
  FileType:byte;
  DataDesc:st80; {The 80 character data description used in version 1 data files.}
  Units:st30; {The 30 character descriptor of the data units used in version 1 data files.}
  FileResolution:TResolution; {The resolution read in from the file.}

  TestFileDesc:st43;
  NullData:double;
  Altitude,
  NumPerDay:word;
  FileData:TFileDataRec;
  StorePacked:TG2StorePacked;
  Hold:TG2Store;
begin
  {First find out what type of file we are dealing with.}
  if not FileExists(FFilename) then raise EG2Binary.Create('the FFilename field has not been set.');
  assignfile(Source,FFilename);
  reset(Source,1);
  blockread(Source,FileType,sizeof(byte));

  {Confirm that the file is telling the truth about the file type.}
  case FileType of
    1:try
        blockread(Source,FFullLocationName,sizeof(st40));
        blockread(Source,FLocationAbbreviation,sizeof(st3));
        blockread(Source,DataDesc,sizeof(st80));
        FDataDescription:=DataDesc;
        blockread(Source,Units,sizeof(st30));
        FDataUnits:=Units;
        blockread(Source,FLatitude,sizeof(double));
        blockread(Source,FLongitude,sizeof(double));

        blockread(Source,NullData,sizeof(double));


        blockread(Source,Altitude,sizeof(word));
        blockread(Source,NumPerDay,sizeof(word));
        if ((Latitude<-90.0) or (Latitude>90.0)) then FileType:=0;
        if ((Longitude<-360.0) or (Longitude>360.0)) then FileType:=0;
      except
        on Exception do FileType:=0;
      end;
    2:begin
        reset(Source,1);
        blockread(Source,FFileVersion,sizeof(TVersionArray));
        blockread(Source,TestFileDesc,sizeof(st43));
        if TestFileDesc<>FFileDesc then FileType:=0;
      end;
    else
      FileType:=0;
  end;
  if FileType=0 then raise EG2Binary.Create('unidentified file type.');

  case FileType of
    1:begin
        FPrecision:=TG2StorePacked;
        while not eof(Source) do
        begin
          blockread(Source,FileData,sizeof(FileData));
          StorePacked:=TG2StorePacked.Create(FileData.Year,self);

          {The next line tells the StorePacked object where to find the data in}
          {the file and those data are only loaded when they are actually needed.}
          StorePacked.FilePosition:=filepos(Source);

          Add(StorePacked);
          seek(Source,FilePos(Source)+2*sizeof(double)+366*NumPerDay*sizeof(word));
        end;
      end;
    2:begin
        blockread(Source,FPrecision,sizeof(TG2GridStoreClass));
        blockread(Source,FileResolution,sizeof(TResolution));
        if FileResolution<>FResolution then raise EG2Binary.Create(
          'the resolution read in from the file did not match the resolution specified for the object.');

        blockread(Source,FFullLocationName,sizeof(st40));
        blockread(Source,FLocationAbbreviation,sizeof(st3));
        blockread(Source,FDataDescription,sizeof(st200));
        blockread(Source,FDataUnits,sizeof(st40));
        blockread(Source,FLatitude,sizeof(double));
        blockread(Source,FLongitude,sizeof(double));
        blockread(Source,FNullValue,sizeof(double));
        blockread(Source,FAltitude,sizeof(double));

        if FResolution=ResRandom then
        begin
//          Add(TG2StoreDateTime.Create(self));
//          (Items[0] as TG2StoreDateTime).




        end
        else
        begin
          while not eof(Source) do
          begin




            Hold:=FPrecision.Create(self);


            Hold.LoadFromFile(Source);

            Add(Hold);
          end;
        end;
      end;
  end;
  closefile(Source);
end; {End of TG2Binary.LoadFromFile.}

procedure TG2Binary.SetNullValue(Value:double);
{------------------------------------------------------------------------------}
{ This procedure sets or resets the value that denotes null data. The new      }
{ value is passed onto all of the objects owned by the object list.
{------------------------------------------------------------------------------}
var
  I:integer;
begin
  {The ordering of the next two lines is important since TG2Store uses the current}
  {setting of FNullValue to determine whether the values it is holding are null.  }
  for I:=0 to Count-1 do (Items[I] as TG2Store).SetNullValue(Value);
  FNullValue:=Value;
end; {End of TG2Binary.SetNullValue.}

procedure TG2Binary.SetFilename(Value:string);
var
  I:integer;
begin
  {The main thing here is that we have to be careful if we are changing file names}
  {and the data storage objects are pointing to specific locations in a different }
  {file, then all hell could break loose. So we load all of the data.}
  if Value<>FFilename then
  for I:=0 to Count-1 do
    if (Items[I] as TG2Store).FilePosition<>0 then
    begin

    end;
  FFilename:=Value;
end; {End of TG2Binary.SetFilename.}

procedure TG2Binary.SetTimeStamp(Value:TG2GridTimeStamp);
begin
  if FResolution=ResRandom then raise EG2Binary.Create(
    'this method cannot be called for random temporal sampled data.');
  FTimeStamp:=Value;
end; {End of TG2Binary.SetTimeStamp.}

end.
