//@***h* BDBP/gb_grib.pas
// 
// NAME
//   gb_grib.pas
//
// PURPOSE
//  Defines an object which manages grib files.
//
// DESCRIPTION
//  This unit defines an object which manages grib files.
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_grib;

{$I UNITDIREC.INC}

interface

uses
  sysutils;

type
  TGribTimeStamp=record
    Year,
    Day,
    Hour,
    Level:word;
  end;
  PGribTimeStamp=^TGribTimeStamp;
  TTimeStampArray=array[1..8760] of PGribTimeStamp;

  TGrib=class
  private
    WGribPath:string;
    NumRecords:word;
    TimeStamp:TTimeStampArray;
  public
    GribFilename:string;
    constructor Create(AFilename,WGribExePath:string;NoGrib:boolean);
    procedure MakeTextFile(AFilename:string;AYear,ADay,AHour,ALevel:word;NoGrib:boolean);
    procedure CheckWGrib;
    destructor Destroy;override;
  end;

  EGrib=class(exception);

implementation

uses
  clipbrd,
  controls,
  dialogs,
  forms,
  windows,
  gb_datetime,
  gb_files,
  gb_strings;

constructor TGrib.Create(AFilename,WGribExePath:string;NoGrib:boolean);
{------------------------------------------------------------------------------}
{ This constructor for TGrib will fail if the grib file named 'AFilename'      }
{ doesn't exist. It is also necessary that the wgrib.exe program exists in the }
{ current directory since the constructor needs to call this to generate an    }
{ inventory file of what data (in terms of dates and times) is available in    }
{ the grib file. If wgrib.exe doesn't exist in the current directory a message }
{ is shown and the user is either given the opportunity to copy the wgrib.exe  }
{ file into the current directory or to exit the program. The routine then     }
{ makes a call to wgrib to generate an inventory file listing the dates and    }
{ times of the data held in the grib file to a temporary text file. The        }
{ routine reads in the dates and times from the inventory file, storing the    }
{ results in the TimeStamp array.                                              }
{------------------------------------------------------------------------------}
var
  Month,  {The month read in from the inventory file.}
  Date,   {The date read in from the inventory file.}
  I:word; {Loop control variable through records.}
  CurrentDirec,  {The current directory.}
  LevelStr,      {The string containing the level value.}
  TempFilename,  {The name of the temporary inventory file.}
  ExecStr,       {The executable string passed to wgrib.exe.}
  Linein:string; {The line of text read in from the inventory file.}
  Source:textfile; {The file handle for the inventory file.}
  hAppProcess,         {The application process handle.}
  hAppThread,          {The application thread handle.}
  hOutputFile:THandle; {The output file handle.}
  StartupInfo:TStartupInfo; {Used in the call to wgrib.exe.}
  SecAttrs:TSecurityAttributes; {The security attributes record for the output file handle creation.}
  ProcessInfo:TProcessInformation; {Used in the call to wgrib.exe.}
begin
  if not FileExists(AFilename) then raise Exception.Create('Something is wrong here');
  getdir(0,CurrentDirec);
  GribFilename:=AFilename;
  WGribPath:=WGribexePath;
  TempFilename:=FindJunkFile;
  ExecStr:='wgrib.exe '+GribFilename+' -s -4yr';
  if WGribPath<>'' then ExecStr:=WGribPath+'\'+ExecStr;
  I:=0;
  repeat
    inc(I);
    if NoGrib then
    begin
      ClipBoard.AsText:=ExecStr+'>'+TempFilename;
      ShowMessage('The NoGrib option was selected. Please execute the following command '+
        'in the '+CurrentDirec+' directory, '+#13#10+ExecStr+'>'+TempFilename+#13#10+
        ' (which has been copied to the clipboard), then press OK.');
    end
    else
    begin
      CheckWGrib;
      hAppProcess:=0;
      hAppThread:=0;
      hOutputFile:=0;
      try
        fillchar(SecAttrs,sizeof(SecAttrs),#0);
        SecAttrs.nLength:=sizeof(SecAttrs);
        SecAttrs.lpSecurityDescriptor:=nil;
        SecAttrs.bInheritHandle:=true;
        hOutputFile:=CreateFile(PChar(TempFilename),generic_read or generic_write,file_share_read or
          file_share_write,@SecAttrs,create_always,file_attribute_normal or file_flag_write_through,0);
        if hOutputFile=invalid_handle_value then raise EGrib.Create(
          'the output file handle couldn''t be created successfully');
        fillchar(StartupInfo,sizeof(StartupInfo),#0);
        StartupInfo.cb:=sizeof(StartupInfo);
        StartupInfo.dwFlags:=STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
        StartupInfo.wShowWindow:=SW_HIDE;
        StartupInfo.hStdOutput:=hOutputFile;
        StartupInfo.hStdInput:=0;
        if CreateProcess(nil,PChar(ExecStr),nil,nil,true,HIGH_PRIORITY_CLASS,
          nil,nil,StartupInfo,ProcessInfo) then
        begin
          WaitForSingleObject(ProcessInfo.hProcess,INFINITE);
          hAppProcess:=ProcessInfo.hProcess;
          hAppThread:=ProcessInfo.hThread;
        end
        else raise EGrib.Create('the CreateProcess call failed');
      finally
        if hOutputFile<>0 then CloseHandle(hOutputFile);
        if hAppProcess<>0 then CloseHandle(ProcessInfo.hProcess);
        if hAppThread<>0 then CloseHandle(ProcessInfo.hThread);
      end;
      if not FileExists(TempFilename) then WaitNSeconds(I*2);
      Application.ProcessMessages;
    end;
  until (FileExists(TempFilename) or (I=5));
  if not FileExists(TempFilename) then raise EGrib.Create('after 5 tries to execute '+
    ExecStr+' there was still no file named '+TempFilename);
  NumRecords:=CountLines(TempFilename);
  assignfile(Source,TempFilename);
  reset(Source);
  for I:=1 to NumRecords do
  begin
    getmem(TimeStamp[I],sizeof(TGribTimeStamp));
    readln(Source,Linein);
    LevelStr:=GetWord(Linein,':',5);
    LevelStr:=GetWord(LevelStr,' ',1);
    LevelStr:=EraseEndChar(LevelStr,'K');
    Linein:=GetWord(Linein,':',3);
    if length(Linein)=10 then
    begin
      TimeStamp[I]^.Year:=strtoint(copy(Linein,3,2));
      if TimeStamp[I]^.Year<30 then TimeStamp[I]^.Year:=
        TimeStamp[I]^.Year+2000 else TimeStamp[I]^.Year:=
        TimeStamp[I]^.Year+1900;
      Linein:=copy(Linein,5,6);
    end
    else
    begin
      TimeStamp[I]^.Year:=strtoint(copy(Linein,3,4));
      Linein:=copy(Linein,7,6);
    end;
    if TimeStamp[I]^.Year<1900 then TimeStamp[I]^.Year:=TimeStamp[I]^.Year+1000;
    Month:=strtoint(copy(Linein,1,2));
    Date:=strtoint(copy(Linein,3,2));
    TimeStamp[I]^.Day:=CalcDay(TimeStamp[I]^.Year,Month,Date);
    TimeStamp[I]^.Hour:=strtoint(copy(Linein,5,2));
    try
      TimeStamp[I]^.Level:=strtoint(LevelStr);
    except
      on EConvertError do TimeStamp[I]^.Level:=65535;
    end;
  end;
  closefile(Source);
  erase(Source);
end; {End of TGrib.Create.}

procedure TGrib.MakeTextFile(AFilename:string;AYear,ADay,AHour,ALevel:word;NoGrib:boolean);
{------------------------------------------------------------------------------}
{ This routine makes up a text file (named 'AFilename') of data extracted from }
{ the grib file for the date and time passed via 'AYear', 'ADay' and 'AHour'.  }
{ If no record is available for the specified date and time, an exception is   }
{ raised.                                                                      }
{------------------------------------------------------------------------------}
var
  RecNum, {The record number for the date and hour requested.}
  I:word; {Loop control variable through records.}
  CurrentDirec,   {The current directory.}
  FinalDirectory, {The directory for the output text file.}
  TempFilename,   {Temporary output file wgrib which is later renamed to AFilename.}
  ExecStr:string; {The executable string passed to wgrib.exe.}
  StartupInfo:TStartupInfo; {Used in the call to wgrib.exe.}
  ProcessInfo:TProcessInformation; {Used in the call to wgrib.exe.}
begin
  I:=0;
  repeat
    inc(I);
    RecNum:=I;
    if ((AYear<>65535) and (TimeStamp[I]^.Year<>AYear)) then RecNum:=0;
    if ((ADay<>65535) and (TimeStamp[I]^.Day<>ADay)) then RecNum:=0;
    if ((AHour<>65535) and (TimeStamp[I]^.Hour<>AHour)) then RecNum:=0;
    if ((ALevel<>65535) and (TimeStamp[I]^.Level<>ALevel)) then RecNum:=0;
  until ((I=NumRecords) or (RecNum<>0));
  if RecNum=0 then raise EGrib.Create('the source grib file ('+GribFilename+')'+
    ' didn''t contain data for hour '+inttostr(AHour)+' of day '+inttostr(ADay)+
    ' of '+inttostr(AYear));
  CheckWGrib;
  I:=0;
  repeat
    inc(I);
    if NoGrib then
    begin
      getdir(0,CurrentDirec);
      ExecStr:='wgrib.exe '+GribFilename+' -d '+inttostr(RecNum)+' -text -o '+AFilename;
      if WGribPath<>'' then ExecStr:=WGribPath+'\'+ExecStr;
      ClipBoard.AsText:=ExecStr;
      ShowMessage('The NoGrib option was selected. Please execute the following command '+
        'in the '+CurrentDirec+' directory, '+#13#10+ExecStr+#13#10+' (which has been '+
        'copied to the clipboard), then press OK.');
    end
    else
    begin
      {Now some (but not all) versions of wgrib.exe can only deal with file names in}
      {the 8.3 format. To make provision for this we use the GetShortFilename       }
      {function defined in gb_files.pas to get the 8.3 format of the source file    }
      {name. We can't do this for the destination file since it doesn't exist yet.  }
      {So we find a temporary file name by calling FindJunkFile defined in          }
      {gb_files.pas, set this as the destination file for the wgrib call and then   }
      {rename the temporary file to the intended output file for wgrib.}
      TempFilename:=FindJunkFile;
      ExecStr:='wgrib.exe '+GetShortFilename(GribFilename)+' -d '+inttostr(RecNum)+
        ' -text -o '+TempFilename;
      if WGribPath<>'' then ExecStr:=WGribPath+'\'+ExecStr;
      fillchar(StartupInfo,sizeof(StartupInfo),#0);
      StartupInfo.cb:=sizeof(StartupInfo);
      StartupInfo.dwFlags:=STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
      StartupInfo.wShowWindow:=SW_HIDE;
      StartupInfo.hStdOutput:=0;
      StartupInfo.hStdInput:=0;
      if not CreateProcess(nil,PChar(ExecStr),nil,nil,false,HIGH_PRIORITY_CLASS,
        nil,nil,StartupInfo,ProcessInfo) then if MessageDlg('The CreateProcess '+
        'call failed',mtError,[mbOK,mbCancel],0)=mrcancel then Application.Terminate;
      WaitForSingleObject(ProcessInfo.hProcess,INFINITE);
      CloseHandle(ProcessInfo.hProcess);
      CloseHandle(ProcessInfo.hThread);
      Application.ProcessMessages;
      if not FileExists(TempFilename) then WaitNMinutes(1); {Wait a minute if the file's not ready.}
      if FileExists(TempFilename) then
      begin
        FinalDirectory:=ExtractFileDir(AFilename);
        if not DirectoryExists(FinalDirectory) then ForceDirectories(FinalDirectory);
        {The text file may have been created in the meantime by another program.}
        if FileExists(AFilename) then SysUtils.DeleteFile(AFilename);
        SysUtils.RenameFile(TempFilename,AFilename);
      end;
    end;
  until (FileExists(AFilename) or (I=5));
  if not FileExists(AFilename) then raise EGrib.Create('the call to '+ExecStr+' failed 5 times');
end; {End of TGrib.MakeTextFile.}

procedure TGrib.CheckWGrib;
{------------------------------------------------------------------------------}
{ Ensures that the wgrib.exe file is available and if it isn't it gives the    }
{ user the option of copying the file to the current directory or exiting the  }
{ program.                                                                     }
{------------------------------------------------------------------------------}
var
  CurrentDirec:string; {The current directory.}
begin
  if WGribPath='' then
  begin
    while not FileExists('wgrib.exe') do
    begin
      getdir(0,CurrentDirec);
      if MessageDlg('There is no wgrib.exe program in the current directory. Please copy this '+
        'program into '+CurrentDirec+' and press OK. Press Cancel to quit the program.',
        mtError,[mbOK,mbCancel],0)=mrcancel then halt;
    end;
  end
  else
    while not FileExists(WGribPath+'\wgrib.exe') do
      if MessageDlg('There is no wgrib.exe program in the '+WGribPath+' directory. Please '+
        'copy this program into this directory and press OK. Press Cancel to quit the program.',
        mtError,[mbOK,mbCancel],0)=mrcancel then halt;
end; {End of TGrib.CheckWGrib.}

destructor TGrib.Destroy;
var
  I:word;
begin
  for I:=1 to NumRecords do freemem(TimeStamp[I],sizeof(TGribTimeStamp));
  inherited Destroy;
end; {End of TGrib.Destroy.}

end.
