//@***h* BDBP/gb_math.pas
// 
// NAME
//   gb_math.pas
//
// PURPOSE
//  Holds all mathematical routines and functions which are not part
//  of the standard Delphi maths routines. 
//
// DESCRIPTION
//  This unit holds all mathematical routines and functions which are not part
//  of the standard Delphi maths routines. The unit makes no use of  
//  other units and is given a priority of zero.     
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_math;

{$I UNITDIREC.INC}

interface

uses
  sysutils,
  gb_array;

const
  MaxMathArray=1048575; {2^23/8-1}
  MaxRealArray=1398100; {2^23/6-1}
  MaxIntArray=4194303; {2^23/2-1}
  MaxWordArray=4194303; {2^23/2-1}
  MaxComplexArray=524287; {2^23/16-1}
  MaxLongArray=2097151; {2^23/4-1}
  VerySmall=1E-300;
  QuiteSmall=1E-20;
  SomewhatSmall=1E-6;

type
  EGBMath=class(exception);

  TMathArray=array[1..MaxMathArray] of double;
  PMathArray=^TMathArray;
  TRealArray=array[1..MaxRealArray] of real;
  PRealArray=^TRealArray;

//  The lines below have been commented out because the definition of TIntegerArray
//  Clashes with the definition of TIntegerArray in gb_array. Since it is not used
//  anywhere in this unit, commenting it out shouldn't be a problem.
//  TIntegerArray=array[1..MaxIntArray] of smallint;
//  PIntegerArray=^TIntegerArray;

  TLongArray=array[1..MaxLongArray] of longint;
  PLongArray=^TLongArray;
  TWordArray=array[1..MaxWordArray] of word;
  PWordArray=^TWordArray;
  T2DArray=array[1..MaxMathArray] of PMathArray;
  P2DArray=^T2DArray;

  TMatrix33=array[1..3,1..3] of double;
  PMatrix33=^TMatrix33;

  Complex=record
    Real,Imag:double;
  end;
  PComplex=^Complex;
  TComplexArray=array[1..MaxComplexArray] of Complex;
  PComplexArray=^TComplexArray;

  TMultiReg=class {The multiple regression object.}
    CoVar,           {The covariance matrix.}
    Beta:P2DArray;   {The second normal equation matrix.}
    NumCoeff:word;   {The number of model coefficients.}
    NumData:longint; {The number of data used in the fit.}
    BasisFunc,    {The array of basis functions at X_i.}
    Sigma,        {The array holding the uncertainty in each measurement.}
    A,            {The model coefficients.}
    X,            {The X data being used in the fit.}
    Y:PMathArray; {The Y data being used in the fit.}
    constructor Create(NumCo:word;NumXY:longint);
    procedure SetSwitches(N:longint);virtual;
    function CalcFunc(X:double):byte;virtual;
    function Run:double;
    destructor Destroy;override;
    property GetNumXY:longint read NumData;
  end;

  TPolyReg=class(TMultiReg)
    function CalcFunc(X:double):byte;override;
    function CalcModelValue(X:double):double;
  end;

  TFourierReg=class(TMultiReg)
    function CalcFunc(X:double):byte;override;
    function CalcModelValue(X:double):double;
  end;

  TIntervalHalver=class
  private
    NumSteps,
    NumCoeff:word;
    StepCount,
    TotalSteps:cardinal;
    StepRange:integer;
    CoeffValues,
    CoeffSens:TDoubleArray;
    Indices:array of integer;
    FUpperOnly,
    FDone:boolean;
    function GetCoeff(I:word):double;
    procedure WriteCoeff(I:word;Value:double);
    function GetCoeffSens(I:word):double;
    procedure WriteCoeffSens(I:word;Value:double);
    function GetPerturbedCoeffs:TDoubleArray;
  public
    constructor Create(ANumCoeff,ANumSteps:word;UpperSideOnly:boolean=true);
    procedure Reset;
    function Step:double;

    property Coeff[I:word]:double read GetCoeff write WriteCoeff;
    property Sens[I:word]:double read GetCoeffSens write WriteCoeffSens;
    property Done:boolean read FDone;
    property Coeffs:TDoubleArray read GetPerturbedCoeffs;
  end;


{Here is an alphabetical list of the other routines available in this unit.}
function  AbsComplex(A:complex):double;
procedure AddComplex(A,B:complex;var RetVal:complex);
function  CalcDet33(Matrix:PMatrix33):double;
function  CalcMean(X:PMathArray;N:cardinal):double;
function  CalcMedian(X:PMathArray;N:cardinal):double;overload;
function  CalcMedian(X:TDoubleArray):double;overload;
function  CalcMin(X:PMathArray;N:cardinal):double;
function  CalcMax(X:PMathArray;N:cardinal):double;
function  CalcStdv(X:PMathArray;N:cardinal):double;
procedure ComplexConjugate(A:Complex;var RetVal:complex);
function  DFT(DFTArray:PMathArray;N:integer):byte;
procedure DivideComplex(A,B:complex;var RetVal:complex);
function  EquivDouble(A,B:double):boolean;
function  EquivSingle(A,B:single):boolean;
function  EquivReal(A,B:real):boolean;
procedure FFT(FFTArray:PMathArray;N:integer);
procedure FreePMathArray(var X:PMathArray;Mem:integer);
function  GaussJordan(A,B:P2DArray;N,M:word):byte;overload;
procedure GaussJordan(Covar:T2DDoubleArray;Beta:T1DDoubleArray);overload;
procedure GaussJordan(Covar:T2DimDoubleArray;Beta:TDoubleArray);overload;
function  GetMoreMemory(var X:PMathArray;N,M:word):boolean;
procedure HeapSort(X:PMathArray;N:word);overload;
procedure HeapSort(var X:TDoubleArray);overload;
procedure HeapSort2Array(X,Y:PMathArray;N:word);
function  HexToDec(ValStr:string):longint;
function  Interpolate(X,Y:PMathArray;var N:word;L1,L2,Res:double):byte;
function  LUBackSubst(A:P2DArray;N:word;Index:PIntegerArray;B:PMathArray):byte;
function  LUDecomposition(A:P2DArray;N:word;Index:PIntegerArray):integer;
procedure MultiplyComplex(A,B:complex;var RetVal:complex);
function  Regression(X,Y:PMathArray;N:integer;var M,C:double;ZeroOff:boolean):double;
function  RobustRegression(X,Y:PMathArray;N:integer;var M,C:double):double;
procedure SubtractComplex(A,B:complex;var RetVal:complex);
procedure SwapDouble(var Num1,Num2:double);
procedure SwapInt(var Num1,Num2:integer);
procedure SwapLongint(var Num1,Num2:longint);
procedure SwapReal(var Num1,Num2:real);
procedure SwapWord(var Num1,Num2:word);

{=============================================================================}
implementation

uses
  math;

constructor TMultiReg.Create(NumCo:word;NumXY:longint);
{------------------------------------------------------------------------------}
{ This is the constructor for the multiple regression object. It assigns the   }
{ memory to all of the arrays required to run this object. This constructor    }
{ may raise an EOutOfMemory exception.                                         }
{------------------------------------------------------------------------------}
var
  I:word; {Loop control variable.}
begin
  NumCoeff:=NumCo;
  NumData:=NumXY;
  getmem(BasisFunc,NumCoeff*sizeof(double));
  getmem(Beta,NumCoeff*sizeof(PMathArray));
  for I:=1 to NumCoeff do getmem(Beta^[I],NumCo*sizeof(double));
  getmem(A,NumCoeff*sizeof(double));
  getmem(X,NumData*sizeof(double));
  getmem(Y,NumData*sizeof(double));
  getmem(Sigma,NumData*sizeof(double));
  getmem(Covar,NumCoeff*sizeof(PMathArray));
  for I:=1 to NumCoeff do getmem(Covar^[I],NumCo*sizeof(double));
end; {End of TMultiReg.Init.}

procedure TMultiReg.SetSwitches(N:longint);
{-----------------------------------------------------------------------------}
{ This routine is always called just before CalcFunc to allow any descedent   }
{ object to set various switches etc.                                         }
{-----------------------------------------------------------------------------}
begin

end; {End of TMultiReg.SetSwitches.}

function TMultiReg.CalcFunc(X:double):byte;
{-----------------------------------------------------------------------------}
{ This routine calculates the output from the NumCoeff basis functions at the }
{ value specified by `X'. The results are assigned directly to the BasisFunc  }
{ array. This virtual method MUST be overwritten in the descendent object     }
{ defined in the calling program. The Ith element in the BasisFunc array will }
{ be the basis function evaluated at `X'. For now this routine just fills the }
{ array elements with zeros.                                                  }
{-----------------------------------------------------------------------------}
var
  I:word; {Loop control variable.}
begin
  for I:=1 to NumCoeff do BasisFunc^[I]:=0.0;
  CalcFunc:=0;
end; {End of TMultiReg.CalcFunc.}

function TMultiReg.Run:double;
{-----------------------------------------------------------------------------}
{ This is the calculating routine that does the multiple regression analysis. }
{ The function will raise a EGBMath exception if any of the basis functions   }
{ couldn't be evaluated.                                                      }
{-----------------------------------------------------------------------------}
var
  I:longint;
  J,K:word; {Loop control variables.}
  Weight,     {The weighting parameter.}
  ChiSquared, {The value of chi squared in the regression analysis.}
  Sigma2i,    {1 divided by sigma squared.}
  Sum:double; {Used in the final chi squared summation.}
  Test:byte;  {Used to test the return from the GaussJordan matrix inversion.}
begin
  for I:=1 to NumCoeff do
  begin
    for J:=1 to NumCoeff do Covar^[I]^[J]:=0.0;
    Beta^[I]^[1]:=0.0;
  end;
  for I:=1 to NumData do
  begin
    SetSwitches(I);
    if CalcFunc(X^[I])<>0 then raise EGBMath.Create('basis function number '+
      inttostr(I)+' couldn''t be evaluated');
    Sigma2i:=1.0/sqr(Sigma^[I]);
    for J:=1 to NumCoeff do
    begin
      Weight:=BasisFunc^[J]*Sigma2i;
      for K:=1 to J do Covar^[J]^[K]:=Covar^[J]^[K]+(Weight*BasisFunc^[K]);
      Beta^[J]^[1]:=Beta^[J]^[1]+(Weight*Y^[I]);
    end;
  end;
  for I:=2 to NumCoeff do
    for J:=1 to I-1 do
      Covar^[J]^[I]:=Covar^[I]^[J];
  Test:=GaussJordan(Covar,Beta,NumCoeff,1);
  if Test<>0 then raise EGBMath.Create('the value returned from GaussJordan was '+
    inttostr(Test));
  for I:=1 to NumCoeff do A^[I]:=Beta^[I]^[1];
  ChiSquared:=0.0;
  for I:=1 to NumData do
  begin
    SetSwitches(I);
    if CalcFunc(X^[I])<>0 then raise EGBMath.Create('basis function number '+
      inttostr(I)+' couldn''t be evaluated');
    Sum:=0.0;
    for J:=1 to NumCoeff do Sum:=Sum+(A^[J]*BasisFunc^[J]);
    ChiSquared:=ChiSquared+sqr((Y^[I]-Sum)/Sigma^[I])
  end;
  Run:=ChiSquared;
end; {End of TMultiReg.Run.}

destructor TMultiReg.Destroy;
{-----------------------------------------------------------------------------}
{ This is the destructor for the multiple regression object. It frees the     }
{ memory that was assigned to some of the object's storage arrays.            }
{-----------------------------------------------------------------------------}
var
  I:word; {Loop control variable.}
begin
  for I:=1 to NumCoeff do
  begin
    FreePMathArray(Beta^[I],NumCoeff*sizeof(double));
    FreePMathArray(Covar^[I],NumCoeff*sizeof(double));
  end;
  if Beta<>nil then freemem(Beta,NumCoeff*sizeof(PMathArray));
  Beta:=nil;
  if Covar<>nil then freemem(Covar,NumCoeff*sizeof(PMathArray));
  Covar:=nil;
  FreePMathArray(BasisFunc,NumCoeff*sizeof(double));
  FreePMathArray(A,NumCoeff*sizeof(double));
  FreePMathArray(X,NumData*sizeof(double));
  FreePMathArray(Y,NumData*sizeof(double));
  FreePMathArray(Sigma,NumData*sizeof(double));
end; {End of TMultiReg.Destroy.}

function TPolyReg.CalcFunc(X:double):byte;
var
  I:word;
begin
  for I:=1 to NumCoeff do BasisFunc^[I]:=power(X,I-1);
  Result:=0; {Tells the ancestor that the basis functions were filled.}
end;

function TPolyReg.CalcModelValue(X:double):double;
var
  I:word;
begin
  Result:=A^[1];
  for I:=2 to NumCoeff do Result:=Result+A^[I]*power(X,I-1)
end;

function TFourierReg.CalcFunc(X:double):byte;
var
  I:word;
begin
  BasisFunc^[1]:=1;
  for I:=1 to ((NumCoeff-1) div 2) do
  begin
    BasisFunc^[I*2]:=sin(I*X);
    BasisFunc^[I*2+1]:=cos(I*X);
  end;
  Result:=0; {Tells the ancestor that the basis functions were filled.}
end;

function TFourierReg.CalcModelValue(X:double):double;
var
  I:word;
begin
  Result:=A^[1];
  for I:=1 to ((NumCoeff-1) div 2) do
  begin
    Result:=Result+A^[I*2]*sin(I*X);
    Result:=Result+A^[I*2+1]*cos(I*X);
  end;
end; {End of TFourierReg.CalcModelValue.}

constructor TIntervalHalver.Create(ANumCoeff,ANumSteps:word;UpperSideOnly:boolean);
begin
  NumCoeff:=ANumCoeff;
  NumSteps:=ANumSteps;
  FUpperOnly:=UpperSideOnly;
  if FUpperOnly then StepRange:=NumSteps-1 else StepRange:=(NumSteps-1) div 2;
  setlength(CoeffValues,NumCoeff);
  setlength(CoeffSens,NumCoeff);
  setlength(Indices,NumCoeff);
  TotalSteps:=round(intpower(NumSteps,NumCoeff));
  Reset;
end; {End of TInternalHalver.Create.}

procedure TIntervalHalver.Reset;
var
  I:integer;
begin
  FDone:=false;
  for I:=1 to NumCoeff do
    if FUpperOnly then Indices[I-1]:=0 else Indices[I-1]:=-StepRange;
  StepCount:=0;
end;

function TIntervalHalver.GetCoeff(I:word):double;
begin
  Result:=CoeffValues[I-1]+(Indices[I-1]*CoeffSens[I-1]);
end;

procedure TIntervalHalver.WriteCoeff(I:word;Value:double);
begin
  CoeffValues[I-1]:=Value;
end;

function TIntervalHalver.GetCoeffSens(I:word):double;
begin
  Result:=CoeffSens[I-1];
end;

procedure TIntervalHalver.WriteCoeffSens(I:word;Value:double);
begin
  CoeffSens[I-1]:=Value;
end;

function TIntervalHalver.GetPerturbedCoeffs:TDoubleArray;
var
  I:integer;
begin
  setlength(Result,NumCoeff);
  for I:=1 to NumCoeff do
    Result[I-1]:=CoeffValues[I-1]+(Indices[I-1]*CoeffSens[I-1]);
end;

function TIntervalHalver.Step:double;
var
  I:integer;
begin
  if FDone then
  begin
    Result:=100.0;
    exit;
  end;
  inc(Indices[0]);
  inc(StepCount);
  I:=0;
  repeat
    if Indices[I]>StepRange then
    begin
      if FUpperOnly then Indices[I]:=0 else Indices[I]:=-StepRange;
      if I=NumCoeff-1 then FDone:=true else inc(Indices[I+1]);
    end;
    inc(I);
  until I=NumCoeff;
  Result:=100*StepCount/TotalSteps;
end; {End of TIntervalHalver.Step.}

function AbsComplex(A:complex):double;
{------------------------------------------------------------------------------}
{ This function returns the absolute value of the complex number passed by A.  }
{------------------------------------------------------------------------------}
begin
  AbsComplex:=sqrt(sqr(A.Real)+sqr(A.Imag));
end; {End of AbsComplex.}

procedure AddComplex(A,B:complex;var RetVal:complex);
{------------------------------------------------------------------------------}
{ This function returns the sum of the two complex numbers passed via A and B. }
{------------------------------------------------------------------------------}
begin
  RetVal.Real:=A.Real+B.Real;
  RetVal.Imag:=A.Imag+B.Imag;
end; {End of AddComplex.}

function CalcDet33(Matrix:PMatrix33):double;
{------------------------------------------------------------------------------}
{ This routine calculates the determinant of a 3X3 matrix passed to it via     }
{ the `Matrix' pointer.                                                        }
{------------------------------------------------------------------------------}
begin
  CalcDet33:=(Matrix^[1,1]*Matrix^[2,2]*Matrix^[3,3])-
             (Matrix^[1,1]*Matrix^[3,2]*Matrix^[2,3])+
             (Matrix^[2,1]*Matrix^[3,2]*Matrix^[1,3])-
             (Matrix^[2,1]*Matrix^[1,2]*Matrix^[3,3])+
             (Matrix^[3,1]*Matrix^[1,2]*Matrix^[2,3])-
             (Matrix^[3,1]*Matrix^[2,2]*Matrix^[1,3]);
end; {End of CalcDet33.}

function CalcMax(X:PMathArray;N:cardinal):double;
{------------------------------------------------------------------------------}
{ This function calculates the maximum of the array specified by 'X'.          }
{------------------------------------------------------------------------------}
var
  I:cardinal;
begin
  Result:=-1E30;
  for I:=1 to N do if X^[I]>Result then Result:=X^[I];
end; {End of CalcMax.}

function CalcMin(X:PMathArray;N:cardinal):double;
{------------------------------------------------------------------------------}
{ This function calculates the minimum of the array specified by 'X'.          }
{------------------------------------------------------------------------------}
var
  I:cardinal;
begin
  Result:=1E30;
  for I:=1 to N do if X^[I]<Result then Result:=X^[I];
end; {End of CalcMin.}

function CalcMean(X:PMathArray;N:cardinal):double;
{------------------------------------------------------------------------------}
{ This function calculates the mean of the array specified by 'X'.             }
{------------------------------------------------------------------------------}
var
  I:cardinal;
begin
  Result:=0;
  for I:=1 to N do Result:=Result+X^[I];
  if N<>0 then Result:=Result/N;
end; {End of CalcMean.}

function CalcMedian(X:PMathArray;N:cardinal):double;
{------------------------------------------------------------------------------}
{ This function calculates the median of the array specified by 'X'.           }
{------------------------------------------------------------------------------}
begin
  HeapSort(X,N);
  if N mod 2=0 then Result:=(X^[N div 2]+X^[(N div 2)+1])/2.0 else Result:=X^[(N div 2)+1];
end; {End of CalcMedian.}

function CalcMedian(X:TDoubleArray):double;
var
  N:integer;
begin
  HeapSort(X);
  N:=length(X);
  if N mod 2=0 then Result:=(X[(N div 2)-1]+X[N div 2])/2.0 else Result:=X[N div 2];
end;

function CalcStdv(X:PMathArray;N:cardinal):double;
{------------------------------------------------------------------------------}
{ This function calculates the standard deviation of the array specified by 'X'}
{------------------------------------------------------------------------------}
var
  Mean:double;
  I:cardinal;
begin
  if N<2 then raise EGBMath.Create('the number of values passed to CalcStdv was less than 2');
  Mean:=0;
  for I:=1 to N do Mean:=Mean+X^[I];
  if N<>0 then Mean:=Mean/N;
  Result:=0;
  for I:=1 to N do Result:=Result+sqr(Mean-X^[I]);
  Result:=sqrt(Result/(N-1));
end; {End of CalcStdv.}

procedure ComplexConjugate(A:Complex;var RetVal:complex);
{------------------------------------------------------------------------------}
{ This function returns the complex conjugate of A.                            }
{------------------------------------------------------------------------------}
begin
  RetVal.Real:=A.Real;
  RetVal.Imag:=-1.0*A.Imag;
end; {End of ComplexConjugate.}

function DFT(DFTArray:PMathArray;N:integer):byte;
{------------------------------------------------------------------------------}
{ This procedure performs a Discrete Fourier Transform (DFT) on the data pass- }
{ ed via `DFTArray'. Note that only the odd numbered positions in this array   }
{ contain data since the imaginary parts of the original data points are zero. }
{ The `N' parameter is the number of original data points which is half the    }
{ number of elements in DFTArray. The DFTArray is replaced by its discrete     }
{ Fourier transform. `N' need not be a power of 2 but must be even. The value  }
{ returned by the function is an error code given by:                          }
{ 0: The DFT was performed successfully.                                       }
{ 1: `N' was not even.                                                         }
{ 2: There was insufficient memory to run this routine.                        }
{ 3: `N' was greater than MaxMathArray.                                        }
{------------------------------------------------------------------------------}
var
  I,J:integer;       {Loop control variables.}
  RealTot,ImagTot,   {The totals of the real and imaginary components.}
  T:double;          {The argument for sine and cosine functions.}
  Hold:PMathArray;   {Temporary heap storage array.}
  MemNeeded:longint; {The amount of memory needed.}
begin
  if (N mod 2)<>0 then
  begin
    DFT:=1;
    exit;
  end;
  MemNeeded:=N+1;
  MemNeeded:=MemNeeded*2*sizeof(double);
  getmem(Hold,MemNeeded);
  for I:=(-1*(N div 2)) to (N div 2) do
  begin
    RealTot:=0.0;
    ImagTot:=0.0;
    for J:=0 to N-1 do
    begin
      T:=2*pi*J*I/N;
      RealTot:=RealTot+((DFTArray^[J*2+1]*cos(T))-(DFTArray^[J*2+2]*sin(T)));
      ImagTot:=ImagTot+((DFTArray^[J*2+2]*cos(T))+(DFTArray^[J*2+1]*sin(T)));
    end;
    Hold^[(I+(N div 2))*2+1]:=RealTot;
    Hold^[(I+(N div 2))*2+2]:=ImagTot;
  end;
  for I:=1 to N do
  begin
    DFTArray^[I*2-1]:=Hold^[I*2-1];
    DFTArray^[I*2]:=Hold^[I*2];
  end;
  freemem(Hold,MemNeeded);
  DFT:=0;
end; {End of DFT.}

procedure DivideComplex(A,B:complex;var RetVal:complex);
{-----------------------------------------------------------------------------}
{ This function divides the complex number A by the complex number B.         }
{-----------------------------------------------------------------------------}
begin
  RetVal.Real:=(A.Real*B.Real+A.Imag*B.Imag)/(sqr(B.Real)+sqr(B.Imag));
  RetVal.Imag:=(A.Imag*B.Real-A.Real*B.Imag)/(sqr(B.Real)+sqr(B.Imag));
end;

function EquivDouble(A,B:double):boolean;
{-----------------------------------------------------------------------------}
{ This function returns true if A=B and false otherwise. Often the normal =   }
{ evaluation doesn't work due to rounding errors etc. This test checks that   }
{ they are with QuiteSmall of each other.                                     }
{-----------------------------------------------------------------------------}
begin
  if abs(A-B)<QuiteSmall then Result:=true else Result:=false;
end; {End of EquivDouble.}

function EquivSingle(A,B:single):boolean;
{-----------------------------------------------------------------------------}
{ This function returns true if A=B and false otherwise. Often the normal =   }
{ evaluation doesn't work due to rounding errors etc. This test checks that   }
{ they are with SomewhatSmall of each other.                                  }
{-----------------------------------------------------------------------------}
begin
  if abs(A-B)<SomewhatSmall then Result:=true else Result:=false;
end; {End of EquivDouble.}


function EquivReal(A,B:real):boolean;
{-----------------------------------------------------------------------------}
{ This function returns true if A=B and false otherwise. Often the normal =   }
{ evaluation doesn't work due to rounding errors etc. This test checks that   }
{ they are with QuiteSmall of each other.                                     }
{-----------------------------------------------------------------------------}
begin
  if abs(A-B)<QuiteSmall then EquivReal:=true else EquivReal:=false;
end; {End of EquivReal.}

procedure FFT(FFTArray:PMathArray;N:integer);
{-----------------------------------------------------------------------------}
{ This function performs a Fast Fourier Transform (FFT) on the data passed    }
{ via `FFTArray'. Note that only the odd numbered positions in this array     }
{ contain data since the imaginary parts of the original data points are      }
{ zero. The `N' parameter is the number of original data points which is half }
{ the number of elements in the FFTArray. The FFTArray is replaced by its     }
{ direct Fourier transform. Note that `N' must be an integer power of 2. No   }
{ error checking is performed to ensure that `N' is an integer power of 2.    }
{-----------------------------------------------------------------------------}
var
  Count1,Count2, {Loop control variables.}
  NumData,       {The number of complex data points.}
  MMax,istep,    {Interval halving variables.}
  I,J:integer;   {Loop control variables.}
  TempReal,TempImag, {Temporary storage variables.}
  wtemp,wr,wpr,wpi,wi,theta:double; {Real and imaginary storage variables.}
begin
  J:=1;
  for Count1:=1 to N do
  begin
    I:=2*Count1-1;
    if (J>I) then
    begin
      TempReal:=FFTArray^[J];
      TempImag:=FFTArray^[J+1];
      FFTArray^[J]:=FFTArray^[I];
      FFTArray^[J+1]:=FFTArray^[I+1];
      FFTArray^[I]:=TempReal;
      FFTArray^[I+1]:=TempImag;
    end;
    NumData:=N;
    while((NumData>=2) and (J>NumData)) do
    begin
      J:=J-NumData;
      NumData:=NumData div 2;
    end;
    J:=J+NumData;
  end;
  MMax:=2;
  while (N*2>MMax) do
  begin
    istep:=2*MMax;
    theta:=6.28318530717959/MMax;
    wpr:=-2.0*sqr(sin(0.5*theta));
    wpi:=sin(theta);
    wr:=1.0;
    wi:=0.0;
    for Count1:=1 to (MMax div 2) do
    begin
      NumData:=2*Count1-1;
      for Count2:=0 to ((N-NumData) div istep) do
      begin
        I:=NumData+Count2*istep;
        J:=I+MMax;
        TempReal:=(wr*FFTArray^[J])-(wi*FFTArray^[J+1]);
        TempImag:=(wr*FFTArray^[J+1])+(wi*FFTArray^[J]);
        FFTArray^[J]:=FFTArray^[I]-TempReal;
        FFTArray^[J+1]:=FFTArray^[I+1]-TempImag;
        FFTArray^[I]:=FFTArray^[I]+TempReal;
        FFTArray^[I+1]:=FFTArray^[I+1]+TempImag;
      end;
      wtemp:=wr;
      wr:=wr*wpr-wi*wpi+wr;
      wi:=wi*wpr+wtemp*wpi+wi;
    end;
    MMax:=istep;
  end;
end; {End of FFT.}

procedure FreePMathArray(var X:PMathArray;Mem:integer);
{------------------------------------------------------------------------------}
{ This procedure frees the 'Mem' bytes of memory associated with 'X' and sets  }
{ the pointer to nil. The routine first checks to see if X is not nil.         }
{------------------------------------------------------------------------------}
begin
  if X<>nil then freemem(X,Mem);
  X:=nil;
end; {End of FreePMathArray.}

function GaussJordan(A,B:P2DArray;N,M:word):byte;
{------------------------------------------------------------------------------}
{ Linear equation solution by Gauss-Jordan elimination. `A' is the input       }
{ matrix of `N' by `N' elements. `B' is an input matrix of `N' by `M' elements }
{ containing the `M' right-hand side vectors. On output, `A' is replaced by    }
{ its matrix inverse, and `B' is replaced by the corresponding set of solution }
{ vectors. The function returns an error code according to:                    }
{ 0: The Gauss-Jordan elimination was successful.                              }
{ 1: `N' or `M' was bigger than MaxMathArray.                                  }
{ 2: The input matrix was singular.                                            }
{ This routine has been tested and appears to work well.                       }
{------------------------------------------------------------------------------}
var
  IRow,
  ICol,
  I,J,K:word;        {Loop control variables.}
  Big,
  PivInv,
  Store:double;    {Storage variable for swapping.}
  IndexC,
  IndexR,
  IPiv:PWordArray;
begin
  ICol:=0;
  IRow:=0;
  getmem(IPiv,N*sizeof(word));
  getmem(IndexC,N*sizeof(word));
  getmem(IndexR,N*sizeof(word));
  for I:=1 to N do IPiv^[I]:=0;
  for I:=1 to N do
  begin
    Big:=0.0;
    for J:=1 to N do
      if (IPiv^[J]<>1) then
        for K:=1 to N do
          if (IPiv^[K]=0) then
          begin
            if (abs(A^[J]^[K])>=Big) then
            begin
              Big:=abs(A^[J]^[K]);
              IRow:=J;
              ICol:=K;
            end;
          end
          else
            if (IPiv^[K]>1) then
            begin
              GaussJordan:=2;
              exit;
            end;
    inc(IPiv^[ICol]);
    if (IRow<>ICol) then
    begin
      for J:=1 to N do
      begin
        Store:=A^[IRow]^[J];
        A^[IRow]^[J]:=A^[ICol]^[J];
        A^[ICol]^[J]:=Store;
      end;
      for J:=1 to M do
      begin
        Store:=B^[IRow]^[J];
        B^[IRow]^[J]:=B^[ICol]^[J];
        B^[ICol]^[J]:=Store;
      end;
    end;
    IndexR^[I]:=IRow;
    IndexC^[I]:=ICol;
    if (A^[ICol]^[ICol]=0.0) then
    begin
      GaussJordan:=2;
      exit;
    end;
    PivInv:=1.0/A^[ICol]^[ICol];
    A^[ICol]^[ICol]:=1.0;
    for J:=1 to N do A^[ICol]^[J]:=A^[ICol]^[J]*PivInv;
    for J:=1 to M do B^[ICol]^[J]:=B^[ICol]^[J]*PivInv;
    for J:=1 to N do
      if (J<>ICol) then
      begin
        Store:=A^[J]^[ICol];
        A^[J]^[ICol]:=0.0;
        for K:=1 to N do A^[J]^[K]:=A^[J]^[K]-(A^[ICol]^[K]*Store);
        for K:=1 to M do B^[J]^[K]:=B^[J]^[K]-(B^[ICol]^[K]*Store);
      end;
  end;
  for I:=N downto 1 do
    if IndexR^[I]<>IndexC^[I] then
      for J:=1 to N do
      begin
        Store:=A^[J]^[IndexR^[I]];
        A^[J]^[IndexR^[I]]:=A^[J]^[IndexC^[I]];
        A^[J]^[IndexC^[I]]:=Store;
      end;
  freemem(IPiv,N*sizeof(word));
  freemem(IndexC,N*sizeof(word));
  freemem(IndexR,N*sizeof(word));
  GaussJordan:=0;
end; {End of GaussJordan.}

procedure GaussJordan(Covar:T2DDoubleArray;Beta:T1DDoubleArray);
{------------------------------------------------------------------------------}
{------------------------------------------------------------------------------}
var
  IndexC,                   {Used for book keeping when pivoting.}
  IndexR,                   {Used for book keeping when pivoting.}
  IPiv:gb_array.TWordArray; {Used for book keeping when pivoting.}
  IRow,
  ICol,
  I,J,K:integer; {Loop control variables.}
  Big,
  PivInv,
  Store:double; {Temporary storage variable.}
begin
  ICol:=0;
  IRow:=0;
  setlength(IPiv,Covar.GetNumX);
  setlength(IndexC,Covar.GetNumX);
  setlength(IndexR,Covar.GetNumX);
  for I:=1 to Covar.GetNumX do IPiv[I-1]:=0;
  for I:=1 to Covar.GetNumX do
  begin
    Big:=0.0;
    for J:=1 to Covar.GetNumX do
      if IPiv[J-1]<>1 then
        for K:=1 to Covar.GetNumX do
          if IPiv[K-1]=0 then
          begin
            if (abs(Covar[J,K])>=Big) then
            begin
              Big:=abs(Covar[J,K]);
              IRow:=J;
              ICol:=K;
            end;
          end
          else
          if (IPiv[K-1]>1) then raise EGBMath.Create('the input matrix was singular.');
    inc(IPiv[ICol-1]);
    if (IRow<>ICol) then
    begin
      for J:=1 to Covar.GetNumX do
      begin
        Store:=Covar[IRow,J];
        Covar[IRow,J]:=Covar[ICol,J];
        Covar[ICol,J]:=Store;
      end;
      Store:=Beta[IRow];
      Beta[IRow]:=Beta[ICol];
      Beta[ICol]:=Store;
    end;
    IndexR[I-1]:=IRow;
    IndexC[I-1]:=ICol;
    if Covar[ICol,ICol]=0.0 then raise EGBMath.Create('the input matrix was singular.');
    PivInv:=1.0/Covar[ICol,ICol];
    Covar[ICol,ICol]:=1.0;
    for J:=1 to Covar.GetNumX do Covar[ICol,J]:=Covar[ICol,J]*PivInv;
    Beta[ICol]:=Beta[ICol]*PivInv;
    for J:=1 to Covar.GetNumX do
      if J<>ICol then
      begin
        Store:=Covar[J,ICol];
        Covar[J,ICol]:=0.0;
        for K:=1 to Covar.GetNumX do Covar[J,K]:=Covar[J,K]-(Covar[ICol,K]*Store);
        Beta[J]:=Beta[J]-(Beta[ICol]*Store);
      end;
  end;
  for I:=Covar.GetNumX downto 1 do
    if IndexR[I-1]<>IndexC[I-1] then
      for J:=1 to Covar.GetNumX do
      begin
        Store:=Covar[J,IndexR[I-1]];
        Covar[J,IndexR[I-1]]:=Covar[J,IndexC[I-1]];
        Covar[J,IndexC[I-1]]:=Store;
      end;
end; {End of TLeastSquaresRegression.GaussJordan.}

procedure GaussJordan(Covar:T2DimDoubleArray;Beta:TDoubleArray);
{------------------------------------------------------------------------------}
{------------------------------------------------------------------------------}
var
  IndexC,                   {Used for book keeping when pivoting.}
  IndexR,                   {Used for book keeping when pivoting.}
  IPiv:gb_array.TWordArray; {Used for book keeping when pivoting.}
  IRow,
  ICol,
  I,J,K:integer; {Loop control variables.}
  Big,
  PivInv,
  Store:double; {Temporary storage variable.}
begin
  ICol:=-1;
  IRow:=-1;
  setlength(IPiv,length(Covar[0]));
  setlength(IndexC,length(Covar[0]));
  setlength(IndexR,length(Covar[0]));
  for I:=0 to length(Covar[0])-1 do IPiv[I]:=0;
  for I:=0 to length(Covar[0])-1 do
  begin
    Big:=0.0;
    for J:=0 to length(Covar[0])-1 do
      if IPiv[J]<>1 then
        for K:=0 to length(Covar[0])-1 do
          if IPiv[K]=0 then
          begin
            if (abs(Covar[J,K])>=Big) then
            begin
              Big:=abs(Covar[J,K]);
              IRow:=J;
              ICol:=K;
            end;
          end
          else
          if (IPiv[K]>1) then raise EGBMath.Create('the input matrix was singular.');
    inc(IPiv[ICol]);
    if (IRow<>ICol) then
    begin
      for J:=0 to length(Covar[0])-1 do
      begin
        Store:=Covar[IRow,J];
        Covar[IRow,J]:=Covar[ICol,J];
        Covar[ICol,J]:=Store;
      end;
      Store:=Beta[IRow];
      Beta[IRow]:=Beta[ICol];
      Beta[ICol]:=Store;
    end;
    IndexR[I]:=IRow;
    IndexC[I]:=ICol;
    if Covar[ICol,ICol]=0.0 then raise EGBMath.Create('the input matrix was singular.');
    PivInv:=1.0/Covar[ICol,ICol];
    Covar[ICol,ICol]:=1.0;
    for J:=0 to length(Covar[0])-1 do Covar[ICol,J]:=Covar[ICol,J]*PivInv;
    Beta[ICol]:=Beta[ICol]*PivInv;
    for J:=0 to length(Covar[0])-1 do
      if J<>ICol then
      begin
        Store:=Covar[J,ICol];
        Covar[J,ICol]:=0.0;
        for K:=0 to length(Covar[0])-1 do Covar[J,K]:=Covar[J,K]-(Covar[ICol,K]*Store);
        Beta[J]:=Beta[J]-(Beta[ICol]*Store);
      end;
  end;
  for I:=length(Covar[0])-1 downto 0 do
    if IndexR[I]<>IndexC[I] then
      for J:=0 to length(Covar[0])-1 do
      begin
        Store:=Covar[J,IndexR[I]];
        Covar[J,IndexR[I]]:=Covar[J,IndexC[I]];
        Covar[J,IndexC[I]]:=Store;
      end;
end; {End of TLeastSquaresRegression.GaussJordan.}

function GetMoreMemory(var X:PMathArray;N,M:word):boolean;
{------------------------------------------------------------------------------}
{ This function get more memory for 'M' more entries in the Matharray pointer  }
{ specified by 'X' which already contains 'N' elements. The function returns   }
{ true if the operation was successful and false otherwise.                    }
{------------------------------------------------------------------------------}
var
  I:word;
  Y:PMathArray;
begin
  if (N+M)>MaxMathArray then Result:=false else
  begin
    Result:=true;
    getmem(Y,(M+N)*sizeof(double));
    for I:=1 to N do Y^[I]:=X^[I];
    freemem(X,N*sizeof(double));
    X:=Y;
  end;
end; {End of GetMoreMemory.}

procedure HeapSort(X:PMathArray;N:word);
{-----------------------------------------------------------------------------}
{ This function uses the heapsort method to sort the `X' array elements into  }
{ ascending order. The number of elements in the array is passed by `N'.      }
{-----------------------------------------------------------------------------}
var
  I,J,          {Loop control variables.}
  IndexX,       {Index for input data array.}
  L:integer;    {Used for interval halving.}
  HoldX:double; {Temporary storage variable.}
  Done:boolean; {Set to true when sorting is complete.}
begin
  if N=1 then exit;
  L:=(N div 2)+1;
  IndexX:=N;
  Done:=false;
  while not Done do
  begin
    if (L>1) then
    begin
      dec(L);
      HoldX:=X^[L];
    end
    else
    begin
      HoldX:=X^[IndexX];
      X^[IndexX]:=X^[1];
      dec(IndexX);
      if IndexX=1 then
      begin
        X^[1]:=HoldX;
        Done:=true;
      end;
    end;
    if not Done then
    begin
      I:=L;
      J:=L+L;
      while (J<=IndexX) do
      begin
        if (J<IndexX) then
          if (X^[J]<X^[J+1]) then inc(J);
        if (HoldX<X^[J]) then
        begin
          X^[I]:=X^[J];
          I:=J;
          J:=J+J;
        end
        else J:=IndexX+1;
      end;
      X^[I]:=HoldX;
    end;
  end;
end; {End of HeapSort.}

procedure HeapSort(var X:TDoubleArray);
{-----------------------------------------------------------------------------}
{ This function uses the heapsort method to sort the `X' array elements into  }
{ ascending order.                                                            }
{-----------------------------------------------------------------------------}
var
  I,J,          {Loop control variables.}
  IndexX,       {Index for input data array.}
  L:integer;    {Used for interval halving.}
  HoldX:double; {Temporary storage variable.}
  Done:boolean; {Set to true when sorting is complete.}
begin
  if length(X)=1 then exit;
  L:=(length(X) div 2)+1;
  IndexX:=length(X);
  Done:=false;
  while not Done do
  begin
    if (L>1) then
    begin
      dec(L);
      HoldX:=X[L-1];
    end
    else
    begin
      HoldX:=X[IndexX-1];
      X[IndexX-1]:=X[0];
      dec(IndexX);
      if IndexX=1 then
      begin
        X[0]:=HoldX;
        Done:=true;
      end;
    end;
    if not Done then
    begin
      I:=L;
      J:=L+L;
      while (J<=IndexX) do
      begin
        if (J<IndexX) then
          if (X[J-1]<X[J]) then inc(J);
        if (HoldX<X[J-1]) then
        begin
          X[I-1]:=X[J-1];
          I:=J;
          J:=J+J;
        end
        else J:=IndexX+1;
      end;
      X[I-1]:=HoldX;
    end;
  end;
end; {End of HeapSort.}

procedure HeapSort2Array(X,Y:PMathArray;N:word);
{-----------------------------------------------------------------------------}
{ This function uses the heapsort method to sort the `X' and `Y' array        }
{ elements into ascending order as indexed by `X'. The number of elements in  }
{ the array is passed by `N'.                                                 }
{-----------------------------------------------------------------------------}
var
  I,J,          {Loop control variables.}
  IndexX,       {Index for input data array.}
  L:integer;    {Used for interval halving.}
  HoldX,        {Temporary storage variable.}
  HoldY:double; {Temporary storage variable.}
  Done:boolean; {Set to true when sorting is complete.}
begin
  if N=1 then exit;
  L:=(N div 2)+1;
  IndexX:=N;
  Done:=false;
  repeat
    if (L>1) then
    begin
      dec(L);
      HoldX:=X^[L];
      HoldY:=Y^[L];
    end
    else
    begin
      HoldX:=X^[IndexX];
      HoldY:=Y^[IndexX];
      X^[IndexX]:=X^[1];
      Y^[IndexX]:=Y^[1];
      dec(IndexX);
      if IndexX=1 then
      begin
        X^[1]:=HoldX;
        Y^[1]:=HoldY;
        Done:=true;
      end;
    end;
    if not Done then
    begin
      I:=L;
      J:=L+L;
      while (J<=IndexX) do
      begin
        if (J<IndexX) then
          if (X^[J]<X^[J+1]) then inc(J);
        if (HoldX<X^[J]) then
        begin
          X^[I]:=X^[J];
          Y^[I]:=Y^[J];
          I:=J;
          J:=J+J;
        end
        else J:=IndexX+1;
      end;
      X^[I]:=HoldX;
      Y^[I]:=HoldY;
    end;
  until Done;
end; {End of HeapSort2Array.}

function HexToDec(ValStr:string):longint;
{-----------------------------------------------------------------------------}
{ This routine converts the hexidecimal string passed by `ValStr' to double.  }
{ If the string passed was not value, the routine returns a value of -1.      }
{-----------------------------------------------------------------------------}
var
  MultFac,         {The multiplication factor.}
  Value,           {The value of the current character.}
  Total:longint;   {The value returned by the function.}
  NumChar,         {The number of characters in the input string.}
  I:word;          {Loop control variable.}
  Chr:char;        {The current character.}
  BadData:boolean; {True if there was an invalid character in `ValStr'.}
begin
  NumChar:=length(ValStr);
  MultFac:=1;
  Total:=0;
  BadData:=false;
  for I:=1 to NumChar do
  begin
    Chr:=ValStr[I];
    if not (Chr in ['0'..'9','a'..'f','A'..'F']) then BadData:=true;
  end;
  if not BadData then
    for I:=1 to NumChar do
    begin
      Chr:=ValStr[NumChar-I+1];
      case Chr of
        '0'..'9':Value:=ord(Chr)-48;
        'a'..'f':Value:=ord(Chr)-87;
        'A'..'F':Value:=ord(Chr)-55;
      else
      begin
        BadData:=true;
        Value:=0;
      end;
      end;
      Total:=Total+MultFac*Value;
      MultFac:=MultFac*16;
    end;
  if BadData then HexToDec:=-1 else HexToDec:=Total;
end; {End of HexToDec.}

function Interpolate(X,Y:PMathArray;var N:word;L1,L2,Res:double):byte;
{------------------------------------------------------------------------------}
{ This routine performs interpolation on the passed `X' and `Y' data arrays to }
{ produce new arrays at the resolution specified by `Res'. The data to be      }
{ interpolated are passed to the routine via the `X' and `Y' pointers. It is   }
{ assumed that the X values are the independent values while the Y values are  }
{ the dependent values ie. the routine will use the X values as the time/      }
{ spatial coordinate. Furthermore, it is assumed that no two X values will be  }
{ the same. If this does occur, the routine will not know which value to use   }
{ for the interpolation and the routine will return an error message (see      }
{ below). The interpolated data arrays are returned via `X' and `Y'. For this  }
{ reason, the X and Y array pointers must have been assigned enough memory to  }
{ contain the new interpolated data arrays. The `L1' and `L2' parameters       }
{ specify the X range over which values must be interpolated while the `N'     }
{ parameter gives the number of values that have been assigned to the X and Y  }
{ arrays. Note that this may be less than the length of the arrays since       }
{ intermediate values are to be interpolated. The value of `N' is modified to  }
{ reflect the new number of data points in `X' and `Y'. Note that the units of }
{ X, L1, L2 and Res (usually spatial or temporal coordinates) must be the same }
{ eg. hours. It is assumed that there are no null data in either of the two    }
{ arrays. The routine first sorts the arrays so that the X values are in       }
{ ascending order. Note that the first entry in the new `X' array will not be  }
{ L1 but L1+Res. The routine returns an error code given by:                   }
{ 0: The interpolation was successful.                                         }
{ 1: There was insufficient memory to perform the interpolation.               }
{ 2: Two X values were the same.                                               }
{------------------------------------------------------------------------------}
var
  NumSteps, {The number of entries in NewX and NewY.}
  XIndex,   {The index of the X data array.}
  I:word;   {Loop control variable.}
  NewX,     {The new X data array.}
  NewY:PMathArray; {The new Y data array.}
begin
  if L2<L1 then SwapDouble(L2,L1);
  NumSteps:=round((L2-L1)/Res);
  HeapSort2Array(X,Y,N);
  for I:=1 to N-1 do
    if X^[I]=X^[I+1] then
    begin
      Interpolate:=2;
      exit;
    end;
  getmem(NewX,NumSteps*sizeof(double));
  getmem(NewY,NumSteps*sizeof(double));
  for I:=1 to NumSteps do
  begin
    NewX^[I]:=((L2-L1)*I/NumSteps)+L1;
    XIndex:=0;
    repeat inc(XIndex) until ((X^[XIndex]>NewX^[I]) or (XIndex=N));
    if XIndex=1 then NewY^[I]:=Y^[1] else
    begin
      if ((XIndex=N) and (X^[XIndex]<NewX^[I])) then NewY^[I]:=Y^[N]
      else NewY^[I]:=((NewX^[I]-X^[XIndex-1])*(Y^[XIndex]-Y^[XIndex-1])/
           (X^[XIndex]-X^[XIndex-1]))+Y^[XIndex-1];
    end;
  end;
  for I:=1 to NumSteps do {I use this technique as     }
  begin                   {there appears to be some-   }
    X^[I]:=NewX^[I];      {thing wrong with the Borland}
    Y^[I]:=NewY^[I];      {Pascal move command.        }
  end;
  freemem(NewX,NumSteps*sizeof(double));
  freemem(NewY,NumSteps*sizeof(double));
  N:=NumSteps;
  Interpolate:=0;
end; {End of Interpolate.}

function LUBackSubst(A:P2DArray;N:word;Index:PIntegerArray;B:PMathArray):byte;
{------------------------------------------------------------------------------}
{ This routine solves the set of N linear equations A.x=B. Here the `A' matrix }
{ is input, not as the matrix of coefficients but rather as its LU decomposit- }
{ ion, determined by the routine LUDecomposition. `B' is input as the right    }
{ hand side vector B, and returns with the solution vector x. `A', `N' and     }
{ `Index' are not modified by this routine. The routine returns an error code  }
{ according to:                                                                }
{ 0: The operation was successful.                                             }
{ 1: `N' was greater than MaxMathArray.                                        }
{ I have tested this routine on a set of 4 equations with 4 unknowns and it    }
{ worked perfectly.                                                            }
{ This routine was taken directly from "Numerical Recipes in Pascal" by Press, }
{ Flannery, Teukolsky and Vetterling, Cambridge University Press, 1992.        }
{------------------------------------------------------------------------------}
var
  I,J,
  II,Ip:integer;
  Sum:double;
begin
  II:=0;
  for I:=1 to N do
  begin
    Ip:=Index^[I];
    Sum:=B^[Ip];
    B^[Ip]:=B^[I];
    if II<>0 then for J:=II to I-1 do Sum:=Sum-A^[I]^[J]*B^[J]
    else if sum<>0.0 then II:=I;
    B^[I]:=Sum;
  end;
  for I:=N downto 1 do
  begin
    Sum:=B^[I];
    for J:=I+1 to N do Sum:=Sum-A^[I]^[J]*B^[J];
    B^[I]:=Sum/A^[I]^[I];
  end;
  LUBackSubst:=0;
end; {End of LUBackSubst.}

function LUDecomposition(A:P2DArray;N:word;Index:PIntegerArray):integer;
{------------------------------------------------------------------------------}
{ Given an N by N matrix `A', this routine replaces it by the LU decomposition }
{ or a rowwise permutation of itself. The `A' matrix is modified by this rout- }
{ ine. Index is an output vector which records the row permutation effected by }
{ the partial pivoting. The function returns code according to:                }
{ -1: The number of row interchanges was odd, the operation was successful.    }
{  0: The matrix was singular and no LU decomposition was performed.           }
{  1: The number of row interchanges was even, the operation was successful.   }
{  2: `N' was bigger than MaxMathArray.                                        }
{  3: There was insufficient memory to run this routine.                       }
{ This routine was taken directly from "Numerical Recipes in Pascal" by Press, }
{ Flannery, Teukolsky and Vetterling, Cambridge University Press, 1992.        }
{------------------------------------------------------------------------------}
var
  I,J,K,              {Loop control variables.}
  IMax:integer;       {Holds the value of I for Dum>Big.}
  Interchange,        {1 if number of row interchanges was even else -1.}
  Sum,                {Used in summations.}
  Dum,                {Dummy storage variable.}
  Big:double;         {Maximum array element.}
  Scaling:PMathArray; {Stores the implicit scaling of each row.}
begin
  IMax:=0;
  getmem(Scaling,N*sizeof(double));
  Interchange:=1.0;
  for I:=1 to N do
  begin
    Big:=0.0;
    for J:=1 to N do
      if abs(A^[I]^[J])>Big then Big:=abs(A^[I]^[J]);
    if Big=0.0 then
    begin
      LUDecomposition:=0;
      exit; {Need to check that it really exits the whole routine here.}
    end;
    Scaling^[I]:=1.0/Big;
  end;
  for J:=1 to N do
  begin
    for I:=1 to J-1 do
    begin
      Sum:=A^[I]^[J];
      for K:=1 to I-1 do Sum:=Sum-A^[I]^[K]*A^[K]^[J];
      A^[I]^[J]:=Sum;
    end;
    Big:=0.0;
    for I:=J to N do
    begin
      Sum:=A^[I]^[J];
      for K:=1 to J-1 do Sum:=Sum-A^[I]^[K]*A^[K]^[J];
      A^[I]^[J]:=Sum;
      Dum:=Scaling^[I]*abs(Sum);
      if Dum>=Big then
      begin
        Big:=Dum;
        IMax:=I;
      end;
    end;
    if J<>IMax then
    begin
      for K:=1 to N do
      begin
        Dum:=A^[IMax]^[K];
        A^[IMax]^[K]:=A^[J]^[K];
        A^[J]^[K]:=Dum;
      end;
      InterChange:=-InterChange;
      Scaling^[IMax]:=Scaling^[J];
    end;
    Index^[J]:=IMax;
    if A^[J]^[J]=0.0 then A^[J]^[J]:=QuiteSmall;
    if J<>N then
    begin
      Dum:=1.0/A^[J]^[J];
      for I:=J+1 to N do A^[I]^[J]:=A^[I]^[J]*Dum;
    end;
  end;
  freemem(Scaling,N*sizeof(double));
  LUDecomposition:=round(Interchange);
end; {End of LUDecomposition.}


procedure MultiplyComplex(A,B:complex;var RetVal:complex);
{-----------------------------------------------------------------------------}
{ This function multiplies the two complex numbers passed via A and B and     }
{ returns the result.                                                         }
{-----------------------------------------------------------------------------}
begin
  RetVal.Real:=(A.Real*B.Real)-(A.Imag*B.Imag);
  RetVal.Imag:=(A.Real*B.Imag)+(A.Imag*B.Real);
end;

function Regression(X,Y:PMathArray;N:integer;var M,C:double;ZeroOff:boolean):double;
{------------------------------------------------------------------------------}
{ This function returns the R coefficient from the regression analysis of the  }
{ data passed via the `X' and `Y' pointers. The number of data values is       }
{ passed by the `N' parameter. The `M' and `C' coefficients from the           }
{ regression analysis are also returned through the variable parameters `M'    }
{ and `C'. If `N' is bigger than MaxMathArray an exception is raised. If       }
{ `ZeroOff' is set to true, then the C value is set to zero and the M value is }
{ calculated accordingly.                                                      }
{------------------------------------------------------------------------------}
var
  I:word; {Loop control variable.}
  SumX,   {The sum of all the X values.}
  SumY,   {The sum of all the Y values.}
  SumXY,  {The sum of the product of the X and Y values.}
  SumXX,  {The sum of all the X values squared.}
  SumYY,  {The sum of all the Y values squared.}
  Hold:double; {Temporary storage variable.}
begin
  if N>MaxMathArray then raise EGBMath.Create('N was greater than MaxMathArray in Regression');
  SumX:=0.0;
  SumY:=0.0;
  SumXY:=0.0;
  SumXX:=0.0;
  SumYY:=0.0;
  for I:=1 to N do
  begin
    SumX:=SumX+X^[I];
    SumY:=SumY+Y^[I];
    SumXY:=SumXY+(X^[I]*Y^[I]);
    SumXX:=SumXX+sqr(X^[I]);
    SumYY:=SumYY+sqr(Y^[I]);
  end;
  Hold:=(N*SumXX)-sqr(SumX);
  if Hold=0 then
  begin
    M:=0.0;
    C:=0.0;
    Regression:=0.0;
    exit;
  end;
  if ZeroOff then
  begin
    if SumXX=0.0 then M:=0.0 
    else M:=SumXY/SumXX;
    C:=0.0;
    Regression:=0.0;
  end
  else
  begin
    M:=((N*SumXY)-(SumX*SumY))/Hold;
    C:=(SumY-M*SumX)/N;
    Hold:=(N*SumXX-sqr(SumX))*(N*SumYY-sqr(SumY));
    if Hold<=0 then Regression:=0.0
    else Regression:=((N*SumXY)-(SumX*SumY))/(sqrt(Hold));
  end;
end; {End of Regression.}

function RobustRegression(X,Y:PMathArray;N:integer;var M,C:double):double;
{------------------------------------------------------------------------------}
{ This function fits Y=MX+C to the data passed via the `X' and `Y' pointer     }
{ using the criterion of least absolute deviations. The function returns the   }
{ mean absolute deviation (in y) of the experimental points from the fitted    }
{ line. The number of data values is passed by the `N' parameter. The M and C  }
{ coefficients from the robust regression analysis are returned through the    }
{ variable parameters `M' and `C'. If `N' is bigger than MaxMathArray as       }
{ exception is raised.                                                         }
{------------------------------------------------------------------------------}
var
  Hold,     {Temporary storage variable.}
  AA,       {Temporary storage variable.}
  BB,       {Temporary storage variable.}
  Abdevt,   {Temporary storage variable.}
  SumX,     {The sum of all the X values.}
  SumY,     {The sum of all the Y values.}
  SumXY,    {The sum of the product of the X and Y values.}
  SumXX,    {The sum of all the X values squared.}
  ChiSq,    {The value of Chi squared.}
  SigB,     {Temporary storage variable.}
  B1,B2,    {Used for interval halving.}
  F1,       {Value returned from CalcRoFunc for B1.}
  F:double; {The value returned by the CalcRoFunc routine for BB.}
  I:word;   {Loop control variable.}
  Arr:PMathArray; {Temporary storage array.}
  Done:boolean;   {Set to true when processing complete.}

  function CalcRoFunc(B:double):double;
  var
    D,           {Temporary storage variable.}
    Sum:double;  {Temporary storage variable.}
    J,           {Loop control variable.}
    N1,          {Temporary storage variable.}
    Nmh,         {Temporary storage variable.}
    Nml:integer; {Temporary storage variable.}
  begin
    N1:=N+1;
    Nml:=N1 div 2;
    Nmh:=N1-Nml;
    for J:=1 to N do Arr^[J]:=Y^[J]-(B*X^[J]);
    HeapSort(Arr,N);
    AA:=0.5*(Arr^[Nml]+Arr^[Nmh]);
    Sum:=0.0;
    Abdevt:=0.0;
    for J:=1 to N do
    begin
      D:=Y^[J]-(B*X^[J]+AA);
      Abdevt:=Abdevt+abs(D);
      if (D>0.0) then Sum:=Sum+X^[J] else Sum:=Sum-X^[J];
    end;
    CalcRoFunc:=Sum;
  end;

begin
  if N>MaxMathArray then
    raise EGBMath.Create('N was greater than MaxMathArray in RobustRegression');
  getmem(Arr,N*sizeof(double));
  SumX:=0.0;
  SumY:=0.0;
  SumXY:=0.0;
  SumXX:=0.0;
  for I:=1 to N do
  begin
    SumX:=SumX+X^[I];
    SumY:=SumY+Y^[I];
    SumXY:=SumXY+(X^[I]*Y^[I]);
    SumXX:=SumXX+sqr(X^[I]);
  end;
  Hold:=(N*SumXX)-sqr(SumX);
  if Hold=0.0 then raise EGBMath.Create('the value of Hold was zero');
  AA:=((SumXX*SumY)-(SumX*SumXY))/Hold;
  BB:=((N*SumXY)-(SumX*SumY))/Hold;
  ChiSq:=0.0;
  for I:=1 to N do ChiSq:=ChiSq+sqr(Y^[I]-(BB*X^[I]+AA));
  SigB:=sqrt(ChiSq/Hold);
  B1:=BB;
  F1:=CalcRoFunc(B1);
  if (F1>0.0) then B2:=BB+abs(3.0*SigB)
  else B2:=BB-abs(3.0*SigB);
  CalcRoFunc(B2);
  Done:=false;
  while ((abs(B2-B1)>Sigb) and not Done) do
  begin
    BB:=0.5*(B1+B2);
    if ((BB=B1) or (BB=B2)) then Done:=true
    else
    begin
      F:=CalcRoFunc(BB);
      if ((F*F1)>=0.0) then
      begin
        F1:=F;
        B1:=BB;
      end
      else B2:=BB;
    end;
  end;
  freemem(Arr,N*sizeof(double));
  C:=AA;
  M:=BB;
  RobustRegression:=abdevt/N;
end; {End of RobustRegression.}

procedure SubtractComplex(A,B:complex;var RetVal:complex);
{-----------------------------------------------------------------------------}
{ This function subtracts the complex number B from the complex number A.     }
{-----------------------------------------------------------------------------}
begin
  RetVal.Real:=A.Real-B.Real;
  RetVal.Imag:=A.Imag-B.Imag;
end; {End of SubtractComplex.}

procedure SwapDouble(var Num1,Num2:double);
{-----------------------------------------------------------------------------}
{ This routine swaps the two numbers passed by `Num1' and `Num2'.             }
{-----------------------------------------------------------------------------}
var
  Hold:double; {Temporary storage variable.}
begin
  Hold:=Num1;
  Num1:=Num2;
  Num2:=Hold;
end; {End of SwapDouble.}

procedure SwapInt(var Num1,Num2:integer);
{-----------------------------------------------------------------------------}
{ This routine swaps the two numbers passed by `Num1' and `Num2'.             }
{-----------------------------------------------------------------------------}
var
  Hold:integer; {Temporary storage variable.}
begin
  Hold:=Num1;
  Num1:=Num2;
  Num2:=Hold;
end; {End of SwapInt.}

procedure SwapLongint(var Num1,Num2:longint);
{-----------------------------------------------------------------------------}
{ This routine swaps the two numbers passed by `Num1' and `Num2'.             }
{-----------------------------------------------------------------------------}
var
  Hold:longint; {Temporary storage variable.}
begin
  Hold:=Num1;
  Num1:=Num2;
  Num2:=Hold;
end; {End of SwapLongint.}

procedure SwapReal(var Num1,Num2:real);
{-----------------------------------------------------------------------------}
{ This routine swaps the two numbers passed by `Num1' and `Num2'.             }
{-----------------------------------------------------------------------------}
var
  Hold:real; {Temporary storage variable.}
begin
  Hold:=Num1;
  Num1:=Num2;
  Num2:=Hold;
end; {End of SwapReal.}

procedure SwapWord(var Num1,Num2:word);
{-----------------------------------------------------------------------------}
{ This routine swaps the two numbers passed by `Num1' and `Num2'.             }
{-----------------------------------------------------------------------------}
var
  Hold:word; {Temporary storage variable.}
begin
  Hold:=Num1;
  Num1:=Num2;
  Num2:=Hold;
end; {End of SwapWord.}

{$WARNINGS OFF}
end. {End of math unit.}
{$WARNINGS ON}


