//@***h* BDBP/gb_ncepncar.pas
// 
// NAME
//   gb_ncepncar.pas
//
// PURPOSE
//  Defines an object which manages global grid data for the NCEP/NCAR reanalysis 
//  project.    
//
// DESCRIPTION
//  This unit defines an object which manages global grid data for the NCEP/NCAR 
//  reanalysis project.     
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_ncepncar;

{$I UNITDIREC.INC}

interface

uses
  stdctrls,
  sysutils,
  gb_array,
  gb_grib;

const
  NumLevels=11;
  Levels:array[1..NumLevels] of word=(270,280,290,300,315,330,350,400,450,550,650);
  PressLevels:array[1..17] of word=(1000,925,850,700,600,500,400,300,250,200,150,100,70,50,30,20,10);

type
  TSpecies=(pv,uw,vw,hgt,tmp,hmd,pottemp,troppres,pressure,mntsf,vvel,ozone);
  TLevelType=(PotTempLevel,PressureLevel,Surface,AltitudeLevel,Column,SeaLevel);

  TNCEPGrid=class(TLargeArray)
  private
    FNCEPNumLong,
    FNCEPNumLat:word;
    WGribPath,       {Specifies where to find wgrib.exe.}
    DataDesc:string; {Description of the data being held by the object.}
    NullValue:real48;
    function Index(X,Y:byte):TRealPtr;
  public
    Year,     {The year of the data being stored.}
    Hour,     {The hour of the day of the data being stored.}
    Day:word; {The day of the year of the data being stored.}
    Grib:TGrib; {The object managing the associated grib file.}
    Level:double; {The pressure or theta surface for the data field.}
    NoGrib:boolean; {Set this to true to stop any calls to CreateProcess in the TGrib object.}
    Species:TSpecies; {The species being stored.}
    LevelType:TLevelType; {Whether the data are on theta or pressure surfaces or neither.}
    ExpectedBinFilesize:longint; {Expected size of the binary data file.}
    constructor Create(ASpecies:TSpecies;ALevelType:TLevelType;ALevel:double);
    procedure SetNull(ANullValue:real48);
    procedure SetAllDataToNull;
    procedure FillWithZero;
    procedure SetGrib(AFilename:string);
    procedure SetWGribPath(Pathname:string);
    function GenerateBinaryFilename(Year,Day,Hour:integer):string; overload;
    function GenerateBinaryFilename(DateTime:TDateTime):string; overload;
    function GenerateTextFilename(Year,Day,Hour:integer):string; overload;
    function GenerateTextFilename(DateTime:TDateTime):string; overload;
    procedure PutValue(LongIndex,LatIndex:byte;Value:real48);
    function GetValue(LongIndex,LatIndex:byte):real48;
    function IsNull(LongIndex,LatIndex:byte):boolean;
    function GetNullValue:real48;
    function Extract(Lat,Long:double;HaltOnBad:boolean):double;
    destructor Destroy;override;
    property Data[LongIndex,LatIndex:byte]:real48 read GetValue write PutValue;default;
    property NumLong:word read FNCEPNumLong;
    property NumLat:word read FNCEPNumLat;
  end;

  ENCEPError=class(exception);

function CalculateLevelInd(Level:word):word;
function CalculateLevel(LevelInd:word):word;
function LoadNCEPData(Data:TNCEPGrid;Year,Day,Hour:integer;CDDirec,CDDumpDirec,
  NCEPDirec,BinDirec:string;DeleteText,SaveBin:boolean;ProgLabel:TLabel):byte; overload;
function LoadNCEPData(Data:TNCEPGrid;DateTime:TDateTime;CDDirec,CDDumpDirec,
  NCEPDirec,BinDirec:string;DeleteText,SaveBin:boolean;ProgLabel:TLabel):byte; overload;
procedure LoadNCEPFromBinary(Data:TNCEPGrid;Filename:string;ProgLabel:TLabel);
procedure SaveNCEPData(Data:TNCEPGrid;Direc:string);
function GetCDYear(CDDirec:string):word;
function GetAbbrSpeciesName(ASpecies:TSpecies):string;
function GetFullSpeciesName(ASpecies:TSpecies):string;
function GetSpeciesFromFilename(Filename:string):TSpecies;
function GetLevelFromFilename(Filename:string):double;
function GetLevelTypeFromFilename(Filename:string):TLevelType;
function InsertSpeciesStr(InStr:string;ASpecies:TSpecies):string;

implementation

uses
  {$IFNDEF DELPHI5}
  DateUtils,
  {$ENDIF}

  controls,
  dialogs,
  {$WARNINGS OFF}
  filectrl,
  {$WARNINGS ON}
  forms,
  math,
  gb_datetime,
  gb_files,
  gb_strings;

function GetCDYear(CDDirec:string):word;
{------------------------------------------------------------------------------}
{ This function returns the year for the CD currently in the CD player.        }
{------------------------------------------------------------------------------}
var
  Source:textfile; {File handle for the date.dat file on this CD.}
  Filename:string; {The name of the file on the CD containing the date.}
begin
  FileMode:=0;
  Filename:=CDDirec+'\date.dat';
  if not FileExists(Filename) then raise ENCEPError.Create(
    'the date.dat file did not exist on the CD.');
  repeat
    assignfile(Source,Filename);
    try
      reset(Source);
    except
      on E:Exception do raise ENCEPError.Create('the year could not be read from'+
        ' the file named '+Filename+'. Exception message was: '+E.Message);
    end;
    read(Source,Result);
    closefile(Source);
  until Result<>0;
  if Result<100 then Result:=Result+1900;
  FileMode:=2;
end; {End of GetCDYear.}

function GetAbbrSpeciesName(ASpecies:TSpecies):string;
{------------------------------------------------------------------------------}
{ This function returns the abbreviated species name.                          }
{------------------------------------------------------------------------------}
begin
  case ASpecies of
    pv:Result:='pv';
    uw:Result:='uw';
    vw:Result:='vw';
    hgt:Result:='hgt';
    tmp:Result:='tmp';
    hmd:Result:='hmd';
    pottemp:Result:='pottemp';
    troppres:Result:='trpprs';
    pressure:Result:='prs';
    mntsf:Result:='mnt';
    vvel:Result:='vvel';
  end;
end; {End of GetAbbrSpeciesName.}

function GetFullSpeciesName(ASpecies:TSpecies):string;
{------------------------------------------------------------------------------}
{ This function returns the full species name.                                 }
{------------------------------------------------------------------------------}
begin
  case ASpecies of
    pv:Result:='potential vorticity';
    uw:Result:='zonal wind';
    vw:Result:='meridional wind';
    hgt:Result:='geopotential height';
    tmp:Result:='temperature';
    hmd:Result:='humidity';
    pottemp:Result:='potential temperature';
    troppres:Result:='tropopause pressure';
    pressure:Result:='pressure';
    mntsf:Result:='Montgomery stream function';
    vvel:Result:='vertical velocity';
  end;
end; {End of GetFullSpeciesName.}

constructor TNCEPGrid.Create(ASpecies:TSpecies;ALevelType:TLevelType;ALevel:double);
{------------------------------------------------------------------------------}
{ The expected binary file size is set in the constructor and the description  }
{ of the data to be stored in the object is set.                               }
{------------------------------------------------------------------------------}
begin
  FNCEPNumLong:=144;
  FNCEPNumLat:=73;
  Species:=ASpecies;
  LevelType:=ALevelType;
  if ((ALevelType=Surface) or (ALevelType=SeaLevel)) then Level:=-1 else Level:=ALevel;
  NullValue:=1E20; {Think very carefully before changing this value.}
  Year:=65535;     {See NCEP/NCAR project|Data arrangement thought in brain first.}
  Day:=65535;
  Hour:=65535;
  DataDesc:='';
  Grib:=nil;
  if Species=troppres then ExpectedBinFileSize:=21024 else ExpectedBinFileSize:=63072;
  case LevelType of
    PotTempLevel:DataDesc:=inttostr(round(Level))+' K ';
    PressureLevel:DataDesc:=inttostr(round(Level))+' hPa ';
    Surface:DataDesc:='surface ';
    SeaLevel:DataDesc:='sealevel ';
    AltitudeLevel:DataDesc:=inttostr(round(Level))+' m ';
  end;
  DataDesc:=DataDesc+GetFullSpeciesName(Species);
  WGribPath:='';
  inherited Create(2000000,FNCEPNumLong*FNCEPNumLat,sizeof(TRealType));
  SetAllDataToNull;
  NoGrib:=false;
end; {End of TNCEPGrid.Create.}

procedure TNCEPGrid.SetNull(ANullValue:real48);
begin
  NullValue:=ANullValue;
end;

procedure TNCEPGrid.SetAllDataToNull;
var
  I,J:word;
begin
  for I:=1 to FNCEPNumLong do
    for J:=1 to FNCEPNumLat do
      Index(I,J)^:=NullValue;
end; {End of TNCEPGrid.SetAllDataToNull.}

procedure TNCEPGrid.FillWithZero;
var
  I,J:word;
begin
  for I:=1 to FNCEPNumLong do
    for J:=1 to FNCEPNumLat do
      Index(I,J)^:=0.0;
end; {End of TNCEPGrid.FillWithZero.}

procedure TNCEPGrid.SetGrib(AFilename:string);
begin
  if not FileExists(AFilename) then raise ENCEPError.Create('the grib file '+
    'name passed to SetGrib doesn''t exist.');
  if Grib<>nil then Grib.Free;
  Grib:=TGrib.Create(AFilename,WGribPath,NoGrib);
end; {End of TNCEPGrid.SetSourceFilename.}

procedure TNCEPGrid.SetWGribPath(Pathname:string);
begin
  WGribPath:=Pathname;
end; {End of TNCEPGrid.SetWGribPath.}

function TNCEPGrid.GenerateBinaryFilename(Year,Day,Hour:integer):string;
var
  LevelStr,
  YearStr,
  HourStr,
  DayStr:string;
begin
  str(Year:4,YearStr);
  str(Day:3,DayStr);
  DayStr:=Replace(DayStr,' ','0');
  if Hour<0 then HourStr:='' else
  begin
    str(Hour:2,HourStr);
    HourStr:=Replace(HourStr,' ','0');
  end;
  if Level<0 then LevelStr:='' else
  begin
    str(round(Level):4,LevelStr);
    LevelStr:=Replace(LevelStr,' ','0');
  end;
  Result:=GetAbbrSpeciesName(Species)+YearStr+DayStr+HourStr+'.';
  case LevelType of
    PotTempLevel:Result:=Result+LevelStr+'K';
    PressureLevel:Result:=Result+LevelStr+'hPa';
    Surface:Result:=Result+'sfc';
    SeaLevel:Result:=Result+'sealevel';
    AltitudeLevel:Result:=Result+LevelStr+'m';
  end;
end;

function TNCEPGrid.GenerateBinaryFilename(DateTime:TDateTime):string;
var
  Year,
  Day,
  Hour:integer;
begin
  Year:=YearOf(DateTime);
  Day:=DayOfTheYear(DateTime);
  Hour:=HourOf(DateTime);
  Result:=GenerateBinaryFilename(Year,Day,Hour);
end;

function TNCEPGrid.GenerateTextFilename(Year,Day,Hour:integer):string;
var
  LevelStr,
  YearStr,
  HourStr,
  DayStr:string;
begin
  str(Year:4,YearStr);
  str(Day:3,DayStr);
  DayStr:=Replace(DayStr,' ','0');
  if Hour<0 then HourStr:='' else
  begin
    str(Hour:2,HourStr);
    HourStr:=Replace(HourStr,' ','0');
  end;
  if Level<0 then LevelStr:='' else
  begin
    str(round(Level):4,LevelStr);
    LevelStr:=Replace(LevelStr,' ','0');
  end;
  Result:=GetAbbrSpeciesName(Species)+YearStr+DayStr+HourStr+'_text.';
  case LevelType of
    PotTempLevel:Result:=Result+LevelStr+'K';
    PressureLevel:Result:=Result+LevelStr+'hPa';
    Surface:Result:=Result+'sfc';
    SeaLevel:Result:=Result+'sealevel';
    AltitudeLevel:Result:=Result+LevelStr+'m';
  end;
end;

function TNCEPGrid.GenerateTextFilename(DateTime:TDateTime):string;
var
  Year,
  Day,
  Hour:integer;
begin
  Year:=YearOf(DateTime);
  Day:=DayOfTheYear(DateTime);
  Hour:=HourOf(DateTime);
  Result:=GenerateTextFilename(Year,Day,Hour);
end;

function TNCEPGrid.Index(X,Y:byte):TRealPtr;
begin
  if X<1 then raise ENCEPError.Create('the X value passed to TNCEPGrid.Index was <1.');
  if X>FNCEPNumLong then raise ENCEPError.Create('the X value passed to TNCEPGrid.Index was >'+inttostr(FNCEPNumLong)+'.');
  if Y<1 then raise ENCEPError.Create('the Y value passed to TNCEPGrid.Index was <1.');
  if Y>FNCEPNumLat then raise ENCEPError.Create('the Y value passed to TNCEPGrid.Index was >'+inttostr(FNCEPNumLat)+'.');
  Index:=TRealPtr(inherited Index((X-1)*FNCEPNumLat+Y-1));
end; {End of TNCEPGrid.Index.}

function TNCEPGrid.IsNull(LongIndex,LatIndex:byte):boolean;
begin
  if abs(Index(LongIndex,LatIndex)^)>=NullValue then Result:=true else Result:=false;
end;

procedure TNCEPGrid.PutValue(LongIndex,LatIndex:byte;Value:real48);
begin
  Index(LongIndex,LatIndex)^:=Value;
end;

function TNCEPGrid.GetValue(LongIndex,LatIndex:byte):real48;
begin
  Result:=Index(LongIndex,LatIndex)^;
end;

function TNCEPGrid.GetNullValue:real48;
begin
  Result:=NullValue;
end;

function TNCEPGrid.Extract(Lat,Long:double;HaltOnBad:boolean):double;
{------------------------------------------------------------------------------}
{ This function uses bilinear interpolation to calculate the field value at    }
{ the latitude and longitude specified by 'Lat' and 'Long' respectively. If    }
{ any of the 4 corner values are null and 'HaltOnBad' is set to true , the     }
{ routine shows an error message and the application halts; if 'HaltOnBad' is  }
{ false an exception is raised.                                                }
{------------------------------------------------------------------------------}
var
  X1,X2,Y:byte;
  D1,D2,W:double;
  BadData:boolean;
begin
  Result:=NullValue;
  if Lat=90.0 then Result:=Index(1,1)^;
  if Lat=-90.0 then Result:=Index(1,FNCEPNumLat)^;
  if ((Lat<90.0) and (Lat>-90.0)) then
  begin
    if Long<0.0 then Long:=Long+360.0;
    X1:=trunc(Long/2.5)+1;
    if X1=FNCEPNumLong then X2:=1 else X2:=X1+1;
    Y:=trunc((90-Lat)/2.5)+1;
    BadData:=false;
    if abs(Index(X1,Y)^)>=NullValue then BadData:=true;
    if abs(Index(X1,Y+1)^)>=NullValue then BadData:=true;
    if abs(Index(X2,Y)^)>=NullValue then BadData:=true;
    if abs(Index(X2,Y+1)^)>=NullValue then BadData:=true;
    if BadData then Result:=NullValue else
    begin
      D1:=frac(Long/2.5)*(Index(X2,Y)^-Index(X1,Y)^)+Index(X1,Y)^;
      D2:=frac(Long/2.5)*(Index(X2,Y+1)^-Index(X1,Y+1)^)+Index(X1,Y+1)^;
      W:=((92.5-Y*2.5)-Lat)/2.5;
      Result:=D1+W*(D2-D1);
    end;
    if Result>=NullValue then
      if HaltOnBad then
      begin
        MessageDlg('Bad data was true in TNCEPGrid.Extract. The program will '+
          'now terminate',mtError,[mbOK],0);
        Application.Terminate;
      end
      else raise ENCEPError.Create('one of the values in the '+DataDesc+
        ' NCEP data grid was null.');
  end;
end; {End of TNCEPGrid.Extract.}

destructor TNCEPGrid.Destroy;
begin
  if Grib<>nil then Grib.Free;
  inherited Destroy;
end;

function CalculateLevelInd(Level:word):word;
begin
  case Level of
    270:Result:=1;
    280:Result:=2;
    290:Result:=3;
    300:Result:=4;
    315:Result:=5;
    330:Result:=6;
    350:Result:=7;
    400:Result:=8;
    450:Result:=9;
    550:Result:=10;
    650:Result:=11;
  else
    Result:=0;
  end;
end;

function CalculateLevel(LevelInd:word):word;
begin
  Result:=Levels[LevelInd];
end;

procedure SetProgLabel(ProgLabel:TLabel;Str:string);
begin
  if ProgLabel<>nil then
  begin
    ProgLabel.Caption:=Str;
    ProgLabel.Refresh;
    Application.ProcessMessages;
  end;
end;

procedure AppendToProgLabel(ProgLabel:TLabel;Str:string);
begin
  if ProgLabel<>nil then
  begin
    ProgLabel.Caption:=ProgLabel.Caption+Str;
    ProgLabel.Refresh;
    Application.ProcessMessages;
  end;
end;

procedure LoadNCEPFromBinary(Data:TNCEPGrid;Filename:string;ProgLabel:TLabel);
{------------------------------------------------------------------------------}
{ This routine loads data from the binary file specified by 'Filename' into    }
{ the NCEP grid object specified by 'Data'. Provision needs to be made for the }
{ possibility that the binary file stores the data as type word and not as     }
{ type real. So we create a word array object, read the data into that, trans- }
{ fer the data across to our real array object (the ancestor for TNCEPGrid)    }
{ and then delete the word array object.                                       }
{------------------------------------------------------------------------------}
var
  I,J:word; {Loop control variables.}
  BinSource:file; {The file handle for the binary file.}
  WordData:T2DWordArray; {Used to read in data from a binary file of 2 byte words.}
begin
  SetProgLabel(ProgLabel,'Loading binary file');
  FileMode:=0;
  assignfile(BinSource,Filename);
  reset(BinSource,1);
  if filesize(BinSource)<>Data.ExpectedBinFileSize then raise ENCEPError.Create(
    'the expected binary file size was '+inttostr(Data.ExpectedBinFileSize)+
    ' but the actual file size was '+inttostr(filesize(BinSource)));
  if Data.ExpectedBinFileSize=21024 then
  begin
    WordData:=T2DWordArray.Create(Data.NumLong,Data.NumLat);
    try
      if WordData.LoadData(BinSource)=1 then raise ENCEPError.Create(
        'the data couldn''t be read in from the binary word data file.');
      for I:=1 to Data.NumLong do
        for J:=1 to Data.NumLat do
          Data.Index(I,J)^:=WordData[I,J];
    finally
      WordData.Free;
    end;
  end
  else if Data.LoadData(BinSource)=1 then raise ENCEPError.Create(
    'the data couldn''t be read in from the binary real data file.');
  closefile(BinSource);
  FileMode:=2;
end; {End of LoadNCEPFromBinary.}

procedure ConfirmCD(Direc,GribFilename:string;Year:word);
var
  CDYear:word; {The year for the data on the current CD.}
  ValidCD:boolean; {True if the CD in the CD directory is an NCEP/NCAR CD.}
begin
  while not DirectoryExists(Direc) do
  begin
    if MessageDlg('There is no CD in the CD drive. Please insert the CD for '+
      inttostr(Year),mtInformation,[mbOK,mbCancel],0)=mrcancel then raise
      ENCEPError.Create('cancel was pressed.');
  end;
  ValidCD:=true;
  CDYear:=0;
  repeat
    try
      CDYear:=GetCDYear(Direc);
    except on ENCEPError do
      begin
        ValidCD:=false;
        if MessageDlg('The CD in the '+Direc+' drive doesn''t appear to be a valid '+
          'NCEP/NCAR CD. Either insert a new CD and press retry or press cancel '+
          'to exit this program.',mtError,[mbRetry,mbCancel],0)=mrCancel then raise
          ENCEPError.Create('cancel was pressed.');
      end;
    end;
  until ValidCD;
  while CDYear<>Year do
  begin
    if MessageDlg('The current CD is for '+inttostr(CDYear)+'. Please insert the CD for '+
      inttostr(Year),mtInformation,[mbOK,mbCancel],0)=mrcancel then raise
          ENCEPError.Create('cancel was pressed.');
    CDYear:=GetCDYear(Direc);
  end;
  if not FileExists(GribFilename) then raise ENCEPError.Create(
    'the source file named '+GribFilename+' couldn''t be found on the CD.');
end; {End of ConfirmCD.}

procedure ConvertSurfTextFile(Filename:string);
var
  Source,
  Dest:textfile;
  Data:T2DDoubleArray;
  X,Y,
  I,J:integer;
  Value,
  Long,
  Lat,
  D1,D2:double;
begin
  Data:=T2DDoubleArray.Create(192,96); {Extra 2 rows for repeating first and last row.}
  try
    assignfile(Source,Filename);
    reset(Source);
    assignfile(Dest,FindJunkFile);
    rewrite(Dest);
    readln(Source);
    writeln(Dest,'144 73');
    for I:=2 to 95 do
      for J:=1 to 192 do
      begin
        readln(Source,Value);
        Data[J,I]:=Value;
      end;
    for J:=1 to 192 do Data[J,1]:=Data[J,2];
    for J:=1 to 192 do Data[J,96]:=Data[J,95];
    closefile(Source);
    erase(Source);
    for I:=1 to 73 do
      for J:=1 to 144 do
      begin
        Long:=(J-1)*2.5;
        Lat:=90-(I-1)*2.5;
        X:=trunc(Long/1.875)+1;
        Y:=trunc((90.74704-Lat)/1.90464)+1;
        D1:=frac(Long/1.875)*(Data[X+1,Y]-Data[X,Y])+Data[X,Y];
        D2:=frac(Long/1.875)*(Data[X+1,Y+1]-Data[X,Y+1])+Data[X,Y+1];
        writeln(Dest,frac((90.44664-Lat)/1.90464)*(D2-D1)+D1:10:4);
      end;
    closefile(Dest);
    rename(Dest,Filename);
  finally
    Data.Free;
  end;
end; {End of ConvertSurfTextFile.}

function InsertSpeciesStr(InStr:string;ASpecies:TSpecies):string;
var
  I:word;
begin
  I:=pos('SSS',InStr);
  while I>0 do
  begin
    Delete(InStr,I,3);
    Insert(GetAbbrSpeciesName(ASpecies),InStr,I);
    I:=pos('SSS',InStr);
  end;
  Result:=InStr;
end;

function GetSpeciesFromFilename(Filename:string):TSpecies;
begin
  Result:=pv;
  if copy(Filename,1,2)='pv' then Result:=pv;
  if copy(Filename,1,2)='uw' then Result:=uw;
  if copy(Filename,1,2)='vw' then Result:=vw;
  if copy(Filename,1,3)='hgt' then Result:=hgt;
  if copy(Filename,1,3)='tmp' then Result:=tmp;
  if copy(Filename,1,7)='pottemp' then Result:=pottemp;
  if copy(Filename,1,6)='trpprs' then Result:=troppres;
  if copy(Filename,1,3)='prs' then Result:=pressure;
  if copy(Filename,1,3)='mnt' then Result:=mntsf;
  if copy(Filename,1,4)='vvel' then Result:=vvel;
end;

function GetLevelFromFilename(Filename:string):double;
begin
  Result:=strtofloat(copy(ExtractFileExt(Filename),2,4));
end;

function GetLevelTypeFromFilename(Filename:string):TLevelType;
var
  FileExt:string;
begin
  Result:=PotTempLevel;
  FileExt:=ExtractFileExt(Filename);
  FileExt:=copy(FileExt,5,length(FileExt)-4);
  if FileExt='K' then Result:=PotTempLevel;
  if FileExt='hPa' then Result:=PressureLevel;
  if FileExt='sfc' then Result:=Surface;
  if FileExt='sealevel' then Result:=SeaLevel;
  if FileExt='m' then Result:=AltitudeLevel;
end;

{$HINTS OFF}
function LoadNCEPData(Data:TNCEPGrid;Year,Day,Hour:integer;CDDirec,CDDumpDirec,
  NCEPDirec,BinDirec:string;DeleteText,SaveBin:boolean;ProgLabel:TLabel):byte;
{------------------------------------------------------------------------------}
{ This function loads data for the specified year, day and possibly hour (if   }
{ the data temporal resolution is higher than 24 hours) into the NCEP data     }
{ object specified by 'Data'. If the data are at 24 hour temporal resolution   }
{ or courser, the value for hour should be set negative so that the names of   }
{ NCEP/NCAR text and binary data files do not include the hour. The routine    }
{ first checks if the binary file is available in the directory specified by   }
{ 'BinDirec'. If it isn't then it checks if the NCEP/NCAR text file is avail-  }
{ able in the directory specified by 'NCEPDirec'. If the text file isn't       }
{ available then a call is made to the MakeTextFile method of the Grib object. }
{ If the grib file is available in the 'CDDumpDirec' directory then that is    }
{ used (the file is assumed to be in CDDumpDirec+'\'+YearStr+'\'), otherwise   }
{ the grib file is assumed to be on the CD whose drive is specified by         }
{ 'CDDirec' (the file is assumed to be in CDDirec+'\data\at00z\' for all files }
{ except geopotential height and CDDirec+'\data\at12z\' for geopotential       }
{ height). If the CD file is required and doesn't exist, an exception is       }
{ raised. The name of the binary grib file is generated automatically. However }
{ the name of the grib file to be used can be overridden by a call to the      }
{ SetGrib method. This makes provision e.g. for one grib file per month. If    }
{ 'DeleteText' is true and the data were read from the NCEP/NCAR text data     }
{ file, the text file will be deleted. If 'SaveBin' is true, and the binary    }
{ files does not exist, the binary file version of the data will be saved. If  }
{ the object stores tropopause height data, it is assumed that the binary file }
{ from which the data will be read will be 21024 bytes in size. An exception   }
{ is raised if the size of the binary file (if it is available) is not what    }
{ was expected given the species that the object stores. The function returns  }
{ the following codes:                                                         }
{ 0: The data were loaded from the binary file.                                }
{ 1: The data were loaded from the text file.                                  }
{ 2: The data were loaded from the grib file.                                  }
{------------------------------------------------------------------------------}
var
  SetYear,  {The year passed to Grib.MakeTextFile.}
  SetDay,   {The day passed to Grib.MakeTextFile.}
  SetHour,  {The hour passed to Grib.MakeTextFile.}
  SetLevel, {The level passed to Grib.MakeTextFile.}
  Month,    {The month number.}
  I,J:word; {Loop control variables through grid indices.}
  TestStr,        {Used to test if the file must be converted.}
  GribFilename1,  {The first possible grib file name - as on CDs.}
  GribFilename2,  {The second possible grib file name - as on NCEP grib store.}
  GribFilename,   {The grib file finally selected.}
  BinFilename,    {The name of the binary data file.}
  TextFilename,   {The name of the text NCEP/NCAR file.}
  LevelStr,       {The 4 digit string denoting the level.}
  MonthStr,       {The 2 digit string denoting the month.}
  YearStr:string; {The 2 digit string for the year.}
  Source:textfile; {The file handle for the input NCEP/NCAR text file.}
  TextFileMade,
  FirstTry:boolean;
begin
  Data.SetAllDataToNull;

  {Work out the name of the binary file.}
  str(Year:4,YearStr);
  CDDirec:=InsertYear(CDDirec,YearStr);
  CDDumpDirec:=InsertYear(CDDumpDirec,YearStr);
  NCEPDirec:=InsertYear(NCEPDirec,YearStr);
  BinDirec:=InsertSpeciesStr(InsertYear(BinDirec,YearStr),Data.Species);
  BinFilename:=BinDirec+'\'+Data.GenerateBinaryFilename(Year,Day,Hour);
  TextFilename:=NCEPDirec+'\'+Data.GenerateTextFilename(Year,Day,Hour);
  if FileExists(BinFilename) then
  begin
    LoadNCEPFromBinary(Data,BinFilename,ProgLabel);
    Data.Year:=Year;
    Data.Day:=Day;
    if Hour>-1 then Data.Hour:=Hour;
    Result:=0;
    if (SaveBin and DeleteText) then DeleteFile(TextFilename);
  end
  else
  begin
    {Generate the name of the text file which may already exist}
    {or will otherwise be created through a call to wgrib.}
    SetProgLabel(ProgLabel,'Loading '+GetFullSpeciesName(Data.Species)+' data from ');
    if FileExists(TextFilename) then
    begin
      Result:=1;
      AppendToProgLabel(ProgLabel,'disk');
    end
    else
    begin
      Result:=2;
      FirstTry:=true;
      repeat
        TextFileMade:=true;

        {If a grib file hasn't been specified then open one. Generate a default}
        {grib file name and store the result in GribFilename1.}
        if Data.Grib=nil then
        begin
          if Data.Level<0 then LevelStr:='' else
            str(round(Data.Level):trunc(log10(Data.Level))+1,LevelStr);
          case Data.Species of
            pv:GribFilename1:='pv'+LevelStr;
            uw:GribFilename1:='ugrd'+LevelStr;
            vw:GribFilename1:='vgrd'+LevelStr;
            hgt:if Data.Level<=100 then GribFilename1:='hgt'+LevelStr else GribFilename1:='hgt';
            tmp:if Data.LevelType=AltitudeLevel then GribFilename1:='tmp.'+LevelStr+'m' else
                GribFilename1:='tmp'+LevelStr;
            pottemp:GribFilename1:='theta'+LevelStr;
            troppres:GribFilename1:='trpprs';
            pressure:GribFilename1:='pres'+LevelStr;
            mntsf:GribFilename1:='mntsf'+LevelStr;
            vvel:GribFilename1:='vvel'+LevelStr;
          end;
          case Data.LevelType of
            PotTempLevel:GribFilename1:=GribFilename1+'.tht';
            PressureLevel:GribFilename1:=GribFilename1+'.prs';
            Surface:GribFilename1:=GribFilename1+'.sfc';
          end;

          if Data.LevelType=PotTempLevel then
          begin
            Month:=CalcMonth(Year,Day);
            str(Month:2,MonthStr);
            MonthStr:=Replace(MonthStr,' ','0');
            str(Year mod 100:2,YearStr);
            YearStr:=Replace(YearStr,' ','0');
            case Data.Species of
              mntsf:GribFilename2:='mntsftht.'+LevelStr+'.'+YearStr+MonthStr;
              pv:GribFilename2:='pv___tht.'+LevelStr+'.'+YearStr+MonthStr;
              uw:GribFilename2:='ugrdtht.'+LevelStr+'.'+YearStr+MonthStr;
              vw:GribFilename2:='vgrdtht.'+LevelStr+'.'+YearStr+MonthStr;
              tmp:GribFilename2:='tmptht.'+LevelStr+'.'+YearStr+MonthStr;
              vvel:GribFilename2:='vveltht.'+LevelStr+'.'+YearStr+MonthStr;
            end;
          end
          else GribFilename2:='';

          {There are 4 possibilities here: GribFilename1 or GribFilename2, CDDirec or CDDumpDirec}
          GribFilename:='';
          I:=0;
          repeat
            inc(I);
            case I of
              1:GribFilename:=CDDumpDirec+'\'+GribFilename1;
              2:begin
                  case Data.Species of
                    pv,uw,vw,tmp:begin
                      if Data.LevelType=PotTempLevel then GribFilename:=CDDirec+'\data\at00z\'+GribFilename1;
                      if Data.LevelType=PressureLevel then GribFilename:=CDDirec+'\data\at12z\'+GribFilename1;
                      if Data.LevelType=AltitudeLevel then GribFilename:=CDDirec+'\data\at00z12z\'+GribFilename1;
                    end;
                    hgt:begin
                      if ((Data.Level>0) and (Data.Level<=100)) then GribFilename:=CDDirec+'\data\at12z\'+GribFilename1;
                      if Data.Level>100 then GribFilename:=CDDirec+'\data\at00z12z\'+GribFilename1;
                      if Data.Level<0 then GribFilename:=CDDirec+'\data\fixed\'+GribFilename1;
                    end;
                    pressure:begin
                      if Data.LevelType=Surface then GribFilename:=CDDirec+'\data\at00z12z\'+GribFilename1;
                    end;
                  end;
                end;
              3:GribFilename:=CDDumpDirec+'\'+GribFilename2;
              4:GribFilename:=CDDirec+'\'+GribFilename2;
            end;
            if FileExists(GribFilename) then
            begin
              case I of
                1,3:AppendToProgLabel(ProgLabel,'cd dump');
                2,4:AppendToProgLabel(ProgLabel,'CD');
              end;
            end
            else GribFilename:='';
          until ((GribFilename<>'') or (I=4));
          if GribFilename='' then raise ENCEPError.Create('neither of the two possible grib files ('+
            GribFilename1+','+GribFilename2+') were available in the two possible directories ('+
            CDDumpDirec+','+CDDirec+')');
          Data.SetGrib(GribFilename);
        end
        else AppendToProgLabel(ProgLabel,'specified grib file');
        SetYear:=Year;
        SetDay:=Day;
        if Hour<0 then SetHour:=65535 else SetHour:=Hour;
        if Data.Level<0.0 then SetLevel:=65535 else SetLevel:=round(Data.Level);
        if ((Data.LevelType=Surface) and (Data.Species=hgt)) then
        begin
          SetYear:=65535;
          SetDay:=65535;
          SetHour:=65535;
        end;
        try
          Data.Grib.MakeTextFile(TextFilename,SetYear,SetDay,SetHour,SetLevel,Data.NoGrib);
        except
          on E:EGrib do
          begin
            if not FirstTry then raise ENCEPError.Create('a grib object containing the '+
            'required data couldn''t be created because '+E.Message);
            FirstTry:=false;
            TextFileMade:=false;
            Data.Grib.Free;
            Data.Grib:=nil;
          end;
        end;
      until TextFileMade;
    end;

    {Now read the data in from the text file.}
    assignfile(Source,TextFilename);
    reset(Source);
    readln(Source,TestStr);
    closefile(Source);
    if ((strtoint(GetWord(TestStr,' ',1))=192) and (strtoint(GetWord(TestStr,' ',2))=94)) then
      ConvertSurfTextFile(TextFilename);
    assignfile(Source,TextFilename);
    reset(Source);
    readln(Source);
    {For information on the format of the text files see programs/wgrib/formats.txt}
    {on any NCEP/NCAR CD.}
    for J:=1 to Data.NumLat do
      for I:=1 to Data.NumLong do
        readln(Source,Data.Index(I,J)^);
    closefile(Source);
    if Data.Species=pv then
    for J:=1 to Data.NumLat do
      for I:=1 to Data.NumLong do {Note need to multiply by theta value to make these values}
        Data.Index(I,J)^:=Data.Level*1E6*Data.Index(I,J)^;     {comparable to UKMO.}
    Data.Year:=Year;
    Data.Day:=Day;
    if Hour>-1 then Data.Hour:=Hour;
    if SaveBin then SaveNCEPData(Data,BinDirec);
    if DeleteText then erase(Source);
  end;
end; {End of LoadNCEPData.}
{$HINTS ON}

function LoadNCEPData(Data:TNCEPGrid;DateTime:TDateTime;CDDirec,CDDumpDirec,
  NCEPDirec,BinDirec:string;DeleteText,SaveBin:boolean;ProgLabel:TLabel):byte;
{------------------------------------------------------------------------------}
{ This function simply provides an alternative for LoadNCEPData above that     }
{ specifies the date and time as a TDateTimevariable.                          }
{------------------------------------------------------------------------------}
{$IFDEF DELPHI5}
var
  Year,Month,Date,Hour,Min,Sec,MSec:word;
{$ENDIF}
begin
  {$IFDEF DELPHI5}
  DecodeDate(DateTime,Year,Month,Date);
  DecodeTime(DateTime,Hour,Min,Sec,MSec);
  Result:=LoadNCEPData(Data,Year,CalcDay(Year,Month,Date),Hour,
    CDDirec,CDDumpDirec,NCEPDirec,BinDirec,DeleteText,SaveBin,ProgLabel);
  {$ELSE}
  Result:=LoadNCEPData(Data,YearOf(DateTime),DayOfTheYear(DateTime),HourOf(DateTime),
    CDDirec,CDDumpDirec,NCEPDirec,BinDirec,DeleteText,SaveBin,ProgLabel);
  {$ENDIF}
end;

procedure SaveNCEPData(Data:TNCEPGrid;Direc:string);
{------------------------------------------------------------------------------}
{ This routine saves the data held in the NCEP grid specified by 'Data' in     }
{ the directory specified by 'Direc'. The name of the binary file is generated }
{ automatically. If the expected binary file size is 21024 then the data are   }
{ saved as two byte words, otherwise they are saved as 6 bytes reals.          }
{------------------------------------------------------------------------------}
var
  WordData:T2DWordArray;
  BinFilename,
  LevelStr,
  YearStr,
  HourStr,
  DayStr:string;
  Dest:file;
  I,J:word;
begin
  str(Data.Year:4,YearStr);
  Direc:=InsertYear(Direc,YearStr);
  str(Data.Day:3,DayStr);
  DayStr:=Replace(DayStr,' ','0');
  if Data.Hour=65535 then HourStr:='' else
  begin
    str(Data.Hour:2,HourStr);
    HourStr:=Replace(HourStr,' ','0');
  end;
  if Data.Level<0 then LevelStr:='' else
  begin
    str(round(Data.Level):4,LevelStr);
    LevelStr:=Replace(LevelStr,' ','0');
  end;
  BinFilename:=GetAbbrSpeciesName(Data.Species)+YearStr+DayStr+HourStr+'.';
  case Data.LevelType of
    PotTempLevel:BinFilename:=BinFilename+LevelStr+'K';
    PressureLevel:BinFilename:=BinFilename+LevelStr+'hPa';
    Surface:BinFilename:=BinFilename+'sfc';
    SeaLevel:BinFilename:=BinFilename+'sealevel';
    AltitudeLevel:BinFilename:=BinFilename+LevelStr+'m';
  end;
  if not DirectoryExists(Direc) then ForceDirectories(Direc);
  assignfile(Dest,Direc+'\'+BinFilename);
  rewrite(Dest,1);
  if Data.ExpectedBinFileSize=21024 then
  begin
    WordData:=T2DWordArray.Create(Data.NumLong,Data.NumLat);
    try
      for I:=1 to Data.NumLong do
        for J:=1 to Data.NumLat do
          WordData[I,J]:=round(Data.Index(I,J)^);
      WordData.SaveData(Dest);
    finally
      WordData.Free;
    end;
  end
  else Data.SaveData(Dest);
  closefile(Dest);
end; {End of SaveNCEPData.}

end.
