//@***h* BDBP/gb_xysignal.pas
// 
// NAME
//   gb_xysignal.pas
//
// PURPOSE
//  Defines a class which is used to and XY data series. It is a 
//  general form of the TVertProfile object defined in gb_profile.pas.
//
// DESCRIPTION
//  This unit defines a class which is used to and XY data series. It is a 
//  general form of the TVertProfile object defined in gb_profile.pas.  
//  
//
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_xysignal;

{$I UNITDIREC.INC}

interface

uses
  sysutils,
  gb_array;

type
  XYRec=record
    XValue,
    YValue:double;
  end;

  TXYSignal=class
  private
    DistToValue,    {The distance to the nearest value when doing interpolation.}
    IntDist:double; {The distance beteen neighbouring levels used in interpolation.}
    IntIndexDist:integer; {The distance between neighbouring indices used in interpolation.}
  protected
    FNullXValue,        {The value denoting the null value for the X data.}
    FNullYValue:double; {The value denoting the null value for the Y data.}
    IndexStep,    {Amount by which the two interpolation indicies are changed.}
    Int1,         {These two interpolation indices are used to}
    Int2:integer; {speed up searches using the GetXValue and GetYValue methods.}
    FHuntWhenInterp,    {True if non-null values must be hunted for when InterpolateYValue is called.}
    FSortedByX,         {True if profile has been sorted by X values.}
    FShowExceptions,    {True if raising of serious exceptions should be displayed in a message box.}
    FRemoveNullXOnInterpolate, {True if null X values should be removed when interpolating.}
    FSortedByY:boolean; {True if profile has been sorted by Y values.}

    {The array below holds pointers to XDataArray in position [0] and YDataArray in}
    {position [1]. The array is protected so that it is available to descendant objects.}
    FData:array[0..1] of PDoubleArray;

    function ProcessArrayIndex(Index:longint):longint;
  public
    constructor Create(Dim:cardinal;ANullXValue,ANullYValue:double); overload;
    constructor Create(ANullXValue,ANullYValue:double); overload;
    constructor Create(Source:TXYSignal); overload;
    procedure Clear;
    procedure DeepCopy(Source:TXYSignal);
    procedure AddCoord(X,Y:double);virtual;
    procedure PutCoord(Index:longint;X,Y:double);overload;
    procedure PutCoord(Index:longint;Coord:XYRec);overload;
    procedure PutXValue(Index:longint;X:double);virtual;
    procedure PutYValue(Index:longint;Y:double);
    procedure AddToYValue(Index:longint;ToAdd:double);
    procedure PutYForGivenX(X,Y:double);
    procedure PutXForGivenY(X,Y:double);
    procedure FillWithNull;
    procedure FillWithZero;
    procedure GaussianFilter(Width:double);
    function  GetCoord(Index:longint):XYRec;
    function  GetFirstCoord:XYRec;
    function  GetLastCoord:XYRec;
    function  GetXValue(Index:longint):double;
    function  GetYValue(Index:longint):double;
    function  CalcMeanYValue:double;
    function  CalcStdvOfYValues:double;
    function  GetNumData:longint;
    function  IsNullX(Index:longint):boolean;
    function  IsNullY(Index:longint):boolean;
    function  InterpolateYValue(X:double;LogInter,ReturnNullOnError:boolean):double;
    function  GetIntDist:double;
    function  GetIntIndexDiff:integer;
    function  GetDistToValue:double;
    procedure SortByX;virtual;
    procedure RegridByX(StartX,EndX,Resolution:double);
    function  RemoveAllNullXEntries:boolean;
    procedure WriteToTextFile(Filename:string);
    destructor Destroy; override;

    {Properties}
    property NullXValue:double read FNullXValue write FNullXValue;
    property NullYValue:double read FNullYValue write FNullYValue;
    property HuntWhenInterpolating:boolean read FHuntWhenInterp write FHuntWhenInterp;
    property ShowExceptions:boolean read FShowExceptions write FShowExceptions;
    property RemoveNullXOnInterpolate:boolean read FRemoveNullXOnInterpolate write FRemoveNullXOnInterpolate;
    property SortedByX:boolean read FSortedByX;
  end;

  EXYSignal=class(Exception);

  TXYSignalArray=array of TXYSignal;

implementation

uses
  Dialogs,
  gb_math;

type
  EInterpol=class(exception);

constructor TXYSignal.Create(Dim:cardinal;ANullXValue,ANullYValue:double);
begin
  Create(ANullXValue,ANullYValue);
  setlength(FData[0]^,Dim);
  setlength(FData[1]^,Dim);
  if Dim=0 then Int2:=0 else Int2:=Dim-1;
  FillWithNull;
end;

constructor TXYSignal.Create(ANullXValue,ANullYValue:double);
begin
  {These are two crucial lines of code.}
  new(FData[0]);
  new(FData[1]);

  Int1:=0;
  Int2:=0;
  IndexStep:=1;
  FSortedByX:=false;
  FSortedByY:=false;
  FNullXValue:=ANullXValue;
  FNullYValue:=ANullYValue;
  FHuntWhenInterp:=true;
  FShowExceptions:=false;
  FRemoveNullXOnInterpolate:=false;
end;

constructor TXYSignal.Create(Source:TXYSignal);
begin
  if Assigned(Source) then
  begin
    {These are two crucial lines of code.}
    new(FData[0]);
    new(FData[1]);

    DeepCopy(Source);
  end
  else raise EXYSignal.Create('the object passed to Create was nil.');
end; {End of TXYSignal.Create.}

destructor TXYSignal.Destroy;
begin
  dispose(FData[0]);
  dispose(FData[1]);
  inherited Destroy;
end;

function TXYSignal.ProcessArrayIndex(Index:longint):longint;
begin
  dec(Index);
  if Index<Low(FData[0]^) then raise EXYSignal.Create(
    'the Index parameter was less than '+inttostr(Low(FData[0]^)));
  if Index>High(FData[0]^) then raise EXYSignal.Create(
    'the Index parameter was greater than '+inttostr(High(FData[0]^)));
  Result:=Index;
end;

procedure TXYSignal.DeepCopy(Source:TXYSignal);
var
  I:integer;
begin
  Int1:=0;
  Int2:=0;
  IndexStep:=1;
  FSortedByX:=false;
  FSortedByY:=false;
  FNullXValue:=Source.NullXValue;
  FNullYValue:=Source.NullYValue;
  FHuntWhenInterp:=Source.FHuntWhenInterp;
  FShowExceptions:=Source.FShowExceptions;
  FRemoveNullXOnInterpolate:=Source.FRemoveNullXOnInterpolate;
  SetLength(FData[0]^,Source.GetNumData);
  SetLength(FData[1]^,Source.GetNumData);
  for I:=1 to length(FData[0]^) do PutCoord(I,Source.GetCoord(I));
end; {End of TXYSignal.DeepCopy.}

procedure TXYSignal.Clear;
begin
  SetLength(FData[0]^,0);
  SetLength(FData[1]^,0);
  Int1:=0;
  Int2:=0;
end;

procedure TXYSignal.AddCoord(X,Y:double);
begin
  Setlength(FData[0]^,length(FData[0]^)+1);
  Setlength(FData[1]^,length(FData[1]^)+1);
  FData[0]^[High(FData[0]^)]:=X;
  FData[1]^[High(FData[1]^)]:=Y;
  FSortedByX:=false;
  FSortedByY:=false;
end;

procedure TXYSignal.PutCoord(Index:longint;X,Y:double);
begin
  dec(Index);
  if Index>High(FData[0]^) then Setlength(FData[0]^,Index+1);
  if Index>High(FData[1]^) then Setlength(FData[1]^,Index+1);
  FData[0]^[Index]:=X;
  FData[1]^[Index]:=Y;
  FSortedByX:=false;
  FSortedByY:=false;
end;

procedure TXYSignal.PutCoord(Index:longint;Coord:XYRec);
begin
  PutCoord(Index,Coord.XValue,Coord.YValue);
end;

procedure TXYSignal.PutXValue(Index:longint;X:double);
begin
  Index:=ProcessArrayIndex(Index);
  FData[0]^[Index]:=X;
  FSortedByX:=false;
end;

procedure TXYSignal.PutYValue(Index:longint;Y:double);
begin
  Index:=ProcessArrayIndex(Index);
  FData[1]^[Index]:=Y;
  FSortedByY:=false;
end;

procedure TXYSignal.AddToYValue(Index:longint;ToAdd:double);
begin
  Index:=ProcessArrayIndex(Index);
  FData[1]^[Index]:=FData[1]^[Index]+ToAdd;
  FSortedByY:=false;
end;

function TXYSignal.GetCoord(Index:longint):XYRec;
begin
  Result.XValue:=GetXValue(Index);
  Result.YValue:=GetYValue(Index);
end;

function TXYSignal.GetFirstCoord:XYRec;
begin
  Result:=GetCoord(1);
end;

function TXYSignal.GetLastCoord:XYRec;
begin
  Result:=GetCoord(length(FData[0]^));
end;

function TXYSignal.GetYValue(Index:longint):double;
begin
  Index:=ProcessArrayIndex(Index);
  Result:=FData[1]^[Index];
end;

function TXYSignal.CalcMeanYValue:double;
var
  I:longint;
begin
  Result:=0.0;
  for I:=Low(FData[1]^) to High(FData[1]^) do Result:=Result+FData[1]^[I];
  Result:=Result/length(FData[1]^);
end;

function TXYSignal.CalcStdvOfYValues:double;
var
  I:longint;
  MeanY:double;
begin
  if length(FData[1]^)<2 then raise EXYSignal.Create(
    'the number of data available was less than 2 in the call to CalcStdvOfYValues.');
  Result:=0.0;
  MeanY:=CalcMeanYValue;
  for I:=Low(FData[1]^) to High(FData[1]^) do Result:=Result+sqr(MeanY-FData[1]^[I]);
  Result:=sqrt(Result/(length(FData[1]^)-1));
end;

function TXYSignal.GetXValue(Index:longint):double;
{This function assumes that the index pass is 1 based not zero based.}
begin
  Index:=ProcessArrayIndex(Index);
  Result:=FData[0]^[Index];
end;

procedure TXYSignal.PutYForGivenX(X,Y:double);
var
  I:longint;
  Found:boolean;
begin
  I:=low(FData[0]^);
  Found:=false;
  while ((not Found) and (I<=High(FData[0]^))) do
  begin
    if FData[0]^[I]=X then
    begin
      Found:=true;
      FData[1]^[I]:=Y;
    end;
    inc(I);
  end;
  if not Found then AddCoord(X,Y);
  FSortedByY:=false;
end;

procedure TXYSignal.PutXForGivenY(X,Y:double);
var
  I:longint;
  Found:boolean;
begin
  I:=low(FData[0]^);
  Found:=false;
  while ((not Found) and (I<=High(FData[0]^))) do
  begin
    if FData[1]^[I]=Y then
    begin
      Found:=true;
      FData[0]^[I]:=X;
    end;
    inc(I);
  end;
  if not Found then raise EXYSignal.Create('there was no Y value to match '+floattostr(Y));
  FSortedByX:=false;
end;

procedure TXYSignal.FillWithNull;
var
  I:longint;
begin
  for I:=Low(FData[0]^) to High(FData[0]^) do
  begin
    FData[0]^[I]:=FNullXValue;
    FData[1]^[I]:=FNullYValue;
  end;
  FSortedByX:=false;
  FSortedByY:=false;
end;

procedure TXYSignal.FillWithZero;
var
  I:longint;
begin
  for I:=Low(FData[0]^) to High(FData[0]^) do FData[1]^[I]:=0.0;
  FSortedByY:=false;
end;

procedure TXYSignal.GaussianFilter(Width:double);
type
  FilterRec=record
    Dist,
    Weight,
    Value:double;
  end;
  TFilterArray=array[-100..100] of FilterRec;
var
  StartIndex,
  EndIndex,
  Count,
  I:longint;
  TotalWeight,
  NewValue,
  MaxDist:double;
  FilterData:TFilterArray; {The array of data to be filtered.}
  FilteredSignal:TXYSignal; {The filtered signal.}
begin
  FilteredSignal:=TXYSignal.Create(self);
  StartIndex:=Low(FData[0]^)+1;
  EndIndex:=High(FData[0]^)+1;
  for Count:=StartIndex to EndIndex do
  if not (IsNullX(Count) or IsNullY(Count)) then
  begin
    for I:=-100 to 100 do FilterData[I].Weight:=-1.0;
    I:=Count;
    MaxDist:=0.0;
    repeat
      if not (IsNullX(I) or IsNullY(I)) then
      with FilterData[I-Count] do
      begin
        Dist:=abs(GetXValue(I)-GetXValue(Count));
        if Dist>MaxDist then MaxDist:=Dist;
        Weight:=exp(-1.0*sqr(Dist/(Width/1.6651092)));
        Value:=GetYValue(I);
      end;
      dec(I);
    until ((I<StartIndex) or ((I-Count)<-100));
    if Count<>EndIndex then
    begin
      I:=Count+1;
      repeat
        if not (IsNullX(I) or IsNullY(I)) then
        with FilterData[I-Count] do
        begin
          Dist:=abs(GetXValue(I)-GetXValue(Count));
          if Dist<=MaxDist then
          begin
            Weight:=exp(-1.0*sqr(Dist/(Width/1.6651092)));
            Value:=GetYValue(I);
          end else I:=EndIndex;
        end;
        inc(I);
      until ((I>EndIndex) or ((I-Count)>100));
    end;
    TotalWeight:=0.0;
    for I:=-100 to 100 do if FilterData[I].Weight>0.0 then TotalWeight:=
      TotalWeight+FilterData[I].Weight;
    NewValue:=0.0;
    for I:=-100 to 100 do if FilterData[I].Weight>0.0 then NewValue:=
      NewValue+FilterData[I].Value*FilterData[I].Weight/TotalWeight;
    FilteredSignal.PutYValue(Count,NewValue);
  end;
  for Count:=StartIndex to EndIndex do PutYValue(Count,FilteredSignal.GetYValue(Count));
  FreeAndNil(FilteredSignal);
end;

function TXYSignal.GetNumData:longint;
begin
  Result:=length(FData[0]^);
end;

function TXYSignal.IsNullX(Index:longint):boolean;
begin
  dec(Index);
  if FNullXValue=0.0 then
    if abs(FData[0]^[Index]-FNullXValue)<1E-4 then Result:=true else Result:=false
  else
    if abs(1.0-(FData[0]^[Index]/FNullXValue))<1E-4 then Result:=true else Result:=false;
end;

function TXYSignal.IsNullY(Index:longint):boolean;
begin
  dec(Index);
  if FNullYValue=0.0 then
    if abs(FData[1]^[Index]-FNullYValue)<1E-4 then Result:=true else Result:=false
  else
    if abs(1.0-(FData[1]^[Index]/FNullYValue))<1E-4 then Result:=true else Result:=false;
end;

{$WARNINGS OFF}
function TXYSignal.InterpolateYValue(X:double;LogInter,ReturnNullOnError:boolean):double;
{------------------------------------------------------------------------------}
{ This routine interpolates a Y value for a given X value. If the HuntWhen-    }
{ Interpolating property is set to true it will scan either side of X until it }
{ finds non-null Y values to interpolate between. If the HuntWhenInterpolating }
{ property is set to false and either of the Y values at the two X values      }
{ bracketting X are null, or if HuntWhenInterpolating is true but no non-null  }
{ Y values either side of X could be found, the routine raises an exception    }
{ or, if ReturnNullOnError is true, will return the Null Y value. After        }
{ calling this function, you can call GetIntDist to get the distance between   }
{ the neighbouring X values that were used to interpolate the Y value and/or   }
{ GetDistToValue to find the distance to the closest X value used in  the      }
{ interpolation. If the X value passed equals one of the X values in the array }
{ and no interpolation is required, the value returned by GetIntDist will be   }
{ zero. If the X value passed to this routine is less than the lowest X value  }
{ the function will return the Y value at the lowest X value if                }
{ HuntWhenInterpolating is true, and in this case GetIntDist will return 1E30. }
{ If HuntWhenInterpolating is false it will raise an exception or return the   }
{ null Y value if ReturnNullOnError is true. Similar behaviour occurs if the   }
{  X value passed is greater than the highest X value. If LogInter is true,    }
{ the routine performs logarithmic interpolation on the Y values. If the Y     }
{ values are zero or negative in this case, an exception is raised or, if      }
{ ReturnNullOnError is true, the routine returns the null Y value.             }
{------------------------------------------------------------------------------}
var
  N,         {This is simply High(Data).}
  I:longint; {General purpose loop control variable.}
  TempXData,
  TempYData:TDoubleArray;
  HadNullXValues:boolean;
begin
  try {Start of finally block.}
    {Put the data series into TempData and remove all null X entries.}
    SortByX;

    if FRemoveNullXOnInterpolate then
    begin
      TempXData:=Copy(FData[0]^);
      TempYData:=Copy(FData[1]^);
      HadNullXValues:=RemoveAllNullXEntries;
    end;

    try
      N:=High(FData[0]^);
      if N=-1 then raise EInterpol.Create('there are no data in the object.');

      {These are the default values returned by the function.}
      IntDist:=1E30;
      DistToValue:=1E30;
      IntIndexDist:=100000;

      {Note that the calls to IsNullY in this routine must pass the index plus 1 since the index }
      {passed to IsNullY must range from 1 to N i.e. it is decremented inside the IsNullY method.}
      {Deal with X value out of range.}
      try
        if ((X<FData[0]^[0]) or (X>FData[0]^[N])) then
        begin
          if not FHuntWhenInterp then raise EInterpol.Create('the X value was out of range.');
          if X<FData[0]^[0] then
          begin
            I:=0;
            while (IsNullY(I+1) and (I<N)) do inc(I);
          end;
          if X>FData[0]^[N] then
          begin
            I:=N;
            while (IsNullY(I+1) and (I>0)) do dec(I);
          end;
          if IsNullY(I+1) then raise EInterpol.Create('no non-null Y values could be found to perform the interpolation.');
          Result:=FData[1]^[I];
          DistToValue:=abs(X-FData[0]^[I]);
          exit;
        end;
      except
        on EInterpol do raise; {Just reraise existing non-critical exceptions.}
        on E:Exception do raise Exception.Create(E.Message+' error when dealing with the X value being out of range');
      end;

      {Adjust the scanning indices. We have <= in the first term because the algorithm}
      {below on traps cases of where X=Data[Int2].Value. So, if X is exactly at the   }
      {lower bound defined by Int1 and Int2, its better to open the scanning range fully.}
      try
        if ((X<=FData[0]^[Int1]) or (X>FData[0]^[Int2])) then
        begin
          inc(IndexStep);
          Int2:=N;
          Int1:=0;
        end
        else if IndexStep>0 then dec(IndexStep);
      except
        on E:Exception do raise Exception.Create(E.Message+' error when adjusting the '+
        'scanning values. The value of Int1 was '+inttostr(Int1)+
        ' and the value of Int2 was '+inttostr(Int2));
      end;

      {Find the indices for the X values spanning the X value passed as a parameter.}
      try
        while (Int2-Int1)>1 do
          if FData[0]^[(Int2+Int1) div 2]<X then Int1:=(Int2+Int1) div 2
          else Int2:=(Int2+Int1) div 2;
        if Int1=Int2 then raise EXYSignal.Create('Int1 was equal to Int2');
      except
        on EXYSignal do raise;
        on E:Exception do raise Exception.Create(E.Message+' occurred when finding the indices '+
        'for the X values spanning the X value passed');
      end;

      {Check if we are right on a level and a non-null Y value is available.}
      try
        if (EquivDouble(X,FData[0]^[Int2]) and (not IsNullY(Int2+1))) then
        begin
          Result:=FData[1]^[Int2];
          IntDist:=0.0;
          DistToValue:=0.0;
          exit;
        end;
      except
        on E:Exception do raise Exception.Create(E.Message+' occurred when checking if we are right in a level');
      end;

      {Scan for non-null Y values if necessary and permitted.}
      try
        if (FHuntWhenInterp and (IsNullY(Int1+1) or IsNullY(Int2+1))) then
        begin
          while (IsNullY(Int1+1) and (Int1>0)) do dec(Int1);
          while (IsNullY(Int2+1) and (Int2<N)) do inc(Int2);
        end;
      except
        on E:Exception do raise Exception.Create(E.Message+' occurred when scanning for non-null Y values');
      end;

      {If we still don't have valid Y values to interpolate between, then raise an exception.}
      if (IsNullY(Int1+1) or IsNullY(Int2+1)) then raise EInterpol.Create(
        'no non-null Y values could be found to perform the interpolation.');

      {We have valid Y values so proceed.}
      try
        IntDist:=FData[0]^[Int2]-FData[0]^[Int1];
        IntIndexDist:=Int2-Int1;
        if abs(X-FData[0]^[Int1])<abs(X-FData[0]^[Int2]) then
          DistToValue:=abs(X-FData[0]^[Int1]) else DistToValue:=abs(X-FData[0]^[Int2]);
        if LogInter then
        begin
          if ((FData[1]^[Int1]<=0.0) or (FData[1]^[Int2]<=0.0)) then raise EInterpol.Create(
            'logarithmic interpolation cannot be applied when Y values are less than or equal to zero');
          Result:=exp((X-FData[0]^[Int1])*(ln(FData[1]^[Int2])-ln(FData[1]^[Int1]))/
            (FData[0]^[Int2]-FData[0]^[Int1])+ln(FData[1]^[Int1]));
        end
        else Result:=(X-FData[0]^[Int1])*(FData[1]^[Int2]-FData[1]^[Int1])/
          (FData[0]^[Int2]-FData[0]^[Int1])+FData[1]^[Int1];
      except
        on E:Exception do raise Exception.Create(E.Message+' occurred when calculating the final returned value');
      end;

      if IndexStep>0 then
      begin
        Int2:=Int2+IndexStep;
        if Int2>N then Int2:=N;
        Int1:=Int1-IndexStep;
        if Int1<0 then Int1:=0;
      end;
    except
      on E:EInterpol do if ReturnNullOnError then Result:=NullYValue else raise EXYSignal.Create(
        'a specific exception occurred in TXYSignal.InterpolateYValue. The error message was '+E.Message);
      on E:EXYSignal do raise EXYSignal.Create('a specific exception occurred in TXYSignal.InterpolateYValue. The error message was '+E.Message);
      on E:Exception do
      begin
        if FShowExceptions then ShowMessage('The following non-specific exception occurred in TXYSignal.InterpolateYValue: '+E.Message);
        raise EXYSignal.Create('a non-specific exception occurred in TXYSignal.InterpolateYValue. The error message was '+E.Message);
      end;
    end;
  finally
    {Restore the original data series.}
    if (FRemoveNullXOnInterpolate and HadNullXValues) then
    begin
      FData[0]^:=Copy(TempXData);
      FData[1]^:=Copy(TempYData);
    end;
  end;
end; {End of TXYSignal.InterpolateYValue.}
{$WARNINGS ON}


function TXYSignal.GetIntDist:double;
begin
  Result:=IntDist;
end;

function TXYSignal.GetIntIndexDiff:integer;
begin
  Result:=IntIndexDist;
end;

function TXYSignal.GetDistToValue:double;
begin
  Result:=DistToValue;
end;

procedure TXYSignal.SortByX;
var
  I,J, {Loop control variables.}
  Index, {Index for input data array.}
  L:longint; {Used for interval halving.}
  Hold:XYRec; {Temporary storage.}
begin
  if FSortedByX then exit;
  if length(FData[0]^)>1 then
  begin
    L:=(length(FData[0]^) div 2)+1;
    Index:=length(FData[0]^);
    FSortedByX:=false;
    repeat
      if (L>1) then
      begin
        dec(L);
        Hold:=GetCoord(L);
      end
      else
      begin
        Hold:=GetCoord(Index);
        PutCoord(Index,GetCoord(1));
        dec(Index);
        if Index=1 then
        begin
          PutCoord(1,Hold);
          FSortedByX:=true;
        end;
      end;
      if not FSortedByX then
      begin
        I:=L;
        J:=L+L;
        while (J<=Index) do
        begin
          if (J<Index) then
          if (FData[0]^[J-1]<FData[0]^[J]) then inc(J);
          if (Hold.XValue<FData[0]^[J-1]) then
          begin
            PutCoord(I,GetCoord(J));
            I:=J;
            J:=J+J;
          end
          else J:=Index+1;
        end;
        PutCoord(I,Hold);
      end;
    until FSortedByX;
  end
  else FSortedByX:=true;
end;

procedure TXYSignal.RegridByX(StartX,EndX,Resolution:double);
var
  Hold:TXYSignal;
  I:integer;
  X:double;
begin
  Hold:=TXYSignal.Create(length(FData[0]^),FNullXValue,FNullYValue);
  try
    for I:=1 to length(FData[0]^) do Hold.PutCoord(I,GetXValue(I),GetYValue(I));
    Setlength(FData[0]^,0);
    Setlength(FData[1]^,0);
    X:=StartX;
    repeat
      AddCoord(X,Hold.InterpolateYValue(X,false,true));
      X:=X+Resolution;
    until (X>EndX);
  finally
    Hold.Free;
  end;
  FSortedByX:=true;
end;

function TXYSignal.RemoveAllNullXEntries:boolean;
var
  I,J:integer;
begin
  Result:=false;
  for I:=high(FData[0]^)+1 downto 1 do
  begin
    if IsNullX(I) then
    begin
      Result:=true;
      for J:=I+1 to High(FData[0]^)+1 do putCoord(J-1,GetCoord(J));
      setlength(FData[0]^,length(FData[0]^)-1);
      setlength(FData[1]^,length(FData[1]^)-1);
    end;
  end;
  Int2:=High(FData[0]^);
end;

procedure TXYSignal.WriteToTextFile(Filename:string);
var
  Dest:textfile;
  I:integer;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  for I:=Low(FData[0]^) to High(FData[0]^) do writeln(Dest,FData[0]^[I]:10:5,' ',FData[1]^[I]:10:5);
  closefile(Dest);
end;

end.


